/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file AIF_adaptive_filter.H
*
* \brief Definition of AIF class
*
*/
//-----------------------------------------------------------------------------

#ifndef ___AIF_ADAPTIVE_FILTER_H__
#define ___AIF_ADAPTIVE_FILTER_H__

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------
#include "MotionField.H"
#include "Frame.H"
#include "RefFrameManager.H"
#include "EntropyEncoder.H"
#include "EntropyDecoder.H"
//-----------------------------------------------------------------------------
/*!
* \brief AIF class
*
* Provides methods for AIF
*/
//-----------------------------------------------------------------------------
class AIF_adaptive_filter 
{
  #define DEFAULT_QUANT 10
  #define FILTER_SIZE 6
  #define FILTER_SIZE_OFFSET 7
  #define FILTER_OFFSET 2
  #define a_pos  0                 //  Full-pel  a_pos  b_pos  c_pos  Full-pel
  #define b_pos  1                 //     d_pos  e_pos  f_pos  g_pos
  #define c_pos  2                 //     h_pos  i_pos  j_pos  k_pos
  #define d_pos  3                 //     l_pos  m_pos  n_pos  o_pos
  #define e_pos  4                 //  Full-pel                       Full-pel
  #define f_pos  5
  #define g_pos  6
  #define h_pos  7
  #define i_pos  8
  #define j_pos  9
  #define k_pos 10
  #define l_pos 11
  #define m_pos 12
  #define n_pos 13
  #define o_pos 14

  public:
  int AIF_FilterFlag;                          
  int AIF_mode_flag1;  
  int AIF_mode_flag2; 
  int FilterFlag_SAIF[15];                     
  int DiffQFilterCoef_SAIF[15][FILTER_SIZE_OFFSET]; 
  int iFilterCoef_SAIF[15][FILTER_SIZE_OFFSET];     
  int aif_filterflag[LISTS][MAX_REF_FRAMES]; 

  AIF_adaptive_filter();
  virtual ~AIF_adaptive_filter();

  void ResetAdaptiveFilter(Frame *cur);
  void ResetAdaptiveFilter_DEC();
  void AIF_calculate_coef_SAIF(Frame *cur, MotionField *mf, RefFrameManager *ref_mgr);
  void CalculateFilterCoefficientsSep();
  void decode_filter_information(EntropyDecoder *edec);
  void encode_filter_information(EntropyEncoder *eenc);

  private: 
  double NxN_Matrix[15][FILTER_SIZE_OFFSET][FILTER_SIZE_OFFSET];
  double N_Vector[15][FILTER_SIZE_OFFSET]; 
  double FilterCoef[15][FILTER_SIZE_OFFSET];   
  double STANDARD_2D_FILTER_SEP[15][FILTER_SIZE_OFFSET]; 
  int iSTANDARD_2D_FILTER_SEP[15][FILTER_SIZE_OFFSET]; 
  int TwoDEquationPattern[15][FILTER_SIZE_OFFSET];
  void SetAdaptiveFilterHor(Frame *cur, PredBlk *curblk, int i, int j, RefFrameManager *ref_mgr);
  void SetAdaptiveFilterVer(Frame *cur, PredBlk *curblk, int i, int j, RefFrameManager *ref_mgr);
  void FindFilterCoefHor();
  void FindFilterCoefVer();
  void InterpolateHor (RefFrameManager *ref_mgr);
  void PredictFilterCoefficientsHor();
  void PredictFilterCoefficientsVer();
  void SAIF_HALF_rd_decision_Hor(Frame *cur, MotionField *mf, RefFrameManager *ref_mgr); 
  void SAIF_HALF_rd_decision_Ver(Frame *cur, MotionField *mf, RefFrameManager *ref_mgr);
  int SATD8x8(int *diff, int mel_shift); 
  void DequantizeFilterCoefficients(int *iQFCoef, int *iFCoef);
  void QuantizeFilterCoefficients(int iFCoef[FILTER_SIZE_OFFSET], int iQFCoef[FILTER_SIZE_OFFSET]);
  int FindPosition(int Dim, int cur_pos, int Offset, int mv, int inter_pel);
};

#endif // ___AIF_ADAPTIVE_FILTER_H__
