/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ArithDecoder.H
*
* \brief Definition of ArithDecoder class
*
*/
//-----------------------------------------------------------------------------

#ifndef __ArithDecoder_H__
#define __ArithDecoder_H__

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"

#include "ProbEst.H"
#include "Buffer.H"
#include "FixedProbEst.H"
#include "SimpleProbEst.H"
#include "UnaryProbEst.H"

//-----------------------------------------------------------------------------
/*!
* \brief Arithmetic decoder class
*
* Provides methods for arithmetic decoding. Usage:
* - Instantiate using ArithDecoder::ArithDecoder
* - Initialize using ArithDecoder::init
* - Decode symbols using ArithDecoder::decode_symbol, ArithDecoder::decode_symbol_half,
*   or ArithDecoder::decode_symbol_unary
* - Release using ArithDecoder::~ArithDecoder
*/
//-----------------------------------------------------------------------------
class ArithDecoder 
{
  unsigned int R; //!< Range register
  unsigned int V; //!< Value register
  int bits; //!< Number of spare bits in value register
  
  double *log_t; //!< Array containing AC_BITS-log2(x). This is used to determine a fractional number of bits used by a decoded symbol for reporting purposes.
  
  public:
  Buffer *buf; //!< Byte buffer from which to decode
#ifdef DEC_TRANSCODER
  Buffer *buf_out; //!< transcoding output.
#endif

  ArithDecoder();
#ifdef CABAC_SYNC_POINTS
  virtual ~ArithDecoder();
#else
  ~ArithDecoder();
#endif
  
#ifdef CABAC_SYNC_POINTS
  virtual void resetSynchro()  {} //!< reset all superblock synchro.
  virtual void initSynchro(int iSB, int jSB) {(void) iSB; (void)jSB; /* unused parameters. */ } //!< initialize superblock synchro.
  virtual void updateSynchro(int iSB) {(void)iSB; /* unused parameter. */} //!< update superblock synchro.
  virtual void terminate() {} //!< terminate frame decode.
#endif

#ifdef DEC_TRANSCODER
  // Functions virtual now.
  virtual void setEncodeThread(int thrd);
  virtual void init(Buffer *b, Buffer *b_out);
  virtual int decode_symbol(SimpleProbEst *ctx);
  virtual int decode_symbol_half(FixedProbEst *ctx);
  virtual int decode_symbol_unary(UnaryProbEst *ctx, FixedProbEst *ctx_suffix);
#ifdef VERBOSE_STATS
  virtual int get_nbits();
#endif
#else
  void init(Buffer *b);
  int decode_symbol(SimpleProbEst *ctx);
  int decode_symbol_half(FixedProbEst *ctx);
  int decode_symbol_unary(UnaryProbEst *ctx, FixedProbEst *ctx_suffix);
#ifdef VERBOSE_STATS
  int get_nbits();
#endif
#endif
};

#endif // __ArithDecoder_H__
