/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ArithTranscoder.H
*
* \brief Definition of ArithTranscoder class
*
*/
//-----------------------------------------------------------------------------

#ifndef __ArithTranscoder_H__
#define __ArithTranscoder_H__

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"

#include "ProbEst.H"
#include "Buffer.H"
#include "FixedProbEst.H"
#include "SimpleProbEst.H"
#include "UnaryProbEst.H"
#include "ArithDecoder.H"


#ifdef DEC_TRANSCODER
//-----------------------------------------------------------------------------
/*!
* \brief Arithmetic transcoder class, derived from ArithDecoder.
*
* During decode, the same symbols that are decoded are re-encoded using a
* supplied ArithEncoder which may have different L & R registers.
*/
//-----------------------------------------------------------------------------
class ArithEncoder;
class EntropyModel;
class ArithTranscoder : public ArithDecoder
{
public:
  ArithTranscoder();
  virtual ~ArithTranscoder();
  
  virtual void setEncodeThread(int thrd);
  virtual void resetSynchro();
  virtual void initSynchro(int iSB, int jSB);
  virtual void updateSynchro(int iSB);
  virtual void terminate();

  virtual void init(Buffer *b_in, Buffer *b_out);
  virtual int decode_symbol(SimpleProbEst *ctx);
  virtual int decode_symbol_half(FixedProbEst *ctx);
  virtual int decode_symbol_unary(UnaryProbEst *ctx, FixedProbEst *ctx_suffix);
  int decoding_unary; // TRUE implies we're decoding unary (which calls decode_symbol, decode_symbol_half).
  Buffer *buf_out; // Buffer given to us as the final output parameter.
  int nthreads;
  int thread_idx;
  ArithEncoder **aenc; // An encoder per thread.
  EntropyModel **cs; // An entropy model per thread.
  Buffer **buf; // A buffer per thread.
  int *sub_stream_size_bits; // Used during buffer output -- # bits to use for coding each emitted size.
  int *sub_stream_size; // Used during buffer output -- the size to write.

private:
  int substream_size(int first_ss, int n_ss);
  int calc_size_bits(int first_emit, int n_emit, int *last_bits = NULL);
  void emit_size(int first_emit, int n_total_emit, int n_emit);
  void emit_thread_bufs(int first_emit, int n_emit);
  int pad_bits;
};
#endif

#endif // __ArithTranscoder_H__
