/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Block.H
*
* \brief Definition of Block class
*
*
*/

#ifndef __Block_H__
#define __Block_H__

#include "Global.H"

//-----------------------------------------------------------------------------
/*!
* \brief MxN block class
*
* Simple class containing a MxN set of numbers. Numbers are represented as
* C-type short. This is to enable MMXization of code that uses this data
* structure.
*/
//-----------------------------------------------------------------------------
class Block
{
public:
  int val[256];

  static const int bwt[8];
  static const int bht[8];
  static const int log_bwt[8];
  static const int log_bht[8];

  static const int ZigZag_Table_16x16[256];
  static const int ZigZag_Table_8x8[64];
  static const int ZigZag_Table_4x4[16];

  static const int scan[6][TX_NCOEFF];

  Block();
  Block(int type);
  Block(int h, int w);
  ~Block();

  void set_size(int h, int w);
  void set_type(int type);

  void zigzagscan();
  void inverse_zigzagscan();

  int get_type() { return m_type; }
  int get_width()   { return width; }
  int get_height()  { return height; }
  int get_stride()  { return width; }
  int get_num_elements() { return TX_NCOEFF_MxN; }

  double  ssd(Block &x);
  void    sub(Block &x);
  void    add(Block &x);

  void zero();
  void set(int v);
  int is_zero();
  int sum();
  int sum2();
  int parity() { return sum() & 1; }
  
  int nnz();
  void sub(int DC);
  void add(int DC);
  void mul(int s);
   
/* Buffer Copying functions */
  void copy       (Block& blk);

  void insert(Block &blk, int i, int j);
  void extract(Block &blk, int i, int j);

  void rescan(int type1, int type2);

/* Private Helper Functions */
private:

/* Data Members */
private:
  const int* get_zz_table();

  int height;
  int width;
  int TX_NCOEFF_MxN;
  int m_type;
};

#endif // __Block_H__

