/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Buffer.H
*
* \brief Definition of Buffer class
*
*
*/


#ifndef __Buffer_H__
#define __Buffer_H__

#include "Global.H"

#define MAX_BUFFER_SIZE 1048576

//-----------------------------------------------------------------------------
/*!
* \brief Byte buffer class
*
* Used to managed bit streams. A buffer may contain up to MAX_BUFFER_SIZE
* bytes.
*/
//-----------------------------------------------------------------------------
class Buffer 
{
  unsigned char *data; //!< Data in buffer
  
public:
  unsigned int p_bit;  //!< Current bit position within current byte
  unsigned int p_byte; //!< Current byte position
  
  unsigned short payload_type; //!< Payload type
  unsigned int payload_size;   //!< Payload size (in bytes)

  enum // defines possible values of payload type
  {
    PLT_HEADER = 0x4ead, //!< Constant defining a header unit
    PLT_FRAME  = 0xda1a, //!< Constant defining a frame data unit
    PLT_END    = 0x0e0e  //!< Constant defining an end of sequence unit
  };

  Buffer();
  ~Buffer();
  
  void reset();
  
  void write_bit(int b);
  int write_bits(int b, int n); 
  
  int read_bit();
  int read_bits(int n);
  int read_byte();
  int read_word();
  
  int write_signed(int b); // cq_start
  int write_unsigned(int b);
  int read_signed();
  int read_unsigned(); // cq_end
  
  int read(FILE *f);
  void write(FILE *f); 
  
  void copy(Buffer *buf);
  void copy_coding_state(Buffer *buf);
#ifdef DEC_TRANSCODER
  static unsigned char rev_bits_tab[256];
  void fill_rev_bits_tab();
  void append_bytes_from(unsigned char *p, int sz);
  void append_bytes_from(Buffer *buf);
  void append_bits_from(Buffer *buf);
  void append_reversed_bits_from(Buffer *buf);
  int  payload_bits() {return p_byte*8+p_bit;}
#endif
#ifdef FRAME_SUBSTREAMS
  int read_bit_at(int bit_offset) {return (data[bit_offset/8]>>(7-bit_offset%8))&1;}
  int read_bits_at(int bit_offset, int bit_length);
  void copy_bits_from(Buffer *buf, int bit_offset, int bit_length);
  void copy_bits_backwards_from(Buffer *buf, int bit_offset, int bit_length);
#endif
#if defined(FRAME_SUBSTREAMS) || defined(DEC_TRANSCODER)
  static int get_coded_offset_size(int total_size_bits);
  static void divide_substreams(int n_ss, int &n_left, int &n_right);
#endif
};

#endif // __Buffer_H__
