/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file Deblocking_Filter.H
*
* \brief Definition of Deblocking_Filter class
*
*/
//-----------------------------------------------------------------------------

#ifndef __DeblockingFilter_H__
#define __DeblockingFilter_H__

#include "Global.H"
#include "EntropyEncoder.H"
#include "EntropyDecoder.H"
#include "Buffer.H"
#include "Param.H"
#include "LoopFilterOvercomplete3.H"
#include "IntraPrediction.H"
#include "LoopFilter.H"
#include "Plane.H"

#define NUM_LOOP_MODES_DBF 11

//-----------------------------------------------------------------------------
/*!
* \brief Base Deblocking_Filter class
*
* Provides methods for deblocking filtering
*/
//-----------------------------------------------------------------------------
class Deblocking_Filter
{
  public:
  int height;
  int width;
  int height2;
  int width2;
  Plane<int> m_luma[2]; //!< Array of filter modes for filtering luma block boundaries
  Plane<int> m_chroma[2][2]; //!< Array of filter modes for filtering chroma block boundaries
  MotionField *str_mf;
  void filter_data_copy(LoopFilter *lf, MotionField *mf) ;  

  Deblocking_Filter(int hh, int ww);

  ~Deblocking_Filter();

  void decision_for_each_sample( int* pos, int a, int b, int p2,  int p1,  int p0,  int q0,  int q1,  int q2 );

  void set_filter_modes();
  void fixed_deblocking_filter(MotionField *mf, unsigned char *oY, unsigned char *xY, int component );
#ifdef DEBLOCK_MT
  void fixed_deblocking_filter_rows(MotionField *mf, unsigned char *oY, Plane<unsigned char> &p_rec, int component, int mb_row_start, int n_mb_rows );
#endif
  void init_alpha_values(int qp_y, int qp_c);

  private:

  int dbf_idx_a[3];
  int dbf_idx_b[3];

  Plane<unsigned char> m_modes[3][2];
  Plane<unsigned char> m_boundaries[3][2];
};

#endif // __DeblockingFilter_H__
