/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file DenoiseTransform.H
*
* \brief Definition of DenoiseTransform class
*
*/
//-----------------------------------------------------------------------------

#ifndef __DenoiseTransform_H__
#define __DenoiseTransform_H__

#include "Global.H"

//-----------------------------------------------------------------------------
/*!
* \brief NDF transform class
*
* Contains methods to do forward/inverse transforms for NDF. This is a base 
* class. Derived classes should be added for a particular transform.
*/
//-----------------------------------------------------------------------------
class DenoiseTransform
{
protected:
  int m_size;            //!< Size of transform
  int m_norm;
  int m_basis_energy;
  int m_width;

  int *m_buffer_fwd;
  int *m_buffer_inv;  
  int *m_block;

  bool m_has_alternate;
  int *m_buffer_alt;
  int *m_block_alt;
  
  int *m_mask;

public:
  DenoiseTransform(int width, int size, int norm, int energy, bool has_alternate);
  virtual ~DenoiseTransform();

  void resetAccumBuffer(int col_begin, int col_end);
  int getSize();
  int getNorm();
  int getBasisEnergy();

  bool setMask(const int *th_row);
  
  int *getBlock(bool use_alternate = false);
  void scaleBlock(int scale);
  bool isZeroBlock();
  
  virtual void doColumnsForward(const int *src, int src_row, int src_pitch, int col_begin, int col_end, bool use_alternate = false) = 0;
  virtual void doRowsForward(int src_row, bool use_alternate = false) = 0;
  virtual void doRowsInverse(int dst_row) = 0;
  virtual void doColumnsInverse(int **dst, int col_begin, int col_end) = 0;
  
  static DenoiseTransform *create(int tsize, bool hadamard, int width);
};

#endif
