/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Encoder.H
*
* \brief Definition of Encoder class
*
*
*/

#ifndef __Encoder_H__
#define __Encoder_H__

#include "EntropyEncoder.H"
#include "MotionField.H"
#include "Frame.H"
#include "Buffer.H"
#include "ContextMap.H"
#include "MLC_Stat.H"
#include "IEncoder6.H"
#include "MotionEst.H"
#include "Global.H"
#include "RefFrameManager.H"
#include "LoopFilter.H"
#include "WienerFilter.H"
#include "Deblocking_Filter.H"
#include "Quantizer.H"
#include "AIF_adaptive_filter.H"

//-----------------------------------------------------------------------------
/*!
* \brief %Encoder class
*
* Contains main encoding loop excluding I/O
*/
//-----------------------------------------------------------------------------

class Encoder 
{
  EntropyEncoder *eenc; //!< Pointer to entropy coder
  MotionField *mf; //!< Pointer to motion field
  
  RefFrameManager *ref_mgr; //!< Pointer to reference frame manager
  LoopFilter *lf; //!< Pointer to loop filter
  WienerFilter *wlf; //!< Pointer to Wiener filter  
  Deblocking_Filter *dbf; //!< Pointer to Deblocking filter  
  AIF_adaptive_filter *saif_half;
  Transform *trans[3]; //!< Pointers to transforms (DCT, Alternate Intra, Alternate Inter)
  
  Buffer *buf_tmp; //!< Byte buffer array used for an encode of one frame
  Frame  *frm_tmp; //!< Reconstructed frame array used for an encode of one frame
  Frame  *frm_tmp_filtered;
  Frame  *frm_tmp_pred; 
  Frame  *frm_tmp_qpe_m;
  Frame  *frm_tmp_qpe_s;
  MLC_Stat *tmp_MLC; //!< Motion/Luma/Chroma stats used for an encode of one frame
  Buffer *(buf_best[2]); //!< Byte buffer array used for best encodes of one frame
  Frame  *(frm_best[2]); //!< Reconstructed frame array used for best encodes of one frame
  MLC_Stat *(best_MLC[2]); //!< Motion/Luma/Chroma stats for best encodes
#ifdef MVCOMP_ZERORES
  MotionField *(mf_best[2]); //!< MF copy of best frame.
  int inter_pel_best[2][LISTS][MAX_REF_FRAMES]; //!< copy of best inter_pel in effect.
  int log_inter_pel_best[2][LISTS][MAX_REF_FRAMES]; //!< copy of best log_inter_pel in effect.
  int ref_list_best[2][LISTS][MAX_REF_FRAMES]; //!< copy of best ref-list in effect.
  int frame_no_best[2][MAX_REF_FRAMES]; //!< copy of associated frame-nos in effect -- not strictly needed.
#endif
  Frame  *(frm_best_filtered [2]);
  Frame  *(frm_best_pred     [2]);
  Frame  *(frm_best_qpe_m    [2]);
  Frame  *(frm_best_qpe_s    [2]);
    
  int intra_cost_update;
  int num_loop_check;
  
  EntropyEncoder *stored_eenc[3]; //!< Pointer to temporal entropy coder (default transform, alternate transform, zero transform)
  Buffer         *stored_buff[3]; //!< Temporal Byte buffer array (default transform, alternate transform, zero transform)

  EntropyEncoder *stored_eencF[3][2][2];
  ZeroTree *stored_ztreeF_P[3][2][2];
  ZeroTree *stored_ztreeF_B[3][2];
  
  ZeroTree *stored_ztree_luma_I; //!< Temporal zero tree coder for luma intra block data
  ZeroTree *stored_ztree_chroma_I; //!< Temporal zero tree coder for chroma intra block data
  ZeroTree *stored_ztree_chroma_P; //!< Temporal zero tree coder for chroma inter block data
  ZeroTree *stored_ztree_luma_P[3]; //!< Temporal zero tree coder for luma inter block data
  ZeroTree *stored_ztree_luma_P_8x8[3]; //!< Temporal zero tree coder for luma 8x8 inter block data
  ZeroTree        *stored_ztree_mv_P; //!< Temporal zero tree coder for 8x8 motion data in P-pic
  ZeroTree        *stored_ztree_mv_B; //!< Temporal zero tree coder for 8x8 motion data in B-pic
  
  ContextMap *zmap[TWO]; //!< Zero block maps for luma (intra and inter)
  ContextMap *zmap_chroma[TWO][TWO]; //!< Zero block maps for chroma (intra and inter)
    
  Quantizer *quant[4]; //!< Array of quantizers (intra/inter and luma/chroma)
  
#ifdef MVCOMP_ZERORES
  int skip_initialize(MotionEst *me, MLC_Stat *mlc, int i, int j, int skip_type);
#endif
  void codemvsbi_blk(MotionEst *me, MLC_Stat *mlc, int i, int j, bool isFinal, EntropyEncoder *tmp_eenc);
  void frame_header(Frame *cur);

  void encode_luma_inter_block(int i, int j, int ii, int jj,int emodel, MLC_Stat  *mlc, int type,
  double lambda_rdoq, double lambda_mode, int &transformIdx);

  void encode_luma_inter_block_8x8(int i, int j, int emodel, WFrame *pred_frame, MLC_Stat  *mlc,
    int type, double lambda_rdoq, double lambda_mode, int &transformIdx);
  void encode_chroma_inter_blocks(int i, int j, int emodel, int type,double lambda_chroma);
  
  void block_mode_decision(Frame *cur, int i8, int j8, MLC_Stat  *mlc, MotionEst *me,int low_zeromv_rdoq, int &trans8x8Idx, int *trans4x4Idx);
  
  double calc_lambda_mode(int qp, int reference_idc, int rdoq, int low_zeromv_rdoq);
  
  void set_dqp(Frame *cur, IEncoder6  *ienc);
  
  int mv_count();
  
  int doIframe(Frame *cur, Buffer *buf, MLC_Stat *mlc);
  
  int get_bit_cost_one_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_orig, Frame *cur_filtered, int tmp_cnt, int frame_type, int me_mode, int single_ref_frame, int pel, int AIF_type);

  void store_prob(Frame *cur, int frame_type, int pel);
  
  int encode_one_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_orig, Frame *cur_filtered, Buffer *buf, MLC_Stat *mlc, int type, int aif_type, int inter_type, int tmp_cnt);  
  
  Frame *last_coded_frame; //!< Pointer to last coded frame
  
  WFrame *pred_frame; //!< motion predicted frame
  
  void keep_best_buf_frm_mlc(int i);
  
public:
  EntropyEncoder *(eenc_best[2]);
  int            low_zeromv_rdoq_best[2];
  const IEncoder6 *getIEncoder() { return m_ienc; }
  int frame_height; //!< %Frame height (luma samples unit)
  int frame_width; //!< %Frame width (chroma samples unit)
  int coded_frames; //!< Number of frames encoded so far
  int coded_inter_frames; //!< Number of inter frames encoded so far
  
  int bits; //!< Number of bits used for encoding a frame.
  int total_bits; //!< Total number of bits used for encoding frames, excluding bits used for byte alignment.
  int htotal_bits; //!< Total number of bits used for encoding frames, including bits used for byte alignment.
  
#ifndef MVCOMP_ZERORES
  int selected_transform_per_mode[10];
  int selected_transform4x4_per_mode[10][4];
#else
  int selected_transform_per_mode[10+MV_MAX_NPREDICTORS];
  int selected_transform4x4_per_mode[10+MV_MAX_NPREDICTORS][4];
#endif

  int total_sad; //!< Unused
  int total_motion_bits; //!< Unused
  
  MLC_Stat mlc_cur; //!< Motion/Luma/Chroma stats for current coded frame
  MLC_Stat mlc_tot; //!< Motion/Luma/Chroma stats for all coded frames
  int sad;

  int change_mv_precision;
  int skip_me_loop_on;
  
  Encoder(int h, int w);
  ~Encoder();
  
  void encode_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_orig, Frame *cur_filtered, Buffer *buf);
  double compute_distortion_filtered_against_orig(Frame *cur, Frame *cur_orig, Frame *cur_filtered);
  
  void set_ref_frame(Frame *cur);

private:
  WFrame *m_cur2;
  WFrame *m_cur2_pred;
  WFrame *m_cur2_qpe;
  WFrame *m_target;
  MotionEst *m_me[4];
  IEncoder6 *m_ienc; //!< Pointer to I-block encoder

#ifdef I16_IN_P_FRAME
  ZeroTree *m_tmpZeroTree16x16[7];
  
  void encode_16x16_blocks(Frame *cur, int i, int j, MLC_Stat *mlc, EntropyEncoder *eenc, int low_zeromv_rdoq);
#endif
  
};


#endif // __Encoder_H__
