/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file EntropyDecoder.H
* \brief Definition of EntropyDecoder class
*/
//-----------------------------------------------------------------------------

#ifndef __EntropyDecoder_H__
#define __EntropyDecoder_H__

#include "Global.H"
#include "EntropyCoder.H"
#include "ArithDecoder.H"
#ifdef DEC_TRANSCODER
#include "ArithTranscoder.H"
#endif
#include "EntropyModel.H"
#include "ZTNode.H"
#include "ZeroTree.H"
#include "PredBlk.H"
#ifdef SUBSTREAM_MT
#include <semaphore.h>
#endif
//-----------------------------------------------------------------------------
/*!
* \brief Entropy decoder class
*
* Contains methods to decode various syntax elements
*/
//-----------------------------------------------------------------------------
class EntropyDecoder : public EntropyCoder
{
  ArithDecoder *adec;    //!< Pointer to arithmetic decoder
  EntropyModel *cs;      //!< Pointer to entropy model
  
  int decode_refindex_lsb0(int ctx, int list);
  int decode_refindex_lsb1(int ctx, int list);
  int decode_refindex_msb(int ctx, int list);
  int decode_scale(int type);
  int decode_offset(int type);
  int decode_mvd(int type);
  
  public:
#ifdef CABAC_SYNC_POINTS
  EntropyDecoder(SynchroState *share = NULL);
#else
  EntropyDecoder();
#endif
  ~EntropyDecoder();
  
    EntropyModel *GetEntropyModel()    { return cs; }
#ifdef CABAC_SYNC_POINTS
#ifdef DEC_TRANSCODER
    void decodingFrameSyntax()         { adec->setEncodeThread(0);}
#endif
    void resetSynchro()                { if (param.superblock_synchro) cs->resetSynchro();         adec->resetSynchro();}
    void initSynchro(int iSB, int jSB) { if (param.superblock_synchro) cs->initSynchro(iSB, jSB);  adec->initSynchro(iSB, jSB); }
    void updateSynchro(int iSB)        { if (param.superblock_synchro) cs->updateSynchro(iSB);     adec->updateSynchro(iSB); }
    void terminate()                   { adec->terminate(); }
#endif
#ifdef MVCOMP_ZERORES
  int proba_mvp(int ctx) {return cs->p_mvp[ctx].p;}
  int count_mvp(int ctx, int cnt) {return cs->p_mvp[ctx].symcount[cnt];}
#endif
#ifdef DEC_TRANSCODER
  void reset(Buffer *b, Buffer *b_out = NULL);
#else
  void reset(Buffer *b);
#endif
  int decode_frame_type();
  int decode_reference_idc();
  int decode_frame_counter();
  int decode_deltaqp();
  int decode_number_of_references();
#ifdef MVCOMP_ZERORES
  int decode_mvcomp_config();
#endif
  int decode_zblk(int past, int chroma);
  int decode_mvp(int ctx);
  int decode_inter_type();

  int decode_I_frame_type ();

  void decode_wlf_flag      (int *wlf_flag);
  void decode_wlf_filtertype(int *filtertype);
  void decode_val           (int *p);
  void decode_wlf_hint_2D(int **coeffs, int *size_rec_x, int *size_rec_y, int *size_pred_x, int *size_pred_y, int *size_qpe_x, int *size_qpe_y, int component, int *filter_precision, int *filter_precision_table);
  void decode_wlf_hint_1D(int **coeffs, int *size_rec_x, int *size_rec_y, int *size_pred_x, int *size_pred_y, int *size_qpe_x, int *size_qpe_y, int component, int *filter_precision, int *filter_precision_table);

  int  decode_dummy();

  int  decode_tsfm_flag();

#ifdef Q_OFFSET
  int decode_q_offset_0();
  int decode_q_offset_1();
  int decode_q_offset8x8_0();
  int decode_q_offset8x8_1();
#endif

  void decode_quantization_set_id_half(); // cq_start
  void read_qmatrix_list_half(int *qmatrixList, int sizeOfQMatrixList, int *useDefaultQMatrixFlag, int matrix_type);
  void read_qmatrix_update_value_half(int *qmatrixList, int matrix_type);
  void read_qmatrix_set_half(int quantization_set_id, int matrix_type);
  void decode_qmatrix_half(int matrix_type); // cq_end

  int decode_FLC(int numbits, FixedProbEst *pe);
  int decode_rpps_type();
  int decode_rpps_id();
  int decode_filter_type();
  int decode_ic_disable();
  int decode_round_type();

  int decode_edge_prediction_flag();
  int decode_threshold_edge();
  
  int decode_ipred(int i, int j);
  int decode_refindex(int n_ref_frames, PredBlk *top, PredBlk *left, int list);
  
  int decode_ztree(ZeroTree *ztree);
  void decode_ztree_64(ZeroTree *ztree,unsigned char* zp);
  
  // functions for decode AIF information
  int decode_AIF_use_all_subpel_positions();
  int decode_AIF_subpel_positions_pattern();
  int decode_AIF_filter_pattern();
  int decode_AIF_diff_qfilter_coef();
  int decode_AIF_mode_flag1();
  int decode_AIF_mode_flag2();
  
  int decode_coeff_level(int type, int k);
  void decode_coeffs_intra_luma_4x4(ZeroTree *ztree, Block *blk, int type, int sip_type);
  void decode_coeffs(ZeroTree *ztree, Block *blk, int type);
  void decode_coeffs_8x8(ZeroTree *ztree, Block *blk, int chroma = 0);
  int decode_coeff_level_64(int k);
  int decode_coeff_level_64_CT8(int k);
  int decode_mode(PredBlk *left, PredBlk *top, int Bframe);
  void decode_mv(PredBlk *pblk);
    
  //{{SUBSTREAM_MT
  // This has been moved from the global param.
//-----------------------------------------------------------------------------
/*!
* \brief Zerotree structure for 8x8 blocks
*
*/
//-----------------------------------------------------------------------------
  typedef struct 
  {
    int nleaves; //!< for ztree size
    int s_nleaves; //!< for ztree split
    int iSoN; //!< start of node
    
    int zpatt; //!< Bit pattern of leaves of tree that have nonzero values
    int *stats; //!< States of zero tree
    
  } ZT_STATS_64;
  
  ZT_STATS_64 zt64[NZTREES64]; //!< Zero tree for 8x8
  int zt_idx[ZTsize64][2]; //!< Index of zero tree
  //SUBSTREAM_MT}}

  void print_stats(FILE *f);
#ifdef FRAME_SUBSTREAMS
  int  get_nbits() {return adec->get_nbits();}
  int  get_payload_bits() {return adec->buf->payload_size*8;}
  void sum_stats_from(EntropyDecoder *src) {cs->sum_stats_from(src->cs);}
#endif
#ifdef VERBOSE_STATS
#ifdef FRAME_SUBSTREAMS
  void print_stats_summary(FILE *f, int verbose, int buffer_payload_bits, int buffer_read_bits);
#else
  void print_stats_summary(FILE *f, int verbose);
#endif
#else
  void print_stats_summary(FILE *f);
#endif
  
  int decode_mvshare_size(int past, int ref1, int ref2, int type);

  int decode_sip_type(int ctx);

  int decode_spatial_intra_prediction_luma(int L_mode_pred_luma);
  int decode_spatial_intra_prediction_chroma(int up_mode, int left_mode, int up_is_inter, int left_is_inter);

  void decode_coeffs_16x16(ZeroTree **ztrees, Block *blk);
  void decode_subztrees(ZeroTree **ztree, unsigned char* rep);
};


#endif // __EntropyDecoder_H__
