/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file EntropyEncoder.H
* \brief Definition of EntropyEncoder class
*/
//-----------------------------------------------------------------------------

#ifndef __EntropyEncoder_H__
#define __EntropyEncoder_H__

#include "Global.H"
#include "EntropyCoder.H"
#include "ArithEncoder.H"
#include "Buffer.H"
#include "EntropyModel.H"
#include "ZTNode.H"
#include "ZeroTree.H"
#include "PredBlk.H"

//-----------------------------------------------------------------------------
/*!
* \brief Entropy encoder class
*
* Contains methods to encode various syntax elements
*/
//-----------------------------------------------------------------------------
class EntropyEncoder : public EntropyCoder
{
  ArithEncoder *aenc;   //!< Pointer to arithmetic decoder
  EntropyModel *cs;     //!< Pointer to entropy model
  
  //  int no_stats_enc;
  
  void encode_refindex_lsb0(int val, int ctx, int list);
  void encode_refindex_lsb1(int val, int ctx, int list);
  void encode_refindex_msb(int val, int ctx, int list);
  int cost_refindex_lsb0(int val, int ctx);
  int cost_refindex_lsb1(int val, int ctx);
  int cost_refindex_msb(int val, int ctx);
  
  void encode_bits(int val, int size, FixedProbEst *pe);
  void encode_unary(int val, UnaryProbEst *pe_level, FixedProbEst *pe_sign);
  void encode_unary2(int val, FixedProbEst *pe_level, FixedProbEst *pe_sign);
  void encode_ztree(ZeroTree *ztree, ZTNode *node);
  
  int cost_unary(int val, UnaryProbEst *pe_level, FixedProbEst *pe_sign);
  
  public:
  Buffer *buf;        //!< Pointer to byte buffer
  int nbits;          //!< Number of bits written to buffer
  int nbits_wlf_coeffs;//!< Number of bits written to buffer

  int bits_luma;      //!< counter for luma bits in IEncoder6
  int bits_chroma;    //!< counter for chroma bits in IEncoder6
  
  EntropyEncoder();
  ~EntropyEncoder();

  EntropyModel *GetEntropyModel() { return cs; }

#ifndef MVCOMP_ZERORES
  void reset(Buffer *b);
#else
  void reset(Buffer *b, int mv_keep_movement = 0);
  void recalc_movement();
#endif
  void set_buffer(Buffer *b);
  void copy_for_wlf(EntropyEncoder *e);
  void encode_frame_type(int val);
  void encode_reference_idc(int val);
  void encode_frame_counter(int val);
  void encode_deltaqp(int val);
  void encode_number_of_references(int val);
#ifdef MVCOMP_ZERORES
  void encode_mvcomp_config(int val);
#endif
  void encode_inter_type(int val);  //TCK8
  void encode_edge_prediction_flag(int val);
  void encode_threshold_edge(int val);
  void encode_I_frame_type (int val);

#ifdef Q_OFFSET
  void encode_q_offsets(int iTheta[2], int iTheta8x8[2]);   
#endif
  
  void encode_unary3(int val, FixedProbEst *pe_level); // cq_start
  void encode_quantization_set_id_half();
  void write_qmatrix_list_half(int *qmatrixList, int sizeOfQMatrixList, int *useDefaultQMatrixFlag, int matrix_type);
  void write_qmatrix_update_value_half(int *qmatrixList, int matrix_type);
  void write_qmatrix_set_half(int quantization_set_id, int matrix_type);
  void encode_qmatrix_half(int matrix_type); // cq_end

  void encode_wlf_flag(int *wlf_flag);
  void encode_wlf_filtertype(int *filtertype);
  void encode_val (int p);
  void encode_wlf_coeffs(int **coeffs, int *size_rec_x, int *size_rec_y, int *size_pred_x, int *size_pred_y, int *size_qpe_x, int *size_qpe_y, int component, int *filter_precision, int *filter_precision_table);

  void encode_dummy();

  void encode_tsfm_flag(int val);

  // functions for encode AIF information
  void encode_AIF_use_all_subpel_positions(int val);
  void encode_AIF_subpel_positions_pattern(int val);
  void encode_AIF_filter_pattern(int val);
  void encode_AIF_diff_qfilter_coef(int val);
  void encode_AIF_mode_flag1 (int val);
  void encode_AIF_mode_flag2 (int val);
  
  void encode_zblk(int val, int past, int chroma);
  void encode_mode(PredBlk *pblk, PredBlk *left, PredBlk *top, int Bframe);
  void encode_mv(PredBlk *pblk, int type, ZeroTree *tmp_ztree);
  void encode_mvp(int val, int ctx);
  void encode_refindex(int val, int n_ref_frames, PredBlk *top, PredBlk *left, int list);
  
  void encode_rpps_type(int val);
  void encode_rpps_id(int val);
  void encode_filter_type(int val);
  void encode_ic_disable(int val);
  void encode_round_type(int val);
  
  void entropycode_sip_type(int val, int ctx);
  int cost_sip_type(int val, int ctx);
  
  void encode_coeffs(ZeroTree *ztree, Block *blk, int type, int sip_type);

  void encode_coeffs(ZeroTree *ztree, Block *blk, int type);
  void encode_coeffs_8x8(ZeroTree *ztree, Block *blk, int chroma=0);
  void encode_coeff_level(int val, int i, int type);
  void encode_mvshare_size(int val, int ref, int past, int type);
  int cost_mvshare_size(int val, int ref, int past, int type);

  int cost_refindex(int val, int n_ref_frames, int x0, int x1);
  int cost_mv(int val, int i);
  int cost_mvp(int val, int ctx);
#ifdef MVCOMP_ZERORES
  int proba_mvp(int ctx) {return cs->p_mvp[ctx].p;}
  int count_mvp(int ctx, int cnt) {return cs->p_mvp[ctx].symcount[cnt];}
#endif
  int cost_ztree(ZTNode *node, int type);
  int cost_scale(int val, int type);
  int cost_offset(int val, int type);
  int cost_mode(int mode, int left, int top, int Bframe);
  
  double cost_one_coeff(int val, int type, int pos);
  double cost_ztree_rdo(Block *blk, int type, int context, ZeroTree *ztree);
  double cost_ztree_rdo8x8(Block *blk, int type, int context, ZeroTree *ztree);
  double cost_one_coeff8x8(int val, int pos);
  int cost_ztree_rdo_base(ZeroTree *ztree, ZTNode *node);
 
  int encode_spatial_intra_prediction_luma(int sip_mode, int L_mode_pred);
  void encode_spatial_intra_prediction_chroma(int sip_mode, int up_mode, int left_mode, int up_is_inter, int left_is_inter);

  void terminate();
  int get_nbits();
  double entropy();
  
  void copy_coding_state(EntropyEncoder *eenc);
  void copy_coding_state_cs(EntropyEncoder *eenc);
  int output_mv_zero(int val);
  int output_prev_intra_pred_mode_flag();
  int output_rem_intra_pred_mode_flag(int ctx);
  int output_intra_chroma_pred_mode(int ctx);
  int output_intra_chroma_pred_mode_1();
  int output_intra_chroma_pred_mode_2();
  int cost_coeff_level(int val, int type, int ctx);
  
   void encode_coeffs_16x16(ZeroTree **ztree, Block *blk);
   void subset_significance(Block *significance, Block *val, int nsubsets, int data_size);
   void copy_vals(Block *dst, Block *src, int offset, int size, int significance);
  
  int cost_spatial_intra_prediction_luma(int sip_mode, int L_mode_pred);
  int cost_spatial_intra_prediction_chroma(int sip_mode, int up_mode, int left_mode, int up_is_inter, int left_is_inter);
  int cost_zblk(int val, int past, int chroma);
  int cost_ztree(ZeroTree *ztree, ZTNode *node); 
  int cost_coeffs(ZeroTree *ztree, Block *blk, int type, int sip_type);
  int cost_coeffs(ZeroTree *ztree, Block *blk, int type);
  int cost_coeffs_8x8(ZeroTree *ztree, Block *blk8x8, int chroma=0);
  int cost_tsfm_flag( int val );
  int cost_coeffs_16x16(ZeroTree **ztree, Block *blk);
};

#endif // __EntropyEncoder_H__
