/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file ExplicitGop.H
*
* \brief Definition of ExplicitGop class and ExplicitGop::GOP_DATA struct
*
* \note right now, this data structure deals with intermediate frames only
* should it be extended to include all frames in a GOP (including last one)?
*
*/

#ifndef _EXPLICIT_GOP_H_
#define _EXPLICIT_GOP_H_

#include "Global.H"
//-----------------------------------------------------------------------------
/*!
* \brief GOP data class
*
*/
//-----------------------------------------------------------------------------
// GOP Pyramid
class ExplicitGop 
{
  int read_number(char **p);
  int read_symbol(char **p);
  
  public:
  
  //-----------------------------------------------------------------------------
  /*!
  * \brief GOP data class
  *
  */
  //-----------------------------------------------------------------------------
  typedef struct
  {
    int frame_type;       //!< Frame type
    int display_no;       //!< GOP Display order
    int reference_idc;    //!< Is reference?
    int frame_qp;         //!< Assigned QP
    int pyramid_layer;    //!< Pyramid layer (used with GOP Pyramid option 2)
  } GOP_DATA;
  
  int intermed_frame; //!< Number of intermediate frames
  GOP_DATA *gop_structure; //!< Array of GOP data for each frame in GOP
  
  void interpret_gop_structure(char *p);

  void create_pyramid();
  
  ExplicitGop(int int_fr);
  ~ExplicitGop();

  void resize(int s);
};

#endif // _EXPLICIT_GOP_H_
