/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Frame.H
*
* \brief Definition of Frame class
*
*
*/
#ifndef __Frame_H__
#define __Frame_H__

#include "Global.H"
#include "Block.H"

//-----------------------------------------------------------------------------
/*!
* \brief %Frame container class
*
* In addition to the actual frame data this also contains data relative to the
* encoding of the frame.
*/
//-----------------------------------------------------------------------------

class Frame 
{
public:
  unsigned char *data;    //!< Data array
  unsigned char *luma;    //!< Pointer to luma data
  unsigned char *crcb[2]; //!< Pointers to chroma data
  
  int size; //!< Number of samples
  int height; //!< Height of frame (luma samples)
  int width; //!< Width of frame (luma samples)
  int height2; //!< Height of frame (chroma samples)
  int width2; //!< Width of frame (chroma samples)
  int height4; //!< Height of frame (4x4 luma blocks)
  int width4; //!< Width of frame (4x4 luma blocks)
  int height8; //!< Height of frame (8x8 luma blocks)
  int width8; //!< Width of frame (8x8 luma blocks)

  double sse_luma; //!< Sum of square errors for luma
  double sse_chroma[2]; //!< Sum of square errors for chroma components
  double mse_luma; //!< Mean of square errors for luma component
  double mse_chroma[2]; //!< Mean of square errors for chroma components
  double snr_luma; //!< PSNR for luma component
  double snr_chroma[2]; //!< PSNR for chroma components

  int intra_frame_type; //!< used when frame_type=I: 0=INTRA, 1=IDR,  3=DDR
 
  int frame_no; //!< %Frame number
  int intermed_idx; //! 0 for edge frame, otherwise coding position of intermediate frame
  int frame_type; //!< %Frame type (I, P, B)
  int reference_idc; //!< Reference frame indicator
  int list_size[2]; //!< Size of each reference list
  int frame_order_counter; //!< %Frame order count
  int qp; //!< QP
  int qpC; //!< QP used for chroma
  int intra_qpC; //!< QP used for intra chroma
  int qpDC; //!< QP used for DC
  int delta_qp; //!< Delta QP
  int delta_qpC; //!< Delta QP for chroma
#ifdef Q_OFFSET
  int iTheta[2];
  int iTheta8x8[2];
#endif
  int qp_matrix[4][16]; //!< QP matrices
  int qp_matrix8x8[4][64]; //!< QP matrices for 8x8 block
  int qp_matrix16x16[4][256]; //!< QP matrices for 16x16 block
  int melq; //!< Motion estimation lambda value
  int lq_offset; //!< Quantized lambda parameter offset
  int lambdaq; //!< Quantized lambda parameter
  int lambdaqC; //!< Quantized lambda parameter for chroma
  float lambda; //!< Lambda parameter
  float lambdaC; //!< Lambda parameter for chroma
  
  int me_mode; //!< Motion estimation mode
  
  int ref_frame_stat[2][MAX_REF_FRAMES]; //!< Reference list index usage statistics
  
  Frame(int h, int w);
  ~Frame();
  
  void clear();
  void clear_stats();
  
  double compute_distortion(Frame *f);
  
  void get_blk_luma(Block *blk, int i, int j);
  void get_blk_chroma(Block *blk, int i, int j, int c);

  void copy(Frame *f);
  void copy_samples(Frame *f);
  void update_rd_params();
  int adjust_qp(int v, int offset); 
  
  void E();
};

#endif // __Frame_H__
