/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Global.H
*
* \brief Global definitions
*
*/

#ifndef __Global_H__
#define __Global_H__

#ifdef WIN32
#pragma warning(disable:4996) // Avoid VC++ warnings related to standard C libraries
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <float.h>
#include <time.h>
#include "Win32.H"

#define BUG_FIX_MANTIS_0000086
#ifdef BUG_FIX_MANTIS_0000086
// These defines control threadability of various filters.
//#define DENOISE_MT
//#define WIENER_MT
//#define DEBLOCK_MT

// Other useful defines present below:
// FRAME_SUBSTREAMS
// DEC_TRANSCODER
// SUBSTREAM_MT
#endif

#define STATS_16x16_FIX

//---------------------------------------------------------------------------------------------------------------
// Add Macros for new Tools below
//---------------------------------------------------------------------------------------------------------------

//#define INCOGNITO //!< Disables all outputs and command-line options in decoder

#define I16_IN_P_FRAME // Experimental

#define BETTER_CHROMA_DC

//#define DBF_BS_FIX

#define PMV0_AT_OUT_OF_BOUNDARY //PMV0 at out of the boundary

#define EXTENDED_PROBY0               // remember to change the probY0.dat to the old one when EXTENDED_PROBY0 is disabled.
#ifdef EXTENDED_PROBY0
#define ZTREE_64_CT8_CTX_READ
#define SIP_CTX_READ
#define MVP_CTX_READ
#endif

#define SCALED_CANDIDATES_IN_REGION_ME //Solution for wobble issue

#define INTRA_QPC_OFF

#define WIENER_FILTER_INLOOP_SWITCH

#define MAX_QUANT_SET_NUM 5 // cq_start, cq_end

// macros on fast mode are moved here
//#define SIXSTEPME_LOG      //!< Check if 6 Step ME is active

#define CABAC_SYNC_POINTS
//#define FRAME_SUBSTREAMS
//#define DEC_TRANSCODER
//#define SUBSTREAM_MT

#define CHROMA_INTRA_ADAPTIVE_LPF

#define MV_CODING_ENHANCE

#define HADAMARD_SHIFT_0 4     //!< sad is first multiplied by param.h0_mult and then divided by 16 by shifting right by 4.

#define AUTO_SELECT_HADAMARD
#ifdef AUTO_SELECT_HADAMARD
#define ASH_WIDTH 832       //!< width at and below which hadamard is automatically selected as 1
#define ASH_HEIGHT 480       //!< width at and below which hadamard is automatically selected as 1
#endif

#define USE_SSD
#ifdef USE_SSD
#define SSD_SHIFT 7     //!< sad is first multiplied by param.ssd_mult and then divided by 128 by shifting right by 7.
#endif

#define FAST_STATE_COPY
#define FAST_INTERPOL
#define FASTER_MVC
//#define VERBOSE_LOG // Enable this macro to get a verbose log

#define NSUBZTREES 4
#define TX_SIZE16 16                       //!< Transform block size for 16x16
#define TX_NCOEFF16 (TX_SIZE16 * TX_SIZE16) //!< Number of transform coefficients in 16x16 block
//#define INTRA_MODE_STATS

//Begin of Wiener Filter parameters
#define MAX_WIENER_FILTER_LENGTH 15
#define FILTER_PRECISION_TABLE_NUMBER 12
//End of Wiener Filter parameters

#define DISABLE_ZPATT_OUT

#define VERBOSE_STATS

#define NO_AIF_CHROMA

#define BLIM 10  //!< number of modes to search shift value in MotionTrellis when motion_sharing_me_trellis_full is FALSE

#define QUANT_OFFSET_STARTING  0 //!< id of quant offset to be stored at begining
#define QUANT_OFFSET_BEST  1     //!< id of quant offset to be stored for best predinction mode
//  #define SIPNxN_OVERALL        //!< for printing SIP statistics
//  #define SIPNxN_FRAME_STAT     //!< Reset statistics at ecery frame

#define Q_OFFSET //!< enable offsets for quantization and reconstruction (in the range of -127..128 -> offset/128*step_size)

#define F_BITS 14 //!< precision of SAIF filter coeffients
#define I_BITS 3  //!< not used
#define BORDER 13 //!< number of pixels on frame border not used to derive SAIF filter coeffients

#define I_PIC_LAMBDA 1.0 //!< not used (weight of lambda for I-picture)
#define P_PIC_LAMBDA 1.0 //!< not used (weight of lambda for P-picture)
#define B_PIC_LAMBDA 1.0 //!< not used (weight of lambda for B-picture)

// boolean constants
#define FALSE 0 //!< Constant equal to 0
#define TRUE 1 //!< Constant equal to 1

// integer constants
#define ZERO 0 //!< Constant equal to 0
#define ONE 1 //!< Constant equal to 1
#define TWO 2 //!< Constant equal to 2
#define THREE 3 //!< Constant equal to 3
#define FOUR 4 //!< Constant equal to 4
#define BITS_IN_BYTE 8 //!< Number of bits in a byte

#define LOG_MOTION_BLK_SIZE 3 //!< Log2 of width/height of a block unit for motion compensation (log2(8) = 3 for 8x8 blocks)
#define MOTION_BLK_SIZE (1 << LOG_MOTION_BLK_SIZE) //!< Width/height of a block unit for motion compensation

// motion stuff
#define LOG_MOTION_SUBBLK_SIZE4 2 //!< Log2 of width/height of a block unit for motion compensation (log2(4) = 2 for 4x4 blocks)
#define MOTION_SUBBLK_SIZE4 (1 << LOG_MOTION_SUBBLK_SIZE4) //!< Width/height of a block unit for motion compensation

#define INTRA_OFFSET_LAMBDA 0.65 //!< weight of RD cost for intra frame when % of intra blocks in inter frame goes beyond 50%

#define LOG_FRACTIONAL_PEL 3  //!< Log2 of precision of motion displacement (log2(8) = 3 for 1/8 pel motion compensation)
#define FPEL  0    //!< Full pel
#define OEPEL  1   //!< One-eighth pel
#define QPEL  2    //!< Quarter pel
#define TEPEL 3    //!< Three-eightth pel
#define HPEL  4    //!< Half pel
#define FEPEL 5    //!< Five-eighth pel
#define TPEL  6    //!< Three-quarter pel
#define SEPEL 7    //!< Seven-eighth pel
#define BILINEAR_FILTER 0 //!< Bilinear filter
#define SIX_TAP_FILTER 1 //!< Six tap filter
#define EIGHT_TAP_FILTER 2 //!< Eight tap filter

#define FRACTIONAL_PEL (1 << LOG_FRACTIONAL_PEL) //!< Precision of motion displacement

#define PEL_ONE_FULL 1 //!< subpel size for integer pel accuracy
#define PEL_ONE_HALF 2 //!< subpel size for half pel accuracy
#define LOG_PEL_ONE_FULL 0 //!< log2 of subpel size for integer pel accuracy
#define LOG_PEL_ONE_HALF 1 //!< log2 of subpel size for half pel accuracy
#define PEL_ONE_QUAD 4        //!< subpel size for quarter pel accuracy
#define PEL_ONE_EIGHTH 8      //!< subpel size for one-eighth pel accuracy 
#define LOG_PEL_ONE_QUAD 2    //!< log2 of subpel size for quater pel accuracy
#define LOG_PEL_ONE_EIGHTH 3  //!< log2 of subpel size for one-eighth pel accuracy

// 8-bit range stuff
#define MID_SAMPLE_VAL 128 //!< Middle sample value
#define MAX_SAMPLE_VAL 255 //!< Maximum sample value

#define SAMPLE_ADD_PREC 6 //!< added precision for transforms (should be 6)

// entropy coding stuff
#define NCTX_ZBLK 3 //!< Number of contexts for coding zero block flags
#define NCTX_ZMV 1 //!< Number of contexts for coding zero MVD flags
#define NCTX_SCALE 3 //!< Number of contexts for coding scale factors

// probabilities use 16-bit fixed point representation
#define PROB_BITS 16 //!< Precision used for representation fixed-point probabilities in arithmetic coder 
// size of arithmetic coder registers
#define AC_BITS 16 //!< Size of registers in bits for arithmetic coding

#define RATE_BITS 10 //!< Number of fractional bits used for param.bitsM

// precision for motion vector rescaling in ME
#define RESCALE_BITS 8 //!< Precision for motion vector rescaling in ME

#define UNDEFINED 0x7fffffff //!< Constant representing an undefined number
#define INT_INF 0x7fffffff //!< Maximum integer value
#define FLT_INF 1e30 //!< Large floating-point value

#define MAX_REF_FRAMES 4 //!< Maximum number of reference frames
#define MAX_LIST_FRAMES 4 //!< Maximum number of reference frames in a list
#define LISTS 2 //!< Number of lists

#define MAX_RPPS_TYPE 2 //!< Maximum number of rpps type
#define LOG_MAX_RPPS_TYPE 1 //!< log2 of Maximum number of rpps type
#define MAX_RPPS_ID 256 //!< Maximum number of rpps id
#define LOG_MAX_RPPS_ID 8 //!< log2 of Maximum number of rpps id
#define MAX_INTER_TYPE 4 //!< Maximum number of inter type
#define LOG_MAX_INTER_TYPE 2 //!< log2 of Maximum number of inter type
#define MAX_FILTER_TYPE 16 //!< Maximum number of interpolation filter type
#define LOG_MAX_FILTER_TYPE 4 //!< log2 of Maximum number of filter type
#define MAX_IC_DISABLE 2 //!< Maximum number of ic disable
#define LOG_MAX_IC_DISABLE 1 //!< log2 of Maximum number of ic disable
#define MAX_ROUND_TYPE 2 //!< Maximum number of rounding type
#define LOG_MAX_ROUND_TYPE 1 //!< log2 of Maximum number of rounding type
// transform-related
#define TX_SIZE 4 //!< Transform block size for 4x4
#define TX_NCOEFF (TX_SIZE * TX_SIZE) //!< Number of transform coefficients in 4x4 block
#define TX_SIZE8 8 //!< Transform block size for 8x8
#define TX_NCOEFF8 (TX_SIZE8 * TX_SIZE8) //!< Number of transform coefficients in 8x8 block

#define MAX_LINE_LEN 256 //!< Maximum length of a line in config file

#define POSITIVE 0 //!< Constant defining a positive number
#define NEGATIVE 1 //!< Constant defining a negative number

#define MAX_P_ENC 128 //!< maximum number of frame encodes
#define MAXNCAND 512 //!< Maximum number of candidates in trellis and region optimization of motion field

#define NZTREES 7 + NSUBZTREES
#define ZTsize64 64 //!< Zero Tree for 64 elements
#define NZTREES64 4 //!< used in decoder, split the 64 zero trees into number of sub trees

#define ZHV 2  //!< step size for 4x4 scanning order, 1 for raster scan, 2 for zig-zag scan. ONLY zig-zag scan (ZHV=2) is functional now!

#define TMA_SEARCH_WINDOW_RIGHT 0  //!< this is the extra search region to the right and above of the right boundary of the current block
// when in zig-zag luma block scan. parts of this search region is not available,  
// Setting to 0 resolves this issue.  The code has not been implemented to cake care of situations 
// when parts of the search region is not available and therefore has mismatch.
// But when ins raster scan, this value can be safely set to 4.
// however some initial experiment show that there is no improvement for setting this to 4.

enum 
{
  EGOP_TWO_LAYER    = 1, //!< Two-layer pyramid
  EGOP_FULL_PYRAMID = 2, //!< Full pyramid
  EGOP_EXPLICIT     = 3  //!< Explicit pyramid
};


// Definitions for Referenced/Non Referenced Pictures
enum 
{
  REFERENCE_PIC = 0,  //!< Reference picture
  DISPOSABLE_PIC = 1, //!< Disposable picture
};

// List Definitions for B frame support
enum 
{
  LIST_0 = 0, //!< List 0 prediction
  LIST_1 = 1, //!< List 1 prediction
  BI_PRED = 3, //!< Biprediction
  BI_PRED_ONLY = 4 //!< Biprediction Only
};

enum 
{
  MODE_P_INTRA = 0, //!< Intra in P-picture
  MODE_P_LIST_0_8x8 =1, //!< 8x8 List 0 in P-picture
  MODE_P_DEF=1 //!< default mode in P-picture
};

enum 
{
  MODE_INTRA = 0, //!< Intra in B-picture
  MODE_LIST_0_8x8 =1, //!< 8x8 List 0 in B-picture
  MODE_LIST_1_8x8 =2, //!< 8x8 List 1 in B-picture
  MODE_BI_PRED = 3,   //!< Bipred mode in B-picture
  MODE_DEF=3 //!< default mode in B-picture
};

// Frame types used when encoding frame type field
enum 
{
  ft_I_frame = 0, //!< I frame
  ft_P_frame = 1, //!< P frame
  ft_B_frame = 3, //!< B frame
  ft_X_frame = 4, //!< This is a dummy type that is used to imply that during encoding both P and B frame types will be tested.
};

enum
{
  frame_pos_edge = 0,
  frame_pos_disposable = 1,
  frame_pos_intermediate = 2
};

// Frame interpolation type used when encoding frame interpolation type field
enum 
{
  fit_HALFPEL_frame = 0, //!< half pel sample
  fit_OneQuad_frame = 1, //!< quarter pel sample
  fit_OneEighth_frame = 2, //!< one eighth sample
};

// Indices for motion
enum 
{
  vertical = 0, //!< Vertical component index
  horizontal = 1 //!< Horizontal component index
};

//-----------------------------------------------------------------------------
// Some basic functions, to be inlined
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Minimum of two integer values
*
* \param[in] a first value
* \param[in] b second value
* \return minimum of a and b
*/
//-----------------------------------------------------------------------------
inline static int imin(int a, int b) 
{
  return ((a) < (b)) ? (a) : (b);
}

//-----------------------------------------------------------------------------
/*!
* \brief Maximum of two integer values
*
* \param[in] a first value
* \param[in] b second value
* \return maximum of a and b
*/
//-----------------------------------------------------------------------------
inline static int imax(int a, int b) 
{
  return ((a) > (b)) ? (a) : (b);
}

//-----------------------------------------------------------------------------
/*!
* \brief Clip integer value to range
*
* Clips x to the range [low,high]
*
* \param[in] x valued to be clipped
* \param[in] low low end of range
* \param[in] high high of range
* \return clipped value
*/
//-----------------------------------------------------------------------------
inline static int clip(int x, int low, int high) 
{
  x = imax(x, low);
  x = imin(x, high);
  
  return x;
}

//-----------------------------------------------------------------------------
/*!
* \brief Absolute value of an integer value
*
* \param[in] x value
* \return absolute value
*/
//-----------------------------------------------------------------------------
inline static int iabs(int x) 
{
  return ((x) < 0) ? -(x) : (x);
}

//-----------------------------------------------------------------------------
/*!
* \brief Sign of an integer value
*
* \param[in] x value
* \return Sign (-1 or 1)
*/
//-----------------------------------------------------------------------------
inline static int sign(int x) 
{
  return ((x) < 0) ? -1 : 1;
}

//-----------------------------------------------------------------------------
/*!
* \brief Right bit shift with rounding
*
* \param[in] val integer value
* \param[in] b number of shifts
* \return shifted value
*/
//-----------------------------------------------------------------------------
inline static int shift_right_round(int val, int b) 
{
  return (val + (1 << (b-1))) >> b;
}

//-----------------------------------------------------------------------------
/*!
* \brief Right bit shift with rounding (unsigned integer)
*
* Same as shift_right_round but for unsigned integer values
*
* \param[in] val integer value
* \param[in] b number of shifts
* \return shifted value
*/
//-----------------------------------------------------------------------------
inline static unsigned int shift_right_round_us(unsigned int val, unsigned int b) 
{
  return (val + (1 << (b-1))) >> b;
}

//-----------------------------------------------------------------------------
/*!
* \brief Integer power of 2
*
* \param[in] x power
* \return 2 to the power x
*/
//-----------------------------------------------------------------------------
inline static int power2(int x) 
{
  return ONE << x;
}

//-----------------------------------------------------------------------------
/*!
* \brief Median value
*
* Computes median value of three integer a, b and c
*
* \param[in] a first integer value
* \param[in] b second integer value
* \param[in] c third integer value
* \return median value
*/
//-----------------------------------------------------------------------------
inline static int median(int a, int b, int c) 
{
  return a + b + c - imin(imin(a, b), c) - imax(imax(a, b), c);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute PSNR
*
* Computes PSNR based on MSE. When the macro JM_PSNR is not defined during compilation, an offset of 1/12 is added to the MSE to avoid infinite PSNRs.
*
* \param[in] mse MSE
* \return PSNR
*/
//-----------------------------------------------------------------------------
inline static double psnr(double mse) 
{
  return 10.0 * log(MAX_SAMPLE_VAL * MAX_SAMPLE_VAL / (mse < 1e-5 ? 1e-5 : mse)) / log(10.0);
}

#define max(a,b)    (((a) > (b)) ? (a) : (b))  //!< Maximum of two values
#define min(a,b)    (((a) < (b)) ? (a) : (b))  //!< Minimum of two values
typedef unsigned char byte;

#define HALF_PEL 1 //!< binary code for half pel accuracy (0, 1)
#define QUAD_PEL 2 //!< binary code for quater pel accuracy (1, 0)
#define EIGHTH_PEL 3 //!< binary code for one-eighth pel accuracy (1, 1)

#define S_FIF 0 //!< binary code for standard fixed interpolation filter (0, 0, 0, 0)
#define AIF_SAIF 1 //!< binary code for separable adaptive interpolation filter (0, 0, 1, 1)
#define N_FIF 5 //!< binary code for n fixed interpolation filter (0, 1, 0, 0)

#define IC_ON 0 //!< binary code for enabling intensity compensation (0)
#define IC_OFF 1 //!< binary code for disabling intensity compensation (1)

#define ROUND_LASTSTEP 0 //!< binary code for rounding at last step (0)
#define ROUND_INTERMED 1 //!< binary code for rounding at last step (1)

// spatial intra prediction stuff
#define NSIP    6 //!< number of SIP type

#define SIP4x4  0 //!< enumeration of 4x4 SIP type
#define SIP2x8  1 //!< enumeration of 2x8 SIP type
#define SIP8x2  2 //!< enumeration of 8x2 SIP type
#define SIP8x8  3 //!< enumeration of 8x8 SIP type 
#define SIP16x16 5 //!< enumeration of 16x16 SIP type

  // Sip type context
#define SIP_INTER  6 //enumeration of non Intra Block
#define NCTX_SIP_TYPE 3 //!< Number of contexts for coding sip type

#define TS_RDOQ_4X4_ENHANCE
#define TS_RDOQ_8X8_ENHANCE

//#define TRACE_DEBUG // Does it work with intra16x16?

// Allow use of a temporal mv predictor, in addition to A and B, and allows zeroing residue
// in block_mode_decision.
#define MVCOMP_ZERORES
#ifdef MVCOMP_ZERORES
// Allow both 'learn' and auto_max_temporal_diff.
// They both require symbol counts that are available
// if TRACE_STATS is set.
  #define MVCOMP_HAVE_SYMBOL_COUNTS
  #if defined(MVCOMP_HAVE_SYMBOL_COUNTS) && !defined(TRACE_STATS)
    #define TRACE_STATS
  #endif
  // The maximum # of predictors available in a stream.
  #define MV_MAX_NPREDICTORS 3
  // Note: these are used as indices.
  #define MV_PRED_LEFT     0
  #define MV_PRED_TOP      1
  #define MV_PRED_TEMPORAL 2
  #define MV_PRED_ZERO     3
  #define MV_PRED_NONE     4
  #define MV_MAX_PRED_INDEX 4
  // Default probability of signalling T when T is in competition.
  #define MV_DEFAULT_TEMPORAL_PROBA 655 // 1%
#endif

//===================================
// IDR coding parameters
// Define flag to turn on 
// Instantaneous Decoder Refresh (IDR)
// and Deferred Decoder Refresh (DDR)
//#define IDR_EXTRACT
// Intra Frame types used when encoding intra frame type field
// used when frame_type==ft_I_frame
enum 
{
  ft_I_noRefresh = 0, //!< I frame normal without refreshing the reference memory
  ft_I_IDR = 1,       //!< I frame with instant decoder refresh
  ft_I_DDR = 3,       //!< I frame with deferred decoder refresh until the next picture with frame order latter than the I frame.
};
//===================================

#define MEM_REDUCTION_FOR_LUMA_INTERPORATION

  enum 
  {
    ICOST_UPDATE = 0,    //!< intra cost refresh
    ICOST_STORED = 1,    //!< intra cost refresh and stored
    ICOST_COPY = 2,      //!< copy stored intra cost
  };

#ifdef CABAC_SYNC_POINTS
// Reset CABAC contexts at the start of each superblock
// using context information from left and upper superblock neighbours.
// Required for substream, transcoder and multithread support.
// Synchronisation points can be added to a stream using the decoder
// in "transcoder" mode if necessary.
#endif

#ifdef FRAME_SUBSTREAMS
// Allow substreams in a frame.  Currently a substream consists
// of lines of superblocks, although more flexible approaches are
// possible later.
// A multi-substream stream can be decoded by any decoder.
// The encoder currently encodes only mono-substream frames.  Substreams
// can only be added by the decoder in "transcoder" mode.
#endif

#ifdef DEC_TRANSCODER
// Introduce transcoder support into the decoder.  A "mono-substream"
// input stream can be recoded (at the arithmetic coder level) into a
// "multi-substream" output stream.
#endif

#ifdef SUBSTREAM_MT
// Allow the decoding of "multi-substream" frames using multiple
// threads to introduce parallelism.  The number of threads used can be
// from 1 up to the number of substreams in the frame and can
// be specified as a decoder parameter.
// This does not affect the encoder at all.
#endif

#if defined(DEC_TRANSCODER) || defined(SUBSTREAM_MT)
#ifndef CABAC_SYNC_POINTS
#define CABAC_SYNC_POINTS
#endif
#ifndef FRAME_SUBSTREAMS
#define FRAME_SUBSTREAMS
#endif
#endif
#if defined(FRAME_SUBSTREAMS)
// We need a delay of 2 superblocks from one superblock line to the next.
// This is due to the treatment of motionfield macroblock data (neighbouring
// macroblocks can be modified during the treatment of a macroblock).
// If this implementation detail is corrected, the delay can be reduced to 1
// which will improve thread parallelism, particularly at lower resolutions.
#ifdef BUG_FIX_MANTIS_0000086
// No TMA funnies in the motion field implies we can reduce the delay here.
#define SUPERBLOCK_DELAY 1
#else
#define SUPERBLOCK_DELAY 2
#endif
#ifndef CABAC_SYNC_POINTS
#define CABAC_SYNC_POINTS
#endif
#endif

#endif // __Global_H__
