/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file IEncoder6.H
* \brief Implementation of IEncoder6 class
*/
//-----------------------------------------------------------------------------

#ifndef __IEncoder6_H__
#define __IEncoder6_H__

#include "Quantizer.H"
#include "WFrame.H"
#include "Transform.H"
#include "ContextMap.H"
#include "ZeroTree.H"
#include "IntraPrediction.H"
#include "MLC_Stat.H"
#include "MotionField.H"

#define INTRA16x16WEIGHTEDCOST 128 // cq_start
#define INTRA8x8WEIGHTEDCOST 128
#define INTRA4x4WEIGHTEDCOST 128
#define INTRA8x2WEIGHTEDCOST 128
#define INTRA2x8WEIGHTEDCOST 128 // cq_end

//-----------------------------------------------------------------------------
/*!
* \brief Main intra encoder class
*
*/
//-----------------------------------------------------------------------------
class IEncoder6
{
public:
  // Data shared with encoder (allocated by encoder)
  WFrame *m_cur2;            //!< WFrame of current picture (input and later reconstructed)
  WFrame *m_cur2_pred;       //!< WFrame of current prediction picture
  WFrame *m_cur2_qpe;        //!< WFrame of current residual picture
  WFrame *m_target;          //!< WFrame of original picture
  void setFrames(WFrame *cur2, WFrame *cur2_pred, WFrame *cur2_qpe, WFrame *target);
  
  MotionField *m_motionField;
  void setMotionField(MotionField *motionField);
  
  PredBlkI *store;
  WFrame *store_ipred_img; //!< WFrame to store residual of intra blocks
  
  PredBlkI *ipic_store;
  
#ifdef TRACE_DEBUG
  int ipass;
#endif
  int     frame_type;       //!< Frame Type
  int    iqp;               //!< QP value for luma
  int    iqpC;             //!< QP value for Chroma
  IntraPrediction *intra_pred;  //!< Pointer to IntraPrediction class where common functions for intra prediction are found
  
#ifdef SIPNxN_FRAME_STAT
  double ss_2x8[11][16];   //!< Sum of Square of coefficients in 2x8 partitions grouped by prediction modes
  int    ss_2x8_count[11]; //!< Counter of 2x8 partitions grouped by prediction modes
  double ss_8x2[11][16];   //!< Sum of Square of coefficients in 8x2 partitions grouped by prediction modes
  int    ss_8x2_count[11]; //!< Counter of 8x2 partitions grouped by prediction modes
  double ss_4x4[11][16];   //!< Sum of Square of coefficients in 4x4 partitions grouped by prediction modes
  int    ss_4x4_count[11]; //!< Counter of 4x4 partitions grouped by prediction modes
  double ss_8x8[11][64];   //!< Sum of Square of coefficients in 8x8 partitions grouped by prediction modes
  int    ss_8x8_count[11]; //!< Counter of 8x8 partitions grouped by prediction modes
#endif 
  
  
  int    sip_type_count[NSIP];  //!< Counter of SIP type
  int    *tmp_mode_y;           //!< temporary mode for luma
  int    *tmp_lmode_y;          //!< temporary L-mode for luma
  
  Block  best_coeff_8x8;       // !< temporal block buffer for storing coefficients of best mode in 8x8 level
  Block  *best_coeff_16x16;// !< temporal block buffer for storing coefficients of best mode in 16x16 level

  EntropyEncoder  stored_eenc;           //!< Temporary Entropy Encoder
  Buffer          stored_buff;           //!< Temporary Bitstream Buffer 
  ZeroTree        *stored_ztree_luma;    //!< Pointer to temporary ZeroTree for luma
  ZeroTree        *stored_ztree_chroma;  //!< Pointer to temporary ZeroTree for chroma
  ZeroTree        *stored_ztree_luma_8x8;//!< Pointer to temporary ZeroTree for luma 8x8

  EntropyEncoder  rd_eenc;               //!< Temporary Entropy Encoder for rd calculation
  ZeroTree        *stored_ztree_chroma_8x8;//!< Pointer to temporary ZeroTree for chroma 8x8
  EntropyEncoder rd_eenc_tmp16;
  Buffer          rd_buff_tmp16;               //!< Temporary Bitstream Buffer for rd calculation

  Buffer          rd_buff;               //!< Temporary Bitstream Buffer for rd calculation
  ZeroTree        *rd_ztree_luma;        //!< Pointer to temporary ZeroTree for luma for rd calculation
  ZeroTree        *rd_ztree_chroma;      //!< Pointer to temporary ZeroTree for chroma for rd calculation
  ZeroTree        *rd_ztree_luma_8x8;  //!< Pointer to temporary ZeroTree for luma 8x8 for rd calculation
  
  ZeroTree        *rd_ztree_chroma_8x8;//!< Pointer to temporary ZeroTree for chroma 8x8 for rd calculation
  ZeroTree        *rd_subztrees_luma[NSUBZTREES];
  ZeroTree        *stored_subztrees_luma[NSUBZTREES];
  
  Transform *m_trans[2];                 //!< Pointers to transforms (DCT and AlternateIntra)
  Quantizer *quant_luma;                 //!< Pointer to luma quantizer
  Quantizer *quant_chroma;               //!< Pointer to chroma quantizer
  ContextMap *zmap;                      //!< Zero block map for luma
  ContextMap *zmap_chroma[2];            //!< Zero block map for chroma
  int sip_mode_count[NSIP][3][IntraPrediction::NO_INTRA_PMODE_LUMA_MxN];         //!< Array of SIP mode counter
  int sip_min_mode_count[NSIP][3][IntraPrediction::NO_INTRA_PMODE_LUMA_MxN];     //!< Array of SIP mode counter for best mode 
  int sip_sel_mode_count[NSIP][3][IntraPrediction::NO_INTRA_PMODE_LUMA_MxN];     //!< Array of SIP mode counter for best mode & L-mode being the same

  int selected_transform;
  int selected_transform_16x16;

  int rd_mode_decision_chroma(EntropyEncoder *eenc, Block min_blk[2], Block org_blk[2], const int i, const int j, const int up_mode, const int left_mode, const int up_is_inter, const int left_is_inter, double lambda, ZeroTree *ztree);

  double get_rate_luma(EntropyEncoder *eenc, Block *coef, int i, int j, int cur_mode, int sip_type);
  int rd_mode_decision_luma(EntropyEncoder *eenc, Block *min_blk, Block *org_blk, int i, int j, double lambda, ZeroTree *ztree, int sip_type, double &min_rdcost);
  double get_rdcost_luma(Block &coeff, int i, int j, EntropyEncoder *eenc, ZeroTree *tmp_ztree, int blkIdx);

  double get_rate_chroma(EntropyEncoder *eenc, Block uv_coef[2], int i, int j, int cur_mode, int up_mode, int left_mode, int up_is_inter, int left_is_inter);
  double calc_lambda_mode(int chroma, int rdoq);

  int rd_mode_decision_luma_16x16(EntropyEncoder *eenc, Block *min_blk, Block *org_blk, const int i, const int j, double lambda);
  double get_rate_luma_16x16(EntropyEncoder  *eenc, Block *coef, int i, int j, int cur_mode);
  double rdcost8, rdcost16;

public:

  IEncoder6(int height, int width, Quantizer *q_luma, Quantizer *q_chroma, ContextMap *zmap_luma, ContextMap *zmap_chroma0, ContextMap *zmap_chroma1);
  ~IEncoder6();

  int  get_transform() {return selected_transform;}
  void reset_zmap();

  void encode_chroma_blocks(int i, int j, EntropyEncoder *eenc, MLC_Stat *mlc, int type, ZeroTree *tmp_ztree, int sip_type);
  void reset_sip_mode_count();
  void print_sip_mode_count() const;

  void encode_luma_block_NxN(int i, int j, EntropyEncoder *eenc, MLC_Stat *mlc, int type, ZeroTree *tmp_ztree, ZeroTree *tmp_ztree_64);

  void encode_luma_block_16x16(int i, int j, EntropyEncoder *eenc, MLC_Stat *mlc, int type, ZeroTree *tmp_ztree, ZeroTree *tmp_ztree_64, ZeroTree **tmp_subztrees);
  void encode_chroma_block_8x8(int i, int j,  EntropyEncoder *eenc, MLC_Stat *mlc);
  double get_rdcost_luma_blk16x16(Block *blk, int i, int j, EntropyEncoder *eenc);
  void entropycode_luma_blk16x16(Block *blk, int i, int j, EntropyEncoder *eenc, int type, ZeroTree **tmp_subztrees, int sip_type);

private:

  void entropycode_luma(Block &data, int i, int j, EntropyEncoder *eenc, int type, ZeroTree *tmp_ztree, int blkIdx);

  void entropycode_luma_sub_block_8x8(Block &coeff, int i, int j, EntropyEncoder *eenc, int type, ZeroTree *tmp_ztree, ZeroTree *tmp_ztree_64, int sip_type);
};

#endif // __IEncoder6_H__
