/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file LumaFrames.H
*
* \brief Definition of LumaFrames class
*
*/
//-----------------------------------------------------------------------------

#ifndef __LumaFrames_H__
#define __LumaFrames_H__

#include "Frame.H"
class AIF_adaptive_filter;

//-----------------------------------------------------------------------------
/*!
* \brief Interpolated luma data class
*
* Contains luma data interpolated with various fractional pel positions.
* This is an abstract class. Derived classes should implement the makemap
* and AIF_makemap methods which generate the interpolated luma data.
*
* \note Interpolated data is extended beyond frame boundaries
*/
//-----------------------------------------------------------------------------
class LumaFrames 
{
  
protected:
  void extend(Frame *ref);
  
public:
  int height; //!< Height of data array, including border
  int width; //!< Width of data array, including border
  unsigned char *data[FRACTIONAL_PEL][FRACTIONAL_PEL]; //!< Pointers to data array for each fractional position
  
  int *intermediate[FRACTIONAL_PEL][FRACTIONAL_PEL]; //!< Pointers to an intermediate data array
  int *buf; //!< Pointer to an array used internally
  
  int border; //!< Size of border added to frame

  int decoder_mode;

  int iFilterCoef_SAIF[15][7];
  int AIF_mode_flag1, AIF_mode_flag2;
  
  virtual void makemap(Frame *ref) = 0; //!< Generate interpolated data \param[in] ref Reference frame
  
  virtual void AIF_makemap(Frame *ref, Frame *cur, int cur_list, int cur_ref, AIF_adaptive_filter *saif_half) = 0;//!< Generate interpolated data using SAIF \param[in] ref Reference frame
  
  virtual ~LumaFrames() 
  {
  } //!< Destructor
  
  static LumaFrames *create(int h, int w, int b, int decoder_mode=0);

  virtual void interpolate_on_the_fly(int k, int l, int u, int v, unsigned char *data) = 0;
  virtual void interpolate_on_the_fly_AIF(int k, int l, int u, int v, unsigned char *data) = 0;
};

#endif // __LumaFrames_H__
