/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file LumaFrames8pel.H
*
* \brief Definition of LumaFrames8pel class
*
*/
//-----------------------------------------------------------------------------

#ifndef __LumaFrames8pel_H__
#define __LumaFrames8pel_H__

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "LumaFrames.H"

//-----------------------------------------------------------------------------
// Class definition
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Interpolated luma data class: 1/8th pel interpolation
*
* Inherits from LumaFrames class and implements 1/8 pel luma interpolation.
*
*/
//-----------------------------------------------------------------------------
class LumaFrames8pel : public LumaFrames 
{
  
  void filter_6tap(int *src, int n);
  void filter_8tap(int *src, int n);
  
  void avg(int *src0, int *src1, int *dst0, unsigned char *dst1);
  void avg_d(int *src0, int *src1, int *dst0, unsigned char *dst1);
  void avg_r(int *src0, int *src1, int *dst0, unsigned char *dst1);
  void avg(unsigned char *src0, unsigned char *src1, unsigned char *dst);
  void avg_d(unsigned char *src0, unsigned char *src1, unsigned char *dst);
  void avg_r(unsigned char *src0, unsigned char *src1, unsigned char *dst);
  
  void print(int startIndex, int endIndex);
  
  void generate_hp();
  void generate_qp();
  void generate_8p_bilinear();
  void generate_8p_multi_tap(int filter);
  void AIF_generate_qp(Frame *cur, int list, int ref);
  void AIF_generate_8p();
  void setAIFcoef(AIF_adaptive_filter *saif_half);
  
public:
  LumaFrames8pel(int h, int w, int b, int decoder_mode);
  virtual ~LumaFrames8pel();
  
  virtual void AIF_makemap(Frame *ref, Frame *cur, int cur_list, int cur_ref, AIF_adaptive_filter *saif_half);
  virtual void makemap(Frame *ref);

  virtual void interpolate_on_the_fly(int k, int l, int u, int v, unsigned char *data);
  virtual void interpolate_on_the_fly_AIF(int k, int l, int u, int v, unsigned char *data);

private:
  void getblock_QP(int k, int l, int u, int v, unsigned char *blk_data);
  void getblock_EP_hor(int k, int l, int u, int v, unsigned char *blk_data);
  void getblock_EP_ver(int k, int l, int u, int v, unsigned char *blk_data);
  void getblock_EP_diag(int k, int l, int u, int v, unsigned char *blk_data);
  void QP_interpolate_fixed_on_the_fly(unsigned char* full_pel_data, unsigned char* sub_pel_data, int* intermed_data, int ver, int hor);
  void avg(int *src0, int *src1, unsigned char *dst1, bool bug = false);

  void getblock_QP_AIF(int k, int l, int u, int v, unsigned char *blk_data);
  void getblock_EP_hor_AIF(int k, int l, int u, int v, unsigned char *blk_data);
  void getblock_EP_ver_AIF(int k, int l, int u, int v, unsigned char *blk_data);
  void getblock_EP_diag_AIF(int k, int l, int u, int v, unsigned char *blk_data);
  void QP_interpolate_adaptive_on_the_fly(unsigned char* full_pel_data, unsigned char* sub_pel_data, int* intermed_data, int ver, int hor);
};

#endif // __LumaFrames8pel_H__

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
