/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MCRefBlk8.H
*
* \brief Definition of MCRefBlk8 class
*
*
*/

#ifndef __MCRefBlk8_H__
#define __MCRefBlk8_H__

#include "Global.H"
#include "PredBlk.H"
#include "Frame.H"
#include "LumaFrames.H"
#include "ChromaFrames.H"
#include "RefFrameManager.H"

#include "MotionField.H"

//-----------------------------------------------------------------------------
/*!
* \brief MC reference block class
*
* Contains an 8x8 reference block (luma and chroma) used for motion estimation.
* Includes methods for computing SAD.
*/
//-----------------------------------------------------------------------------

class MCRefBlk8 
{
  
  void get_luma(Frame *ref, int u, int v);
  void get_chroma(Frame *ref, int u, int v, int c);
  unsigned char *get_chroma_ptr(PredBlk *blk, int list, int i, int j, unsigned char *data = NULL);
  
public:
  unsigned char val[96]; //!< 64 luma and 32 chroma samples of block to be matched
  
  
  int sum_luma; //!< Sum of luma values in array val
  RefFrameManager *ref_mgr; //!< Pointer to reference frame manager
  int mel_shift; //!< Bit shift count to be applied to luma SAD
  int chroma_weight; //!< Weight to be applied to chroma SAD
  
  unsigned char *get_luma_ptr(PredBlk *blk, int list, int i, int j);
    
  void get_data(Frame *ref, int u, int v);
  int sad(int i,int j,int maxsad,PredBlk *mv);
  int sad_wp(int i,int j,int maxsad,PredBlk *mv);
  int distortion(int i,int j,int maxsad,PredBlk *mv);    
  int sad(int i,int j,PredBlk *mv);
  int sad_w_intra_check(int i,int j,PredBlk *mv);
  int sad_fic(int i,int j,PredBlk *mv);
  int get_mc_sum_luma(int i, int j, PredBlk *blk, int list);
  int bipred_sad(int i,int j,int maxsad,PredBlk *mv);
  int bipred_sad_wp(int i,int j,int maxsad,PredBlk *mv);
  int bipred_sad(int i,int j,PredBlk *mv);
  
  int mel_shift_sad(int sad, int mult, int shift);

  int transform(int *x);
  int transform8x8(int *x);
  int transform2x8(int *x);
  int transform8x2(int *x);
  int transformC(int *x);

  int distortion_mvshare(int i, int j, PredBlk *cur, PredBlk *top, PredBlk *left, int BiPred);
  int sad_mvshare(int i, int j, PredBlk *blk, PredBlk *top, PredBlk *left);
  int sad_wp_mvshare(int i, int j, PredBlk *blk, PredBlk *top, PredBlk *left);
  int bipred_sad_mvshare(int i, int j, PredBlk *blk, PredBlk *top, PredBlk *left);
  int bipred_sad_wp_mvshare(int i, int j, PredBlk *blk, PredBlk *top, PredBlk *left);
    
private:
  int distortion_luma(int *diff);
};

#endif // __MCRefBlk8_H__

