/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MLC_Stat.H
*
* \brief Definition of MLC_Stat class
*
*/

#ifndef __MLC_Stat_H__
#define __MLC_Stat_H__

#include "Global.H"
#include "IntraPrediction.H"

//-----------------------------------------------------------------------------
/*!
* \brief Motion/Luma/Chroma statistics class
*
* Contains various statistics for reporting purposes.
*/
//-----------------------------------------------------------------------------
class MLC_Stat 
{
  public:
  int motion_bits; //!< Number of bits for motion data
  int luma_bits; //!< Number of bits for luma data
  int chroma_bits; //!< Number of bits for chroma data
  int aif_bits; //!< bits for SAIF information
  int zblks; //!< Number of zero inter blocks
  int nzblks; //!< Number of nonzero inter blocks
  int iblks; //!< Number of intra blocks
  
  int bit_count_luma; //!< bits for luma
  int bit_count_chroma; //!< bits for chroma
  
#ifdef INTRA_MODE_STATS
  int sip_type_count[NSIP];
  int sip_mode_count[NSIP][3][NO_INTRA_PMODE_LUMA_MAX];
#endif //INTRA_MODE_STATS

  int intra_tma_L5; //!< Number of intra tma L5 blocks
  int intra_tma_L9; //!< Number of intra tma L9 blocks
  int intra_tma_L5c; //!< Number of intra tma L5 blocks for chroma
  int intra_tma_L9c; //!< Number of intra tma L9 blocks for chroma

  int mvshare_two;
  int mvshare_four;
  int mvshare_zero;
  int mvshare_off;

  int list_stat[5];  //!< Prediction mode usage (LIST_0;0, LIST_1;1, BI_PRED;3, intra;4)
  int ref_frame_stat[LISTS][MAX_REF_FRAMES]; //!< Reference frame usage
  
  MLC_Stat();
  
  void reset();
  void add(MLC_Stat *mlc);
  
  int list_stat_norm();
  int ref_frame_stat_norm(int list);

#ifdef INTRA_MODE_STATS
  void print_intra_mode_stats(int sip_type, MLC_Stat* mlc_tot, int num_of_frame);
#endif //INTRA_MODE_STATS
};

#endif // __MLC_Stat_H__
