/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEst.H
*
* \brief Definition of MotionEst class
*
*/

#ifndef __MotionEst_H__
#define __MotionEst_H__

#include "MotionField.H"
#include "Frame.H"
#include "WFrame.H"
#include "RefFrameManager.H"
#include "EntropyEncoder.H"
#include "IEncoder6.H"

//-----------------------------------------------------------------------------
/*!
* \brief Base motion estimator class
*
*/
//-----------------------------------------------------------------------------
class MotionEst 
{
  int c_mv_zero[16]; //!< \brief Cost for indicating which motion data components are nonzero in a single-list block;
  //!< 16 combinations arise from the fact that the motion data has 4 components.
  int c_mv_zero_bi[128]; //!< \brief Cost for indicating which motion data components are nonzero in a bipred block;
  //!< 128 combinations arise from the fact that the motion data has 7 components.
  int *c_mv_level[2]; //!< Array of costs for coding displacement differences (vertical and horizontal)
  int *c_mv_lvl[2]; //!< Pointer to allocated memory for c_mv_level
  
  int c_listidx[4][4][4]; //!< Cost for coding a list index (mode)
  int c_listidx_P[2][2][2]; //!< Cost for coding a list index for P-picture(mode)

  int c_siptype[NSIP][NCTX_SIP_TYPE]; //!< Index of spatial intra prediction type(8x8,2x8,8x2)
  
  int c_mvshare[5][5][3][7]; //!< Cost for coding a partition size for motion sharing

  int c_refidx[LISTS][MAX_REF_FRAMES][MAX_REF_FRAMES][MAX_REF_FRAMES]; //!< Cost for coding a reference index
  
  int *c_scl[NCTX_SCALE]; //!< Pointer to allocated memory for c_scale
  int *c_ofs[2]; //!< Pointer to allocated memory for c_offset
  int *c_scale[NCTX_SCALE]; //!< Cost for coding a scale factor (intensity compensation)
  int *c_offset[2]; //!< Cost for coding an offset (intensity compensation)
  
  int c_pmode[2]; //!< Estimated bit for intra block mode
  int c_rmode[5][2]; //!< Estimated bit for intra block mode
  int c_cmode[3][2]; //!< Estimated bit for intra block mode
  int c_cmode_1[2]; //!< Estimated bit for intra block mode
  int c_cmode_2[2]; //!< Estimated bit for intra block mode
  int mel; //!< Lambda parameter used for motion estimation
  
  int mult_mel(int x);
  int mult_mel_double(int x);
  
  PredBlk pleft; //!< Motion data of block to left of current block
  PredBlk ptop; //!< Motion data of block above current block
  PredBlk ptleft; //!< Motion data of block to upper left of current block

  public:

#ifndef MVCOMP_ZERORES
  int c_mvp[2][2];  //!< Cost for coding motion data prediction direction
#else
  int c_mvp[3][2];  //!< Cost for coding motion data prediction direction
  int mv_temporal_thresh; // |T-x| threshold in effect for tossing temporal.
                          // Negative implies toss close values.  Positive far values.
                          // 0 no thresh.
#endif
  MotionField *mf;  //!< Motion field
  int nloops; //!< Number of loops to perform in estimation
  RefFrameManager *ref_mgr; //!< Pointer to reference frame manager
  int mel_shift; //!< Number of bits luma SAD is shifted for R-D cost computation
  int chroma_weight; //!< Factor by which chroma SAD is shifted for R-D cost computation
  int me_range; //!< Motion search range (fractional pel units)
  
  int intra_cost_update;

  MotionEst(MotionField *mf, RefFrameManager *ref_mgr);  
  virtual ~MotionEst();
  
  virtual void reset(Frame *cur);
  void reset_cost(EntropyEncoder *eenc);
  
  int mv_cost_d(int d0, int d1, int d_scale, int d_offset);
  int mv_cost_d_bipred(int d0_0, int d0_1, int d1_0, int d1_1, int d_scale, int d_soffset, int d_offset);

#ifndef MVCOMP_ZERORES
  int bit_cost(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft);
#else
  int bit_cost(PredBlk *cur, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft);
#endif
   
  int bit_cost_intra(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft);
#ifndef MVCOMP_ZERORES
  int bit_cost_inter(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft); 
#else
  int bit_cost_inter(PredBlk *cur, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft);
#endif
    
  int bit_cost_inter1(PredBlk *cur, PredBlk *left);
#ifndef MVCOMP_ZERORES
  int bit_cost_inter2(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft);
#else
  int bit_cost_inter2(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft, int extra_pred);
#endif

  int bit_cost_refindex(PredBlk *cur, PredBlk *left, PredBlk *top, int list);
  int bit_cost_listindex(PredBlk *cur, PredBlk *left, PredBlk *top);
  
  int bit_cost_mvshare(PredBlk *cur, PredBlk *top, PredBlk *left);

  int bit_cost_listindex_P(PredBlk *cur, PredBlk *left, PredBlk *top);
  
  virtual void estimate(Frame *cur, int list) = 0;
  void intra_init(IEncoder6 *ienc, EntropyEncoder *eenc);
      
};

#endif // __MotionEst_H__
