/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEstRegion.H
*
* \brief Definition of MotionEstRegion class
*
*/


#ifndef __MotionEstRegion_H__
#define __MotionEstRegion_H__

#include "MotionEst.H"
#include "MCRefBlk8.H"

//-----------------------------------------------------------------------------
/*!
* \brief Region motion estimator class
*
* This estimator optimizes a motion field by operating on contiguous blocks
* that share the same motion data.
*/
//-----------------------------------------------------------------------------
class MotionEstRegion : public MotionEst 
{
  int *region; //!< Array of region indices
  int nregions; //!< Number of regions
  PredBlk *cand; //!< Candidate list
  PredBlk *cand_sl; //!< Candidate short list
  int ncand; //!< Number of candidates
  int ncand_sl; //!< Number of candidates in short list
  int max_ncand; //!< Maximum number of candidates
  int bbox_top; //!< Top row index of bounding box 
  int bbox_bottom; //!< Bottom row index of bounding box
  int bbox_left; //!< Left column index of bounding box
  int bbox_right; //!< Right column index of bounding box
  
  int get_region(int i, int j);
  int test_and_fill(PredBlk *blk, int i, int j);
  void fill(PredBlk *blk,int *same_region, int *nstored);
  double cost(Frame *cur, int r, double mincost);
  void insert_cand(PredBlk *blk);
  void insert_cand_vs_slist(PredBlk *blk);
  int insert_cand_short_list(PredBlk *blk);
  void insert_cands(PredBlk *blk, int extend);
  void find_region_bbox(int r);

  int total_cost_mvshare(int i, int j, PredBlk *curblk, MotionField *mf, MCRefBlk8 *blk, int Bframe);
  
  public:
  
  MotionEstRegion(MotionField *mf, RefFrameManager *ref_mgr);
  ~MotionEstRegion();
  
  void estimate(Frame *cur, int list);
};

#endif // __MotionEstRegion_H__
