/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEstTrellis.H
*
* \brief Definition of MotionEstTrellis class
*
*/

#ifndef __MotionEstTrellis_H__
#define __MotionEstTrellis_H__

#include "MotionEst.H"
#include "Global.H"
#include "MCRefBlk8.H"

typedef int candarray[MAXNCAND]; //!< Pointers used for backtracking

//-----------------------------------------------------------------------------
/*!
* \brief Trellis motion estimator class
*
* This estimator optimizes a motion field by operating on rows and columns
* of 8x8 blocks.
*/
//-----------------------------------------------------------------------------
class MotionEstTrellis : public MotionEst 
{
  PredBlk **mvcand; //!< Array of candidates
  candarray *backtrack; //!< Pointers used for backtracking
  int *ncand; //!< Number of candidates
  int size; //!< Max of height and width of frame (in 8x8 block units)
  
  void reorder(int z, int *cost);
  int insert_default_cands(PredBlk *mvcand, int i, int j, int list);
  void insert_neighbour_cands(PredBlk *mvcand, int *n, int i, int j, int di, int dj, int remaining);
  
  void insert_cand(PredBlk *mvcand, int *n, PredBlk *cand);
  int treillisH(Frame *cur, int first, int step, int list);
  int treillisV(Frame *cur, int first, int step, int list);
  
  int total_cost_mvshare(int type, int i, int j, PredBlk *curblk, PredBlk *top_left_blk, MotionField *mf, MCRefBlk8 *blk, int Bframe);

  public:
  
  MotionEstTrellis(MotionField *mf, RefFrameManager *ref_mgr);

  ~MotionEstTrellis();
  
  void estimate(Frame *cur, int list);
};

#endif // __MotionEstTrellis_H__
