/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEstUZSearchBiPred.H
*
* \brief Definition of MotionEstUZSearchBiPred class, 
*  and MotionEstUZSearchBiPred::SPoint and MotionEstUZSearchBiPred::FastMEPattern structs
*
*/


#ifndef __MotionEstUZSearchBiPred_H__
#define __MotionEstUZSearchBiPred_H__

#include "MotionEst.H"
#include "MCRefBlk8.H"

#define MAXMEBLK 16

//-----------------------------------------------------------------------------
/*!
* \brief Zonal search motion estimator class
*
* This estimator should be used for the initial ME, which may later be refined
* using other estimators.
*/
//-----------------------------------------------------------------------------
class MotionEstUZSearchBiPred : public MotionEst 
{
  PredBlk *row; //!< One row storage of mvs.
  
  //-----------------------------------------------------------------------------
  /*!
  * \brief Search point
  *
  */
  //-----------------------------------------------------------------------------
  typedef struct
  {
    short x; //!< Vertical delta
    short y; //!< Horizontal delta
    short start_nmbr; //!< Start number
    short next_points; //!< Number of points in next iteration
  }
  SPoint;
  
  //-----------------------------------------------------------------------------
  /*!
  * \brief Search pattern
  *
  * \note
  * This should be changed into 2 dimensional so it may contain all possible structures 
  * and would be easier to use combining methods.
  */
  //-----------------------------------------------------------------------------
  typedef struct MEPatternNode
  {
    int    searchPoints; //!< Number of search points
    SPoint point[128]; //!< Array of search points
    int    stopSearch; //!< Flag indicating to stop search after current iteration
    int    nextLast; //!< Flag indicating that next pattern is last
    struct MEPatternNode *nextpattern; //!< Pointer to pattern to use in next iteration
  }
  FastMEPattern;
  
  FastMEPattern searchPattern; //!< Search pattern
  FastMEPattern searchPatternD; //!< Dual search pattern 
  FastMEPattern predictor; //!< List of predictors \note Name was changed from predictor to avoid aliasing 
  FastMEPattern fixed_predictor; //!< Set of fixed predictors describing logarithmic search
  
  FastMEPattern sdiamond; //!< Small diamond search pattern
  FastMEPattern square; //!< Square search pattern 
  FastMEPattern ediamond; //!< Extended diamond search pattern
  FastMEPattern ldiamond;  //!< Large diamond search pattern
  FastMEPattern sbdiamond; //!< Half-pel accurary diamond search pattern
  
  int medianthres; //!< Threshold for early termination
    
  int uZonalSearchBlockIndex; //!< Time stamp
   
  void assign_pattern(FastMEPattern *pattern, int data[][4], int searchPoints, int stopSearch, int nextLast, FastMEPattern *nextpattern);
  void set_pattern(FastMEPattern *pattern, int shape);
  void UnifiedZonalSearchInit();
  void UnifiedZonalSearchFixedPredictorInit( int search_range, FastMEPattern *predictor);
  void UnifiedZonalSearchFixedPredictors   ( PredBlk medianMV, FastMEPattern *predictor, int *prednum, int list);
    
  
  int spatial_predictors(int ref, PredBlk  *block_a,PredBlk  *block_b, PredBlk  *block_c, PredBlk  *block_d, FastMEPattern *predictor, int list);
  void row_predictors(int ref, int i,int j, int *pred, FastMEPattern *predictor, int list);
  
  int mvset_size; //!< Number of motion vectors in search range
  int *motionMap; //!< Ordered indices
  int motionMapWidth;
  int motionMapMinValue;
  int motionMapMaxValue;

  bool isInBounds(int val[2]);
  bool testAndMark(int val[2]);
  void forceIntegerPel(PredBlk *pblk);

  int total_cost_mvshare(int i, int j, int sadin, PredBlk *curblk, MotionField *mf, MCRefBlk8 *blk, int Bframe);

  PredBlk estimate(int ref, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft, PredBlk *topright, MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], int *bs, int list, int u, int v, int n);
  PredBlk estimate_one_list(int list, PredBlk *block_a, PredBlk *block_b, PredBlk *block_c, PredBlk *block_d, int i, int j, MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], int u, int v, int n);

  void add_to_list(int val, int *list, int *list_size);
  void search_subpel(int list, PredBlk *imv, PredBlk *cand, PredBlk *block_a, PredBlk *block_b, PredBlk *block_c, MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], int *bsad, int k_range, int l_range, int i, int j, int u, int v, int n);

public:
  
  MotionEstUZSearchBiPred(MotionField *mf, RefFrameManager *ref_mgr);
  ~MotionEstUZSearchBiPred();
  
  virtual void reset(Frame *cur);

  void estimate(Frame *cur, int list);

  static MotionField *prevMf;
  static void initPrevMf(int height, int width);
  static void freePrevMf();

  void setMv(PredBlk *pblk, int i, int j, int u, int v, int n);
  int bitcostN(PredBlk *pblk, int i, int j, int u, int v, int n);
  int sadN(MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], PredBlk *pblk, int i, int j, int u, int v, int n);
  int estimateR(int i, int j, MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], int u, int v, int n, int list, Frame *cur);
};

#endif // __MotionEstUZSearchBiPred_H__
