/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionField.H
*
* \brief Definition of MotionField class 
*
*/

#ifndef __MotionField_H__
#define __MotionField_H__

#include "Frame.H"
#include "Global.H"
#include "PredBlk.H"
#include "RefFrameManager.H"
#include "WFrame.H"
#include "ContextMap.H"

//-----------------------------------------------------------------------------
/*!
* \brief Motion field class
*
* Contains motion field information
*/
//-----------------------------------------------------------------------------
class MotionField 
{
  PredBlk *blk_array;
  
  unsigned char *get_luma_ptr(RefFrameManager *ref_mgr, PredBlk *blk, int list, int i, int j, unsigned char *data = NULL);
  unsigned char *get_chroma_ptr(RefFrameManager *ref_mgr, PredBlk *blk, int list, int i, int j, unsigned char *data = NULL);

  void predict_blk_part_mvshare(short luma[64], short chroma[2][16], RefFrameManager *ref_mgr, int ii_t, int jj_t, int ii_b, int jj_b);

public:
  int width; //!< Width of motion field in 8x8 block units
  int height; //!< Height of motion field in 8x8 block units
  int frame_type; //!< %Frame type (ft_I_frame, ft_P_frame, ft_B_frame)
  PredBlkI *blk_array_I; //!< Array of intra-related info for each 8x8 block
    
  MotionField(int h, int w);
  ~MotionField();

#ifdef MVCOMP_ZERORES
  void copy_field_from(MotionField *mf_src);
#endif

  void set_all_intra();
  
  void reset_motion_field();
  void change_resolution_motion_field(RefFrameManager *ref_mgr, int frame_type);
  
  PredBlk *get_pblk_4x4(int i2, int j2);
  PredBlk *get_pblk(int i, int j);
  PredBlk *get_pblk_unsafe(int i, int j);
    
  void predict(WFrame *cur, RefFrameManager *ref_mgr);
  void predict_blk(WFrame *cur, RefFrameManager *ref_mgr, int i, int j); 
  
  int mvshare_type(int i, int j);
  int mvshare_type_simple(PredBlk *cur, PredBlk *top, PredBlk *left);
  int get_mvshare_size_ctx(int mvshare_type, PredBlk *top, PredBlk *left);

  void predict_blk_mvshare(WFrame *pred, RefFrameManager *ref_mgr, int i, int j);

  void set_block_intra(int i, int j);

  void fix_lists();

  int get_transform_blocksize(int i, int j);
  void print_info(FILE *f, int frame_no);
  void write_stat_files(Frame *cur);
};


#endif // __MotionField_H__
