/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Param.H
*
* \brief Definition of Param class
*
*/


#ifndef __Param_H__
#define __Param_H__

#include "Buffer.H"
#include "ZeroTree.H"
#include "Global.H"

#ifdef MVCOMP_ZERORES
//-----------------------------------------------------------------------------
/*!
* \brief Predictor parameter info.
*
*/
//-----------------------------------------------------------------------------
class MVCompConfig
{
public:
  MVCompConfig();
  ~MVCompConfig();
  void print(int i);

  int mv_zero_residue; //!< try a zero residue in block_mode_decision?
  char mv_predictor_strings[MV_MAX_NPREDICTORS][MAX_LINE_LEN]; //!< config file strings.
  int mv_npredictors;
  int mv_predictors[MV_MAX_NPREDICTORS]; //!< what set of predictors are being used?
  int mv_predictor_used[MV_MAX_PRED_INDEX+1]; //!< index with predictor to determine if used.
  int mv_temporal_proba; //!< fixed proba for signalling T. 655 = 1%.
  int mv_keep_learned_movement; //!< avoid reset on movement statistics in encoder?
  int mv_max_temporal_diff; //!< if we're looking at x&T as diff preds, we ignore T if |T-x| > diff (diff +ve)
                            //!< or |T-x| < -diff (diff -ve) to reduce signalling. 0 => no special treatment.
  int mv_auto_max_temporal_diff; //!< the above value is generated automatically from proba(T).
};
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Global parameter class
*
*/
//-----------------------------------------------------------------------------
class Param
{
public:
  // Extension parameters
  int use_extension;
  int bugfix87;
  int bugfix88;
  int bugfix89;
  int bugfix90;
  int alt8x8_trans_idx;
  int bugfix93;
  int ndf_lc_luma;
  int ndf_lc_chroma;
#ifdef I16_IN_P_FRAME
  int intra16x16p;
#endif
  
  enum
  {
    tag_end = 0,
    tag_bugfix87 = 1,
    tag_bugfix88 = 2,
    tag_bugfix89 = 3,
    tag_alt8x8_trans_idx = 4,
    tag_ndf_lc = 5,
    // 6 not valid
    tag_edge_pred_bug_fix = 7,
#ifdef I16_IN_P_FRAME
    tag_intra16x16p = 8,
#endif
    tag_bugfix93 = 9
  };
  
  /*input*/
  int lambdaq_base; //!< Quantized lambda parameter
  int melq_base; //!< Quantized lambda parameter for ME
  int lambda_mode_chroma; //!< Use chroma-qp-adjust lambda for block mode decision

  int qp; //!< Quantization parameter
  int qpC; //!< Quantization parameter for chroma
  int q_rdoffset; //!< Enables Rate distortion optimized quantization
  int q_rdoffset8x8; //!< Enables Rate distortion optimized quantization for 8x8 block
  int nframes; //!< Number of frames to encode
  int f_header; //!< Size of header in source image file
  int frame_offset; //!< Number of frames to skip at begining of sequence
  int frame_height; //!< %Frame height in luma samples
  int frame_width; //!< %Frame width in luma samples
  int use_chroma_me; //!< Flag to use chroma in ME
  int chroma_on_fly; //!< Flag to use chroma interpolation on the fly
  int pyramid_coding; //!< Pyramid coding mode
  int no_reordering; //!< Flag to disable reordering in pyramid coding to create Hierarchical P.
  int me_range; // motion estimation range
  int me_loops1; //!< Number of iterations for refine ME
  int me_loops2; //!< Number of iterations for trellis ME
  int me_loops3; //!< Number of iterations for region ME
  int me_repeat_edge; //!< Number of iterations for additonal trellis and region ME
  int me_repeat_intref_p; //!< Number of iterations for additonal trellis and region ME
  int me_repeat_intref_b; //!< Number of iterations for additonal trellis and region ME
  int me_repeat_disposable_p; //!< Number of iterations for additonal trellis and region ME
  int me_repeat_disposable_b; //!< Number of iterations for additonal trellis and region ME
  int me_loops2_edge; //!< Number of iterations for trellis ME
  int me_loops3_edge; //!< Number of iterations for region ME
  int me_loops2_intref; //!< Number of iterations for trellis ME
  int me_loops3_intref; //!< Number of iterations for region ME
  int me_loops2_disposable; //!< Number of iterations for trellis ME
  int me_loops3_disposable; //!< Number of iterations for region ME
  int open_loop; //!< Flag to use open-loop coding (use original frames instead of reconstructed ones for MC)
  int better_color; //!< Adjustment for chroma QP
#ifdef BETTER_CHROMA_DC
  int chroma_dc_scale; //!< Adjustment for chroma QP DC. controls the rate of reduction of the chroma DC qp
  int chroma_dc_th; //!< Adjustment for chroma QP DC. controls the point where you start reducing the chroma DC qp
#endif
  int bc_offset; //!< Defines the offset between the chroma QP value and the intermediate chroma QP values.
  int q_adaptoffset; //!< Offset adaptation mode
  int q_offset[6]; // offset used for rounding during quantization
  int loss_rate; //!< Adaptation rate of arithmetic coder
  int ac_exp_limit; // !< Exponential limit of arithmetic coder
  int lime_size;
  int luma_mc; //!< Luma MC interpolation mode
  int chroma_mc; //!< Chroma MC interpolation mode
  int treillis_ncand; //!< Max number of candidates for trellis ME
  int region_ncand; //!< Max number of candidates for region ME
  int mc_round; //!< Use alternate rounding for chroma MC
  int onlyI; //!< Flag to code all frames as I
  int idr_interval; //!< interval of Decoder Refresh frames, in terms of I frames. 2 means every other I frame is IDR or DDR.
  int enableB; //!< Flag to enable B frames
  int adapt_frame_type; //!< Flag to enable adaptation of frame type (P/B).
  int lumaDC; //!< QP adaptation for DC
  int candidate_search_region; //!< defines the candidate search region for Intra TMA
  int candidate_search_region_intra16x16; //!< defines the candidate search region for Intra TMA
  int num_candidate_averaged;  //!< number of candidates used in the MultiHypothesis
  int candidate_sad_threshold;   //!< defines the threshold for candidate selection
    
  int mvs_realcost; //!< Flag to use real cost for selecting best MV predictor when coding
  int ndf_mode; //!< NDF mode index
  int ndf_trsf_8x8;

  int deblockingfilter;
  int dbf_offset_a;
  int dbf_offset_b;

#ifdef WIENER_FILTER_INLOOP_SWITCH
  int wienerfilter_inloop;
#endif  
  int wienerfilter;
  int wlf_enable[3];
  int fs_max_rec;
  int fs_max_qpe;
  int fs_max_pred;
  
  // motion sharing
  int motion_sharing_enable; //!< Flag to enable motion sharing
  int mvshare_two_enable;  //!< Flag to enable motion sharing of 2
  int mvshare_two_enable_me;  //!< Flag to enable motion sharing of 2 in ME loop
  int motion_sharing_me_init_offset; //!< Flag to enable the search of mvshare size with IC parameter in initial ME
  int motion_sharing_me_refine;  //!< Flag to enable the basic search of mvshare size in refine ME
  int motion_sharing_me_refine_full;  //!< Flag to enable the full search of mvshare size with MV in refine ME
  int motion_sharing_me_trellis;  //!< Flag to enable the basic search of mvshare size with MV in trellis ME
  int motion_sharing_me_trellis_full;  //!< Flag to enable the full search of mvshare size with MV in trellis ME
  int motion_sharing_me_region;  //!< Flag to enable the search of mvshare size with MV in Region ME
  int motion_sharing_log;  //!< Flag to print log of block modes with motion sharing

  double learn_offset; //!< Penalty for coding a symbol
  int trellis_mode; //!< Trellis optimization order
  int bugfix80;
  int bugfix94; //!< bug fix in EntropyEncoder::cost_mode
  int frame_skip; //!< Number of frames to skip in source file between two coded frames
  int intermed_frame; //!< Flag to enable intermediate frame (frame reordering)
  int idr_intermed_frame; //!< Flag to enable IDR intermediate frame (frame reordering)
  int intermed_disp; //!< Flag to enable disposable frames

  int mpm_default; //!< Default value for most probable mode
  int bugfix92; //!< Enable inclusion of rate for sip_type in R-D cost

  int edge_prediction_ipd; //!< Flag to enable the high precision edge prediction for intra
  int threshold_edge_vector;  //!< Threshold for the decision between edge prediction and DC mode
  
  int fastME; //!< Search pattern index for fast ME
  int fastMEDual; //!< Dual search pattern index for fast ME
  int fastMEFixed; //!< Flag to use fixed predictors in fast ME
  
  // End of addition
  int nref_frames; //!< Number of reference frames
  
  // intensity compensation related fields
  int enable_intcomp; //!< Flag to enable intensity compensation
  int DC_mult; //!< Multiplier applied to DC offsets in intensity compensation
  int weight_diff_scale; //!< Shift value applied for weight difference (bipred ic)
  int weight_diff_max;   //!< Maximum absolute value of weight difference (bipred ic)

  int me_refine_s0; //!< Local search range used in MotionEstRefine
  int me_refine_s1; //!< Local search range used in MotionEstRefine
  int me_refine_s2; //!< Local search range used in MotionEstRefine
  int me_trellis_s0; //!< Local search range used in MotionEstTrellis
  int me_trellis_s1; //!< Local search range used in MotionEstTrellis
  int me_trellis_s2; //!< Local search range used in MotionEstTrellis
  int me_trellis_o0; //!< Local search range used in MotionEstTrellis
  int me_trellis_o1; //!< Local search range used in MotionEstTrellis
  int me_region_s0; //!< Local search range used in MotionEstRegion
  int me_region_s1; //!< Local search range used in MotionEstRegion
  int me_region_s2; //!< Local search range used in MotionEstRegion
  int me_region_o0; //!< Local search range used in MotionEstRegion
  int me_region_o1; //!< Local search range used in MotionEstRegion
  
  int trans8x8_enable;  //!< Flag to enable use of 8x8 transform for 8x8 MC blocks.
  
  int ndf_two_rounds; //!< Flag to enable two rounds in NDF (overcomplete)
  int ndf_weighting; //!< Flag to enable weighting in NDF (overcomplete)
  int ndf_signif_cnt; //!< Threshold: number of significant coefficients (overcomplete)
  int ndf_4x4_mult; //!< multiplier for number significant coefficients in 4x4 block (overcomplete)
  int ndf_aggressive; //!< Sets all filter mask extent values to2 (when ndf_mode is equal to 3)
  int ndf_ic_refine; //!< Enables ic specific modes (when ndf_mode is equal to 3)
  int ndf_mvd_th; //!< Threshold: motion vector difference (overcomplete)
  int ndf_s_pred_quant; //!< Strength of filter in pred_quant mode (overcomplete)
  int ndf_s_pred_signif; //!< Strength of filter in pred_signif mode (overcomplete)
  int ndf_s_pred_mot; //!< Strength of filter in pred_mot mode (overcomplete)
  int ndf_s_intra_quant; //!< Strength of filter in intra_quant mode (overcomplete)
  int ndf_s_mot_disc; //!< Strength of filter in mot_disc mode (overcomplete)
  int ndf_s_ic_strong; //!< Strength of filter for ic_strong mode (when ndf_mode is equal to 3)
  int ndf_s_ic_intermed; //!< Strength of filter for ic_intermed mode (when ndf_mode is equal to 3)
  int ndf_s_ic_weak; //!< Strength of filter for ic_weak mode (when ndf_mode is equal to 3)
  int ndf_s_chroma; //!< Strenght of filter for chroma (overcomplete)
  int ndf_x_pred_quant; //!< Extent of filter mask for intra_quant mode
  int ndf_x_pred_signif; //!< Extent of filter mask for pred_signif mode
  int ndf_x_pred_mot; //!< Extent of filter mask for pred_mot mode
  int ndf_x_intra_quant; //!< Extent of filter mask for pred_quant mode
  int ndf_x_mot_disc; //!< Strength of filter for mot_disc mode (when ndf_mode is equal to 3)
  int ndf_x_ic_strong; //!< Extent of filter mask for ic_strong mode (when ndf_mode is equal to 3)
  int ndf_x_ic_intermed; //!< Extent of filter mask for ic_intermed mode
  int ndf_x_ic_weak; //!< Extent of filter mask for ic_weak mode
  
  int mel_offset; //!< Adjustement to ME lambda value for initial ME
  int adapt_list_size; //!< Flag to limit list sizes to a single temporal direction
  int lambdaq_offset; //!< Lambda value offset for disposable frames
  int qp_off_intdisp; //!< QP offset for intermediate disposable frames
  int qp_off_intra; //!< QP offset for the intra frames in the P or B coded frames
  int qp_off_onlyI; //!< QP offset for the intra frames in the only I coded frames
  int qp_off_intra_scene_cut; //!< QP offset for the intra frames at scene cut
  int qp_off_intra_disposable_pic; //!< QP offset for the intra frames in the disposable coded frames
  int qp_off_intref; //!< QP offset for intermediate reference frames
  int qp_frame_mod; //!< Flag for enabling QP modulation on frame basis
  int me_subpel; //!< Indicates which subpel position is searched in the me process
  int aif; //!< Separable adaptive interpolation filter
  int P_subpel;  //!< Subpel precision(s) to use for P pictures
  int B_subpel;  //!< Subpel precision(s) to use for B pictures
  int gen8pel_filter; //!< Type of 1/8 pel filter to use
  
  int num_ctx_read; //!< Number of reading contexts
  // 8x8 transform stuff
  int hadamard; //!< Selects method of hadamard transform mode
  int h0_mult; //!< Selects the scaling for hadamard transform mode 0, multiplied by h0_mult and divided by 2^HADAMARD_SHIFT_0
#ifdef USE_SSD
  int ssd_mult; //!< Selects the scaling for hadamard transform mode 0, multiplied by ssd_mult and divided by 2^SSD_SHIFT_0
  int ssd_qp_P;  //!< value of the qp above which the SAD for P pictures uses the SSD calculation
  int ssd_qp_B;  //!< value of the qp above which the SAD for B pictures uses the SSD calculation
  int use_ssd;  //!< internal flag for using SSD instead of SAD, set in update_rd_params
  int ssd_disp_only;  //!< apply SSD to disposable pictures only
#endif
  int copy_offset; //!< Index of copy offset
  
  // zerotree stuff
  int ztree_pattern_64[5*ZTsize64];  //!< Pattern of zero tree for 8x8
  ZeroTree *ztree_64; //!< Zero tree for 8x8
  
  int max_quantization_set_num_minus_one; // cq_start
  int quantization_set_id;
  int already_update[2];
  //! qmatrix/qoffsetmatrix variables, [0]: qmatrix; [1]: qoffsetmatrix
  int qmatrix_enable_flag[2];                             //!< Flag to enable customizable quantization
  int qmatrix_overwrite_flag[2][MAX_LINE_LEN];            //!< Flag to overwrite common matrix
  char qmatrixfile[2][MAX_LINE_LEN];                      //!< Name of file containing matrix
  int qmatrix_tmp[2][MAX_QUANT_SET_NUM][16][260];         //!< Matrix read from cfg file
  int qmatrix[2][MAX_QUANT_SET_NUM][16][260];             //!< Matrix for quantization
  int qmatrix_check[2][MAX_QUANT_SET_NUM][16];            //!< Flag to indicate if the matrix is sent in configuration file
  int qmatrix_update_check[2][MAX_QUANT_SET_NUM][16];     //!< Check if scale, slope values for a matrix exist
  int UseDefaultQMatrixFlag[2][16];  // cq_end          //!< Flag to indicate if a defalut matrix is selected 

  int intra16x16;
  int intra16x16_org;
  int transform_selection_16x16;
  int chroma8x8;
  ZeroTree *ztree_64_CT8; //!< Zero tree for chroma 8x8

  int sb_height; //!< Superblock height (in 8 sample units)
  int sb_width;  //!< Superblock width  (in 8 sample units)
  //SUBSTREAM_MT zt64 and zt_idx stats have been moved to the EntropyDecoder.
  
  int ztree_pattern[NZTREES][5*17]; //!< Zero tree definitions read from file
  ZeroTree *ztree[NZTREES]; //!< Zero trees
  
  /*output*/
  int bitsM[2][65536]; //!< Array containing -(1<<RATE_BITS)*log_2(i/65536)
  float bitsF[65536]; //!< Array containing -log_2(i/65536)
  
  char srcfile[MAX_LINE_LEN]; //!< Filename of source sequence
  char dstfile[MAX_LINE_LEN]; //!< Filename of coded sequence
  char recfile[MAX_LINE_LEN]; //!< Filename of reconstructed sequence
  char sadfile[MAX_LINE_LEN]; //!< Name of file containing SAD data
  char reffile[MAX_LINE_LEN]; //!< Name of file containing reference sequence (for open loop encoding)
  char prob_file[MAX_LINE_LEN]; //!< Name of file containing probability data
  char ztreefile[MAX_LINE_LEN]; //!< Name of file containing zero tree data
  char ztreefile_8x8[MAX_LINE_LEN]; //!< Name of file containing zero tree data for 8x8
  
  char qtabfile_4x4[MAX_LINE_LEN]; //!< quantization table for 4x4
  char qtabfile_8x8[MAX_LINE_LEN]; //!< quantization table for 8x8
  char qtabfile_2x8[MAX_LINE_LEN]; //!< quantization table for 2x8
  char qtabfile_8x2[MAX_LINE_LEN]; //!< quantization table for 8x2
  
  char parfile[MAX_LINE_LEN]; //!< Name of file containing QP and lambda parameters for individual frames
  char logfile[MAX_LINE_LEN]; //!< Filename of log
  char sadlogfile[MAX_LINE_LEN]; //!< Filename of SAD log
  char expPrmdFrmt[MAX_LINE_LEN]; //!< String containing explicit pyramid parameters
  char expPrmdFrmtIdr[MAX_LINE_LEN]; //!< String containing explicit pyramid parameters
  int deferred_decoder_refresh; //!< flag to indicate a deferred decode refresh is pending.
  
#ifdef USE_SSD
  int ssd_lookup[256];
#endif

  char mode_stats_fname[MAX_LINE_LEN]; //!< Name of file containing mode statistics (decoder)
  char mv_stats_fname[MAX_LINE_LEN]; //!< Name of file containing motion statistics (decoder)
  FILE *mode_stats_file; //!< Pointer to file containing mode statistics (decoder)
  FILE *mv_stats_file; //!< Pointer to file containing motion statistics (decoder) 

#ifdef Q_OFFSET
  int iTheta[2];
  int iTheta8x8[2];
#endif
#ifdef MVCOMP_ZERORES
  MVCompConfig mv_comp_config[2]; // Two possible configs.
  int mv_default_comp_config_idx; // For now, the competition config is chosen from the config file.
  int mv_cur_comp_config_idx;
  MVCompConfig *mv_cur_comp_config() {return &mv_comp_config[mv_cur_comp_config_idx];}
  static char const *mv_pred_name[MV_MAX_PRED_INDEX+2]; //!< predictor names.  terminated with NULL.
  int get_mv_predictor(char const *str);
#endif
#ifdef CABAC_SYNC_POINTS
  int  superblock_synchro; //!< encode using super-block synchronisation points?
  int  superblock_nsubstreams; //!< number of substreams.
  int  decode_nthreads; //!< number of threads we're using for decoding.
  int  decode_inputref; //!< are we decoding a reference stream (no synchro/thread indicators)?
#endif
#ifdef DEC_TRANSCODER
  int  decode_transcode; //!< are we transcoding?
  int  decode_transcode_nthreads; //!< how many output threads do we want if we're transcoding?
#endif

  int  intra8x8;
  int  transform_selection;
  Param();
  ~Param();
  
  void update();
#ifdef FRAME_SUBSTREAMS
  void init_ztree(int encoder);
#else
  void init_ztree();
#endif  

  int write(Buffer *buf);
  void read(Buffer *buf);
  
  int read_config_file(char *filename);
  int read_prob(FILE *f, char const *name, int qp);
  void read_prob_file();
  
  EntropyModel *initialProbEntropyModel; //!< stores initial proba settings.

  char* GetConfigFileContent (char *Filename); // cq_start
  void matrix_scan(int *qmatrixList, int quantization_list_id, int matrix_size);
  void matrix_inv_scan(int *qmatrixList, int quantization_list_id, int matrix_size);
  void ini_qmatrix(int matrix_type);
  void dupli_qmatrix0(int matrix_type);
  int read_qmatrix_file(char* filename, int matrix_type);
  int get_array_id(int type, int size, int yuv);
  int get_value_arrage(int value, int size);
  int strindex(char *str,const char *substr);
  void check_qmatrix_para (char *s, int *set_id, int *array_id, int *array_off, int *range);
  void parse_qmatrix (char *buf, int bufsize, int matrix_type);
  void patch_qmatrix(int quantization_set_id, int matrix_type);
  void qmatrix_generate(int *qmatrixList_tmp, int *qmatrixList, int sizeOfQMatrixList, int matrix_type);
  void update_qmatrix(int quantization_set_id, int matrix_type);
  int write_qmatrix_list(Buffer *buf, int *qmatrixList, int sizeOfQMatrixList, int *useDefaultQMatrixFlag);
  int write_qmatrix_update_value(Buffer *buf, int *qmatrixList);
  int write_qmatrix_set(Buffer *buf, int quantization_set_id, int matrix_type);
  int write_qmatrix_set_all(Buffer *buf, int matrix_type);
  void read_qmatrix_list(Buffer *buf, int *qmatrixList, int sizeOfQMatrixList, int *useDefaultQMatrixFlag);
  void read_qmatrix_update_value(Buffer *buf, int *qmatrixList);
  void read_qmatrix_set(Buffer *buf, int quantization_set_id, int matrix_type);
  void read_qmatrix_set_all(Buffer *buf, int matrix_type); // cq_end

#ifdef DENOISE_MT
  int  decode_ndenoisethreads; //!< number of threads we're using for denoising.
#endif
#ifdef WIENER_MT
  int  decode_nwfthreads; //!< number of threads we're using for wiener filter.
#endif
#ifdef DEBLOCK_MT
  int  decode_ndeblockthreads; //!< number of threads we're using for deblocking.
#endif
  
  int enable_simple_lambdas;
  int sl_trans_sel;
  double sl_weight16x16;
  double sl_rate_offset;
  double sl_lambda_I_luma;
  double sl_lambda_I_luma_rdoq;
  double sl_lambda_I_chroma;
  double sl_lambda_I_chroma_rdoq;
  double sl_lambda_P_luma;
  double sl_lambda_P_luma_rdoq;
  double sl_lambda_P_chroma;
  double sl_lambda_P_chroma_rdoq;
};

extern Param param;

#endif
