/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ProbEst.H
* \brief Definition of ProbEst class
*/
//-----------------------------------------------------------------------------
#ifndef __ProbEst_H__
#define __ProbEst_H__

#include "Global.H"

// Bits assigned to classify probabilities for reading, writing
// and printing.
#define CTXT_CLASS_ALL               0x10000000
#define CTXT_CLASS_8X8               0x20000000 // we distinguish 8x8 contexts for stats writing, for now.
#define CTXT_CLASS_DUMMY             0x40000000
#define CTXT_CLASS_NORW              0x80000000 // Initialized 'specially' -- not read from or written to external files.

#define CTXT_CLASS_PBCS_MASK         0x00000007 // mask for all print-by-category sumX values.
#define CTXT_CLASS_PBCS_SUMA         0x00000001
#define CTXT_CLASS_PBCS_SUMB         0x00000002
#define CTXT_CLASS_PBCS_SUMC         0x00000003
#define CTXT_CLASS_PBCS_SUMD         0x00000004

#define CTXT_CLASS_PBC_MASK          0x000000f0 // mask for all print-by-category non-sumX values.
#define CTXT_CLASS_PBC_REFINDEX_LSB  0x00000010
#define CTXT_CLASS_PBC_REFINDEX_MSB  0x00000020
#define CTXT_CLASS_PBC_INTRA_BLOCK   0x00000030
#define CTXT_CLASS_PBC_INTCOMP       0x00000040
#define CTXT_CLASS_PBC_MVPDIR        0x00000050
#define CTXT_CLASS_PBC_MVDSIGN       0x00000060
#define CTXT_CLASS_PBC_MVDZERO       0x00000070
#define CTXT_CLASS_PBC_MVD           0x00000080

#define CTXT_CLASS_PBCF_MASK         0x0000ff00 // mask for all print-by-category-full values.
#define CTXT_CLASS_PBCF_RSHIFT       8          // right shift to convert PCBF class to counter.
#define CTXT_CLASS_PBCF_HEADER       0x00000100
#define CTXT_CLASS_PBCF_ZBLK0        0x00000200
#define CTXT_CLASS_PBCF_ZBLK1        0x00000300
#define CTXT_CLASS_PBCF_ZBLK2        0x00000400
#define CTXT_CLASS_PBCF_ZBLK3        0x00000500

#define CTXT_CLASS_PBCF_COEFF0       0x00000600
#define CTXT_CLASS_PBCF_COEFF1       0x00000700
#define CTXT_CLASS_PBCF_COEFF2       0x00000800
#define CTXT_CLASS_PBCF_COEFF3       0x00000900
#define CTXT_CLASS_PBCF_COEFF4       0x00000a00
#define CTXT_CLASS_PBCF_COEFF5       0x00000b00
#define CTXT_CLASS_PBCF_ZTREE0       0x00000c00
#define CTXT_CLASS_PBCF_ZTREE1       0x00000d00
#define CTXT_CLASS_PBCF_ZTREE2       0x00000e00
#define CTXT_CLASS_PBCF_ZTREE3       0x00000f00
#define CTXT_CLASS_PBCF_ZTREE4       0x00001000
#define CTXT_CLASS_PBCF_ZTREE5       0x00001100
#define CTXT_CLASS_PBCF_ZTREE6       0x00001200
#define CTXT_CLASS_PBCF_ZTREE7       0x00001300
#define CTXT_CLASS_PBCF_ZTREE8       0x00001400

#define CTXT_CLASS_PBCF_ZMV          0x00001500
#define CTXT_CLASS_PBCF_MVVAL        0x00001700
#define CTXT_CLASS_PBCF_MVP          0x00001800
#define CTXT_CLASS_PBCF_REFINDEX     0x00001900
#define CTXT_CLASS_PBCF_MVSHARE       0x00001a00
#define CTXT_CLASS_PBCF_IC           0x00001b00
#define CTXT_CLASS_PBCF_PMODE        0x00001c00
#define CTXT_CLASS_PBCF_SIP          0x00001d00
#define CTXT_CLASS_PBCF_SIPLUMA      0x00001e00
#define CTXT_CLASS_PBCF_SIPCHROMA    0x00001f00
#define CTXT_CLASS_PBCF_AIF          0x00002000
#define CTXT_CLASS_PBCF_RPPS         0x00002100
#define CTXT_CLASS_PBCF_WLF          0x00002200
#define CTXT_CLASS_PBCF_TSFM         0x00002300
#define CTXT_CLASS_PBCF_CLUSTER      0x00002400
#define CTXT_CLASS_PBCF_LF           0x00002500
#define CTXT_CLASS_PBCF_QMATRIX      0x00002600
#define CTXT_CLASS_PBCF_DUMMY1       0x00002700
#ifdef STATS_16x16_FIX
#define CTXT_CLASS_PBCF_ZTREE9       0x00002800
#define CTXT_CLASS_PBCF_LASTCLASS    0x00002800 // class limit.
#else
#define CTXT_CLASS_PBCF_LASTCLASS    0x00002700 // class limit.
#endif // STATS_16x16_FIX


//-----------------------------------------------------------------------------
/*!
* \brief Base probability estimator class
*
* Base class from which others are derived
*/
//-----------------------------------------------------------------------------
class ProbEst 
{
  
  public:
  unsigned int p; //!< Probability estimate for symbol 0
  int symcount[2]; //!< Symbol counts (for reporting purposes)
  int bitcount[2]; //!< Renormalization counts (approximate bit count)
  double fbitcount; //!< More accurate bit count (fractional)
  
  unsigned int i_p; //!< Storage of default Probality
  unsigned char type; //!< Type of ProbEst class
  int idCtx; //!< Context ID for current probability model
  char const *idName; //!< Context name
  unsigned int idClasses; //!< Set of class bits for context.
#ifdef CABAC_SYNC_POINTS
  static int const LOSS_RATE_SYNCHRO = 5;
  static int const SYNCHRO_MAX_CONFIDENCE = 14;
  static int const SYNCHRO_SHIFT_CONFIDENCE = 10;
#endif

  ProbEst()   { idCtx=-1; type='P'; } //!< Constructor
  virtual ~ProbEst() 
  {
  } //!< Destructor
  
  virtual void reset() = 0; //!< Reset internal variables
  virtual void set_prob(int p) = 0; //!< Set probability estimate \param[in] p Probability estimate
  
  virtual void copyInitialProba(ProbEst *src);
  virtual void set_default_prob(unsigned int p); //!< our initial probability
  virtual void restoreDefaultState(); //!< restore current Probabiliy status from default value
#ifdef CABAC_SYNC_POINTS
  virtual void setExternalData(unsigned int *p2d, unsigned int *mnb);
#endif
  double entropy();
  
  void reset_counts();
  void print(FILE *f);
  
#ifdef FAST_STATE_COPY
  virtual void copy_est(ProbEst &pe) { p = pe.p; }
#endif
};

#endif
