/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file RefFrame.H
*
* \brief Definition of RefFrame class
*
*
*/

#ifndef __RefFrame_H__
#define __RefFrame_H__

#include "LumaFrames.H"
#include "ChromaFrames.H"
#include "Frame.H"
class AIF_adaptive_filter;
#ifdef MVCOMP_ZERORES
class MotionField;
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Reference frame container class
*
* This frame class contains eighth-pel or fractional-pel-interpolated arrays.
*/
//-----------------------------------------------------------------------------

class RefFrame 
{
  
  public:
#ifdef FAST_INTERPOL
  int interpol_frame_no;
#endif
  Frame *frame; //!< Reference frame
  LumaFrames *luma; //!< Interpolated luma arrays
  int *luma_hor;    //!< Horizontal interpolated luma arrays
  ChromaFrames *chroma; //!< Interpolated chroma arrays
  int frame_no; //!< %Frame number
  int empty; //!< Flag indicating whether in use
  int decoder_mode;
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  int malloc_enable; //!< Flag indicating whether chroma interpolated signal is needed

#ifdef MVCOMP_ZERORES
  MotionField *saved_mf; //!< motion vector storage.
  int saved_ref_list[2][MAX_REF_FRAMES]; //!< refs in effect.
  int saved_ref_framenos[MAX_REF_FRAMES]; //!< framenos of refs in effect.
  int saved_inter_pel[2][MAX_REF_FRAMES]; //!< 1/8 or 1/4 accuracy in effect.
  int saved_log_inter_pel[2][MAX_REF_FRAMES]; //!< 1/8 or 1/4 accuracy in effect.
#endif

  RefFrame(int h, int w, int mem_on, int decoder_mode=0);
#else
#ifdef NO_AIF_CHROMA
  int chroma_enable; //!< Flag indicating whether chroma interpolated signal is needed

#ifdef MVCOMP_ZERORES
  MotionField *saved_mf; //!< motion vector storage.
  int saved_ref_list[2][MAX_REF_FRAMES]; //!< refs in effect.
  int saved_ref_framenos[MAX_REF_FRAMES]; //!< framenos of refs in effect.
  int saved_inter_pel[2][MAX_REF_FRAMES]; //!< 1/8 or 1/4 accuracy in effect.
  int saved_log_inter_pel[2][MAX_REF_FRAMES]; //!< 1/8 or 1/4 accuracy in effect.
#endif

  RefFrame(int h, int w, int chroma, decoder_mode=0);
#else
  RefFrame(int h, int w, decoder_mode=0);
#endif 
#endif
  
  RefFrame(int h, int w);
  ~RefFrame();
  
  void interpolate();
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  void interpolate_chroma();
#endif
  
  void init ();
  
  void AIF_interpolate(Frame *cur, int list, int ref, AIF_adaptive_filter *saif_half);  
};

#endif
