/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file RefFrameManager.H
*
* \brief Definition of RefFrameManager class
*
*
*/

#ifndef __RefFrameManager_H__
#define __RefFrameManager_H__

#include "RefFrame.H"
#include "Global.H"
#include "RefProcess.H"

//-----------------------------------------------------------------------------
/*!
* \brief Reference frame buffer class
*
*/
//-----------------------------------------------------------------------------

class RefFrameManager 
{
  public:
  RefFrame *ref[MAX_REF_FRAMES]; //!< Reference frames in buffer
  
  RefProcess *ref_process; //!< reference frame processing
    
  RefFrame *AIF_ref[MAX_REF_FRAMES];       //!< Reference frames to calculate AIF
  int aif_loop[LISTS][MAX_REF_FRAMES];     //!< indicator to use AIF or not for reference frames
  int iFilterCoef_SAIF[15][7]; // estimated filter coefficients
  int AIF_mode_flag1, AIF_mode_flag2;
  
  int list[LISTS][MAX_REF_FRAMES]; //!< Reference lists
  int lsize[LISTS]; //!< Reference lists
  int lRelation[LISTS][MAX_REF_FRAMES]; //!< Mapping of reference indices between lists 
  int n_ref_frames; //!< Number of reference frames in buffer
  int single_ref_frame; //!< Flag to limit size of reference lists to 1
  int mv_scale[LISTS][MAX_REF_FRAMES][MAX_REF_FRAMES]; //!< Scaling factors for motion vectors  

#ifdef MV_CODING_ENHANCE
  int inv_dist[LISTS][MAX_REF_FRAMES];
  int dist[LISTS][MAX_REF_FRAMES];
#endif

#ifdef MVCOMP_ZERORES
  int cur_frame_no;
  int cur_frame_type;
#endif

  int decoder_mode;
  RefFrameManager(int h, int w, int decoder_mode=0);
  ~RefFrameManager();

#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  void standard_interpolation();
#endif
#ifndef MVCOMP_ZERORES
  void add_frame(Frame *f);
#else
  void add_frame(Frame *f, MotionField *mf_frame, int *inter_pel, int *log_inter_pel, int *saved_ref_list, int *saved_frame_nos);
#endif

  void set_inter_pel(int list, int ref, int interpel);
  
  void AIF_replace_frame(Frame *cur, AIF_adaptive_filter *saif_half);
  void getAIFcoef(AIF_adaptive_filter *saif_half); 
  void setAIFcoef(LumaFrames *lf); 
  
  void prepare_list(int frame_type, int list, int frame_no);
  LumaFrames *get_luma(int index, int cur_list);
  ChromaFrames *get_chroma(int index, int cur_list);
  
  LumaFrames *AIF_get_luma(int index, int cur_list);
  ChromaFrames *AIF_get_chroma(int index, int cur_list);

  int target_rpps_id;
  int prev_rpps_id;
  
  void process_list(Frame * /*cur*/, AIF_adaptive_filter *saif_half);
  void init_rpps_id();
  int find_rpps_id();
  int temp_rpps_id[MAX_P_ENC]; //!< temporal RPPS id
  int ref_pic_parameter_set_id_used[MAX_RPPS_ID]; //!< flag showing RPPS id is used or not
  int ref_pic_parameter_set_id_shared[MAX_RPPS_ID]; //!< flag showing RPPS id is shared in both encoder and decoder
  unsigned char inter_type_set[MAX_RPPS_ID][LISTS][MAX_REF_FRAMES]; //!< interpolation pel accuracy
  unsigned char filter_type_set[MAX_RPPS_ID][LISTS][MAX_REF_FRAMES]; //!< interpolation filter type 
  unsigned char ic_disable_set[MAX_RPPS_ID][LISTS][MAX_REF_FRAMES]; //!< disabling intensity compensation
  unsigned char round_type_set[MAX_RPPS_ID][LISTS][MAX_REF_FRAMES]; //!< rounding type in interpolation process
  
  int deferred_refresh_flag; //!< Indicates a deferred refresh is pending.
  int deferred_refresh_frame_number; //!< The frame number beyond which the deferred refresh is to take place.
  void init (void);
  void perform_deferred_refresh (void); //!< Performs deferred refresh. One frame stays.

  void print_lists();
  void get_lRelation(); 
    
  unsigned char *get_luma_ptr(int list, int ref_index, int *mv, int i, int j, int h, int w, unsigned char *data = NULL);
  unsigned char *get_chroma_ptr(int list, int ref_index, int *mv, int i, int j, int h, int w, unsigned char *data = NULL);
  int get_stride();

private:
  int *intermediate[FRACTIONAL_PEL][FRACTIONAL_PEL]; //!< %Frame buffer (temp)
  int *buf; //!< Line buffer (temp)

  void allocate_intermediate_memory(int h, int w);
  void free_intermediate_memory();
};

#endif
