/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file SimpleProbEst.H
* \brief Definition of SimpleProbEst class
*/
//-----------------------------------------------------------------------------

#ifndef __SimpleProbEst_H__
#define __SimpleProbEst_H__

#include "ProbEst.H"

#ifdef CABAC_SYNC_POINTS
// Max number of superblocks per line of superblocks.
// 350 allows a resolution of 2600 pixels with a superblock every macroblock.
#define MAX_AC_SYNCHRO 350 
#define MAX_MNB (1<<PROB_BITS)
#define MAX_NBLOG (1<<SYNCHRO_MAX_CONFIDENCE)
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Simple probability estimator class
*
* Adapts probability based on exponential decay function.
*/
//-----------------------------------------------------------------------------
class SimpleProbEst : public ProbEst 
{

  public:
  
#ifdef CABAC_SYNC_POINTS
  unsigned int nb; //!< nb symbols for synchro bloc 
  unsigned int *mnb; //!< mean number of symbols
  unsigned int *p2d; //!< stored probas

  SimpleProbEst()   { type = 'S'; mnb=0; p2d=0; }
#else
  SimpleProbEst()   { type = 'S'; }
#endif

  virtual void reset();
  virtual void set_prob(int p);
  
  virtual void update(int v);

  virtual void copyInitialProba(ProbEst *src);
  virtual void restoreDefaultState();
#ifdef CABAC_SYNC_POINTS
  virtual void resetSynchro();
  virtual void updateSynchro(int i);
  virtual void initSynchro(int i);
  void averaging(int i);
  void setExternalData(unsigned int *_p2d, unsigned int *_mnb)  { if (p2d == NULL) {p2d=_p2d; mnb=_mnb;}}
  static int tLog[MAX_NBLOG];
  int KFunc(int mn0, int mn1);

#endif
#ifdef FAST_STATE_COPY
#ifdef CABAC_SYNC_POINTS
  virtual void copy_est(SimpleProbEst &pe) { ProbEst::copy_est(pe); nb = pe.nb; }
#else
  virtual void copy_est(SimpleProbEst &pe) { ProbEst::copy_est(pe); }
#endif
#endif
};

#endif
