/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Trace.H
*
* \brief Definition of Trace class
*
*
*/
#ifndef __Trace_H__
#define __Trace_H__

#include "Global.H"
#ifdef TRACE_DEBUG

class Trace
{
public:
  Trace()
  {
#ifndef FRAME_SUBSTREAMS
    m_pTraceFile = NULL;
#endif
    m_ptmpTraceFile = NULL;
    m_pbestTraceFile = NULL;
    m_bWriteTrace = false;
    m_bDecoderTrace = false;
    m_dBitsCount = 0;
    m_iMode = 0;
#ifdef FRAME_SUBSTREAMS
    m_nSubStreams = 1; //!< number of substreams being decoded (decoder only)
#endif
  }
  virtual ~Trace();

#ifdef FRAME_SUBSTREAMS
  // Add stream indicator as parameter.
  static void openTrace(char* pcString, int iMode, int nSubStreams); //iMode specifies file to open: 0 TraceFile, 1 tmpTraceFile, 2 bestTraceFile
#else
  static void openTrace(char* pcString, int iMode); //iMode specifies file to open: 0 TraceFile, 1 tmpTraceFile, 2 bestTraceFile
#endif
  static void closeTrace(int iMode);
  static void printTrace(char* pcString, int val, int iMode);
  static void printTrace(char* pcString, int val, int ctx, int iMode);
  static void printTrace(int val, int iMode);
  static void printTrace(char* pcString, int iMode);
  static void printTrace(int iMode);

  static void enableWritingTrace() { m_bWriteTrace = true;}  //for encoder only: because of multiple loop of rd opt
  static void disableWritingTrace() { m_bWriteTrace = false;}

  static void setDecoderTrace(bool b) { m_bDecoderTrace = b;}
  static void copyFile(int iSrcMode, int iDstMode);

  static void initFrameTrace(int iVal, int iMode, int iDec);
  static void initBlockTrace(int iValx, int iValy, int iMode);

  static void countBits(double d);
  static void PrintBits();
  static void setMode(int i) { m_iMode = i;}

#ifdef FRAME_SUBSTREAMS
#ifndef SUBSTREAM_MT
  static void set_trace_ss(int ss) {edec_ss = ss;}
  static int edec_ss;
#endif
#endif
protected:
#ifdef FRAME_SUBSTREAMS
  static FILE* m_pTraceFiles[100];
#else
  static FILE* m_pTraceFile; 
#endif
  static FILE* m_ptmpTraceFile;
  static FILE* m_pbestTraceFile;
  static bool  m_bWriteTrace;
  static bool  m_bDecoderTrace;
  static double m_dBitsCount;
  static int m_iMode;
#ifdef FRAME_SUBSTREAMS
  static int m_nSubStreams;
  static char *replace_percent_x(char const *src, char const *percent, int val, char *res, size_t res_size);
#endif
};


#define SETMODE(i)         Trace::setMode(i)
#ifdef FRAME_SUBSTREAMS
#define OPEN_ETRACE        Trace::openTrace("TraceEncoder-s%d.txt", 0, 0)
#define OPEN_ETRACETMP     Trace::openTrace("TraceEncoderTmp.txt", 1, 0)
#define OPEN_ETRACEBEST    Trace::openTrace("TraceEncoderBest.txt", 2, 0)
#ifndef SUBSTREAM_MT
#define SET_TRACE_SS(ss)   Trace::set_trace_ss(ss)
#endif // SUBSTREAM_MT
#else
#define OPEN_ETRACE        Trace::openTrace("TraceEncoder.txt", 0)
#endif // FRAME_SUBSTREAMS

#ifndef FRAME_SUBSTREAMS
#define OPEN_ETRACETMP     Trace::openTrace("TraceEncoderTmp.txt", 1)
#define OPEN_ETRACEBEST    Trace::openTrace("TraceEncoderBest.txt", 2)
#define SET_TRACE_SS(ss)
#endif

#define CLOSE_ETRACE       Trace::closeTrace(0)
#define CLOSE_ETRACETMP    Trace::closeTrace(1)
#define CLOSE_ETRACEBEST   Trace::closeTrace(2)

#define SVPRINT_ETRACE(s,x)  Trace::printTrace(s,x,1)
#define SVVPRINT_ETRACE(s,x,y)  Trace::printTrace(s,x,y,1)
#define SPRINT_ETRACE(s)  Trace::printTrace(s,1)
#define VPRINT_ETRACE(x)  Trace::printTrace(x,1)
#define RPRINT_ETRACE  Trace::printTrace(1)

#define INITFRAME_ETRACE(x) Trace::initFrameTrace(x,1, 1)
#define INITBLOCK_ETRACE(j,i) Trace::initBlockTrace(j,i,1)
#define ENABLE_ETRACE Trace::enableWritingTrace()
#define DISABLE_ETRACE Trace::disableWritingTrace()

#define SAVE_ETRACEBEST    Trace::copyFile(1, 2)
#define SAVE_ETRACE        Trace::copyFile(2, 0)

#ifdef FRAME_SUBSTREAMS
#define OPEN_DTRACE(nss)   Trace::openTrace("TraceDecoder-s%d.txt", 0, nss)
#else
#define OPEN_DTRACE        Trace::openTrace("TraceDecoder.txt", 0)
#endif
#define CLOSE_DTRACE       Trace::closeTrace(0)
#define SVPRINT_DTRACE(s,x)  Trace::printTrace(s,x,0)
#define SVVPRINT_DTRACE(s,x,y)  Trace::printTrace(s,x,y,0)
#define SPRINT_DTRACE(s)  Trace::printTrace(s,0)
#define VPRINT_DTRACE(x)  Trace::printTrace(x,0)
#define RPRINT_DTRACE  Trace::printTrace(0)
#define ENABLE_DTRACE Trace::setDecoderTrace(true)
#define DISABLE_DTRACE Trace::setDecoderTrace(false)

#define INITFRAME_DTRACE(x) Trace::initFrameTrace(x,0, 0)
#define INITBLOCK_DTRACE(j,i) Trace::initBlockTrace(j,i,0)

#define COUNT_BITS(d) Trace::countBits(d)
#define PRINT_BITS():: Trace::PrintBits()

#endif
#endif //__Trace_H__
