/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file UnaryProbEst.H
* \brief Definition of UnaryProbEst class
*/
//-----------------------------------------------------------------------------

#ifndef __UnaryProbEst_H__
#define __UnaryProbEst_H__

#include "SimpleProbEst.H"

//-----------------------------------------------------------------------------
/*!
* \brief Unary probability estimator
*
* This estimator is used for coding integer data. An integer is coded using
* a prefix and a suffix. The prefix uses a unary representation whereas the
* suffix is a fixed-length binary word. The size of the suffix adapts to
* the underlying distribution. The probability estimate for the prefix is
* adaptive whereas the probability estimate for the suffix is fixed to 50/50.
*/
//-----------------------------------------------------------------------------
class UnaryProbEst : public SimpleProbEst 
{
  
  public:
  int suffix; //!< Suffix value
  int suffix_size; //!< Suffix size
#ifdef CABAC_SYNC_POINTS
  unsigned int *s2d; // stored suffix size
#endif 
  
#ifdef CABAC_SYNC_POINTS
  UnaryProbEst()  { type = 'U'; s2d = 0; reset(); BuildTabUpDnSuff(); }
  void BuildTabUpDnSuff();
#else
  UnaryProbEst()  { type = 'U';}
#endif

  virtual void reset();
  virtual void set_prob(int p);
  void adjust_suffix();
  virtual void copyInitialProba(ProbEst *src);
  virtual void set_default_prob(unsigned int p);
  virtual void restoreDefaultState();
  virtual void update(int v);
#ifdef CABAC_SYNC_POINTS
  virtual void updateSynchro(int i);
  virtual void resetSynchro();
  virtual void initSynchro(int i);
  void saveraging(int i);
  void setExternalData(unsigned int *_p2d, unsigned int *_mnb, unsigned int *_s2d)   { if (p2d == NULL) {p2d=_p2d; mnb=_mnb; s2d=_s2d;} }
#endif

#ifdef FAST_STATE_COPY
#ifdef CABAC_SYNC_POINTS
  virtual void copy_est(UnaryProbEst &pe) { ProbEst::copy_est(pe); suffix_size = pe.suffix_size;  nb = pe.nb; }
#else
  virtual void copy_est(UnaryProbEst &pe) { ProbEst::copy_est(pe); suffix_size = pe.suffix_size; }
#endif
#endif
};

#endif
