/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file WFrame.H
*
* \brief Definition of WFrame class
*
*
*/

#ifndef __WFrame_H__
#define __WFrame_H__

#include "Frame.H"
#include "Block.H"

//-----------------------------------------------------------------------------
/*!
* \brief Wide frame container class
*
* This frame class uses 16-bit signed samples. A 16-bit sample x is related
* to an 8-bit sample y by x = (y - MID_SAMPLE_VAL) << SAMPLE_ADD_PREC.
*/
//-----------------------------------------------------------------------------

class WFrame 
{
  short *m_data; //!< data array
  short *m_luma; //!< pointer to luma
  short *m_crcb[2]; //!< pointers to chroma
  int m_size; //!< Number of samples in frame
  
public:
  
  int height; //!< Height of frame (luma samples)
  int width; //!< Width of frame (luma samples)
  int height2; //!< Height of frame (chroma samples)
  int width2; //!< Width of frame (chroma samples)
  int height4; //!< Height of frame (4x4 luma blocks)
  int width4; //!< Width of frame (4x4 chroma blocks)
  int height8; //!< Height of frame (8x8 luma blocks/4x4 chroma blocks)
  int width8; //!< Width of frame (8x8 luma blocks/4x4 chroma blocks)
  
  WFrame(int h, int w);
  ~WFrame();
  
  short *get_luma_ptr(int i, int j) { return m_luma + width * i + j; }
  short *get_chroma_ptr(int i, int j, int c) { return m_crcb[c] + width2 * i + j; }
  
  void get_blk_luma(Block *blk, int i, int j);
  void put_blk_luma(Block *blk, int i, int j);
  void get_blk_chroma(Block *blk, int i, int j, int c);
  void put_blk_chroma(Block *blk, int i, int j, int c);
  
  void put_mc_block(int i, int j, short lu[64], short ch[2][16]);
  void put_mc_block_mvshare(int i, int j, short lu[64], short ch[2][16], int shift_y, int shift_x);
  void put_mc_block_chroma(int i, int j, short ch[2][16]);
  
  void clear();
  void copy(WFrame *f);
  
  void add(WFrame *f);
  void sub(WFrame *f);
  double sad(WFrame *f);
  double sse(WFrame *f);
  
  void expand(Frame *f); 
  void collapse(Frame *f); 
  void collapse_sign(Frame *f); 
  void collapse_modulus(Frame *f); 
};

#endif
