/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file WienerFilter.H
*
* \brief Definition of WienerFilter class
*
*/
//-----------------------------------------------------------------------------

#ifndef __WienerFilter_H__
#define __WienerFilter_H__

#include "Global.H"

#include "Plane.H"
#include "Frame.H"
#include "Param.H"
#include "WFrame.H"
#include "EntropyDecoder.H"
#include "EntropyEncoder.H"
#include "Buffer.H"

//-----------------------------------------------------------------------------
/*!
* \brief Base Wiener filter class
*
* Provides methods for wiener filtering
*/
//-----------------------------------------------------------------------------
class WienerFilter 
{
public:
  int store_low_zeromv_rdoq;
  
  WienerFilter(int hh, int ww, Param *param);
  ~WienerFilter();
  
  void estimate(Frame *in, Frame *orig, Frame *pred, Frame *qpe_m, Frame *qpe_s, double lambdaY, double lambdaC, EntropyEncoder *eenc, Param *param);
  void filter(Frame *in, Frame *out, Frame *pred, Frame *qpe_m, Frame *qpe_s);
  
  void encode_params(EntropyEncoder *eenc);
  void decode_params(EntropyDecoder *edec);

private:
  int filter_precision_table      [FILTER_PRECISION_TABLE_NUMBER];
  int filter_precision_table_shift[FILTER_PRECISION_TABLE_NUMBER];
  int filter_precision_table_half [FILTER_PRECISION_TABLE_NUMBER];

  int height;
  int width;
  int height2;
  int width2;  
  int fs_max_rec;
  int fs_max_qpe;
  int fs_max_pred;  
  int filter_length_RD_rec_x [3];
  int filter_length_RD_rec_y [3];
  int filter_length_RD_pred_x[3];
  int filter_length_RD_pred_y[3];
  int filter_length_RD_qpe_x [3];
  int filter_length_RD_qpe_y [3];
  int filter_precision[3];
  int filtertype[3];
  int wlf_flag[3];
  int **coeffs;

  EntropyEncoder *rd_wlf_eenc;
  Buffer         *rd_wlf_buff;

  double lambda;   
  double *h;
  double *rho;
  double *rho_n;
  double *rho_RD;  
  double **a_final;
  double *r_final;
  double *h_final;  
  double **a;
  double **an;
  double **a_RD;
  double **a_1;
  double *r1;
  double *h1;
  int  *null;
  int  *coeffs_save;
  int  *coeffs_best;
  int  *coeffs_best_precision;
  
  unsigned char *image;  
  
  double estimate_filter_2D_equal_size(unsigned char *dY, unsigned char *xY, unsigned char *pY, unsigned char *qpe_m, unsigned char *qpe_s, EntropyEncoder *eenc, int component);
  void filter_2D_equal_size(unsigned char *xY, unsigned char *oY, unsigned char *pY, unsigned char *qpe_m, unsigned char *qpe_s,int component);
  
  void set_lambda(double set_lambda);
  
  void calc_mean_cross        (int on, int width_f, int height_f, unsigned char *dY);

  void set_lines_and_columns_of_auto_to_zero(int tap, int tap_minus_one_half, int tap_check, int on);
  
  void   sgaus  (double **a, double *x, double *b, int n);
  double get_mse(unsigned char *image1,unsigned char *image2, int size);

#ifdef WIENER_MT
  public:
  void filter_2D_equal_size_rows(const Plane<unsigned char> &xY, unsigned char *oY, const Plane<unsigned char> &pY, const Plane<short> &qpe, int component, int first_row, int nrows);
  private:
#endif
  void filter_2D_equal_size(const Plane<unsigned char> &xY, unsigned char *oY, const Plane<unsigned char> &pY, const Plane<short> &qpe, int component);
  double filter_2D_equal_size_d(double *rounding, unsigned char *org, const Plane<unsigned char> &xY, unsigned char *oY, const Plane<unsigned char> &pY, const Plane<short> &qpe, int component);
  template <typename T, typename U>
  void calc_auto_correlation(int on, int om, int tap1, int tap2, int tap_minus_one_half1, int tap_minus_one_half2, const Plane<T> &im1, const Plane<U> &im2);
  template <typename T>
  void calc_cross_correlation(int on, int tap, int tap_minus_one_half, const Plane<T> &xY, unsigned char *dY);
  template <typename T>
  void calc_mean_auto(int on, int om, int tap, int tap_minus_one_half, const Plane<T> &im);

  int  get_mem1Dint(int **array1D, int rows);
  int  get_mem2Dint(int ***array2D, int rows, int columns);
  int  get_mem1Ddouble(double **array1D, int rows);  
  int  get_mem2Ddouble(double ***array2D, int rows, int columns);
  void free_mem1Dint(int *array1D);
  void free_mem2Dint(int **array2D);
  void free_mem1Ddouble(double *array1D);  
  void free_mem2Ddouble(double **array2D);    
  void no_mem_exit(const char *where);
};

#endif // __WienerFilter_H__
