/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ZeroTree.H
*
* \brief Definition of ZeroTree class
*
*/
//-----------------------------------------------------------------------------

#ifndef __ZeroTree_H__
#define __ZeroTree_H__

#include "EntropyModel.H"
#include "ZTNode.H"
#include "Block.H"
#include "ArithEncoder.H"

//-----------------------------------------------------------------------------
/*!
* \brief Zero tree class
*
*/
//-----------------------------------------------------------------------------
class ZeroTree 
{
  int tree_idx; //!< Index into tree structure of pattern (temp)
  int coeff_idx; //!< Index into coeff structure of pattern (temp)
  int node_idx; //!< Node counter (temp)
  
  
  ZTNode *nodes; //!< Array of nodes in tree
  int *bit_costi[2][2]; //!< Array of rate cost for coding node. Indexed by [mode][symbol][node].
  
  ZTNode *grow_children(int *pattern); // allocates memory
  void reset_bit_costs();
  
public:
  ZTNode *root; //!< Pointer to root node
  int nleaves; //!< Number of leaf nodes in tree
  int *i_prob; //!< Array of initial probabilities used for estimators
#ifdef FRAME_SUBSTREAMS
  // name change.
  SimpleProbEst *local_pe; //!< Array of probability estimators used for coding tree
  // Return estimators stored in entropymodel if they're stored there (decoder, transcoder), otherwise our locals (encoder).
  SimpleProbEst *get_pe(EntropyModel *cs) {return first_ctx_id < 0 ? local_pe : &cs->p_ztree_pe[first_ctx_id];}
#else
  SimpleProbEst *pe; //!< Array of probability estimators used for coding tree
#endif
  int *stats; //!< Array of statistics from coding tree
  
#ifdef FRAME_SUBSTREAMS
  // indication of context storage for all ztrees.
  // -1 indicates we're just using the local pe, not the entropymodel one.
  int first_ctx_id;
  int n_ctxs;
#endif

  // used for decoding
  int size; //!< Number of nodes in tree
  int *m_structure; //!< Bit pattern of tree structure
  int *leaf_index; //!< Mapping of leaves indices to array of data represented by tree

#ifdef FRAME_SUBSTREAMS
  ZeroTree(int *pattern, int do_reset = 1, int ctx0 = -1);
#else
  ZeroTree(int *pattern, int do_reset = 1);
#endif
  ~ZeroTree();
  
  void copy_initial_probs_back(EntropyModel *cs_src, int *pattern);
  void copy_initial_probs_from(ZeroTree *zt_src, EntropyModel *cs_src);

  void update_values(Block *val);
  int get_ztcost(ZTNode *node);
#ifdef FRAME_SUBSTREAMS
  void update_bit_costs(SimpleProbEst *pe);
#else
  void update_bit_costs();
#endif
  
  void copy_coding_state(ZeroTree *ztree);

#ifdef FRAME_SUBSTREAMS
  int cost_nodes(Block *blk, ArithEncoder *aenc, EntropyModel *cs);
#else
  int cost_nodes(Block *blk, ArithEncoder *aenc);
#endif
};

#endif

