/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

#ifdef WIN32

#include <windows.h>
#include <semaphore.h>

#define MAX_SEM_COUNT 0xffff

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Create a new semaphore.
*
* \param[out] sem Identifies the new semaphore.
* \param[in]  shared Not supported.
* \param[in]  value Initial semaphore value.
*
* The maximum count available for the semaphore is controlled by the
* local define here MAX_SEM_COUNT.
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int sem_init(sem_t *sem, int shared, unsigned int value)
{
  if (sem == NULL)
    return -1;
  if (shared)
    return -1; // not supported.

  *sem = CreateSemaphore(NULL, value, MAX_SEM_COUNT, NULL);
  return *sem == NULL ? -1 : 0;
}


//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Delete an existing semaphore.
*
* \param[in] sem Identifies the semaphore.
*
* The semaphore identifier is no longer valid when this function returns.
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int sem_destroy(sem_t *sem)
{
  if (sem == NULL)
    return -1;
  return CloseHandle(*sem) ? 0 : -1;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Increment the counter associated with a semaphore.
*
* \param[in] sem Identifies the semaphore.
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int sem_post(sem_t *sem)
{
  if (sem == NULL)
    return -1;
  return !ReleaseSemaphore(*sem, 1, NULL) ? -1 : 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Decrement the count of a semaphore, waiting if necessary.
*
* \param[in] sem Identifies the semaphore.
*
0* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int sem_wait(sem_t *sem)
{
  if (sem == NULL)
    return -1;
  return WaitForSingleObject(*sem, INFINITE) == WAIT_OBJECT_0 ? 0 : -1;
}

#endif
