/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file AIF_adaptive_filter.cpp
*
* \brief Implementation of AIF_adaptive_filter class
*
*/
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------
#include "AIF_adaptive_filter.H"
#include "Encoder.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*/
//-----------------------------------------------------------------------------
AIF_adaptive_filter::AIF_adaptive_filter() 
{
  int i, j, sign;
  double filter_tables[15][FILTER_SIZE_OFFSET] = 
  {
    {1.0 / 64.0, -5.0 / 64.0, 52.0 / 64.0, 20.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // a_pos
    {1.0 / 32.0, -5.0 / 32.0, 20.0 / 32.0, 20.0 / 32.0, -5.0 / 32.0, 1.0 / 32.0, 0.0}, // b_pos
    {1.0 / 64.0, -5.0 / 64.0, 20.0 / 64.0, 52.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // c_pos
    {1.0 / 64.0, -5.0 / 64.0, 52.0 / 64.0, 20.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // d_pos
    {1.0 / 64.0, -5.0 / 64.0, 52.0 / 64.0, 20.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // e_pos
    {1.0 / 64.0, -5.0 / 64.0, 52.0 / 64.0, 20.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // f_pos
    {1.0 / 64.0, -5.0 / 64.0, 52.0 / 64.0, 20.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // g_pos
    {1.0 / 32.0, -5.0 / 32.0, 20.0 / 32.0, 20.0 / 32.0, -5.0 / 32.0, 1.0 / 32.0, 0.0}, // h_pos
  {1.0 / 32.0, -5.0 / 32.0, 20.0 / 32.0, 20.0 / 32.0, -5.0 / 32.0, 1.0 / 32.0, 0.0}, // i_pos
    {1.0 / 32.0, -5.0 / 32.0, 20.0 / 32.0, 20.0 / 32.0, -5.0 / 32.0, 1.0 / 32.0, 0.0}, // j_pos
    {1.0 / 32.0, -5.0 / 32.0, 20.0 / 32.0, 20.0 / 32.0, -5.0 / 32.0, 1.0 / 32.0, 0.0}, // k_pos
    {1.0 / 64.0, -5.0 / 64.0, 20.0 / 64.0, 52.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // l_pos
    {1.0 / 64.0, -5.0 / 64.0, 20.0 / 64.0, 52.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // m_pos
    {1.0 / 64.0, -5.0 / 64.0, 20.0 / 64.0, 52.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // n_pos
    {1.0 / 64.0, -5.0 / 64.0, 20.0 / 64.0, 52.0 / 64.0, -5.0 / 64.0, 1.0 / 64.0, 0.0}, // o_pos
  };
  int eq_tables[15][FILTER_SIZE_OFFSET]  =    // get equation number for symmetric filter coefficients
  {
    { 0,  1,  2,  3,  4,  5, 0}, // a_pos
    { 0,  1,  2,  2,  1,  0, 0}, // b_pos
    { 5,  4,  3,  2,  1,  0, 0}, // c_pos
    { 0,  1,  2,  3,  4,  5, 0}, // d_pos
    { 0,  1,  2,  3,  4,  5, 0}, // e_pos
    { 0,  1,  2,  2,  1,  0, 0}, // f_pos
    { 5,  4,  3,  2,  1,  0, 0}, // g_pos
    { 0,  1,  2,  2,  1,  0, 0}, // h_pos
    { 0,  3,  6,  9, 12, 15, 0}, // i_pos
    { 0,  1,  2,  2,  1,  0, 0}, // j_pos
    {15, 12,  9,  6,  3,  0, 0}, // k_pos
    { 5,  4,  3,  2,  1,  0, 0}, // l_pos
    { 5, 10, 14, 17, 19, 20, 0}, // m_pos
    {15, 16, 17, 17, 16, 15, 0}, // n_pos
    {20, 19, 17, 14, 10,  5, 0}  // o_pos
  }; 

  for (i = 0; i < 15; i ++)
    for (j = 0; j < FILTER_SIZE_OFFSET; j++)
  {
      STANDARD_2D_FILTER_SEP[i][j] = filter_tables[i][j];
    TwoDEquationPattern[i][j] = eq_tables[i][j];
  }  
  for (i=0;i<15;i++)
    for (j=0;j<FILTER_SIZE_OFFSET;j++)
    {
      sign = STANDARD_2D_FILTER_SEP[i][j] >= 0 ? 1:-1;
      iSTANDARD_2D_FILTER_SEP[i][j] = sign * ((int)(fabs(STANDARD_2D_FILTER_SEP[i][j]) * (1 << F_BITS) + 0.5));
      iFilterCoef_SAIF[i][j]= iSTANDARD_2D_FILTER_SEP[i][j];
    }
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory allocated by constructor.
*/
//-----------------------------------------------------------------------------
AIF_adaptive_filter::~AIF_adaptive_filter() 
{
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
/*!
* \brief Inverse quantization of filter coefficients for SAIF 
*
* \param[in] iQFCoef Pointer to filter coefficients to be dequantized
* \param[in,out] iFCoef Pointer to dequantized filter coefficients
*/
//-----------------------------------------------------------------------------
void AIF_adaptive_filter::DequantizeFilterCoefficients(int *iQFCoef, int *iFCoef)
{
  int j;
  int sign;
  int increment_steps = F_BITS - (DEFAULT_QUANT-1);  // 1 bit for sign, DEFAULT_QUANT-1 for amplitude
  for(j=0; j < FILTER_SIZE_OFFSET; j++)
  {
    sign = iQFCoef[j] >= 0 ? 1:-1;
    iFCoef[j]  = sign * abs(iQFCoef[j]) * (1 << increment_steps);    // reconstructed (quantized) filter coefficients with additional step up to be in line with IAIF
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Quantization of filter coefficients for AIF FULL
*
* \param[in,out] iFCoef Pointer to filter coefficients to be quantized and dequantized
* \param[in,out] iQFCoef Pointer to quantized filter coefficients
*/
//-----------------------------------------------------------------------------
void AIF_adaptive_filter::QuantizeFilterCoefficients(int iFCoef[FILTER_SIZE_OFFSET], int iQFCoef[FILTER_SIZE_OFFSET])
{
  int number_of_steps = F_BITS - (DEFAULT_QUANT-1);  // 1 bit for sign, NumberOfQBits-1 for amplitude
  int sign;
  int j;
  for(j=0; j < FILTER_SIZE_OFFSET; j++)
  {
    sign = iFCoef[j] >= 0 ? 1:-1;
    iQFCoef[j] = sign*((abs(iFCoef[j])+(1<<(number_of_steps-1))) >> number_of_steps);    // step number
    iFCoef[j] = sign*(abs(iQFCoef[j]) << number_of_steps); // reconstructed (quantized) filter coefficients
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Find integer pixel position after motion compensation
*
* \param[in] Dim number of horizontal/vertical pixels
* \param[in] cur_pos horizontal/vertical position of target pixel
* \param[in] Offset target pixel position in 6x6 tap 2D array
* \param[in] mv value of horizontal/vertical motion vector
* \param[in] inter_pel subpel size
*
* \return motion compensated pixel position
*/
//-----------------------------------------------------------------------------
int AIF_adaptive_filter::FindPosition(int Dim, int cur_pos, int Offset, int mv, int inter_pel)
{
  int position;
  if(mv < 0 && abs(mv)%inter_pel!=0) 
    mv = mv-inter_pel;
  mv =mv/inter_pel;
  position = cur_pos-FILTER_OFFSET+Offset+mv;
  return max(0, min(Dim-1, position ) );
}

//-----------------------------------------------------------------------------
/*!
* \brief Reset of AIF filter coefficients at encoder
*
* Set coefficients to zero at encoder
*
* \param[in] cur Pointer to current frame
*/
//-----------------------------------------------------------------------------
void AIF_adaptive_filter::ResetAdaptiveFilter(Frame *cur)
{
  int k,i,j;

  if(cur->frame_type == ft_P_frame || cur->frame_type == ft_B_frame)
  {
    for(k = 0; k < 15; k++)
    {
      for(i = 0; i < FILTER_SIZE_OFFSET; i++)
      {
        for(j = 0; j < FILTER_SIZE_OFFSET; j++)
           NxN_Matrix[k][i][j] = 0.0;
        N_Vector[k][i]       = 0.0;
        FilterCoef[k][i]       = 0.0;
        iFilterCoef_SAIF[k][i] = 0;
      }
      FilterFlag_SAIF[k] = 0;
    }   
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Reset of AIF filter coefficients at decoder
*
* Set coefficients to zero at docoder
*
*/
//-----------------------------------------------------------------------------
void AIF_adaptive_filter::ResetAdaptiveFilter_DEC()
{ 
  int sub_pel,i;

  for(sub_pel = 0; sub_pel < 15; sub_pel++) 
    for(i = 0; i < FILTER_SIZE_OFFSET; i++)
    {
      DiffQFilterCoef_SAIF[sub_pel][i] = 0;
      iFilterCoef_SAIF[sub_pel][i] = 0;
    }
}

//-----------------------------------------------------------------------------
/*!
* \brief Calculate AIF filter coefficients for SAIF_HALF with RD optimization
*
* \param[in] cur Pointer to current frame
* \param[in] mf Pointer to motion field
* \param[in] ref_mgr Pointer to reference frame manager
*/
//-----------------------------------------------------------------------------
void AIF_adaptive_filter::AIF_calculate_coef_SAIF(Frame *cur, MotionField *mf, RefFrameManager *ref_mgr)
{
  int i, j, sub_pos;
  PredBlk *curblk;

  for(i=a_pos; i<= o_pos; i++)
    FilterFlag_SAIF[i] = 1;    
  for (sub_pos = a_pos; sub_pos <= o_pos; sub_pos ++)
  {
    if (!FilterFlag_SAIF[sub_pos])
      for (i = 0; i < FILTER_SIZE_OFFSET; i++) 
      {
        FilterCoef[sub_pos][i] = STANDARD_2D_FILTER_SEP[sub_pos][i];
        iFilterCoef_SAIF[sub_pos][i] = iSTANDARD_2D_FILTER_SEP[sub_pos][i];
      }
    for(i = 0; i < FILTER_SIZE_OFFSET; i++)
    {
      for(j = 0; j < FILTER_SIZE_OFFSET; j++)
        NxN_Matrix[sub_pos][i][j] = 0.0;
      N_Vector[sub_pos][i]       = 0.0;
    }
  }
  for (j=0; j<mf->height; j++) 
    for (i=0; i<mf->width; i++) 
    {
      curblk=mf->get_pblk(j,i);
      if(curblk->intra==0)
        SetAdaptiveFilterHor(cur, curblk, i, j, ref_mgr);
    }
  FindFilterCoefHor();
  SAIF_HALF_rd_decision_Hor (cur, mf, ref_mgr); 
  InterpolateHor (ref_mgr);
  for (j=0; j<mf->height; j++) 
    for (i=0; i<mf->width; i++) 
    {
      curblk=mf->get_pblk(j,i);
      if(curblk->intra==0)
        SetAdaptiveFilterVer(cur, curblk, i, j, ref_mgr);
    }
  FindFilterCoefVer();
  if (AIF_mode_flag1)
    SAIF_HALF_rd_decision_Ver (cur, mf, ref_mgr); 
}
/*!
************************************************************************
* \brief
*    Predict and calculate filter coefficients for separable filter
************************************************************************
*/
void AIF_adaptive_filter::CalculateFilterCoefficientsSep(void)
{
  int i,sub_pos;
  for(sub_pos = a_pos; sub_pos <= o_pos; sub_pos++)
  {
    if(FilterFlag_SAIF[sub_pos])
    {
      DequantizeFilterCoefficients(DiffQFilterCoef_SAIF[sub_pos], iFilterCoef_SAIF[sub_pos]);
      for(i=0; i < FILTER_SIZE_OFFSET; i++)
        iFilterCoef_SAIF[sub_pos][i] += iSTANDARD_2D_FILTER_SEP[sub_pos][i];
    }
    else
      for(i=0; i < FILTER_SIZE_OFFSET; i++)
        iFilterCoef_SAIF[sub_pos][i] = iSTANDARD_2D_FILTER_SEP[sub_pos][i];
  }
}
/*!
************************************************************************
* \brief
*    sets necessary horizontal filter data for current macroblock
************************************************************************
*/
void AIF_adaptive_filter::SetAdaptiveFilterHor(Frame *cur, PredBlk *curblk, int ii, int jj, RefFrameManager *ref_mgr)
{
  int i, x, y, xj, yi;
  float b_scaling=1.0;
  int param_ic_full;
  int run=0, rerun=0;
  int equation=0, eq_new = 0, i_new = 0;
  int mvx=0,  mvy=0;
  int mvxF=0, mvyF=0;
  int mvxB=0, mvyB=0;
  int mvxF_sub=0, mvyF_sub=0;
  int mvxB_sub=0, mvyB_sub=0;
  int sub_pos=0;
  int sub_posF=0;
  int sub_posB=0;
  int ref_frame=0;
  int ref_frameF=-1;
  int ref_frameB=-1;
  int filter_pos_x1, filter_pos_y1, filter_pos_x2;
  int x_orig, y_orig;       
  int number_equation;
  int scaleF=0, offsetF=0;  
  int scaleB=0, offsetB=0;
  int scale, offset;
  unsigned char *RecY  = NULL;
  unsigned char *RecYF = NULL;
  unsigned char *RecYB = NULL;
  int temp_inter_pel;
  param_ic_full=param.enable_intcomp;
  param.enable_intcomp=0;

  if(curblk->intra) //intra macroblocks are not used for calculation of the filter coeffs.
    return;

  x = x_orig = MOTION_BLK_SIZE*ii;
  y = y_orig = MOTION_BLK_SIZE*jj;
  mvxF=curblk->mv[LIST_0][horizontal];
  mvyF=curblk->mv[LIST_0][vertical];

  if(param.enable_intcomp)
  {
    scaleF=curblk->scale[LIST_0];
    offsetF=curblk->DC[LIST_0];
  }
  temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_0][curblk->ref_index[LIST_0]];
  if(mvxF >= 0)
    mvxF_sub = (mvxF%temp_inter_pel);      // x-sub-coordinate in a 4x4block 
  else
    mvxF_sub = ((temp_inter_pel-abs(mvxF)%temp_inter_pel)%temp_inter_pel);
  if(mvyF >= 0)
    mvyF_sub = (mvyF%temp_inter_pel);      // y-sub-coordinate in a 4x4block
  else
    mvyF_sub = ((temp_inter_pel-abs(mvyF)%temp_inter_pel)%temp_inter_pel);

  if (temp_inter_pel == PEL_ONE_EIGHTH) 
  {   // this is a simple mapping of 1/8 pel position into 1/4 pel AIF filter
    mvxF_sub = mvxF_sub>>1;
    mvyF_sub = mvyF_sub>>1;
  }
  else if (temp_inter_pel == PEL_ONE_HALF) 
  {   // this is a simple mapping of 1/2 pel position into 1/4 pel AIF filter
    mvxF_sub = mvxF_sub<<1;
    mvyF_sub = mvyF_sub<<1;
  }
  sub_posF = mvxF_sub + 4*mvyF_sub-1;    // pos 0..14 in a 4x4 block 
  ref_frameF=ref_mgr->list[LIST_0][curblk->ref_index[LIST_0]];
  if(ref_frameF != -1)
    RecYF  = ref_mgr->AIF_ref[ref_frameF]->frame->data;
  if(cur->frame_type == ft_B_frame)
  {
    mvxB=curblk->mv[LIST_1][horizontal];
    mvyB=curblk->mv[LIST_1][vertical];
    if(param.enable_intcomp)
    {
      scaleB=curblk->scale[LIST_1];
      offsetB=curblk->DC[LIST_1];
    }
    temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_1][curblk->ref_index[LIST_1]];
    if(mvxB >= 0)
      mvxB_sub = (mvxB%temp_inter_pel);      // x-sub-coordinate in a 4x4block 
    else
      mvxB_sub = ((temp_inter_pel-abs(mvxB)%temp_inter_pel)%temp_inter_pel);
    if(mvyB >= 0)
      mvyB_sub = (mvyB%temp_inter_pel);      // y-sub-coordinate in a 4x4block
    else
      mvyB_sub = ((temp_inter_pel-abs(mvyB)%temp_inter_pel)%temp_inter_pel);

    if (temp_inter_pel == PEL_ONE_EIGHTH) 
    {   // this is a simple mapping of 1/8 pel position into 1/4 pel AIF filter
      mvxB_sub = mvxB_sub>>1;
      mvyB_sub = mvyB_sub>>1;
    }
    else if (temp_inter_pel == PEL_ONE_HALF) 
    {   // this is a simple mapping of 1/2 pel position into 1/4 pel AIF filter
      mvxB_sub = mvxB_sub<<1;
      mvyB_sub = mvyB_sub<<1;
    }
    sub_posB = mvxB_sub + 4*mvyB_sub-1;    // pos 0..14 in a 4x4 block 
    ref_frameB=ref_mgr->list[LIST_1][curblk->ref_index[LIST_1]];
    if(ref_frameB != -1)
      RecYB  = ref_mgr->AIF_ref[ref_frameB]->frame->data;
    rerun=1;
  }
  else
    rerun=0;

  if(ref_frameF != -1 && ref_frameB != -1)
    b_scaling = 0.5;
  else
    b_scaling = 1.0;

  for(run=0; run<= rerun; run++)
  {
    if(run==0)
    {
      sub_pos=sub_posF;
      mvx=mvxF;
      mvy=mvyF;
      ref_frame=ref_frameF;
      scale=scaleF;
      offset=offsetF;
      RecY = RecYF;
      temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_0][curblk->ref_index[LIST_0]];
    }
    else
    {
      sub_pos=sub_posB;
      mvx=mvxB;
      mvy=mvyB;
      ref_frame=ref_frameB;
      scale=scaleB;
      offset=offsetB;
      RecY = RecYB;
      temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_1][curblk->ref_index[LIST_1]];
    }
    if (sub_pos != -1 && ref_frame != -1 && sub_pos < d_pos) 
    {
      number_equation = FILTER_SIZE;
      for(yi = 0; yi < 8; yi++)    
      {
        filter_pos_y1 = FindPosition(cur->height,y+yi,FILTER_OFFSET,mvy, temp_inter_pel); 
        for(xj = 0; xj < 8; xj++)  
        {
          for(equation = 0; equation < number_equation; equation++)
          {
            if(sub_pos == a_pos || sub_pos == c_pos) 
            {
              filter_pos_x1 = FindPosition(cur->width, x+xj,equation,mvx, temp_inter_pel); 
              for(i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_x2 = FindPosition(cur->width, x+xj, i,mvx, temp_inter_pel); 
                NxN_Matrix[sub_pos][equation][i] += b_scaling*
                  RecY[filter_pos_y1*cur->width+filter_pos_x1]*
                  RecY[filter_pos_y1*cur->width+filter_pos_x2];               
              }
              N_Vector[sub_pos][equation] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)]*
                RecY[filter_pos_y1*cur->width+filter_pos_x1];                
            }
            else // b_pos
            {
              filter_pos_x1 = FindPosition(cur->width, x+xj,equation,mvx, temp_inter_pel); 
              eq_new = TwoDEquationPattern[b_pos][equation];
              for(i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_x2 = FindPosition(cur->width, x+xj, i,mvx, temp_inter_pel); 
                i_new = TwoDEquationPattern[b_pos][i];
                NxN_Matrix[sub_pos][eq_new][i_new] += b_scaling*
                  RecY[filter_pos_y1*cur->width+filter_pos_x1]*
                  RecY[filter_pos_y1*cur->width+filter_pos_x2];               
              }
              N_Vector[sub_pos][eq_new] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)]*
                RecY[filter_pos_y1*cur->width+filter_pos_x1];                
            }
            // one more coefficient to solve to get filter offset
            if(sub_pos == a_pos || sub_pos == c_pos) 
            {
              for(i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_x2 = FindPosition(cur->width, x+xj, i,mvx, temp_inter_pel); 
                NxN_Matrix[sub_pos][FILTER_SIZE][i] += b_scaling*
                  RecY[filter_pos_y1*cur->width+filter_pos_x2];    
                NxN_Matrix[sub_pos][i][FILTER_SIZE] += b_scaling*
                  RecY[filter_pos_y1*cur->width+filter_pos_x2];                     
              }
              NxN_Matrix[sub_pos][FILTER_SIZE][FILTER_SIZE] += 1.; 
              N_Vector[sub_pos][FILTER_SIZE] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)];              
            }
            else // b_pos
            {
              for(i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_x2 = FindPosition(cur->width, x+xj, i,mvx, temp_inter_pel); 
                i_new = TwoDEquationPattern[sub_pos][i];
                eq_new = 3;
                NxN_Matrix[sub_pos][eq_new][i_new] += b_scaling*
                  RecY[filter_pos_y1*cur->width+filter_pos_x2];    
                NxN_Matrix[sub_pos][i_new][eq_new] += b_scaling*
                  RecY[filter_pos_y1*cur->width+filter_pos_x2];                     
              }
              NxN_Matrix[sub_pos][eq_new][eq_new] += 1.; 
              N_Vector[sub_pos][eq_new] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)];              
            }
          }
        }
      }
    }
  }
  param.enable_intcomp=param_ic_full;
}

/*!
************************************************************************
* \brief
*    sets necessary vertical filter data for current macroblock
************************************************************************
*/
void AIF_adaptive_filter::SetAdaptiveFilterVer(Frame *cur, PredBlk *curblk, int ii, int jj, RefFrameManager *ref_mgr)
{
  int i, x, y, xj, yi;
  float b_scaling=1.0;
  int param_ic_full;
  int run=0, rerun=0;
  int equation=0, eq_new = 0, i_new = 0;
  int mvx=0,  mvy=0;
  int mvxF=0, mvyF=0;
  int mvxB=0, mvyB=0;
  int mvxF_sub=0, mvyF_sub=0;
  int mvxB_sub=0, mvyB_sub=0;
  int scaleF=0, offsetF=0;     
  int scaleB=0, offsetB=0;     
  int scale, offset;
  int sub_pos=0;
  int sub_posF=0;
  int sub_posB=0;
  int ref_frame=0;
  int ref_frameF=-1;
  int ref_frameB=-1;
  int temp_inter_pel;
  int filter_pos_x1, filter_pos_y1, filter_pos_y2;
  int x_orig, y_orig;       
  int number_equation;
  int h1, h2;
  int *new_eq, *new_filter_pos;
  int *RecY=NULL; 
  int *RecYF=NULL; 
  int *RecYB=NULL; 
  unsigned char *RecY_FullPel = NULL;
  unsigned char *RecYF_FullPel = NULL;
  unsigned char *RecYB_FullPel = NULL;
  int border = 16;
  int border4 = 64;
  int width4border = cur->width * 4+border4*2;
  new_filter_pos = &h1;
  new_eq = &h2;
  param_ic_full=param.enable_intcomp;
  param.enable_intcomp=0;

  if(curblk->intra) //intra macroblocks are not used for calculation of the filter coeffs.
    return;
  
  x = x_orig = MOTION_BLK_SIZE*ii;
  y = y_orig = MOTION_BLK_SIZE*jj;
  mvxF=curblk->mv[LIST_0][horizontal];
  mvyF=curblk->mv[LIST_0][vertical];

  if(param.enable_intcomp)
  {
    scaleF=curblk->scale[LIST_0];
    offsetF=curblk->DC[LIST_0];
  }
  temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_0][curblk->ref_index[LIST_0]];
  if(mvxF >= 0)
    mvxF_sub = (mvxF%temp_inter_pel);      // x-sub-coordinate in a 4x4block 
  else
    mvxF_sub = ((temp_inter_pel-abs(mvxF)%temp_inter_pel)%temp_inter_pel);
  if(mvyF >= 0)
    mvyF_sub = (mvyF%temp_inter_pel);      // y-sub-coordinate in a 4x4block
  else
    mvyF_sub = ((temp_inter_pel-abs(mvyF)%temp_inter_pel)%temp_inter_pel);

  if (temp_inter_pel == PEL_ONE_EIGHTH) 
  {   // this is a simple mapping of 1/8 pel position into 1/4 pel AIF filter
    mvxF_sub = mvxF_sub>>1;
    mvyF_sub = mvyF_sub>>1;
  }
  else if (temp_inter_pel == PEL_ONE_HALF) 
  {   // this is a simple mapping of 1/2 pel position into 1/4 pel AIF filter
    mvxF_sub = mvxF_sub<<1;
    mvyF_sub = mvyF_sub<<1;
  }
  sub_posF = mvxF_sub + 4*mvyF_sub-1;    // pos 0..14 in a 4x4 block 
  ref_frameF=ref_mgr->list[LIST_0][curblk->ref_index[LIST_0]];
  if(ref_frameF != -1)
  {
    RecYF = ref_mgr->AIF_ref[ref_frameF]->luma_hor;
    RecYF_FullPel = ref_mgr->AIF_ref[ref_frameF]->frame->data;
  }
  if(cur->frame_type == ft_B_frame)
  { 
    mvxB=curblk->mv[LIST_1][horizontal];
    mvyB=curblk->mv[LIST_1][vertical];
    if(param.enable_intcomp)
    {
      scaleB=curblk->scale[LIST_1];
      offsetB=curblk->DC[LIST_1];
    }
    temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_1][curblk->ref_index[LIST_1]];
    if(mvxB >= 0)
      mvxB_sub = (mvxB%temp_inter_pel);      // x-sub-coordinate in a 4x4block 
    else
      mvxB_sub = ((temp_inter_pel-abs(mvxB)%temp_inter_pel)%temp_inter_pel);
    if(mvyB >= 0)
      mvyB_sub = (mvyB%temp_inter_pel);      // y-sub-coordinate in a 4x4block
    else
      mvyB_sub = ((temp_inter_pel-abs(mvyB)%temp_inter_pel)%temp_inter_pel);

  if (temp_inter_pel == PEL_ONE_EIGHTH) 
    {   // this is a simple mapping of 1/8 pel position into 1/4 pel AIF filter
      mvxB_sub = mvxB_sub>>1;
      mvyB_sub = mvyB_sub>>1;
    }
    else if (temp_inter_pel == PEL_ONE_HALF) 
    {   // this is a simple mapping of 1/2 pel position into 1/4 pel AIF filter
      mvxB_sub = mvxB_sub<<1;
      mvyB_sub = mvyB_sub<<1;
    }
    sub_posB = mvxB_sub + 4*mvyB_sub-1;    // pos 0..14 in a 4x4 block 
    ref_frameB=ref_mgr->list[LIST_1][curblk->ref_index[LIST_1]];
    if(ref_frameB != -1)
    {
      RecYB = ref_mgr->AIF_ref[ref_frameB]->luma_hor;
      RecYB_FullPel = ref_mgr->AIF_ref[ref_frameB]->frame->data;
    }
    rerun=1;
  }
  else
    rerun=0;

  if(ref_frameF != -1 && ref_frameB != -1)
    b_scaling = 0.5;
  else
    b_scaling = 1.0;

  for(run=0; run<= rerun; run++)
  {
    if(run==0)
    {
      sub_pos=sub_posF;
      mvx=mvxF;
      mvy=mvyF;
      ref_frame=ref_frameF;
      RecY = RecYF;
      RecY_FullPel = RecYF_FullPel;
      scale=scaleF;
      offset=offsetF;
      temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_0][curblk->ref_index[LIST_0]];
    }
    else
    {
      sub_pos=sub_posB;
      mvx=mvxB;
      mvy=mvyB;
      ref_frame=ref_frameB;
      RecY = RecYB;
      RecY_FullPel = RecYB_FullPel;
      scale=scaleB;
      offset=offsetB;
      temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_1][curblk->ref_index[LIST_1]];
    }      
    if (sub_pos != -1 && ref_frame != -1 && sub_pos > c_pos) 
    {
      number_equation = FILTER_SIZE;      
      if (sub_pos == d_pos)
      {
        for (xj = 0; xj < 8; xj++)   
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          for (yi = 0; yi < 8; yi++)      
          {
            for (equation = 0; equation < number_equation; equation++)
            {
              filter_pos_y1 = FindPosition (cur->height, y + yi, equation, mvy, temp_inter_pel); 
              for (i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
                NxN_Matrix[sub_pos][equation][i] += b_scaling*
                  RecY_FullPel[filter_pos_y1*cur->width+filter_pos_x1]*
                  RecY_FullPel[filter_pos_y2*cur->width+filter_pos_x1];    
              }
              N_Vector[sub_pos][equation] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)]*
                RecY_FullPel[filter_pos_y1*cur->width+filter_pos_x1];                
            }
          }
        }
      }
      else if (sub_pos == h_pos)
      {
        for (xj = 0; xj < 8; xj++)   
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          for (yi = 0; yi < 8; yi++) 
          {
            for (equation = 0; equation < number_equation; equation++)
            {
              filter_pos_y1 = FindPosition (cur->height, y + yi, equation, mvy, temp_inter_pel); 
              eq_new = TwoDEquationPattern[b_pos][equation]; 
              for (i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
                i_new = TwoDEquationPattern[b_pos][i]; 
                NxN_Matrix[sub_pos][eq_new][i_new] += b_scaling*
                  RecY_FullPel[filter_pos_y1*cur->width+filter_pos_x1]*
                  RecY_FullPel[filter_pos_y2*cur->width+filter_pos_x1];                
              }
              N_Vector[sub_pos][eq_new] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)]*
                RecY_FullPel[filter_pos_y1*cur->width+filter_pos_x1];                
            }
          }
        }
      }
      else if (sub_pos == l_pos)
      {
        for (xj = 0; xj < 8; xj++)  
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          for (yi = 0; yi < 8; yi++)  
          {
            for (equation = 0; equation < number_equation; equation++)
            {
              filter_pos_y1 = FindPosition (cur->height, y + yi, equation, mvy, temp_inter_pel); 
              for (i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
                NxN_Matrix[d_pos][number_equation-equation-1][FILTER_SIZE-i-1] += b_scaling*
                  RecY_FullPel[filter_pos_y1*cur->width+filter_pos_x1]*
                  RecY_FullPel[filter_pos_y2*cur->width+filter_pos_x1];    
              }
              N_Vector[d_pos][number_equation-equation-1] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)]*
                RecY_FullPel[filter_pos_y1*cur->width+filter_pos_x1];                
            }
          }
        }
      }
      else if (sub_pos > l_pos)   
      {
        for (xj = 0; xj < 8; xj++)    
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          filter_pos_x1 = (filter_pos_x1 << 2) + ((sub_pos + 1) % 4); 
          for (yi = 0; yi < 8; yi++)   
          {
            for (equation = 0; equation < number_equation; equation++)
            {
              filter_pos_y1 = FindPosition (cur->height, y + yi, equation, mvy, temp_inter_pel); 
              for (i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
                NxN_Matrix[sub_pos-8][number_equation-equation-1][FILTER_SIZE-i-1] += b_scaling*
                  RecY[(filter_pos_y1+border)*width4border+filter_pos_x1+border4]/32.0*
                  RecY[(filter_pos_y2+border)*width4border+filter_pos_x1+border4]/32.0;    
              }
              N_Vector[sub_pos-8][number_equation-equation-1] += b_scaling*cur->data[(y+yi)*cur->width+(x+xj)]*
                RecY[(filter_pos_y1+border)*width4border+filter_pos_x1+border4]/32.0;                
            }
          }
        }
      }
      else if (sub_pos == i_pos || sub_pos == j_pos || sub_pos == k_pos)
      {
        for (xj = 0; xj < 8; xj++) 
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          filter_pos_x1 = (filter_pos_x1 << 2) + ((sub_pos + 1) % 4); 

          for (yi = 0; yi < 8; yi++)     
          {
            for (equation = 0; equation < number_equation; equation++)
            {
              filter_pos_y1 = FindPosition (cur->height, y + yi, equation, mvy, temp_inter_pel); 
              eq_new = TwoDEquationPattern[b_pos][equation];  
              for (i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
                i_new = TwoDEquationPattern[b_pos][i]; 
                NxN_Matrix[sub_pos][eq_new][i_new] += b_scaling*
                  RecY[(filter_pos_y1+border)*width4border+filter_pos_x1+border4]/32.0*
                  RecY[(filter_pos_y2+border)*width4border+filter_pos_x1+border4]/32.0;           
              }
              N_Vector[sub_pos][eq_new] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)]*
                RecY[(filter_pos_y1+border)*width4border+filter_pos_x1+border4]/32.0;                
            }
          }
        }
      }
      else if (sub_pos == e_pos || sub_pos == f_pos || sub_pos == g_pos)
      {
        for (xj = 0; xj < 8; xj++)  
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          filter_pos_x1 = (filter_pos_x1 << 2) + ((sub_pos + 1) % 4);  
          for (yi = 0; yi < 8; yi++)  
          {
            for (equation = 0; equation < number_equation; equation++)
            {
              filter_pos_y1 = FindPosition (cur->height, y + yi, equation, mvy, temp_inter_pel); 
              for (i = 0; i < FILTER_SIZE; i++)
              {
                filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
                NxN_Matrix[sub_pos][equation][i] += b_scaling*
                  RecY[(filter_pos_y1+border)*width4border+filter_pos_x1+border4]/32.0*
                  RecY[(filter_pos_y2+border)*width4border+filter_pos_x1+border4]/32.0;                      
              }
              N_Vector[sub_pos][equation] += b_scaling*
                cur->data[(y+yi)*cur->width+(x+xj)]*
                RecY[(filter_pos_y1+border)*width4border+filter_pos_x1+border4]/32.0;                
            }
          }
        }
      }
      // one more coefficient to solve to get filter offset
      if(sub_pos == d_pos || sub_pos == l_pos) 
      {
        for (xj = 0; xj < 8; xj++)  
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          for (yi = 0; yi < 8; yi++)  
          {
            for (i = 0; i < FILTER_SIZE; i++)
            {
              filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
              NxN_Matrix[sub_pos][FILTER_SIZE][i] += b_scaling*
                RecY_FullPel[filter_pos_y2*cur->width+filter_pos_x1];    
              NxN_Matrix[sub_pos][i][FILTER_SIZE] += b_scaling*
                RecY_FullPel[filter_pos_y2*cur->width+filter_pos_x1]; 
            }
            NxN_Matrix[sub_pos][FILTER_SIZE][FILTER_SIZE] += 1.; 
            N_Vector[sub_pos][FILTER_SIZE] += b_scaling*
              cur->data[(y+yi)*cur->width+(x+xj)];                
          }
        }
      }
      else if(sub_pos == h_pos)
      {
        for (xj = 0; xj < 8; xj++)   
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          for (yi = 0; yi < 8; yi++) 
          {
            for (i = 0; i < FILTER_SIZE; i++)
            {
              filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
              i_new = TwoDEquationPattern[b_pos][i];
              eq_new = 3;
              NxN_Matrix[sub_pos][eq_new][i_new] += b_scaling*
                RecY_FullPel[filter_pos_y2*cur->width+filter_pos_x1]; 
              NxN_Matrix[sub_pos][i_new][eq_new] += b_scaling*
                RecY_FullPel[filter_pos_y2*cur->width+filter_pos_x1];                     
            }
            NxN_Matrix[sub_pos][eq_new][eq_new] += 1.; 
            N_Vector[sub_pos][eq_new] += b_scaling*
              cur->data[(y+yi)*cur->width+(x+xj)];                
          }
        }
      }
      else if(sub_pos == i_pos || sub_pos == j_pos || sub_pos == k_pos)
      {
        for (xj = 0; xj < 8; xj++)     
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          filter_pos_x1 = (filter_pos_x1 << 2) + ((sub_pos + 1) % 4); 
          for (yi = 0; yi < 8; yi++)      
          {
            for (i = 0; i < FILTER_SIZE; i++)
            {
              filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
              i_new = TwoDEquationPattern[b_pos][i];
              eq_new = 3;

              NxN_Matrix[sub_pos][eq_new][i_new] += b_scaling*
                RecY[(filter_pos_y2+border)*width4border+filter_pos_x1+border4]/32.0;     
              NxN_Matrix[sub_pos][i_new][eq_new] += b_scaling*
                RecY[(filter_pos_y2+border)*width4border+filter_pos_x1+border4]/32.0;                     
            }
            NxN_Matrix[sub_pos][eq_new][eq_new] += 1.; 
            N_Vector[sub_pos][eq_new] += b_scaling*
              cur->data[(y+yi)*cur->width+(x+xj)];                
          }
        }
      }
      else if(sub_pos > d_pos) 
      {
        for (xj = 0; xj < 8; xj++)   
        {
          filter_pos_x1 = FindPosition (cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel); 
          filter_pos_x1 = (filter_pos_x1 << 2) + ((sub_pos + 1) % 4);  
          for (yi = 0; yi < 8; yi++)      
          {
            for (i = 0; i < FILTER_SIZE; i++)
            {
              filter_pos_y2 = FindPosition (cur->height, y + yi, i, mvy, temp_inter_pel); 
              NxN_Matrix[sub_pos][FILTER_SIZE][i] += b_scaling*
                RecY[(filter_pos_y2+border)*width4border+filter_pos_x1+border4]/32.0;      
              NxN_Matrix[sub_pos][i][FILTER_SIZE] += b_scaling*
                RecY[(filter_pos_y2+border)*width4border+filter_pos_x1+border4]/32.0;                     
            }
            NxN_Matrix[sub_pos][FILTER_SIZE][FILTER_SIZE] += 1.; 
            N_Vector[sub_pos][FILTER_SIZE] += b_scaling*
              cur->data[(y+yi)*cur->width+(x+xj)];                
          }
        }
      }
    } 
  }
  param.enable_intcomp=param_ic_full;
}

/*!
************************************************************************
* \brief
*    calculation of filter coefficients for horizontal filter 
************************************************************************
*/
void AIF_adaptive_filter::FindFilterCoefHor (void)
{
  int i, j, k, sign;
  double alpha[FILTER_SIZE_OFFSET][FILTER_SIZE_OFFSET]; 
  double beta[FILTER_SIZE_OFFSET][FILTER_SIZE_OFFSET];
  double y[FILTER_SIZE_OFFSET];
  int N = FILTER_SIZE_OFFSET;
  int sub_pos;
  int start_pos, end_pos;
  double CalculatedFilter[15][FILTER_SIZE_OFFSET]; 
  start_pos = a_pos;
  end_pos = c_pos;

  for (sub_pos = start_pos; sub_pos <= end_pos; sub_pos++)
  {
    N = (sub_pos == b_pos) ? FILTER_SIZE>>1 : FILTER_SIZE;
    N++;  
    //1.Step : LU Decomposition
    //initialize the ii-components
    for (j = 0; j < N; j++)
    {
      for (i = 0; i < N; i++)
      {
        beta[i][j] = 0.0;
        alpha[i][j] = i == j ? 1.0 : 0.0;
      }
      y[j] = 0.0;
      CalculatedFilter[sub_pos][j] = 0.0;
    }
    for (j = 0; j < N; j++)
    {
      for (i = 0; i <= j; i++)
      {
        beta[i][j] = NxN_Matrix[sub_pos][i][j];
        for (k = 0; k < i; k++)
        {
          beta[i][j] -= alpha[i][k] * beta[k][j];
        }
      }
      if (fabs (beta[j][j]) < 0.000001)
      {
        for (i = 0; i <= FILTER_SIZE; i++) 
        {
          FilterCoef[sub_pos][i] = STANDARD_2D_FILTER_SEP[sub_pos][i];
          iFilterCoef_SAIF[sub_pos][i] = iSTANDARD_2D_FILTER_SEP[sub_pos][i];
        }
        continue; 
      }
      else
      {
        for (i = j + 1; i < N; i++)
        {
          alpha[i][j] = NxN_Matrix[sub_pos][i][j];
          for (k = 0; k < j; k++)
          {
            alpha[i][j] -= alpha[i][k] * beta[k][j];
          }
          alpha[i][j] = alpha[i][j] / beta[j][j];
        }
      }
    }
    y[0] = N_Vector[sub_pos][0];
    for (i = 1; i < N; i++)
    {
      y[i] = N_Vector[sub_pos][i];
      for (j = 0; j < i; j++)
      {
        y[i] -= alpha[i][j] * y[j];
      }
    }
    //3. Step: Solving U*x = y, backsubstitution
    CalculatedFilter[sub_pos][N - 1] = y[N - 1] / beta[N - 1][N - 1];
    for (i = N - 2; i >= 0; i--)
    {
      CalculatedFilter[sub_pos][i] = y[i];
      for (j = i + 1; j < N; j++)
        CalculatedFilter[sub_pos][i] -= beta[i][j] * CalculatedFilter[sub_pos][j];
      CalculatedFilter[sub_pos][i] = CalculatedFilter[sub_pos][i] / beta[i][i];
    }
    if (sub_pos == b_pos)
    {
      FilterCoef[b_pos][0] = CalculatedFilter[b_pos][0];
      FilterCoef[b_pos][1] = CalculatedFilter[b_pos][1];
      FilterCoef[b_pos][2] = CalculatedFilter[b_pos][2];
      FilterCoef[b_pos][6] = CalculatedFilter[b_pos][3];
      FilterCoef[b_pos][3] = CalculatedFilter[b_pos][2];
      FilterCoef[b_pos][4] = CalculatedFilter[b_pos][1];
      FilterCoef[b_pos][5] = CalculatedFilter[b_pos][0];
    }
    else
    {
      for (i = 0; i <= FILTER_SIZE; i++)
        FilterCoef[sub_pos][i] = CalculatedFilter[sub_pos][i];
    }
    for (i = 0; i <= FILTER_SIZE; i++)
    {
      if (FilterFlag_SAIF[sub_pos])
      {
        sign = FilterCoef[sub_pos][i] >= 0 ? 1:-1;
        iFilterCoef_SAIF[sub_pos][i]= sign*((int)(fabs(FilterCoef[sub_pos][i]) * (1 << F_BITS) + 0.5)); 
      }
    }
  }
  PredictFilterCoefficientsHor ();    
}

/*!
************************************************************************
* \brief
*    calculation of filter coefficients for vertical filter 
************************************************************************
*/
void AIF_adaptive_filter::FindFilterCoefVer (void)
{
  int i, j, k, sign;
  double alpha[FILTER_SIZE_OFFSET][FILTER_SIZE_OFFSET]; 
  double beta[FILTER_SIZE_OFFSET][FILTER_SIZE_OFFSET];
  double y[FILTER_SIZE_OFFSET];
  int N = FILTER_SIZE_OFFSET;
  int sub_pos;
  int start_pos, end_pos, incr;
  double CalculatedFilter[15][FILTER_SIZE_OFFSET]; 
  start_pos = d_pos;
  end_pos = k_pos;
  incr = 1;
  for (sub_pos = start_pos; sub_pos <= end_pos; sub_pos+=incr)
  {
    if (sub_pos == h_pos || sub_pos == i_pos || sub_pos == j_pos || sub_pos == k_pos)
      N = FILTER_SIZE>>1;          
    else
      N = FILTER_SIZE; 
    N++;  
    //1.Step : LU Decomposition
    //initialize the ii-components
    for (j = 0; j < N; j++)
    {
      for (i = 0; i < N; i++)
      {
        beta[i][j] = 0.0;
        alpha[i][j] = i == j ? 1.0 : 0.0;
      }
      y[j] = 0.0;
      CalculatedFilter[sub_pos][j] = 0.0;
    }
    for (j = 0; j < N; j++)
    {
      for (i = 0; i <= j; i++)
      {
        beta[i][j] = NxN_Matrix[sub_pos][i][j];
        for (k = 0; k < i; k++)
          beta[i][j] -= alpha[i][k] * beta[k][j];
      }
      if (fabs (beta[j][j]) < 0.000001)
      {
        for (i = 0; i <= FILTER_SIZE; i++) 
        {
          FilterCoef[sub_pos][i] = STANDARD_2D_FILTER_SEP[sub_pos][i];
          iFilterCoef_SAIF[sub_pos][i] = iSTANDARD_2D_FILTER_SEP[sub_pos][i];
        }
        continue;
      }
      else
      {
        for (i = j + 1; i < N; i++)
        {
          alpha[i][j] = NxN_Matrix[sub_pos][i][j];
          for (k = 0; k < j; k++)
            alpha[i][j] -= alpha[i][k] * beta[k][j];
          alpha[i][j] = alpha[i][j] / beta[j][j];
        }
      }
    }
    y[0] = N_Vector[sub_pos][0];
    for (i = 1; i < N; i++)
    {
      y[i] = N_Vector[sub_pos][i];
      for (j = 0; j < i; j++)
        y[i] -= alpha[i][j] * y[j];
    }
    //3. Step: Solving U*x = y, backsubstitution
    CalculatedFilter[sub_pos][N - 1] = y[N - 1] / beta[N - 1][N - 1];
    for (i = N - 2; i >= 0; i--)
    {
      CalculatedFilter[sub_pos][i] = y[i];
      for (j = i + 1; j < N; j++)
        CalculatedFilter[sub_pos][i] -= beta[i][j] * CalculatedFilter[sub_pos][j];
      CalculatedFilter[sub_pos][i] = CalculatedFilter[sub_pos][i] / beta[i][i];
    }
    if (sub_pos == h_pos || sub_pos == i_pos || sub_pos == j_pos || sub_pos == k_pos)
    {
      FilterCoef[sub_pos][0] = CalculatedFilter[sub_pos][0];
      FilterCoef[sub_pos][1] = CalculatedFilter[sub_pos][1];
      FilterCoef[sub_pos][2] = CalculatedFilter[sub_pos][2];
      FilterCoef[sub_pos][6] = CalculatedFilter[sub_pos][3];
      FilterCoef[sub_pos][3] = CalculatedFilter[sub_pos][2];
      FilterCoef[sub_pos][4] = CalculatedFilter[sub_pos][1];
      FilterCoef[sub_pos][5] = CalculatedFilter[sub_pos][0];
    }
    else
    {
      for (i = 0; i <= FILTER_SIZE; i++)
        FilterCoef[sub_pos][i] = CalculatedFilter[sub_pos][i]; 
    }
  }
  for (sub_pos = l_pos; sub_pos <= o_pos; sub_pos++)
  {
    for (i = 0; i < FILTER_SIZE; i++) 
      FilterCoef[sub_pos][FILTER_SIZE-i-1] = FilterCoef[sub_pos-8][i];
    FilterCoef[sub_pos][FILTER_SIZE] = CalculatedFilter[sub_pos-8][FILTER_SIZE];
    FilterFlag_SAIF[sub_pos] = FilterFlag_SAIF[sub_pos-8]; 
  }
  for (sub_pos = d_pos; sub_pos <= o_pos; sub_pos++)
  {
    for (i = 0; i <= FILTER_SIZE; i++)
    {
      if (FilterFlag_SAIF[sub_pos])
      {
        sign = FilterCoef[sub_pos][i] >= 0. ? 1:-1;
        iFilterCoef_SAIF[sub_pos][i]= sign*((int)(fabs(FilterCoef[sub_pos][i]) * (1 << F_BITS) + 0.5));
      }
    }
  }
  PredictFilterCoefficientsVer ();  
}

/*!
************************************************************************
* \brief
*    predict filter coefficients for horizontal filter
*
************************************************************************
*/
void AIF_adaptive_filter::PredictFilterCoefficientsHor (void)      
{
  int i, pos;
  int iDiffFilterCoef_SAIF[15][7];  
  int start_pos, end_pos;
  start_pos = a_pos;
  end_pos = c_pos;
  for (pos = start_pos; pos <= end_pos; pos++)
  {
    if(FilterFlag_SAIF[pos])
    {
      for (i = 0; i <= FILTER_SIZE; i++)
        iDiffFilterCoef_SAIF[pos][i] = iFilterCoef_SAIF[pos][i] - iSTANDARD_2D_FILTER_SEP[pos][i];
      QuantizeFilterCoefficients (iDiffFilterCoef_SAIF[pos], DiffQFilterCoef_SAIF[pos]);
      for (i = 0; i <= FILTER_SIZE; i++)
        iFilterCoef_SAIF[pos][i] = iDiffFilterCoef_SAIF[pos][i] + iSTANDARD_2D_FILTER_SEP[pos][i];
    }
    else 
      for (i = 0; i <= FILTER_SIZE; i++)
        iFilterCoef_SAIF[pos][i]=iSTANDARD_2D_FILTER_SEP[pos][i];
  }
}

/*!
************************************************************************
* \brief
*    predict filter coefficients for vertical filter
*
************************************************************************
*/
void AIF_adaptive_filter::PredictFilterCoefficientsVer (void)     
{
  int i, pos;
  int iDiffFilterCoef_SAIF[15][7];  
  int start_pos, end_pos, incr;
  start_pos = d_pos;
  end_pos = o_pos;
  incr = 1;
  for (pos = start_pos; pos <= end_pos; pos+=incr)
  {
    if(FilterFlag_SAIF[pos])
    {
      for (i = 0; i <= FILTER_SIZE; i++)
        iDiffFilterCoef_SAIF[pos][i] = iFilterCoef_SAIF[pos][i] - iSTANDARD_2D_FILTER_SEP[pos][i];
      QuantizeFilterCoefficients (iDiffFilterCoef_SAIF[pos], DiffQFilterCoef_SAIF[pos]);
      for (i = 0; i <= FILTER_SIZE; i++)
        iFilterCoef_SAIF[pos][i] = iDiffFilterCoef_SAIF[pos][i] + iSTANDARD_2D_FILTER_SEP[pos][i];
    }
    else
      for (i = 0; i <= FILTER_SIZE; i++)
        iFilterCoef_SAIF[pos][i]=iSTANDARD_2D_FILTER_SEP[pos][i];
  }
}
/*!
************************************************************************
* \brief
*    Horizontal upsampling with new calculated filter coefficients.
************************************************************************
*/
void AIF_adaptive_filter::InterpolateHor (RefFrameManager *ref_mgr)
{
  int i, j, jj, l, ref;
  int *out4Y_aif_hor; 
  unsigned char *refY;
  int img_width, img_width4, img_height;
  int sub_pos, pos_x;
  int is;
  int border = TWO * MOTION_BLK_SIZE;
  int border4 = border << 2;
  int shift = 5;
  int f_bits_minus_shift = F_BITS - shift;
  int f_bits_minus_shift_minus_1 = F_BITS - shift - 1;
  int f_bits_minus_1 = F_BITS - 1;
  int shift_f_bits_minus_shift_minus_1 = (1 << (f_bits_minus_shift_minus_1));
  
  for(l=0;l<LISTS;l++)
  {
    for(ref=0;ref<MAX_REF_FRAMES;ref++)
    {
      if (ref_mgr->ref[ref_mgr->list[l][ref]]->empty != TRUE)
      {
        img_width     = ref_mgr->AIF_ref[ref_mgr->list[l][ref]]->frame->width;
        img_width4    = img_width<<2;
        img_height    = ref_mgr->AIF_ref[ref_mgr->list[l][ref]]->frame->height;
        out4Y_aif_hor = ref_mgr->AIF_ref[ref_mgr->list[l][ref]]->luma_hor; 
        refY          = ref_mgr->AIF_ref[ref_mgr->list[l][ref]]->frame->data;
        border = TWO * MOTION_BLK_SIZE;
        border4 = border << 2;
        for (i = -border; i < border + img_height; i++)
        {
          for (j = -border4; j < border4 + img_width4; j++) 
          {
            sub_pos = (j + border4) % 4;  
            is = 0;
            if (sub_pos)
            {
              for (jj = 0; jj < FILTER_SIZE; jj++)
              {
                pos_x = max (0, min (img_width - 1, j / 4 - FILTER_OFFSET + jj));
                if (j < 0 && pos_x > 0)
                  pos_x -= 1;

                if (AIF_mode_flag1 || (sub_pos - 1 == b_pos))
                  is += (iFilterCoef_SAIF[sub_pos - 1][jj] * refY[(max (0, min (img_height - 1, i)))*img_width + pos_x]);
                else
                  is += ((iFilterCoef_SAIF[b_pos][jj] * refY[(max (0, min (img_height - 1, i)))*img_width + pos_x] + 1) >> 1);
              }
              if (AIF_mode_flag1 || (sub_pos - 1 == b_pos))
              {
                is += iFilterCoef_SAIF[sub_pos-1][FILTER_SIZE];
                out4Y_aif_hor[(i + border) * (2*border4 + img_width4) + j + border4] = (is + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 
              }
              else
              {
                is += ((iFilterCoef_SAIF[b_pos][FILTER_SIZE] + 1) >> 1);
                if (sub_pos - 1 == a_pos)
                  is += (refY[(max (0, min (img_height - 1, i)))*img_width + max (0, min (img_width - 1, j / 4))] << f_bits_minus_1);
                else 
                  is += (refY[(max (0, min (img_height - 1, i)))*img_width + max (0, min (img_width - 1, j / 4 + 1))] << f_bits_minus_1);
                out4Y_aif_hor[(i + border) * (2*border4 + img_width4) + j + border4] = (is + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 
              }
            }
            else
              out4Y_aif_hor[(i + border) * (2*border4 + img_width4) + j + border4] = 
                refY[(max (0, min (img_height - 1, i)))*img_width+(max (0, min (img_width - 1, j / 4)))] << shift;      
          }
        }
      }
    }
  }
}

/*!
************************************************************************
* \brief
*    sets necessary vertical filter data for current macroblock
************************************************************************
*/
void AIF_adaptive_filter::SAIF_HALF_rd_decision_Hor(Frame *cur, MotionField *mf, RefFrameManager *ref_mgr)
{
  int x, y, xj, yi, ii, jj, fi, i, p;
  float b_scaling=1.0;
  int param_ic_full;
  int run=0, rerun=0;
  int mvx=0,  mvy=0;
  int mvxF=0, mvyF=0;
  int mvxB=0, mvyB=0;
  int mvxF_sub=0, mvyF_sub=0;
  int mvxB_sub=0, mvyB_sub=0;
  int scaleF=0, offsetF=0;     
  int scaleB=0, offsetB=0;     
  int scale, offset;
  int sub_pos=0;
  int sub_posF=0;
  int sub_posB=0;
  int ref_frame=0;
  int ref_frameF=-1;
  int ref_frameB=-1;
  int pos_y = 0, pos_x = 0;
  double cost_mode1, cost_mode2;
  int diff_blk[2][64];
  int pred_blk[2][64];
  double SATD[15][2];
  double nr8x8blocks = (double)mf->height*(double)mf->width;
  double scalingThres = 6240.0;
  int temp_inter_pel;
  int x_orig, y_orig;       
  int h1, h2;
  int *new_eq, *new_filter_pos;
  int *RecY=NULL; 
  int *RecYF=NULL; 
  int *RecYB=NULL; 
  unsigned char *RecY_FullPel = NULL;
  unsigned char *RecYF_FullPel = NULL;
  unsigned char *RecYB_FullPel = NULL;
  int coef_rate[15];
  int block_cnt[15];
  PredBlk *curblk;
  double pix_mode1, pix_mode2;
  int mel_precision = 8;   
  int mel_shift = mel_precision + 2 - (cur->melq >> 4);
  EntropyEncoder *eenc_aif;
  Buffer         *eenc_aif_buf;
  eenc_aif = new EntropyEncoder; 
  eenc_aif_buf = new Buffer;
  eenc_aif->reset(eenc_aif_buf);
  new_filter_pos = &h1;
  new_eq = &h2;  
  param_ic_full=param.enable_intcomp;
  param.enable_intcomp=0;

  for (sub_pos = 0; sub_pos < 15; sub_pos ++)
  {
    SATD[sub_pos][0] = SATD[sub_pos][1] = 0.0;
    block_cnt[sub_pos] = coef_rate[sub_pos] = 0;
  }
  for (jj=0; jj < mf->height; jj++) 
  { 
    for (ii=0; ii < mf->width; ii++) 
    {
      curblk=mf->get_pblk(jj,ii);
      if(curblk->intra==0)
      {
        x_orig = MOTION_BLK_SIZE*ii;
        y_orig = MOTION_BLK_SIZE*jj;
        x = x_orig;
        y = y_orig;
    
        mvxF=curblk->mv[LIST_0][horizontal];
        mvyF=curblk->mv[LIST_0][vertical];
    
        if(param.enable_intcomp)
        {
          scaleF=curblk->scale[LIST_0];
          offsetF=curblk->DC[LIST_0];
        }
    
        temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_0][curblk->ref_index[LIST_0]];
        if(mvxF >= 0)
          mvxF_sub = (mvxF%temp_inter_pel);      // x-sub-coordinate in a 4x4block 
        else
          mvxF_sub = ((temp_inter_pel-abs(mvxF)%temp_inter_pel)%temp_inter_pel);
        if(mvyF >= 0)
          mvyF_sub = (mvyF%temp_inter_pel);      // y-sub-coordinate in a 4x4block
        else
          mvyF_sub = ((temp_inter_pel-abs(mvyF)%temp_inter_pel)%temp_inter_pel);
    
        if (temp_inter_pel == PEL_ONE_EIGHTH) 
        {   
          mvxF_sub = mvxF_sub>>1;
          mvyF_sub = mvyF_sub>>1;
        }
        else if (temp_inter_pel == PEL_ONE_HALF) 
        {   
          mvxF_sub = mvxF_sub<<1;
          mvyF_sub = mvyF_sub<<1;
        }    
        sub_posF = mvxF_sub + 4*mvyF_sub-1;    // pos 0..14 in a 4x4 block 
        ref_frameF=ref_mgr->list[LIST_0][curblk->ref_index[LIST_0]];

        if(ref_frameF != -1)
        {
          RecYF = ref_mgr->AIF_ref[ref_frameF]->luma_hor;
          RecYF_FullPel = ref_mgr->AIF_ref[ref_frameF]->frame->data;
        }
        if(cur->frame_type == ft_B_frame)
        {
          mvxB=curblk->mv[LIST_1][horizontal];
          mvyB=curblk->mv[LIST_1][vertical];
          if(param.enable_intcomp)
          {
            scaleB=curblk->scale[LIST_1];
            offsetB=curblk->DC[LIST_1];
          }    
          temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_1][curblk->ref_index[LIST_1]];
          if(mvxB >= 0)
            mvxB_sub = (mvxB%temp_inter_pel);      // x-sub-coordinate in a 4x4block 
          else
            mvxB_sub = ((temp_inter_pel-abs(mvxB)%temp_inter_pel)%temp_inter_pel);
          if(mvyB >= 0)
            mvyB_sub = (mvyB%temp_inter_pel);      // y-sub-coordinate in a 4x4block
          else
            mvyB_sub = ((temp_inter_pel-abs(mvyB)%temp_inter_pel)%temp_inter_pel);
      
          if (temp_inter_pel == PEL_ONE_EIGHTH) 
          {   
            mvxB_sub = mvxB_sub>>1;
            mvyB_sub = mvyB_sub>>1;
          }
          else if (temp_inter_pel == PEL_ONE_HALF) 
          {   
            mvxB_sub = mvxB_sub<<1;
            mvyB_sub = mvyB_sub<<1;
          }      
          sub_posB = mvxB_sub + 4*mvyB_sub-1;    // pos 0..14 in a 4x4 block 
          ref_frameB=ref_mgr->list[LIST_1][curblk->ref_index[LIST_1]];

          if(ref_frameB != -1)
          {
            RecYB = ref_mgr->AIF_ref[ref_frameB]->luma_hor;
            RecYB_FullPel = ref_mgr->AIF_ref[ref_frameB]->frame->data;
          }
          rerun=1;
        }
        else
          rerun=0;
    
        if(ref_frameF != -1 && ref_frameB != -1)
          b_scaling = 0.5;
        else
          b_scaling = 1.0;

        for (p = 0; p < 64; p ++)
        {
          diff_blk[0][p] = diff_blk[1][p] = 0;
          pred_blk[0][p] = pred_blk[1][p] = 0;
        }

        for(run=0; run<= rerun; run++)
        {
          if(run==0)
          {
            sub_pos=sub_posF;
            mvx=mvxF;
            mvy=mvyF;
            ref_frame=ref_frameF;
            RecY = RecYF;
            RecY_FullPel = RecYF_FullPel;
            scale=scaleF;
            offset=offsetF;
            temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_0][curblk->ref_index[LIST_0]];
          }
          else
          {
            sub_pos=sub_posB;
            mvx=mvxB;
            mvy=mvyB;
            ref_frame=ref_frameB;
            RecY = RecYB;
            RecY_FullPel = RecYB_FullPel;
            scale=scaleB;
            offset=offsetB;
            temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_1][curblk->ref_index[LIST_1]];
          }      
          if(sub_pos != -1 && sub_pos < d_pos && ref_frame != -1) 
          {
            for (xj = 0; xj < 8; xj++)   
            {
              for (yi = 0; yi < 8; yi++)
              {         
                pix_mode1 = pix_mode2 = 0.0;
                {
                  pos_y = FindPosition (cur->height, y + yi, FILTER_OFFSET, mvy, temp_inter_pel); 
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_x = FindPosition (cur->width, x + xj, fi, mvx, temp_inter_pel); 
                    pix_mode2 += FilterCoef[sub_pos][fi]*RecY_FullPel[pos_y*cur->width+pos_x];
                    if (sub_pos == a_pos)
                    {
                      if (fi == 2)
                        pix_mode1 += (FilterCoef[b_pos][fi] * 0.5 + 0.5) * RecY_FullPel[pos_y*cur->width+pos_x];  
                      else
                        pix_mode1 += FilterCoef[b_pos][fi] * 0.5 * RecY_FullPel[pos_y*cur->width+pos_x]; 
                    }
                    else if (sub_pos == b_pos)
                    {
                      pix_mode1 += FilterCoef[sub_pos][fi]*RecY_FullPel[pos_y*cur->width+pos_x];
                    }
                    else if (sub_pos == c_pos)
                    {
                      if (fi == 3)
                        pix_mode1 += (FilterCoef[b_pos][fi] * 0.5 + 0.5) * RecY_FullPel[pos_y*cur->width+pos_x];  
                      else
                        pix_mode1 += FilterCoef[b_pos][fi] * 0.5 * RecY_FullPel[pos_y*cur->width+pos_x];
                    }
                  }
                  pix_mode2 += FilterCoef[sub_pos][FILTER_SIZE];
                  if (sub_pos == b_pos)
                    pix_mode1 += FilterCoef[sub_pos][FILTER_SIZE];
                  else if (sub_pos == a_pos || sub_pos == c_pos)
                    pix_mode1 += FilterCoef[sub_pos][FILTER_SIZE] * 0.5;
                }
                pred_blk[1][yi * 8 + xj] += (int)(pix_mode2 * b_scaling + 0.5);
                pred_blk[0][yi * 8 + xj] += (int)(pix_mode1 * b_scaling + 0.5);
              }
            }
          } 
        }
        if(sub_pos >= a_pos && sub_pos < d_pos && ref_frame != -1) 
        { 
          for (xj = 0; xj < 8; xj++)   
          {
            for (yi = 0; yi < 8; yi++)
            {  
              diff_blk[1][yi * 8 + xj] = (cur->data[(y+yi)*cur->width+(x+xj)] - pred_blk[1][yi * 8 + xj]) << SAMPLE_ADD_PREC; 
              diff_blk[0][yi * 8 + xj] = (cur->data[(y+yi)*cur->width+(x+xj)] - pred_blk[0][yi * 8 + xj]) << SAMPLE_ADD_PREC;
            }
          }
          block_cnt[sub_pos] ++;
          SATD[sub_pos][1] += (double)SATD8x8(diff_blk[1], mel_shift);
          SATD[sub_pos][0] += (double)SATD8x8(diff_blk[0], mel_shift);  
        } 
      }
    }
  }
  for (sub_pos = a_pos; sub_pos < d_pos; sub_pos ++)
  {
    if (block_cnt[sub_pos])
    {
      SATD[sub_pos][1] = SATD[sub_pos][1] / (double)block_cnt[sub_pos];
      SATD[sub_pos][0] = SATD[sub_pos][0] / (double)block_cnt[sub_pos];
    }
    else
    {
      SATD[sub_pos][1] = 0.0;
      SATD[sub_pos][0] = 0.0;
    }
  }
  eenc_aif->reset(eenc_aif_buf); 
  for(i=0;i<FILTER_SIZE;i++) // a_pos
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[a_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[a_pos][FILTER_SIZE]); 
  coef_rate[a_pos] = eenc_aif->get_nbits();
  

  eenc_aif->reset(eenc_aif_buf); 
  for(i=0;i<(FILTER_SIZE>>1);i++) // b_pos
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[b_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[b_pos][FILTER_SIZE]); 
  coef_rate[b_pos] = eenc_aif->get_nbits();


  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<FILTER_SIZE;i++) // c_pos
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[c_pos][i]-DiffQFilterCoef_SAIF[a_pos][FILTER_SIZE-i-1]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[c_pos][FILTER_SIZE]); 
  coef_rate[c_pos] = eenc_aif->get_nbits();

  FilterFlag_SAIF[b_pos] = 1;

  // a_pos and c_pos
  if (nr8x8blocks < scalingThres)
  {
    cost_mode1 = SATD[a_pos][0] + SATD[b_pos][0] + SATD[c_pos][0] + ((int)((double)coef_rate[b_pos] * scalingThres/nr8x8blocks + 0.5));
    cost_mode2 = SATD[a_pos][1] + SATD[b_pos][1] + SATD[c_pos][1] + ((int)((double)(coef_rate[a_pos] + coef_rate[b_pos] + coef_rate[c_pos]) * scalingThres/nr8x8blocks + 0.5));  
  }
  else
  {
    cost_mode1 = SATD[a_pos][0] + SATD[b_pos][0] + SATD[c_pos][0] + coef_rate[b_pos];
    cost_mode2 = SATD[a_pos][1] + SATD[b_pos][1] + SATD[c_pos][1] + coef_rate[a_pos] + coef_rate[b_pos] + coef_rate[c_pos];  
  }
  if (cost_mode2 < cost_mode1)
    AIF_mode_flag1 = 1;
  else
    AIF_mode_flag1 = 0;

  if (!AIF_mode_flag1) // AIF mode 1
  {
    FilterFlag_SAIF[b_pos] = FilterFlag_SAIF[h_pos] = FilterFlag_SAIF[j_pos] = 1;

    FilterFlag_SAIF[a_pos] = FilterFlag_SAIF[c_pos] = FilterFlag_SAIF[d_pos] =
    FilterFlag_SAIF[e_pos] = FilterFlag_SAIF[f_pos] = FilterFlag_SAIF[g_pos] =
    FilterFlag_SAIF[i_pos] = FilterFlag_SAIF[k_pos] = FilterFlag_SAIF[l_pos] = 
    FilterFlag_SAIF[m_pos] = FilterFlag_SAIF[n_pos] = FilterFlag_SAIF[o_pos] = 0;
  } 
  else  // AIF mode 2 / 3
    FilterFlag_SAIF[a_pos] = FilterFlag_SAIF[b_pos] = FilterFlag_SAIF[c_pos] = 1;

  param.enable_intcomp=param_ic_full;
  delete eenc_aif_buf;
  delete eenc_aif;
}

/*!
************************************************************************
* \brief
*    sets necessary vertical filter data for current macroblock
************************************************************************
*/
void AIF_adaptive_filter::SAIF_HALF_rd_decision_Ver(Frame *cur, MotionField *mf, RefFrameManager *ref_mgr)
{
  int x, y, xj, yi, ii, jj, fi, i, p;
  float b_scaling=1.0;
  int param_ic_full;
  int run=0, rerun=0;
  int mvx=0,  mvy=0;
  int mvxF=0, mvyF=0;
  int mvxB=0, mvyB=0;
  int mvxF_sub=0, mvyF_sub=0;
  int mvxB_sub=0, mvyB_sub=0;
  int scaleF=0, offsetF=0;     
  int scaleB=0, offsetB=0;     
  int scale, offset;
  int sub_pos=0;
  int sub_posF=0;
  int sub_posB=0;
  int ref_frame=0;
  int ref_frameF=-1;
  int ref_frameB=-1;
  int pos_y = 0, pos_x = 0;
  double cost_mode2, cost_mode3;
  int diff_blk[2][64];
  int pred_blk[2][64];
  double SATD[15][2];
  double nr8x8blocks = (double)mf->height*(double)mf->width;
  double scalingThres = 6240.0;
  int temp_inter_pel;
  int x_orig, y_orig;       
  int number_equation;
  int h1, h2;
  int *new_eq, *new_filter_pos;
  int *RecY=NULL; 
  int *RecYF=NULL; 
  int *RecYB=NULL; 
  unsigned char *RecY_FullPel = NULL;
  unsigned char *RecYF_FullPel = NULL;
  unsigned char *RecYB_FullPel = NULL;
  int border = 16;
  int border4= 64;
  int coef_rate[15];
  int block_cnt[15];
  PredBlk *curblk;
  double pix_mode2, pix_mode3;
  int mel_precision = 8;   
  int mel_shift = mel_precision + 2 - (cur->melq >> 4);
  EntropyEncoder *eenc_aif;
  Buffer         *eenc_aif_buf;
  eenc_aif = new EntropyEncoder; 
  eenc_aif_buf = new Buffer;
  eenc_aif->reset(eenc_aif_buf);
  new_filter_pos = &h1;
  new_eq = &h2;
  param_ic_full=param.enable_intcomp;
  param.enable_intcomp=0;

  for (sub_pos = 0; sub_pos < 15; sub_pos ++)
  {
    SATD[sub_pos][0] = 0.0; 
    SATD[sub_pos][1] = 0.0;
    block_cnt[sub_pos] = 0;
    coef_rate[sub_pos] = 0;
  }

  for (jj=0; jj < mf->height; jj++) 
  { 
    for (ii=0; ii < mf->width; ii++) 
    {
      curblk=mf->get_pblk(jj,ii);
      if(curblk->intra==0)
      {
        x_orig = MOTION_BLK_SIZE*ii;
        y_orig = MOTION_BLK_SIZE*jj;
    
        x = x_orig;
        y = y_orig;
    
        mvxF=curblk->mv[LIST_0][horizontal];
        mvyF=curblk->mv[LIST_0][vertical];
    
        if(param.enable_intcomp)
        {
          scaleF=curblk->scale[LIST_0];
          offsetF=curblk->DC[LIST_0];
        }
    
        temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_0][curblk->ref_index[LIST_0]];
        if(mvxF >= 0)
          mvxF_sub = (mvxF%temp_inter_pel);      // x-sub-coordinate in a 4x4block 
        else
          mvxF_sub = ((temp_inter_pel-abs(mvxF)%temp_inter_pel)%temp_inter_pel);
        if(mvyF >= 0)
          mvyF_sub = (mvyF%temp_inter_pel);      // y-sub-coordinate in a 4x4block
        else
          mvyF_sub = ((temp_inter_pel-abs(mvyF)%temp_inter_pel)%temp_inter_pel);
    
        if (temp_inter_pel == PEL_ONE_EIGHTH) 
        {   
          mvxF_sub = mvxF_sub>>1;
          mvyF_sub = mvyF_sub>>1;
        }
        else if (temp_inter_pel == PEL_ONE_HALF) 
        {   
          mvxF_sub = mvxF_sub<<1;
          mvyF_sub = mvyF_sub<<1;
        }
    
        sub_posF = mvxF_sub + 4*mvyF_sub-1;    
    
        ref_frameF=ref_mgr->list[LIST_0][curblk->ref_index[LIST_0]];

        if(ref_frameF != -1)
        {
          RecYF = ref_mgr->AIF_ref[ref_frameF]->luma_hor;
          RecYF_FullPel = ref_mgr->AIF_ref[ref_frameF]->frame->data;
        }
        if(cur->frame_type == ft_B_frame)
        {
          mvxB=curblk->mv[LIST_1][horizontal];
          mvyB=curblk->mv[LIST_1][vertical];
      
          if(param.enable_intcomp)
          {
            scaleB=curblk->scale[LIST_1];
            offsetB=curblk->DC[LIST_1];
          }
      
          temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_1][curblk->ref_index[LIST_1]];
          if(mvxB >= 0)
            mvxB_sub = (mvxB%temp_inter_pel);      // x-sub-coordinate in a 4x4block 
          else
            mvxB_sub = ((temp_inter_pel-abs(mvxB)%temp_inter_pel)%temp_inter_pel);
          if(mvyB >= 0)
            mvyB_sub = (mvyB%temp_inter_pel);      // y-sub-coordinate in a 4x4block
          else
            mvyB_sub = ((temp_inter_pel-abs(mvyB)%temp_inter_pel)%temp_inter_pel);
      
          if (temp_inter_pel == PEL_ONE_EIGHTH) 
          {   
            mvxB_sub = mvxB_sub>>1;
            mvyB_sub = mvyB_sub>>1;
          }
          else if (temp_inter_pel == PEL_ONE_HALF) 
          {   
            mvxB_sub = mvxB_sub<<1;
            mvyB_sub = mvyB_sub<<1;
          }
      
          sub_posB = mvxB_sub + 4*mvyB_sub-1;    
      
          ref_frameB=ref_mgr->list[LIST_1][curblk->ref_index[LIST_1]];

          if(ref_frameB != -1)
          {
            RecYB = ref_mgr->AIF_ref[ref_frameB]->luma_hor;
            RecYB_FullPel = ref_mgr->AIF_ref[ref_frameB]->frame->data;
          }
          rerun=1;
        }
        else
          rerun=0;
    
        if(ref_frameF != -1 && ref_frameB != -1)
          b_scaling = 0.5;
        else
          b_scaling = 1.0;
 
        for (p = 0; p < 64; p ++)
        {
          diff_blk[0][p] = 0;
          diff_blk[1][p] = 0;
          pred_blk[0][p] = 0;
          pred_blk[1][p] = 0;
        }
        for(run=0; run<= rerun; run++)
        {
          if(run==0)
          {
            sub_pos=sub_posF;
            mvx=mvxF;
            mvy=mvyF;
            ref_frame=ref_frameF;
            RecY = RecYF;
            RecY_FullPel = RecYF_FullPel;
            scale=scaleF;
            offset=offsetF;
            temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_0][curblk->ref_index[LIST_0]];
          }
          else
          {
            sub_pos=sub_posB;
            mvx=mvxB;
            mvy=mvyB;
            ref_frame=ref_frameB;
            RecY = RecYB;
            RecY_FullPel = RecYB_FullPel;
            scale=scaleB;
            offset=offsetB;
            temp_inter_pel = ref_mgr->ref_process->inter_pel[LIST_1][curblk->ref_index[LIST_1]];
          }      
      
          if(sub_pos != -1 && sub_pos > c_pos && ref_frame != -1 && FilterFlag_SAIF[sub_pos]) 
          {
            number_equation = FILTER_SIZE; 
            for (xj = 0; xj < 8; xj++)   
            {
              for (yi = 0; yi < 8; yi++)
              {         
                pix_mode3 = 0.0;
                pix_mode2 = 0.0;
                if (sub_pos == d_pos || sub_pos == h_pos || sub_pos == l_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET,mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[sub_pos][fi]*RecY_FullPel[pos_y*cur->width+pos_x];
                    pix_mode3 += FilterCoef[sub_pos][fi]*RecY_FullPel[pos_y*cur->width+pos_x];
                  }
                  pix_mode2 += FilterCoef[sub_pos][FILTER_SIZE];
                  pix_mode3 += FilterCoef[sub_pos][FILTER_SIZE];
                }
                else if (sub_pos == e_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[h_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4]/32.0; // h_pos
                    pix_mode2 += FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0; // j_pos
                    pix_mode3 += FilterCoef[e_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+1]/32.0;
                  }
                  pix_mode2 += (double)RecY[(pos_y+border-3)*(cur->width*4+2*border4)+pos_x*4+border4]/32.0; // full_pel
                  pix_mode2 += (double)RecY[(pos_y+border-3)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0; // b_pos
                  pix_mode2 += FilterCoef[h_pos][FILTER_SIZE] + FilterCoef[j_pos][FILTER_SIZE]; // offsets
                  pix_mode2 = pix_mode2 / 4.0;
                  pix_mode3 += FilterCoef[e_pos][FILTER_SIZE];
                }
                else if (sub_pos == f_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0;   
                    pix_mode3 += FilterCoef[f_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0;
                  }
                  pix_mode2 += (double)RecY[(pos_y+border-3)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0 + FilterCoef[j_pos][FILTER_SIZE];
                  pix_mode2 = pix_mode2 / 2.0;
                  pix_mode3 += FilterCoef[f_pos][FILTER_SIZE];
                }
                else if (sub_pos == g_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[h_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+4]/32.0; // h_pos
                    pix_mode2 += FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0; // j_pos
                    pix_mode3 += FilterCoef[g_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+3]/32.0;
                  }
                  pix_mode2 += (double)RecY[(pos_y+border-3)*(cur->width*4+2*border4)+pos_x*4+border4+4]/32.0; // full_pel
                  pix_mode2 += (double)RecY[(pos_y+border-3)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0; // b_pos
                  pix_mode2 += FilterCoef[h_pos][FILTER_SIZE] + FilterCoef[j_pos][FILTER_SIZE]; // offsets
                  pix_mode2 = pix_mode2 / 4.0;
                  pix_mode3 += FilterCoef[g_pos][FILTER_SIZE];
                }
                else if (sub_pos == i_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[h_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4]/32.0 +
                                 FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0;   
                    pix_mode3 += FilterCoef[i_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+1]/32.0;
                  }
                  pix_mode2 += FilterCoef[h_pos][FILTER_SIZE] + FilterCoef[j_pos][FILTER_SIZE];
                  pix_mode2 = pix_mode2 / 2;
                  pix_mode3 += FilterCoef[i_pos][FILTER_SIZE];
                }
                else if (sub_pos == j_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0;
                    pix_mode3 += FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0;
                  }
                  pix_mode2 += FilterCoef[j_pos][FILTER_SIZE];
                  pix_mode3 += FilterCoef[j_pos][FILTER_SIZE];
                }
                else if (sub_pos == k_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[h_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+4]/32.0 +
                                 FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0;   
                    pix_mode3 += FilterCoef[k_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+3]/32.0;
                  }
                  pix_mode2 += FilterCoef[h_pos][FILTER_SIZE] + FilterCoef[j_pos][FILTER_SIZE];
                  pix_mode2 = pix_mode2 / 2;
                  pix_mode3 += FilterCoef[k_pos][FILTER_SIZE];
                }
                else if (sub_pos == m_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[h_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4]/32.0; // h_pos
                    pix_mode2 += FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0; // j_pos
                    pix_mode3 += FilterCoef[m_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+1]/32.0;
                  }
                  pix_mode2 += (double)RecY[(pos_y+border-2)*(cur->width*4+2*border4)+pos_x*4+border4]/32.0; // full_pel
                  pix_mode2 += (double)RecY[(pos_y+border-2)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0; // b_pos
                  pix_mode2 += FilterCoef[h_pos][FILTER_SIZE] + FilterCoef[j_pos][FILTER_SIZE]; // offsets
                  pix_mode2 = pix_mode2 / 4.0;
                  pix_mode3 += FilterCoef[m_pos][FILTER_SIZE];
                }
                else if (sub_pos == n_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0;   
                    pix_mode3 += FilterCoef[n_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0;
                  }
                  pix_mode2 += (double)RecY[(pos_y+border-2)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0 + FilterCoef[j_pos][FILTER_SIZE];
                  pix_mode2 = pix_mode2 / 2;
                  pix_mode3 += FilterCoef[n_pos][FILTER_SIZE];
                }
                else if (sub_pos == o_pos)
                {
                  pos_x = FindPosition(cur->width, x + xj, FILTER_OFFSET, mvx, temp_inter_pel);
                  for(fi = 0; fi < FILTER_SIZE; fi++)
                  {
                    pos_y = FindPosition (cur->height, y + yi, fi, mvy, temp_inter_pel); 
                    pix_mode2 += FilterCoef[h_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+4]/32.0; // h_pos
                    pix_mode2 += FilterCoef[j_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0; // j_pos
                    pix_mode3 += FilterCoef[o_pos][fi]*(double)RecY[(pos_y+border)*(cur->width*4+2*border4)+pos_x*4+border4+3]/32.0;
                  }
                  pix_mode2 += (double)RecY[(pos_y+border-2)*(cur->width*4+2*border4)+pos_x*4+border4+4]/32.0; // full_pel
                  pix_mode2 += (double)RecY[(pos_y+border-2)*(cur->width*4+2*border4)+pos_x*4+border4+2]/32.0; // b_pos
                  pix_mode2 += FilterCoef[h_pos][FILTER_SIZE] + FilterCoef[j_pos][FILTER_SIZE]; // offsets
                  pix_mode2 = pix_mode2 / 4.0;
                  pix_mode3 += FilterCoef[o_pos][FILTER_SIZE];
                }
                pred_blk[0][yi * 8 + xj] += (int)(pix_mode2 * b_scaling + 0.5);
                pred_blk[1][yi * 8 + xj] += (int)(pix_mode3 * b_scaling + 0.5);
              }
            }
          }
        }
        if(sub_pos > c_pos && ref_frame != -1 && FilterFlag_SAIF[sub_pos]) 
        {
          for (xj = 0; xj < 8; xj++)   
          {
            for (yi = 0; yi < 8; yi++)
            {  
              diff_blk[0][yi * 8 + xj] = (cur->data[(y+yi)*cur->width+(x+xj)] - pred_blk[0][yi * 8 + xj]) << SAMPLE_ADD_PREC; 
              diff_blk[1][yi * 8 + xj] = (cur->data[(y+yi)*cur->width+(x+xj)] - pred_blk[1][yi * 8 + xj]) << SAMPLE_ADD_PREC; 
            }
          }
          block_cnt[sub_pos] ++;
          SATD[sub_pos][0] += (double)SATD8x8(diff_blk[0], mel_shift);
          SATD[sub_pos][1] += (double)SATD8x8(diff_blk[1], mel_shift);  
        }
      }
    }
  }
  for (sub_pos = d_pos; sub_pos <= o_pos; sub_pos ++)
  {
    if (block_cnt[sub_pos])
    {
      SATD[sub_pos][0] = SATD[sub_pos][0] / (double)block_cnt[sub_pos];
      SATD[sub_pos][1] = SATD[sub_pos][1] / (double)block_cnt[sub_pos];
    }
    else
    {
      SATD[sub_pos][0] = 0.0;
      SATD[sub_pos][1] = 0.0;
    }
  }
  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<FILTER_SIZE;i++)
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[d_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[d_pos][FILTER_SIZE]); 
  coef_rate[d_pos] = eenc_aif->get_nbits();

  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<FILTER_SIZE;i++)
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[e_pos][i]-DiffQFilterCoef_SAIF[d_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[e_pos][FILTER_SIZE]); 
  coef_rate[e_pos] = eenc_aif->get_nbits();

  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<FILTER_SIZE;i++)
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[f_pos][i]-DiffQFilterCoef_SAIF[e_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[f_pos][FILTER_SIZE]); 
  coef_rate[f_pos] = eenc_aif->get_nbits();

  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<FILTER_SIZE;i++)
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[g_pos][i]-DiffQFilterCoef_SAIF[f_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[g_pos][FILTER_SIZE]);
  coef_rate[g_pos] = eenc_aif->get_nbits();

  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<(FILTER_SIZE>>1);i++)
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[h_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[h_pos][FILTER_SIZE]); 
  coef_rate[h_pos] = eenc_aif->get_nbits();

  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<(FILTER_SIZE>>1);i++)
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[i_pos][i]-DiffQFilterCoef_SAIF[h_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[i_pos][FILTER_SIZE]); 
  coef_rate[i_pos] = eenc_aif->get_nbits();

  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<(FILTER_SIZE>>1);i++)
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[j_pos][i]-DiffQFilterCoef_SAIF[i_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[j_pos][FILTER_SIZE]); 
  coef_rate[j_pos] = eenc_aif->get_nbits();


  eenc_aif->reset(eenc_aif_buf);
  for(i=0;i<(FILTER_SIZE>>1);i++)
  {
    eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[k_pos][i]-DiffQFilterCoef_SAIF[j_pos][i]);
  }
  eenc_aif->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[k_pos][FILTER_SIZE]); 
  coef_rate[k_pos] = eenc_aif->get_nbits();

  cost_mode2 = cost_mode3 = 0.0;
  for (sub_pos = d_pos; sub_pos <= o_pos; sub_pos ++)
  {
    cost_mode2 += SATD[sub_pos][0];
    cost_mode3 += SATD[sub_pos][1];
  }

  if (nr8x8blocks < scalingThres)
  {
    cost_mode2 += ((int)((double)(coef_rate[d_pos] + coef_rate[h_pos] + coef_rate[j_pos]) * scalingThres/nr8x8blocks + 0.5));

    cost_mode3 += ((int)((double)(coef_rate[d_pos] + coef_rate[e_pos] + coef_rate[f_pos] + coef_rate[g_pos] + coef_rate[h_pos] + 
                    coef_rate[i_pos] + coef_rate[j_pos] + coef_rate[k_pos]) * scalingThres/nr8x8blocks + 0.5));
  }
  else
  {
    cost_mode2 += (coef_rate[d_pos] + coef_rate[h_pos] + coef_rate[j_pos]);

    cost_mode3 += (coef_rate[d_pos] + coef_rate[e_pos] + coef_rate[f_pos] + coef_rate[g_pos] + coef_rate[h_pos] + 
                    coef_rate[i_pos] + coef_rate[j_pos] + coef_rate[k_pos]);  
  }

  if (cost_mode3 < cost_mode2)
    AIF_mode_flag2 = 1;
  else
    AIF_mode_flag2 = 0;

  if (!AIF_mode_flag2) // AIF mode 2
  {
    FilterFlag_SAIF[a_pos] = FilterFlag_SAIF[b_pos] = FilterFlag_SAIF[c_pos] = 
    FilterFlag_SAIF[d_pos] = FilterFlag_SAIF[h_pos] = FilterFlag_SAIF[j_pos] = 
    FilterFlag_SAIF[l_pos] = 1;
      
    FilterFlag_SAIF[e_pos] = FilterFlag_SAIF[f_pos] = FilterFlag_SAIF[g_pos] =
    FilterFlag_SAIF[i_pos] = FilterFlag_SAIF[k_pos] = FilterFlag_SAIF[m_pos] = 
    FilterFlag_SAIF[n_pos] = FilterFlag_SAIF[o_pos] = 0;
  }
  else // AIF mode 3
  {
    for(sub_pos = a_pos; sub_pos <= o_pos; sub_pos++)
      FilterFlag_SAIF[sub_pos] = 1;
  }
  param.enable_intcomp=param_ic_full;
  delete eenc_aif_buf;
  delete eenc_aif;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SATD
*
* Applies a Hadamard transform and computes the SAD in the transformed domain.
*
* \param[in] diff Block array (addressed by [8*i+j]), mel_shift
* \return SATD
*/
//-----------------------------------------------------------------------------
int AIF_adaptive_filter::SATD8x8(int *diff, int mel_shift) 
{
  int i, j, jj, satd=0;
  int m1[8][8], m2[8][8], m3[8][8];

  //horizontal
  for (j=0; j < 8; j++)
  {
    jj = j << 3;
    m2[j][0] = diff[jj  ] + diff[jj+4];
    m2[j][1] = diff[jj+1] + diff[jj+5];
    m2[j][2] = diff[jj+2] + diff[jj+6];
    m2[j][3] = diff[jj+3] + diff[jj+7];
    m2[j][4] = diff[jj  ] - diff[jj+4];
    m2[j][5] = diff[jj+1] - diff[jj+5];
    m2[j][6] = diff[jj+2] - diff[jj+6];
    m2[j][7] = diff[jj+3] - diff[jj+7];

    m1[j][0] = m2[j][0] + m2[j][2];
    m1[j][1] = m2[j][1] + m2[j][3];
    m1[j][2] = m2[j][0] - m2[j][2];
    m1[j][3] = m2[j][1] - m2[j][3];
    m1[j][4] = m2[j][4] + m2[j][6];
    m1[j][5] = m2[j][5] + m2[j][7];
    m1[j][6] = m2[j][4] - m2[j][6];
    m1[j][7] = m2[j][5] - m2[j][7];

    m2[j][0] = m1[j][0] + m1[j][1];
    m2[j][1] = m1[j][0] - m1[j][1];
    m2[j][2] = m1[j][2] + m1[j][3];
    m2[j][3] = m1[j][2] - m1[j][3];
    m2[j][4] = m1[j][4] + m1[j][5];
    m2[j][5] = m1[j][4] - m1[j][5];
    m2[j][6] = m1[j][6] + m1[j][7];
    m2[j][7] = m1[j][6] - m1[j][7];
  }

  //vertical
  for (i=0; i < 8; i++)
  {
    m3[0][i] = m2[0][i] + m2[4][i];
    m3[1][i] = m2[1][i] + m2[5][i];
    m3[2][i] = m2[2][i] + m2[6][i];
    m3[3][i] = m2[3][i] + m2[7][i];
    m3[4][i] = m2[0][i] - m2[4][i];
    m3[5][i] = m2[1][i] - m2[5][i];
    m3[6][i] = m2[2][i] - m2[6][i];
    m3[7][i] = m2[3][i] - m2[7][i];

    m1[0][i] = m3[0][i] + m3[2][i];
    m1[1][i] = m3[1][i] + m3[3][i];
    m1[2][i] = m3[0][i] - m3[2][i];
    m1[3][i] = m3[1][i] - m3[3][i];
    m1[4][i] = m3[4][i] + m3[6][i];
    m1[5][i] = m3[5][i] + m3[7][i];
    m1[6][i] = m3[4][i] - m3[6][i];
    m1[7][i] = m3[5][i] - m3[7][i];

    m2[0][i] = m1[0][i] + m1[1][i];
    m2[1][i] = m1[0][i] - m1[1][i];
    m2[2][i] = m1[2][i] + m1[3][i];
    m2[3][i] = m1[2][i] - m1[3][i];
    m2[4][i] = m1[4][i] + m1[5][i];
    m2[5][i] = m1[4][i] - m1[5][i];
    m2[6][i] = m1[6][i] + m1[7][i];
    m2[7][i] = m1[6][i] - m1[7][i];
  }
  for (j=0; j < 8; j++)
    for (i=0; i < 8; i++) 
      satd += abs(m2[j][i]);

  if (mel_shift >= SAMPLE_ADD_PREC - ONE - FOUR + 6)
    return satd << (mel_shift + ONE - SAMPLE_ADD_PREC + FOUR - 6);
  else
    return shift_right_round(satd, SAMPLE_ADD_PREC - ONE - mel_shift - FOUR + 6);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode adaptive interpolation filter information
*
* \param[in] eenc: pointer to entropy encoder
* \return void
*/
//-----------------------------------------------------------------------------
void AIF_adaptive_filter::encode_filter_information(EntropyEncoder *eenc) 
{
  int i;
  eenc->encode_AIF_mode_flag1(AIF_mode_flag1);
  if (!AIF_mode_flag1) // AIF mode 1
  {
    FilterFlag_SAIF[b_pos] = FilterFlag_SAIF[h_pos] = FilterFlag_SAIF[j_pos] = 1;
    FilterFlag_SAIF[a_pos] = FilterFlag_SAIF[c_pos] = FilterFlag_SAIF[d_pos] = 
    FilterFlag_SAIF[e_pos] = FilterFlag_SAIF[f_pos] = FilterFlag_SAIF[g_pos] =
    FilterFlag_SAIF[i_pos] = FilterFlag_SAIF[k_pos] = FilterFlag_SAIF[l_pos] = 
    FilterFlag_SAIF[m_pos] = FilterFlag_SAIF[n_pos] = FilterFlag_SAIF[o_pos] = 0;
  } 
  else 
  {
    eenc->encode_AIF_mode_flag2(AIF_mode_flag2);
    if (!AIF_mode_flag2) // AIF mode 2
    {
      FilterFlag_SAIF[a_pos] = FilterFlag_SAIF[b_pos] = FilterFlag_SAIF[c_pos] = 
      FilterFlag_SAIF[d_pos] = FilterFlag_SAIF[h_pos] = FilterFlag_SAIF[j_pos] = 
      FilterFlag_SAIF[l_pos] = 1;
      FilterFlag_SAIF[e_pos] = FilterFlag_SAIF[f_pos] = FilterFlag_SAIF[g_pos] = 
      FilterFlag_SAIF[i_pos] = FilterFlag_SAIF[k_pos] = FilterFlag_SAIF[m_pos] = 
      FilterFlag_SAIF[n_pos] = FilterFlag_SAIF[o_pos] = 0;
    }
    else // AIF mode 3
      for(i=a_pos; i<= o_pos; i++)
        FilterFlag_SAIF[i] = 1;
  }
  if(FilterFlag_SAIF[a_pos])
  {
    for(i=0;i<FILTER_SIZE;i++)
      eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[a_pos][i]);
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[a_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[b_pos])
  {
    for(i= 0; i < (FILTER_SIZE >> 1); i++)
      eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[b_pos][i]);
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[b_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[c_pos])
  {
    if(FilterFlag_SAIF[a_pos])
    {
      for(i=0;i<FILTER_SIZE;i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[c_pos][i]-
        DiffQFilterCoef_SAIF[a_pos][FILTER_SIZE-i-1]);
    }
    else
    {
      for(i=0;i<FILTER_SIZE;i++)
         eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[c_pos][i]);
    }
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[c_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[d_pos])
  {
    for(i=0;i<FILTER_SIZE;i++)
      eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[d_pos][i]);
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[d_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[e_pos])
  {
    if(FilterFlag_SAIF[d_pos])
    {
      for(i=0;i<FILTER_SIZE;i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[e_pos][i]-
      DiffQFilterCoef_SAIF[d_pos][i]);
    }
    else
    {
      for(i=0;i<FILTER_SIZE;i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[e_pos][i]);
    }
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[e_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[f_pos])
  {
    if(FilterFlag_SAIF[e_pos])
    {
      for(i=0;i<FILTER_SIZE;i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[f_pos][i]-
        DiffQFilterCoef_SAIF[e_pos][i]);
    }
    else
    {
      for(i=0;i<FILTER_SIZE;i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[f_pos][i]);
    }
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[f_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[g_pos])
  {
    if(FilterFlag_SAIF[f_pos])
    {
      for(i=0;i<FILTER_SIZE;i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[g_pos][i]-
        DiffQFilterCoef_SAIF[f_pos][i]);
    }
    else
    {
      for(i=0;i<FILTER_SIZE;i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[g_pos][i]);
    }
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[g_pos][FILTER_SIZE]);
  }
  if(FilterFlag_SAIF[h_pos])
  {
    for(i=0;i<(FILTER_SIZE>>1);i++)
      eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[h_pos][i]);
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[h_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[i_pos])
  {
    if(FilterFlag_SAIF[h_pos])
    {
      for(i=0;i<(FILTER_SIZE>>1);i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[i_pos][i]-
        DiffQFilterCoef_SAIF[h_pos][i]);
    }
    else
    {
      for(i=0;i<(FILTER_SIZE>>1);i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[i_pos][i]);
    }
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[i_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[j_pos])
  {
    if(FilterFlag_SAIF[i_pos])
    {
      for(i=0;i<(FILTER_SIZE>>1);i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[j_pos][i]-
        DiffQFilterCoef_SAIF[i_pos][i]);
    }
    else
    {
      for(i=0;i<(FILTER_SIZE>>1);i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[j_pos][i]);
    }
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[j_pos][FILTER_SIZE]); 
  }
  if(FilterFlag_SAIF[k_pos])
  {
    if(FilterFlag_SAIF[j_pos])
    {
      for(i=0;i<(FILTER_SIZE>>1);i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[k_pos][i]-
        DiffQFilterCoef_SAIF[j_pos][i]);
    }
    else
    {
      for(i=0;i<(FILTER_SIZE>>1);i++)
        eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[k_pos][i]);
    }
    eenc->encode_AIF_diff_qfilter_coef(DiffQFilterCoef_SAIF[k_pos][FILTER_SIZE]); 
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode adaptive interpolation filter information
*
* \param[in] edec: pointer to entropy decoder
* \return void
*/
//-----------------------------------------------------------------------------
void AIF_adaptive_filter::decode_filter_information(EntropyDecoder *edec) 
{
  int i;
  ResetAdaptiveFilter_DEC();
  if(AIF_FilterFlag == 2)
  {
    AIF_mode_flag1 = edec->decode_AIF_mode_flag1();    
    if (!AIF_mode_flag1) // AIF mode 1
    {
      FilterFlag_SAIF[b_pos] = FilterFlag_SAIF[h_pos] = FilterFlag_SAIF[j_pos] = 1;
      FilterFlag_SAIF[a_pos] = FilterFlag_SAIF[c_pos] = FilterFlag_SAIF[d_pos] = 
    FilterFlag_SAIF[e_pos] = FilterFlag_SAIF[f_pos] = FilterFlag_SAIF[g_pos] =
      FilterFlag_SAIF[i_pos] = FilterFlag_SAIF[k_pos] = FilterFlag_SAIF[l_pos] = 
    FilterFlag_SAIF[m_pos] = FilterFlag_SAIF[n_pos] = FilterFlag_SAIF[o_pos] = 0;
      AIF_mode_flag2 = 0;
    } 
    else 
    {
      AIF_mode_flag2=edec->decode_AIF_mode_flag2();      
      if (!AIF_mode_flag2) // AIF mode 2
      {
        FilterFlag_SAIF[a_pos] = FilterFlag_SAIF[b_pos] = FilterFlag_SAIF[c_pos] = 
    FilterFlag_SAIF[d_pos] = FilterFlag_SAIF[h_pos] = FilterFlag_SAIF[j_pos] = 
        FilterFlag_SAIF[l_pos] = 1;        
        FilterFlag_SAIF[e_pos] = FilterFlag_SAIF[f_pos] = FilterFlag_SAIF[g_pos] = 
    FilterFlag_SAIF[i_pos] = FilterFlag_SAIF[k_pos] = FilterFlag_SAIF[m_pos] = 
        FilterFlag_SAIF[n_pos] = FilterFlag_SAIF[o_pos] = 0;
      }
      else // AIF mode 3
        for(i=a_pos; i<= o_pos; i++)
          FilterFlag_SAIF[i] = 1;
    }
    if(FilterFlag_SAIF[a_pos])
    {
      for(i= 0; i < FILTER_SIZE; i++)
        DiffQFilterCoef_SAIF[a_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[a_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef(); 
    }
    if(FilterFlag_SAIF[b_pos])
    {
      for(i= 0; i < (FILTER_SIZE >> 1); i++)
        DiffQFilterCoef_SAIF[b_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[b_pos][3] = DiffQFilterCoef_SAIF[b_pos][2];
      DiffQFilterCoef_SAIF[b_pos][4] = DiffQFilterCoef_SAIF[b_pos][1];                 
      DiffQFilterCoef_SAIF[b_pos][5] = DiffQFilterCoef_SAIF[b_pos][0]; 
      DiffQFilterCoef_SAIF[b_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef();
    }
    if(FilterFlag_SAIF[c_pos])
    {
      if (FilterFlag_SAIF[a_pos])
        for(i= 0; i < FILTER_SIZE; i++)
          DiffQFilterCoef_SAIF[c_pos][i] = edec->decode_AIF_diff_qfilter_coef() + 
      DiffQFilterCoef_SAIF[a_pos][FILTER_SIZE-i-1];
      else
        for(i= 0; i < FILTER_SIZE; i++)
          DiffQFilterCoef_SAIF[c_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[c_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef(); 
    }
    if(FilterFlag_SAIF[d_pos])
    {
      for(i= 0; i < FILTER_SIZE; i++)
        DiffQFilterCoef_SAIF[d_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[d_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef(); 
    }
    if(FilterFlag_SAIF[e_pos])
    {
      if (FilterFlag_SAIF[d_pos])
        for(i= 0; i < FILTER_SIZE; i++)
          DiffQFilterCoef_SAIF[e_pos][i] = edec->decode_AIF_diff_qfilter_coef() + 
      DiffQFilterCoef_SAIF[d_pos][i];
      else
        for(i= 0; i < FILTER_SIZE; i++)
          DiffQFilterCoef_SAIF[e_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[e_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef(); 
    }
    if(FilterFlag_SAIF[f_pos])
    {
      if (FilterFlag_SAIF[e_pos])
        for(i= 0; i < FILTER_SIZE; i++)
          DiffQFilterCoef_SAIF[f_pos][i] = edec->decode_AIF_diff_qfilter_coef() + 
      DiffQFilterCoef_SAIF[e_pos][i];
      else
        for(i= 0; i < FILTER_SIZE; i++)
          DiffQFilterCoef_SAIF[f_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[f_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef(); 
    }
    if(FilterFlag_SAIF[g_pos])
    {
      if (FilterFlag_SAIF[f_pos])
        for(i= 0; i < FILTER_SIZE; i++)
          DiffQFilterCoef_SAIF[g_pos][i] = edec->decode_AIF_diff_qfilter_coef() + 
      DiffQFilterCoef_SAIF[f_pos][i];
      else
        for(i= 0; i < FILTER_SIZE; i++)
          DiffQFilterCoef_SAIF[g_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[g_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef();
    }
    if(FilterFlag_SAIF[h_pos])
    {
      for(i= 0; i < FILTER_SIZE>>1; i++)
        DiffQFilterCoef_SAIF[h_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[h_pos][3] = DiffQFilterCoef_SAIF[h_pos][2];
      DiffQFilterCoef_SAIF[h_pos][4] = DiffQFilterCoef_SAIF[h_pos][1];                 
      DiffQFilterCoef_SAIF[h_pos][5] = DiffQFilterCoef_SAIF[h_pos][0]; 
      DiffQFilterCoef_SAIF[h_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef(); 
    }
    if(FilterFlag_SAIF[i_pos])
    {
      if(FilterFlag_SAIF[h_pos])
        for(i= 0; i < FILTER_SIZE>>1; i++)
          DiffQFilterCoef_SAIF[i_pos][i] = edec->decode_AIF_diff_qfilter_coef() + DiffQFilterCoef_SAIF[h_pos][i];
      else
        for(i= 0; i < FILTER_SIZE>>1; i++)
          DiffQFilterCoef_SAIF[i_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[i_pos][3] = DiffQFilterCoef_SAIF[i_pos][2];
      DiffQFilterCoef_SAIF[i_pos][4] = DiffQFilterCoef_SAIF[i_pos][1];                 
      DiffQFilterCoef_SAIF[i_pos][5] = DiffQFilterCoef_SAIF[i_pos][0]; 
      DiffQFilterCoef_SAIF[i_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef(); 
    }
    if(FilterFlag_SAIF[j_pos])
    {
      if(FilterFlag_SAIF[i_pos])
        for(i= 0; i < FILTER_SIZE>>1; i++)
          DiffQFilterCoef_SAIF[j_pos][i] = edec->decode_AIF_diff_qfilter_coef() + DiffQFilterCoef_SAIF[i_pos][i];
      else
        for(i= 0; i < FILTER_SIZE>>1; i++)
          DiffQFilterCoef_SAIF[j_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[j_pos][3] = DiffQFilterCoef_SAIF[j_pos][2];
      DiffQFilterCoef_SAIF[j_pos][4] = DiffQFilterCoef_SAIF[j_pos][1];                 
      DiffQFilterCoef_SAIF[j_pos][5] = DiffQFilterCoef_SAIF[j_pos][0]; 
      DiffQFilterCoef_SAIF[j_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef();
    }
    if(FilterFlag_SAIF[k_pos])
    {
      if(FilterFlag_SAIF[j_pos])
        for(i= 0; i < FILTER_SIZE>>1; i++)
          DiffQFilterCoef_SAIF[k_pos][i] = edec->decode_AIF_diff_qfilter_coef() + DiffQFilterCoef_SAIF[j_pos][i];
      else
        for(i= 0; i < FILTER_SIZE>>1; i++)
          DiffQFilterCoef_SAIF[k_pos][i] = edec->decode_AIF_diff_qfilter_coef();
      DiffQFilterCoef_SAIF[k_pos][3] = DiffQFilterCoef_SAIF[k_pos][2];
      DiffQFilterCoef_SAIF[k_pos][4] = DiffQFilterCoef_SAIF[k_pos][1];                 
      DiffQFilterCoef_SAIF[k_pos][5] = DiffQFilterCoef_SAIF[k_pos][0]; 
      DiffQFilterCoef_SAIF[k_pos][FILTER_SIZE] = edec->decode_AIF_diff_qfilter_coef(); 
    }
    if(FilterFlag_SAIF[l_pos])
    {
      for(i= 0; i < FILTER_SIZE; i++)
        DiffQFilterCoef_SAIF[l_pos][i] = DiffQFilterCoef_SAIF[d_pos][FILTER_SIZE-i-1];
      DiffQFilterCoef_SAIF[l_pos][FILTER_SIZE] = DiffQFilterCoef_SAIF[d_pos][FILTER_SIZE];
    }
    if(FilterFlag_SAIF[m_pos])
    {
      for(i= 0; i < FILTER_SIZE; i++)
        DiffQFilterCoef_SAIF[m_pos][i] = DiffQFilterCoef_SAIF[e_pos][FILTER_SIZE-i-1];
      DiffQFilterCoef_SAIF[m_pos][FILTER_SIZE] = DiffQFilterCoef_SAIF[e_pos][FILTER_SIZE];
    }
    if(FilterFlag_SAIF[n_pos])
    {
      for(i= 0; i < FILTER_SIZE; i++)
        DiffQFilterCoef_SAIF[n_pos][i] = DiffQFilterCoef_SAIF[f_pos][FILTER_SIZE-i-1];
      DiffQFilterCoef_SAIF[n_pos][FILTER_SIZE] = DiffQFilterCoef_SAIF[f_pos][FILTER_SIZE];
    }
    if(FilterFlag_SAIF[o_pos])
    {
      for(i= 0; i < FILTER_SIZE; i++)
        DiffQFilterCoef_SAIF[o_pos][i] = DiffQFilterCoef_SAIF[g_pos][FILTER_SIZE-i-1]; 
      DiffQFilterCoef_SAIF[o_pos][FILTER_SIZE] = DiffQFilterCoef_SAIF[g_pos][FILTER_SIZE];
    }    
    CalculateFilterCoefficientsSep();
  }
}
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
