/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file Deblocking_Filter.cpp
*
* \brief Implementation of Deblocking_Filter class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"
#include "EntropyEncoder.H"
#include "Buffer.H"
#include "EntropyEncoder.H"
#include "Deblocking_Filter.H"
#include "MotionField.H"

#ifdef DEBLOCK_MT
#include <pthread.h>
#endif

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

int alpha_tab[128] =
{
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  3,  3,  3,  3,  3,  3,  4,  4,  4,
  4,  5,  5,  5,  5,  6,  6,  6,  6,  7,  7,  7,  8,  8,  9,  9,
  9, 10, 10, 11, 11, 12, 12, 13, 14, 14, 15, 16, 16, 17, 18, 19,
 20, 20, 21, 22, 23, 24, 26, 27, 28, 29, 31, 32, 33, 35, 36, 38,
 40, 42, 44, 45, 48, 50, 52, 54, 57, 59, 62, 65, 68, 71, 74, 77,
 80, 84, 88, 92, 96,100,105,109,114,119,125,130,136,142,148,155,
162,169,176,184,193,201,210,219,229,239,250,255,255,255,255,255
};
int beta_tab[128] =
{
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4,
 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7,
 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10 ,10,
10,10,10,11,11,11,11,11,12,12,12,12,12,12,13,13,
13,13,13,14,14,14,14,14,15,15,15,15,15,15,16,16,
16,16,16,17,17,17,17,17,18,18,18,18,18,18,19,19
};

int clip_tab[128] =
{
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5,
 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9,
10,10,11,11,12,12,12,13,13,13,14,14,14,15,16,17
};

int mode_to_bs[NUM_LOOP_MODES3 + 2] = {0, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0}; // t1

int clip_tab_bs[4][128] =
{
{ // BS = 0
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
},
{ // BS = 1
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 
 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 
 4, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 
 8, 8, 8, 9, 9, 9,10,10,11,11,12,12,12,13,13,13
},
{ // BS = 2
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5,
 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9,
10,10,11,11,12,12,12,13,13,13,14,14,14,15,16,17
},
{ // BS = 3
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 
 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 
 4, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 
 8, 8, 8, 9, 9, 9,10,10,11,11,12,12,12,13,13,13,
14,14,15,15,16,16,17,17,18,19,20,20,21,22,23,25
}
};

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in 4x4 luma block units
* \param[in] w width of frame in 4x4 luma block units
*/
//-----------------------------------------------------------------------------
Deblocking_Filter::Deblocking_Filter(int h, int w)
{
  height  = h;
  width   = w;
  height2 = h>>1;
  width2  = w>>1;

  int dirIdx;
  for (dirIdx = 0; dirIdx < 2; dirIdx++)
  {
    m_luma[dirIdx].set_size(w>>2, h>>2, 0);
    m_chroma[0][dirIdx].set_size(w>>3, h>>3, 0);
    m_chroma[1][dirIdx].set_size(w>>3, h>>3, 0);
    m_modes[0][dirIdx].set_size(width, height, 0);
    m_modes[1][dirIdx].set_size(width2, height2, 0);
    m_modes[2][dirIdx].set_size(width2, height2, 0);
  }
  str_mf = new MotionField(h, w);
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
Deblocking_Filter::~Deblocking_Filter() 
{
  delete str_mf;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Decision of filter samples
*
* \param[in] pos  sample position to be filtered
* \param[in] a threshold value
* \param[in] b threshold value
* \param[in] p2 sample value
* \param[in] p1 sample value
* \param[in] p0 sample value
* \param[in] q0 sample value
* \param[in] q1 sample value
* \param[in] q2 sample value
*/
//-----------------------------------------------------------------------------
void Deblocking_Filter::decision_for_each_sample( int* pos, int a, int b, int p2, int p1, int p0, int q0, int q1, int q2 )
{
  pos[0] = 0;
  pos[1] = 0;
  pos[2] = 0;
  pos[3] = 0;
  
  if (abs(p0 - q0) < a && abs(p1 - p0) < b && abs(q1 - q0) < b)
  {
    pos[1] = 2;
    pos[2] = 3;

    if (abs(p2 - p0) < b && abs(q2 - q0) < b)
    {
      pos[0] = 1;
      pos[3] = 4;
    }
  }
}


//-----------------------------------------------------------------------------
/*!
 * \brief Setup filter modes
 *
 */
//-----------------------------------------------------------------------------

void Deblocking_Filter::set_filter_modes()
{
  int x,y;
  int h=height/4;
  int w=width/4;
  int h2=height2/4;
  int w2=width2/4;
  int m,m1,m2;
  int c;   
  int map_modes[128];

#ifdef DBF_BS_FIX
  for (c=0;c<128;c++)
  {
    map_modes[c]=c&15;
  }
#else
  for (c=0;c<128;c++)
  {
    map_modes[c]=0;
  }
  map_modes[0]=0;
  map_modes[1]=1;
  map_modes[2]=2;
  map_modes[3]=3;
  map_modes[4]=4;
  map_modes[5]=5;
  map_modes[6]=6;
  map_modes[7]=7;
  map_modes[8]=8;
  for (c=32;c<64;c++)
  {
   map_modes[c]=9;
  }
  for (c=64;c<128;c++)
  {
   map_modes[c]=10;
  }
#endif
   
  m_modes[0][0].set(0, 0, 0, height, width);
  m_modes[0][1].set(0, 0, 0, height, width);

  for (y=0;y<h;y++)
  {
    for (x=1;x<w;x++)
    {
      m=map_modes[m_luma[0][y][x]&(DCM_DENOISE_THIS_BLOCK-1)];
      if (m!=0)
      {
        m_modes[0][0].set(4*y, 4*x-2, m, 4, 4);
      }
    }
  }
  for (y=1;y<h;y++)
  {
    for (x=0;x<w;x++)
    {
      m=map_modes[m_luma[1][y][x]&(DCM_DENOISE_THIS_BLOCK-1)];
      if (m!=0)
      {
        m_modes[0][1].set(4*y-2, 4*x, m, 4, 4);
      }
    }
  }

  m_modes[1][0].set(0, 0, 0, height2, width2);
  m_modes[1][1].set(0, 0, 0, height2, width2);
  m_modes[2][0].set(0, 0, 0, height2, width2);
  m_modes[2][1].set(0, 0, 0, height2, width2);

  for (y=0;y<h2;y++)
  {
    for (x=1;x<w2;x++)
    {
      m1=map_modes[m_chroma[0][0][y][x]&(DCM_DENOISE_THIS_BLOCK-1)];
      m2=map_modes[m_chroma[0][1][y][x]&(DCM_DENOISE_THIS_BLOCK-1)];
      if (m1!=0 || m2!=0)
      {
        m_modes[1][0].set(4*y, 4*x-2, m1, 4, 4);
        m_modes[2][0].set(4*y, 4*x-2, m2, 4, 4);
      }
    }
  }
  for (y=1;y<h2;y++)
  {
    for (x=0;x<w2;x++)
    {
      m1=map_modes[m_chroma[1][0][y][x]&(DCM_DENOISE_THIS_BLOCK-1)];
      m2=map_modes[m_chroma[1][1][y][x]&(DCM_DENOISE_THIS_BLOCK-1)];

      if (m1!=0 || m2!=0)
      {
        m_modes[1][1].set(4*y-2, 4*x, m1, 4, 4);
        m_modes[2][1].set(4*y-2, 4*x, m2, 4, 4);
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Initialization of threshold values
 *
 * \param[in] qp_y QP value for luminance
 * \param[in] qp_c QP value for chrominance
 */
//-----------------------------------------------------------------------------

void Deblocking_Filter::init_alpha_values(int qp_y, int qp_c)
{
  int dbf_offset_a = param.dbf_offset_a-16;
  int dbf_offset_b = param.dbf_offset_b-16;
    
  dbf_idx_a[0] = clip( (qp_y+dbf_offset_a), 0, 127);
  dbf_idx_b[0] = clip( (qp_y+dbf_offset_b),0, 127);
  dbf_idx_a[1] = clip( (qp_c+dbf_offset_a), 0, 127);
  dbf_idx_b[1] = clip( (qp_c+dbf_offset_b),0, 127);
  dbf_idx_a[2] = clip( (qp_c+dbf_offset_a), 0, 127);
  dbf_idx_b[2] = clip( (qp_c+dbf_offset_b),0, 127);
}

#ifdef DEBLOCK_MT
struct DeblockingThreadArgs
{
  pthread_t tid;
  Deblocking_Filter *parent;
  MotionField *mf;
  unsigned char *oY;
  Plane<unsigned char> *p_rec;
  int component;
  int mb_row_start;
  int n_mb_rows;
};

//-----------------------------------------------------------------------------
/*!
 * \brief Launch deblocking threads
 *
 * \param[in] arg Pointer to thread parameters, indicating what to deblock.
 */
//-----------------------------------------------------------------------------

static void *launch_fixed_deblock(void *arg)
{
  struct DeblockingThreadArgs *targ = (DeblockingThreadArgs *)arg;
  targ->parent->fixed_deblocking_filter_rows(targ->mf,
                       targ->oY,
                       *targ->p_rec,
                       targ->component,
                       targ->mb_row_start,
                       targ->n_mb_rows);
  return NULL;
}

#endif

//-----------------------------------------------------------------------------
/*!
 * \brief DeblockingFilter process
 *
 * \param[in] mf pointer to motion field
 * \param[in] oY pointer to output image signal
 * \param[in] xY reference to reconstructed image signal
 * \param[in] component color component
 */
//-----------------------------------------------------------------------------

void Deblocking_Filter::fixed_deblocking_filter(MotionField *mf, unsigned char *oY, unsigned char *xY, int component )
{
  int width_f, height_f;
  int i, j;

  if (component == 0)
  {
    width_f = width;
    height_f = height;
  }
  else
  {
    width_f = width2;
    height_f = height2;
  }
  Plane<unsigned char> p_rec(width_f, height_f, 4);

  for (i = 0; i < height_f; i++)
  {
    for (j = 0; j < width_f; j++)
    {
      p_rec[i][j] = xY[width_f * i + j];
    }
  }

  p_rec.mirror();

#ifdef DEBLOCK_MT
  // Launch the filtering of this component.
  struct DeblockingThreadArgs *targs = new DeblockingThreadArgs[param.decode_ndeblockthreads];
  int hmbs = height_f >> 3; // 8x8 blocks.
  int mb_rows_per_thread = (hmbs+param.decode_ndeblockthreads-1)/param.decode_ndeblockthreads;

  for (int t = 0; t < param.decode_ndeblockthreads; t++)
  {
  targs[t].mb_row_start = t*mb_rows_per_thread*2; // 4x4 blocks.
    targs[t].n_mb_rows = mb_rows_per_thread*2; // 4x4 blocks.
     if (t == param.decode_ndeblockthreads-1 || targs[t].mb_row_start+targs[t].n_mb_rows > height_f/4)
      targs[t].n_mb_rows = height_f/4-targs[t].mb_row_start;

    targs[t].parent = this;
    targs[t].mf = mf;
  targs[t].oY = oY;
  targs[t].p_rec = &p_rec;
  targs[t].component = component;
  }

  // Launch.
  if (param.decode_ndeblockthreads == 1)
  {
    // direct call.
    launch_fixed_deblock(&targs[0]);
  }
  else
  {
    for (int t = 0; t < param.decode_ndeblockthreads; t++)
    {
      if (targs[t].n_mb_rows <= 0)
        break;
      pthread_create(&targs[t].tid, NULL, launch_fixed_deblock, &targs[t]);
    }
    // Wait.
    for (int t = 0; t < param.decode_ndeblockthreads; t++)
    {
      if (targs[t].n_mb_rows <= 0)
        break;
      pthread_join(targs[t].tid, NULL);
    }
  }

  delete[] targs;
}

//-----------------------------------------------------------------------------
/*!
 * \brief DeblockingFilter process used by a single thread.
 *
 * \param[in] mf pointer to motion field
 * \param[in] oY pointer to output image signal
 * \param[in] xY reference to reconstructed image signal
 * \param[in] component color component
 * \param[in] mb_row_start First 4x4 block row to deblock.
 * \param[in] n_mb_rows Number of 4x4 block rows to deblock.
 *
 * Like fixed_deblocking_filter, except a row-range is given in which to deblock.
 */
//-----------------------------------------------------------------------------

void Deblocking_Filter::fixed_deblocking_filter_rows(MotionField *mf, unsigned char *oY, Plane<unsigned char> &p_rec, int component, int mb_row_start, int n_mb_rows )
{
  int i;
  int width_f, height_f;
  if (component == 0)
  {
    width_f = width;
    height_f = height;
  }
  else
  {
    width_f = width2;
    height_f = height2;
  }

  int h = mb_row_start+n_mb_rows; // 4x4 blocks, mb_row_start is on an 8x8 boundary.
#else

  int h = height_f >> 2;
#endif
  int w = width_f >> 2;

  Plane<unsigned char> &modes_h = m_modes[component][0];
  Plane<unsigned char> &modes_v = m_modes[component][1];

  Plane<int> &blk_modeH = (component == 0) ? m_luma[0] : m_chroma[0][component-1];
  Plane<int> &blk_modeV = (component == 0) ? m_luma[1] : m_chroma[1][component-1];

  int x,y;
  int type;
  int tmp, delta;
  unsigned char tmp_p[4] = { 0, 0, 0, 0 };
  int *clip_tab_local;
  int p2,p1,p0,q0,q1,q2;
  int filter_pos[4];
  int alpha, beta;
  alpha = alpha_tab[dbf_idx_a[component]];
  beta  = beta_tab[dbf_idx_b[component]];

  // vertical
  for( x=0; x<w; x++ )
  {
#ifdef DEBLOCK_MT
    // Each loop can write y*4+2, 3, 4, 5, ie, two into the "next" oY[] row.
  // These updated oYs are used by the subsequent loop.
    // The logic of the first loop is changed to only update those rows
    // used by the second loop when regarding the window being processed
    // in order to prevent thread-border conflicts.
    int backed_up = 0;
  int row_start = mb_row_start;
    if (row_start > 0)
  {
    // back up one row, during the processing of which we'll write only the 'next'.
    row_start -= 1;
    backed_up = 1;
  }
  //if (y == h-1)
  //{
  //  // the last row -- we won't write any 'next'.
  //}
    for( y=row_start; y<(h == height_f/4 ? height_f/4-1 : h); y++, backed_up = 0 )
#else
    for( y=0; y<h-1; y++ )
#endif
    {
#ifdef DEBLOCK_MT
    // ensure initial transform_blocksize is picked up.
      if( component == 0 && ( y%2==0 || backed_up ) )
#else
      if( component == 0 && y%2==0 )
#endif
      {
        type = mf->get_transform_blocksize( (y>>1), (x>>1) );
        if( type == 1 )
        {
#ifdef DEBLOCK_MT
      if (!backed_up)
      {
        y++;
        if (y == height_f/4-1)
          break;
      }
#else
          y++;
          if( y == h-1 )
          {
            break;
          }
#endif
        }
      }
      if( blk_modeV[y+1][x] != DCM_SKIP )
      {
        for( i=0; i<4; i++ )
        {
          p2 = p_rec[(y<<2)+1][(x<<2)+i];
          p1 = p_rec[(y<<2)+2][(x<<2)+i];
          p0 = p_rec[(y<<2)+3][(x<<2)+i];
          q0 = p_rec[(y<<2)+4][(x<<2)+i];
          q1 = p_rec[(y<<2)+5][(x<<2)+i];
          q2 = p_rec[(y<<2)+6][(x<<2)+i];
          decision_for_each_sample( filter_pos, alpha, beta,  p2, p1, p0, q0, q1, q2 );
#ifdef DEBLOCK_MT
      if (!backed_up)
#endif
          if( filter_pos[1]!=0 )
          {
            if (param.deblockingfilter&8)
              clip_tab_local = clip_tab_bs[mode_to_bs[modes_v[(y<<2)+3][(x<<2)+i]]];    // overiding the global clip_tab;
            else
              clip_tab_local = clip_tab;
            tmp = (p1+((p0+q0)<<2)-q1+4)>>3;
            delta = min(max(-clip_tab_local[dbf_idx_a[component]], (tmp-p0)), clip_tab_local[dbf_idx_a[component]]);
            oY[((y<<2)+3)*width_f+(x<<2)+i]=min(255,max(0,p0+delta));
          }
#ifdef DEBLOCK_MT
      if (y != h-1)
#endif
          if( filter_pos[2]!=0 )
          {
            if (param.deblockingfilter&8)
              clip_tab_local = clip_tab_bs[mode_to_bs[modes_v[(y<<2)+4][(x<<2)+i]]];    // overiding the global clip_tab;
            else
              clip_tab_local = clip_tab;
            tmp = (q1+((q0+p0)<<2)-p1+4)>>3;
            delta = min(max(-clip_tab_local[dbf_idx_a[component]], (tmp-q0)), clip_tab_local[dbf_idx_a[component]]);
            oY[((y<<2)+4)*width_f+(x<<2)+i]=min(255,max(0,q0+delta));
          }
#ifdef DEBLOCK_MT
      if (!backed_up)
#endif
          if( filter_pos[0]!=0 && component == 0 )
          {
            if (param.deblockingfilter&8)
              clip_tab_local = clip_tab_bs[mode_to_bs[modes_v[(y<<2)+2][(x<<2)+i]]];    // overiding the global clip_tab;
            else
              clip_tab_local = clip_tab;
            if (param.deblockingfilter & 4)
            {
              tmp = ((p2*7) +(p1*3)+(p0*5)+q0+8)>>4;
            }
            else
            {
              tmp = ( (p2<<1)+ p0+q0 + 2 )>>2;
            }
            delta = min(max(-clip_tab_local[dbf_idx_a[component]], (tmp-p1)), clip_tab_local[dbf_idx_a[component]]);
            oY[((y<<2)+2)*width_f+(x<<2)+i]=min(255,max(0,p1+delta));
          }
#ifdef DEBLOCK_MT
      if (y != h-1)
#endif
          if( filter_pos[3]!=0 && component == 0 )
          {
            if (param.deblockingfilter&8)
              clip_tab_local = clip_tab_bs[mode_to_bs[modes_v[(y<<2)+5][(x<<2)+i]]];    // overiding the global clip_tab;
            else
              clip_tab_local = clip_tab;
            if (param.deblockingfilter & 4)
            {
              tmp = ((q2*7) +(q1*3)+(q0*5)+p0+8)>>4;
            }
            else
            {
              tmp = ( (q2<<1)+ q0+p0 + 2 )>>2;
            }
            delta = min(max(-clip_tab_local[dbf_idx_a[component]], (tmp-q1)), clip_tab_local[dbf_idx_a[component]]);
            oY[((y<<2)+5)*width_f+(x<<2)+i]=min(255,max(0,q1+delta));
          }
        }
      }
    }
  }
  // horizontal
#ifdef DEBLOCK_MT
  for( y=mb_row_start*4; y<(mb_row_start+n_mb_rows)*4; y++ )
#else
  for( y=0; y<height_f; y++ )
#endif
  {
    for( x=0; x<w-1; x++ )
    {
      if( component == 0 && x%2==0 )
      {
        type = mf->get_transform_blocksize( (y>>3), (x>>1) );        
        if( type == 1 )
        {
          if( x!= 0 )
          {
            oY[y*width_f+(x<<2)-2]=tmp_p[0];
            oY[y*width_f+(x<<2)-1]=tmp_p[1];
            oY[y*width_f+(x<<2)]   =tmp_p[2];
            oY[y*width_f+(x<<2)+1]=tmp_p[3];
          }
          tmp_p[0] = oY[y*width_f+(x<<2)+2];
          tmp_p[1] = oY[y*width_f+(x<<2)+3];
          tmp_p[2] = oY[y*width_f+(x<<2)+4];
          tmp_p[3] = oY[y*width_f+(x<<2)+5];
          x++;
          if( x== w-1 )
          {
            oY[y*width_f+(x<<2)-2]=tmp_p[0];
            oY[y*width_f+(x<<2)-1]=tmp_p[1];
            oY[y*width_f+(x<<2)]   =tmp_p[2];
            oY[y*width_f+(x<<2)+1]=tmp_p[3];
            break;
          }
        }
      }
      if( blk_modeH[y>>2][x+1] != DCM_SKIP )
      {
        p2 = p_rec[y][(x<<2)+1];
        p1 = p_rec[y][(x<<2)+2];
        p0 = p_rec[y][(x<<2)+3];
        q0 = p_rec[y][(x<<2)+4];
        q1 = p_rec[y][(x<<2)+5];
        q2 = p_rec[y][(x<<2)+6];
        decision_for_each_sample( filter_pos, alpha, beta,  p2, p1, p0, q0, q1, q2 );
// filtering
        p2 = oY[y*width_f + (x<<2)+1];
        p1 = oY[y*width_f + (x<<2)+2];
        p0 = oY[y*width_f + (x<<2)+3];
        q0 = oY[y*width_f + (x<<2)+4];
        q1 = oY[y*width_f + (x<<2)+5];
        q2 = oY[y*width_f + (x<<2)+6];
        if( x!= 0 )
        {
          oY[y*width_f+(x<<2)-2]=tmp_p[0];
          oY[y*width_f+(x<<2)-1]=tmp_p[1];
          oY[y*width_f+(x<<2)]   =tmp_p[2];
          oY[y*width_f+(x<<2)+1]=tmp_p[3];
        }
        clip_tab_local = clip_tab;
        if( filter_pos[1]!=0 )
        {
          if (param.deblockingfilter&8)
            clip_tab_local = clip_tab_bs[mode_to_bs[modes_h[y][(x<<2)+3]]];    // overiding the global clip_tab;
          else
            clip_tab_local = clip_tab;
          tmp = (p1+((p0+q0)<<2)-q1+4)>>3;
          delta = min(max(-clip_tab_local[dbf_idx_a[component]], (tmp-p0)), clip_tab_local[dbf_idx_a[component]]);
          tmp_p[1]=min(255,max(0, p0+delta));          
        }
        else
        {
          tmp_p[1]=p0;
        }
        if( filter_pos[2]!=0 )
        {
          if (param.deblockingfilter&8)
            clip_tab_local = clip_tab_bs[mode_to_bs[modes_h[y][(x<<2)+4]]];    // overiding the global clip_tab;
          else
            clip_tab_local = clip_tab;
          tmp = (q1+((q0+p0)<<2)-p1+4)>>3;
          delta = min(max(-clip_tab_local[dbf_idx_a[component]], (tmp-q0)), clip_tab_local[dbf_idx_a[component]]);
          tmp_p[2]=min(255,max(0, q0+delta));
        }
        else
        {
          tmp_p[2]=q0;
        }
        if( filter_pos[0]!=0 && component == 0 )
        {
          if (param.deblockingfilter&8)
            clip_tab_local = clip_tab_bs[mode_to_bs[modes_h[y][(x<<2)+2]]];    // overiding the global clip_tab;
          else
            clip_tab_local = clip_tab;

          if (param.deblockingfilter & 4)
          {
            tmp = ((p2*7) +(p1*3)+(p0*5)+q0+8)>>4;
          }
          else
          {
            tmp = ( (p2<<1)+ p0+q0 + 2 )>>2;
          }
          delta = min(max(-clip_tab_local[dbf_idx_a[component]], (tmp-p1)), clip_tab_local[dbf_idx_a[component]]);
          tmp_p[0]=min(255,max(0, p1+delta));
        }
        else
        {
          tmp_p[0]=p1;
        }
        if( filter_pos[3]!=0 && component == 0 )
        {
          if (param.deblockingfilter&8)
            clip_tab_local = clip_tab_bs[mode_to_bs[modes_h[y][(x<<2)+5]]];    // overiding the global clip_tab;
          else
            clip_tab_local = clip_tab;

          if (param.deblockingfilter & 4)
          {
            tmp = ((q2*7) +(q1*3)+(q0*5)+p0+8)>>4;
          }
          else
          {
            tmp = ( (q2<<1)+ q0+p0 + 2 )>>2;
          }
          delta = min(max(-clip_tab_local[dbf_idx_a[component]], (tmp-q1)), clip_tab_local[dbf_idx_a[component]]);
          tmp_p[3]=min(255,max(0, q1+delta));
        }
        else
        {
          tmp_p[3]=q1;
        }
      }
      else
      {
        if( x!= 0 )
        {
          oY[y*width_f+(x<<2)-2]=tmp_p[0];
          oY[y*width_f+(x<<2)-1]=tmp_p[1];
          oY[y*width_f+(x<<2)]   =tmp_p[2];
          oY[y*width_f+(x<<2)+1]=tmp_p[3];
        }
        tmp_p[0] = oY[y*width_f+(x<<2)+2];
        tmp_p[1] = oY[y*width_f+(x<<2)+3];
        tmp_p[2] = oY[y*width_f+(x<<2)+4];
        tmp_p[3] = oY[y*width_f+(x<<2)+5];
      }
    }
    oY[y*width_f+(x<<2)-2]=tmp_p[0];
    oY[y*width_f+(x<<2)-1]=tmp_p[1];
    oY[y*width_f+(x<<2)]   =tmp_p[2];
    oY[y*width_f+(x<<2)+1]=tmp_p[3];
  }
}


//-----------------------------------------------------------------------------
/*!
 * \brief Copy 
 *
 * \param[in] lf pointer to loop filter
 * \param[in] mf pointer to motion field
 */
//-----------------------------------------------------------------------------
void Deblocking_Filter::filter_data_copy(LoopFilter *lf, MotionField *mf) 
{
  memcpy(m_luma[0][0],lf->lumaH,sizeof(int)*lf->width*lf->height) ;
  memcpy(m_luma[1][0],lf->lumaV,sizeof(int)*lf->width*lf->height) ;
  memcpy(m_chroma[0][0][0],lf->chromaH[0],sizeof(int)*lf->width2*lf->height2) ;
  memcpy(m_chroma[0][1][0],lf->chromaH[1],sizeof(int)*lf->width2*lf->height2) ;
  memcpy(m_chroma[1][0][0],lf->chromaV[0],sizeof(int)*lf->width2*lf->height2) ;
  memcpy(m_chroma[1][1][0],lf->chromaV[1],sizeof(int)*lf->width2*lf->height2) ;
  str_mf->copy_field_from(mf);
}
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
