/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
 * \file DenoiseTransformDCT8.cpp
 *
 * \brief Implementation of DenoiseTransformDCT8 class
 *
 */
//-----------------------------------------------------------------------------

#include "DenoiseTransformDCT8.H"


DenoiseTransformDCT8::DenoiseTransformDCT8(int width, bool has_alternate) : DenoiseTransform(width, 8, 338 * 338 >> 6, 338, has_alternate)
{
}

void DenoiseTransformDCT8::doColumnsForward(const int *src, int src_row, int src_pitch, int col_begin, int col_end, bool use_alternate)
{
  int col;
  const int *src_ptr = src + src_pitch * src_row;
  int *dst_ptr = (use_alternate) ? m_buffer_alt : m_buffer_fwd;
  
  for (col = col_begin; col <= col_end; col++)
  {
    if (m_mask[col])
    {
      int a0, a1, a2, a3, a4, a5, a6, a7;
      
      a0 = src_ptr[0 * src_pitch + col];
      a1 = src_ptr[1 * src_pitch + col];
      a2 = src_ptr[2 * src_pitch + col];
      a3 = src_ptr[3 * src_pitch + col];
      a4 = src_ptr[4 * src_pitch + col];
      a5 = src_ptr[5 * src_pitch + col];
      a6 = src_ptr[6 * src_pitch + col];
      a7 = src_ptr[7 * src_pitch + col];
      
      int b0, b1, b2, b3, b4, b5, b6, b7;
      
      b0 = a0 + a7;
      b1 = a1 + a6;
      b2 = a2 + a5;
      b3 = a3 + a4;
      b4 = a3 - a4;
      b5 = a2 - a5;
      b6 = a1 - a6;
      b7 = a0 - a7;
      
      int c0, c1, c2, c3, c4, c5, c6, c7;
      
      c0 = b0 + b3;
      c1 = b1 + b2;
      c2 = b1 - b2;
      c3 = b0 - b3;
      c4 = b4;
      c5 = b5;
      c6 = b6;
      c7 = b7;
      
      int d0, d1, d2, d3, d4, d5, d6, d7;
      
      d0 = 13 * (c0 + c1);
      d1 = 13 * (c0 - c1);
      d2 = 17 * c3 + 7 * c2;
      d3 = 7 * c3 - 17 * c2;
      d4 = 19 * c7 + 15 * c6 +  9 * c5 +  3 * c4;
      d5 = 15 * c7 - 19 * c6 -  3 * c5 +  9 * c4;
      d6 =  9 * c7 +  3 * c6 - 19 * c5 - 15 * c4;
      d7 =  3 * c7 -  9 * c6 + 15 * c5 - 19 * c4;
      
      dst_ptr[8 * col + 0] = shift_right_round(d0, 1);
      dst_ptr[8 * col + 1] = shift_right_round(d1, 1);
      dst_ptr[8 * col + 2] = shift_right_round(d2, 1);
      dst_ptr[8 * col + 3] = shift_right_round(d3, 1);
      dst_ptr[8 * col + 4] = shift_right_round(d4, 1);
      dst_ptr[8 * col + 5] = shift_right_round(d5, 1);
      dst_ptr[8 * col + 6] = shift_right_round(d6, 1);
      dst_ptr[8 * col + 7] = shift_right_round(d7, 1);
    }
  }
}

void DenoiseTransformDCT8::doRowsForward(int src_col, bool use_alternate)
{
  int row;
  const int *src_ptr = ((use_alternate) ? m_buffer_alt : m_buffer_fwd) + 8 * src_col;
  int *dst_ptr = (use_alternate) ? m_block_alt : m_block;
  
  for (row = 0; row < 8; row++)
  {
    int a0, a1, a2, a3, a4, a5, a6, a7;
    
    a0 = src_ptr[0 * 8 + row];
    a1 = src_ptr[1 * 8 + row];
    a2 = src_ptr[2 * 8 + row];
    a3 = src_ptr[3 * 8 + row];
    a4 = src_ptr[4 * 8 + row];
    a5 = src_ptr[5 * 8 + row];
    a6 = src_ptr[6 * 8 + row];
    a7 = src_ptr[7 * 8 + row];
    
    int b0, b1, b2, b3, b4, b5, b6, b7;
    
    b0 = a0 + a7;
    b1 = a1 + a6;
    b2 = a2 + a5;
    b3 = a3 + a4;
    b4 = a3 - a4;
    b5 = a2 - a5;
    b6 = a1 - a6;
    b7 = a0 - a7;
    
    int c0, c1, c2, c3, c4, c5, c6, c7;
    
    c0 = b0 + b3;
    c1 = b1 + b2;
    c2 = b1 - b2;
    c3 = b0 - b3;
    c4 = b4;
    c5 = b5;
    c6 = b6;
    c7 = b7;
    
    int d0, d1, d2, d3, d4, d5, d6, d7;
    
    d0 = 13 * (c0 + c1);
    d1 = 13 * (c0 - c1);
    d2 = 17 * c3 + 7 * c2;
    d3 = 7 * c3 - 17 * c2;
    d4 = 19 * c7 + 15 * c6 +  9 * c5 +  3 * c4;
    d5 = 15 * c7 - 19 * c6 -  3 * c5 +  9 * c4;
    d6 =  9 * c7 +  3 * c6 - 19 * c5 - 15 * c4;
    d7 =  3 * c7 -  9 * c6 + 15 * c5 - 19 * c4;
    
    dst_ptr[0 * 8 + row] = shift_right_round(d0, 1);
    dst_ptr[1 * 8 + row] = shift_right_round(d1, 1);
    dst_ptr[2 * 8 + row] = shift_right_round(d2, 1);
    dst_ptr[3 * 8 + row] = shift_right_round(d3, 1);
    dst_ptr[4 * 8 + row] = shift_right_round(d4, 1);
    dst_ptr[5 * 8 + row] = shift_right_round(d5, 1);
    dst_ptr[6 * 8 + row] = shift_right_round(d6, 1);
    dst_ptr[7 * 8 + row] = shift_right_round(d7, 1);
  }
}

void DenoiseTransformDCT8::doRowsInverse(int dst_col)
{
  int row;
  int *dst_ptr = m_buffer_inv + 8 * dst_col;
  
  for (row = 0; row < 8; row++)
  {
    int d0, d1, d2, d3, d4, d5, d6, d7;

    d0 = m_block[0 * 8 + row];
    d1 = m_block[1 * 8 + row];
    d2 = m_block[2 * 8 + row];
    d3 = m_block[3 * 8 + row];
    d4 = m_block[4 * 8 + row];
    d5 = m_block[5 * 8 + row];
    d6 = m_block[6 * 8 + row];
    d7 = m_block[7 * 8 + row];
    
    int c0, c1, c2, c3, c4, c5, c6, c7;
    
    c0 = 13 * (d0 + d1);
    c1 = 13 * (d0 - d1);
    c2 = 7 * d2 - 17 * d3;
    c3 = 17 * d2 + 7 * d3;
    c4 =  3 * d4 +  9 * d5 - 15 * d6 - 19 * d7;
    c5 =  9 * d4 -  3 * d5 - 19 * d6 + 15 * d7;
    c6 = 15 * d4 - 19 * d5 +  3 * d6 -  9 * d7;
    c7 = 19 * d4 + 15 * d5 +  9 * d6 +  3 * d7;
    
    int b0, b1, b2, b3, b4, b5, b6, b7;
    
    b0 = c0 + c3;
    b1 = c1 + c2;
    b2 = c1 - c2;
    b3 = c0 - c3;
    b4 = c4;
    b5 = c5;
    b6 = c6;
    b7 = c7;
    
    int a0, a1, a2, a3, a4, a5, a6, a7;
    
    a0 = b0 + b7;
    a1 = b1 + b6;
    a2 = b2 + b5;
    a3 = b3 + b4;
    a4 = b3 - b4;
    a5 = b2 - b5;
    a6 = b1 - b6;
    a7 = b0 - b7;
    
    dst_ptr[0 * 8 + row] += shift_right_round(a0, 1);
    dst_ptr[1 * 8 + row] += shift_right_round(a1, 1);
    dst_ptr[2 * 8 + row] += shift_right_round(a2, 1);
    dst_ptr[3 * 8 + row] += shift_right_round(a3, 1);
    dst_ptr[4 * 8 + row] += shift_right_round(a4, 1);
    dst_ptr[5 * 8 + row] += shift_right_round(a5, 1);
    dst_ptr[6 * 8 + row] += shift_right_round(a6, 1);
    dst_ptr[7 * 8 + row] += shift_right_round(a7, 1);
  }
}

void DenoiseTransformDCT8::doColumnsInverse(int **dst, int col_begin, int col_end)
{
  int col;
  
  for (col = col_begin; col <= col_end; col++)
  {
    if (m_mask[col])
    {
      int d0, d1, d2, d3, d4, d5, d6, d7;
      
      d0 = m_buffer_inv[8 * col + 0] >> 6;
      d1 = m_buffer_inv[8 * col + 1] >> 6;
      d2 = m_buffer_inv[8 * col + 2] >> 6;
      d3 = m_buffer_inv[8 * col + 3] >> 6;
      d4 = m_buffer_inv[8 * col + 4] >> 6;
      d5 = m_buffer_inv[8 * col + 5] >> 6;
      d6 = m_buffer_inv[8 * col + 6] >> 6;
      d7 = m_buffer_inv[8 * col + 7] >> 6;
      
      int c0, c1, c2, c3, c4, c5, c6, c7;
      
      c0 = 13 * (d0 + d1);
      c1 = 13 * (d0 - d1);
      c2 = 7 * d2 - 17 * d3;
      c3 = 17 * d2 + 7 * d3;
      c4 =  3 * d4 +  9 * d5 - 15 * d6 - 19 * d7;
      c5 =  9 * d4 -  3 * d5 - 19 * d6 + 15 * d7;
      c6 = 15 * d4 - 19 * d5 +  3 * d6 -  9 * d7;
      c7 = 19 * d4 + 15 * d5 +  9 * d6 +  3 * d7;
      
      int b0, b1, b2, b3, b4, b5, b6, b7;
      
      b0 = c0 + c3;
      b1 = c1 + c2;
      b2 = c1 - c2;
      b3 = c0 - c3;
      b4 = c4;
      b5 = c5;
      b6 = c6;
      b7 = c7;
      
      int a0, a1, a2, a3, a4, a5, a6, a7;
      
      a0 = b0 + b7;
      a1 = b1 + b6;
      a2 = b2 + b5;
      a3 = b3 + b4;
      a4 = b3 - b4;
      a5 = b2 - b5;
      a6 = b1 - b6;
      a7 = b0 - b7;
      
      dst[0][col] += shift_right_round(a0, 1);
      dst[1][col] += shift_right_round(a1, 1);
      dst[2][col] += shift_right_round(a2, 1);
      dst[3][col] += shift_right_round(a3, 1);
      dst[4][col] += shift_right_round(a4, 1);
      dst[5][col] += shift_right_round(a5, 1);
      dst[6][col] += shift_right_round(a6, 1);
      dst[7][col] += shift_right_round(a7, 1);
    }
  }
}
