/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MLC_Stat.cpp
*
* \brief Implementation of MLC_Stat class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MLC_Stat.H"
#include "Global.H"

//-----------------------------------------------------------------------------
// Constructor/Destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
*/
//-----------------------------------------------------------------------------
MLC_Stat::MLC_Stat()
{
  reset();
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Reset
*
* Resets fields to 0.
*/
//-----------------------------------------------------------------------------
void MLC_Stat::reset()
{
  int i,j;
  
  motion_bits = ZERO;
  luma_bits   = ZERO;
  chroma_bits = ZERO;
  aif_bits = ZERO;
  
  bit_count_luma   = ZERO;
  bit_count_chroma = ZERO;
  
#ifdef INTRA_MODE_STATS
  memset(sip_type_count, 0, sizeof(int)*NSIP);

 int ii, jj;
  for (jj=0; jj<NSIP; jj++ ) // number of SIP 
  { 
    for (ii=0; ii<NO_INTRA_PMODE_LUMA_MAX; ii++ ) 
    {
      sip_mode_count[jj][0][ii] = 0;       
      sip_mode_count[jj][1][ii] = 0;       
      sip_mode_count[jj][2][ii] = 0;       
    }
  }
#endif //INTRA_MODE_STATS

  intra_tma_L5=ZERO;
  intra_tma_L9=ZERO;
  intra_tma_L5c=ZERO;
  intra_tma_L9c=ZERO;
  
  mvshare_two=ZERO;
  mvshare_four=ZERO;
  mvshare_zero=ZERO;
  mvshare_off=ZERO;

  zblks  = ZERO;
  nzblks = ZERO;
  iblks  = ZERO;
  
  for (i=0; i<5; i++)
    list_stat[i] = ZERO;
  
  for (i=0; i<LISTS; i++)
  {
    for (j=0; j<MAX_REF_FRAMES; j++)
    {
      ref_frame_stat[i][j] = ZERO;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Combine statistics
*
* \param[in] mlc Statistics to be added
*/
//----------------------------------------------------------------------------
void MLC_Stat::add(MLC_Stat *mlc)
{
  int i,j;
  
  motion_bits += mlc->motion_bits;
  luma_bits   += mlc->luma_bits;
  chroma_bits += mlc->chroma_bits;
  aif_bits += mlc->aif_bits;
  
  bit_count_luma   += mlc->bit_count_luma;
  bit_count_chroma += mlc->bit_count_chroma;
  
  intra_tma_L5 += mlc->intra_tma_L5;
  intra_tma_L9 += mlc->intra_tma_L9;
  intra_tma_L5c += mlc->intra_tma_L5c;
  intra_tma_L9c += mlc->intra_tma_L9c;
  
  mvshare_two += mlc->mvshare_two;
  mvshare_four += mlc->mvshare_four;
  mvshare_zero += mlc->mvshare_zero;
  mvshare_off += mlc->mvshare_off;
  
  zblks  += mlc->zblks;
  nzblks += mlc->nzblks;
  iblks  += mlc->iblks;
  
  for (i=0; i<5; i++)
    list_stat[i] += mlc->list_stat[i];
  
  for (i=0; i<LISTS; i++)
  {
    for (j=0; j<MAX_REF_FRAMES; j++)
    {
      ref_frame_stat[i][j] += mlc->ref_frame_stat[i][j];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Get number of samples in list_stat
*
* \return Number of samples
*/
//-----------------------------------------------------------------------------
int MLC_Stat::list_stat_norm()
{
  int i;
  int sum;
  
  sum = ZERO;
  
  for (i=0; i<5; i++)
    sum += list_stat[i];
  
  return sum;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get number of samples in ref_frame_stat
*
* \param[in] list List to consider (LIST_0 or LIST_1)
*
* \return Number of samples
*/
//-----------------------------------------------------------------------------
int MLC_Stat::ref_frame_stat_norm(int list)
{
  int i;
  int sum;
  
  sum = ZERO;
  
  for (i=0; i<4; i++)
    sum += ref_frame_stat[list][i];
  
  return sum;
}
#ifdef INTRA_MODE_STATS
//-----------------------------------------------------------------------------
/*!
* \brief Print the intra mode statistics in log file
*
* \param[in] sip_type the sip type being printed for
* \param[in] mlc_tot the pointer to the MLC statistics class
* \param[in] num_of_frame total number of frames have been encoded
* \return Number of samples
*/
//-----------------------------------------------------------------------------
#ifdef VERBOSE_LOG
void MLC_Stat::print_intra_mode_stats(int sip_type, MLC_Stat* mlc_tot, int num_of_frame)
#else
void MLC_Stat::print_intra_mode_stats(int /*sip_type*/, MLC_Stat* /*mlc_tot*/, int /*num_of_frame*/)
#endif
{
#ifdef VERBOSE_LOG
  int i, max_mode_num;
  int sip_count = sip_type_count[sip_type];
  int tot_blks  = mlc_tot->iblks/num_of_frame/4;
  int mode_blks_count = (sip_count<<2);

  if (mode_blks_count == 0)  mode_blks_count = 1;

  switch (sip_type)
  {
  case SIP8x8:
    {
      max_mode_num = NO_INTRA_PMODE_LUMA_8x8-2;
      printf("INTRA8x8:   ");
      break;
    }
  case SIP16x16:
    {
      max_mode_num = NO_INTRA_PMODE_LUMA_16x16;
      printf("INTRA16x16: ");
      break;
    }
  case SIP8x2:
    {
      max_mode_num = NO_INTRA_PMODE_LUMA_8x2;
      printf("INTRA8x2:   ");
      break;
    }
  case SIP2x8:
    {
      max_mode_num = NO_INTRA_PMODE_LUMA_2x8;
      printf("INTRA2x8:   ");
      break;
    }
  case SIP4x4:
    {
      max_mode_num = NO_INTRA_PMODE_LUMA_4x4;
      printf("INTRA4x4:   ");
      break;
    }
  default:
    {
      printf("sip_type is not correct.\n");
      exit(1);
    }
  }

 printf("%-6.2f%% ", sip_count*100.0/tot_blks);

 for(i=0; i<max_mode_num; i++)
 {
   if ( i== max_mode_num-1 )
   {
      printf("%d: %-6.2f%%\n", i, sip_mode_count[sip_type][0][i]*100.0/mode_blks_count);
   }
   else
   {
      printf("%d: %-6.2f%%, ", i, sip_mode_count[sip_type][0][i]*100.0/mode_blks_count);
   }
 }
#endif
}
#endif //INTRA_MODE_STATS
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
