/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file MainDecoder.cpp
*
* \brief Main decoding program
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"
#include "Decoder.H"
#include "Quantizer.H"
#include "Param.H"
#include "FastDCT.H"
#include "AIF_adaptive_filter.H"

#ifdef TRACE_DEBUG
#include "Trace.H"
#endif
//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

static int help_mode = FALSE; //!< Flag for using help mode
#ifdef INCOGNITO
static int silent_mode = TRUE; //!< Flag for using silent mode
#else
static int silent_mode = FALSE; //!< Flag for using silent mode
#endif
static int file_mode = FALSE; //!< Flag for outputting stat files
static int output_mode = TRUE; //!< Flag for outputting decoded sequence
#ifdef VERBOSE_STATS
static int verbose_mode = FALSE; //!< Flag for using verbose statistic mode
#endif
int aif_loop[LISTS][MAX_REF_FRAMES];//!< Variable used to determine whether AIF interpolated data is used for prediction

//-----------------------------------------------------------------------------
// Main function
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Main decoding function
*
* \param[in] argc Argument count
* \param[in] argv Array of argument values
* \return EXIT_SUCCESS or EXIT_FAILURE
*/
//-----------------------------------------------------------------------------
int main(int argc, char **argv) 
{
  FILE *fcomp; // compressed file
  int fdec = -1;  // decoded file descriptor
  FILE *f_stat;
  
  Buffer *buf;  // byte buffer
#ifdef DEC_TRANSCODER
  Buffer *buf_out = NULL; // When transcoding.
  FILE   *ftranscoded = NULL;
#endif
  Decoder *dec; // decoder
  Frame *curr; // current frame
  Frame *curr_filtered;
  Frame *curr_pred;
  Frame *curr_qpe_m;
  Frame *curr_qpe_s;
  int frame; // current frame number
  int i, j;
  int last_frame_number = 0;
  int a;
  
  // start here
  
  for (a=1; a<argc; a++) 
  {
    if (argv[a][ZERO] == '-') 
    {
      if (strcmp(argv[a], "-h") == ZERO)
        help_mode = TRUE;
#ifndef INCOGNITO
      else if (strcmp(argv[a], "-s") == ZERO)
        silent_mode = TRUE;
#ifdef VERBOSE_STATS
      else if (strcmp(argv[a], "-v") == ZERO)
        verbose_mode = TRUE;
#endif
      else if (strcmp(argv[a], "-f") == ZERO)
        file_mode = FALSE;
      else if (strcmp(argv[a], "-F") == ZERO)
        file_mode = TRUE;
      else if (strcmp(argv[a], "-N") == ZERO)
        output_mode = FALSE;
#ifdef SUBSTREAM_MT
      else if (strcmp(argv[a], "-dt") == ZERO && argv[a+1] != NULL && atoi(argv[a+1]) > 0)
      {
          param.decode_nthreads = atoi(argv[a+1]);
          a++;
      }
#endif
#ifdef DENOISE_MT
      else if (strcmp(argv[a], "-dnt") == ZERO && argv[a+1] != NULL && atoi(argv[a+1]) > 0)
      {
          param.decode_ndenoisethreads = atoi(argv[a+1]);
          a++;
      }
#endif
#ifdef WIENER_MT
      else if (strcmp(argv[a], "-wft") == ZERO && argv[a+1] != NULL && atoi(argv[a+1]) > 0)
      {
          param.decode_nwfthreads = atoi(argv[a+1]);
          a++;
      }
#endif
#ifdef DEBLOCK_MT
      else if (strcmp(argv[a], "-dbt") == ZERO && argv[a+1] != NULL && atoi(argv[a+1]) > 0)
      {
          param.decode_ndeblockthreads = atoi(argv[a+1]);
          a++;
      }
#endif
#ifdef CABAC_SYNC_POINTS
      else if (strcmp(argv[a], "-inputref") == ZERO)
          param.decode_inputref = TRUE;
#endif
#ifdef DEC_TRANSCODER
      else if (strcmp(argv[a], "-tt") == ZERO && argv[a+1] != NULL && atoi(argv[a+1]) > 0)
      {
          param.decode_transcode = TRUE;
          param.decode_transcode_nthreads = atoi(argv[a+1]);
          a++;
      }
#endif
#endif // INCOGNITO
      else
        help_mode = TRUE;
    }
    else
      break;
  }
  
  if (a != argc - 2 && a != argc - 1) // Two for bitstream file and output file
    help_mode = TRUE;

#ifndef INCOGNITO
  if (help_mode == TRUE)
    silent_mode = FALSE;
  
  if (silent_mode == FALSE)
    printf("CDCM video decoder\n");
#endif
  
  if (help_mode == TRUE) 
  {
    printf("Usage: %s [options] <bitstream file>.bit [<output file>.yuv]\n", argv[ZERO]);
    printf("Options:\n");
    printf("-h : help mode\n");
#ifndef INCOGNITO
    printf("-s : silent mode\n");
#ifdef CABAC_SYNC_POINTS
    printf("-inputref: the input is a 2.5.3 reference stream (no synchro, no threads)\n");
#endif
#ifdef DEC_TRANSCODER
    printf("-tt n: transcode to a new stream containing n substreams per frame.\n");
#endif
#ifdef SUBSTREAM_MT
    printf("-dt n: decode stream with at most n threads.\n");
#endif
#ifdef DENOISE_MT
    printf("-dnt n: denoise stream with n threads.\n");
#endif
#ifdef WIENER_MT
    printf("-wft n: wiener filter with n threads.\n");
#endif
#ifdef DEBLOCK_MT
    printf("-dbt n: deblock with n threads.\n");
#endif
#ifdef VERBOSE_STATS
    printf("-v : verbose stats mode\n");
#endif
#endif
    exit(EXIT_FAILURE);
  }
  
  buf = new Buffer;
  
  if (file_mode == TRUE) 
  {
    f_stat = fopen("stats_dec.txt", "wt");
  }
  else 
  {
    f_stat = NULL;
  }
  
  fcomp = fopen(argv[a],"rb");
  if (!fcomp) 
  {
    fprintf(stderr, "Could not open file [%s].\n", argv[a]);
    exit(EXIT_FAILURE);
  }
  
  if (output_mode == TRUE) 
  {
    const char *outFilename = (a == argc - 2) ? argv[a+1] : "dec.yuv";
    fdec = open(outFilename, OPENFLAGS_WRITE, OPEN_PERMISSIONS);
    if (fdec == -1) 
    {
      fprintf(stderr, "Could not open file.\n");
      exit(EXIT_FAILURE);
    }
  }
    
#ifdef DEC_TRANSCODER    
  if (param.decode_transcode)
  {
      ftranscoded = fopen("transcoded.dcm", "wb");
      if (ftranscoded == NULL)
      {
          fprintf(stderr, "Coult not open transcode output file 'transcoded.dcm'\n");
          exit(EXIT_FAILURE);
      }
      buf_out = new Buffer;
      buf_out->reset();
  }
#endif

  param.ini_qmatrix(0); // cq_start
  param.ini_qmatrix(1); // cq_end

  buf->reset();
  buf->read(fcomp);
  if (buf->payload_type != Buffer::PLT_HEADER)
  {
    fprintf(stderr, "Could not find header packet in bitstream file.\n");
    exit(EXIT_FAILURE);
  }

  param.read(buf);
#ifdef DEC_TRANSCODER
  if (param.decode_transcode)
  {
      // Limit number of substreams to one per superblock line to prevent
      // writing empty substream data that's never used.
      int nthread_max = ( (param.frame_height+7)/8 + param.sb_height-1 )/param.sb_height;
      if (param.decode_transcode_nthreads > nthread_max)
      {
        printf("Transcode: number of output streams limited to %d (not -tt parameter %d) due to frame height of %d pixels\n",
               nthread_max, param.decode_transcode_nthreads, param.frame_height);
        param.decode_transcode_nthreads = nthread_max;
      }
      // We write our input parameters, possibly changing the number of threads and adding
      // synchronisation.
      int old_synchro = param.superblock_synchro;
      int old_nthreads = param.superblock_nsubstreams;
      param.superblock_synchro = 1;
      param.superblock_nsubstreams = param.decode_transcode_nthreads;
      param.write(buf_out);
      param.superblock_synchro = old_synchro;
      param.superblock_nsubstreams = old_nthreads;
      buf_out->payload_type = Buffer::PLT_HEADER;
      buf_out->write(ftranscoded);
      buf_out->reset();
  }
#endif
  param.update();
  
  curr = new Frame(param.frame_height, param.frame_width);
  curr_filtered = new Frame(param.frame_height, param.frame_width);  
  curr_pred = new Frame(param.frame_height, param.frame_width);
  curr_qpe_m= new Frame(param.frame_height, param.frame_width);
  curr_qpe_s= new Frame(param.frame_height, param.frame_width);
  dec = new Decoder(param.frame_height, param.frame_width);

#ifdef TRACE_STATS
  if (strlen(param.mode_stats_fname) == 0)
    strcpy(param.mode_stats_fname,"mode_stats.md");
  if (file_mode)
    param.mode_stats_file = fopen(param.mode_stats_fname, "wb");
  else
    param.mode_stats_file = NULL;
  
  if (strlen(param.mv_stats_fname) == 0)
    strcpy(param.mv_stats_fname,"mv_stats.mv");
  if (file_mode)
    param.mv_stats_file = fopen(param.mv_stats_fname, "wb");
  else
    param.mv_stats_file = NULL;

#endif
#ifdef TRACE_DEBUG
#ifdef FRAME_SUBSTREAMS
  OPEN_DTRACE(param.superblock_nsubstreams);
#else
  OPEN_DTRACE;
#endif
#endif  
  if (silent_mode == FALSE)
  {
    printf("----------------------------------------------\n");
    printf("Frame    "); // 9
    printf("QP"); // 4
    printf("\n");
  }
  for (frame = 0; TRUE; frame++) 
  {
    if (buf->read(fcomp) == FALSE)
      break; // stop if read error
    
    if (buf->payload_type == Buffer::PLT_END)
      break; // stop if end of stream signal
    
    curr->frame_no = last_frame_number;
#ifdef DEC_TRANSCODER
    dec->decode_frame(curr, curr_pred, curr_qpe_m, curr_qpe_s, curr_filtered, buf, buf_out);
#else
    dec->decode_frame(curr, curr_pred, curr_qpe_m, curr_qpe_s, curr_filtered, buf);
#endif
    last_frame_number = curr->frame_no;
    
    if (silent_mode == FALSE) 
    {
      switch (curr->frame_type)  
      {
        case ft_I_frame:
          if (curr->intra_frame_type==ft_I_IDR)
          {
            printf("%05d(R) ",curr->frame_no);
            break;
          }
          else if (curr->intra_frame_type==ft_I_DDR)
          {
            printf("%05d(D) ",curr->frame_no);
            break;
          }
          if (curr->reference_idc)
            printf("%05d(i) ",curr->frame_no);
          else
            printf("%05d(I) ",curr->frame_no);
          break;
        case ft_P_frame:
          if (curr->reference_idc)
            printf("%05d(p) ",curr->frame_no);
          else
            printf("%05d(P) ",curr->frame_no);
          break;
        case ft_B_frame:
        default:
          if (curr->reference_idc)
            printf("%05d(b) ",curr->frame_no);
          else
            printf("%05d(B) ",curr->frame_no);
          break;
      }
      printf("%03d\n", curr->qp);
      #ifdef VERBOSE_STATS 
      dec->print_stats_summary(stdout, verbose_mode);
      #else
      dec->print_stats_summary(stdout);
      #endif
    }
    
#ifdef DEC_TRANSCODER
    if (param.decode_transcode)
    {
        buf_out->payload_type = Buffer::PLT_FRAME;
        buf_out->write(ftranscoded);
        buf_out->reset();
    }
#endif

    dec->print_stats(f_stat);
    
    if (fdec != -1) 
    {
      lseek (fdec, (int64) curr->size*curr->frame_no, SEEK_SET);

      if (param.wienerfilter!=0)
      {
        write(fdec, curr_filtered->data, curr->size);
      }
      else
      {
        write(fdec, curr->data, curr->size);
      }

    }
  }

#ifdef TRACE_STATS
  if (param.mode_stats_file)  
    fclose(param.mode_stats_file);
  if (param.mv_stats_file)
    fclose(param.mv_stats_file);
#endif
#ifdef TRACE_DEBUG
  CLOSE_DTRACE;
#endif  
  fclose(fcomp);
  if (fdec != -1)
    close(fdec);

#ifdef DEC_TRANSCODER
  if (param.decode_transcode)
  {
    buf_out->reset();
    buf_out->payload_type = Buffer::PLT_END;
    buf_out->payload_size = 0;
    buf_out->write(ftranscoded);
    fclose(ftranscoded);
  }
#endif

  //{{SUBSTREAM_MT
  // The code below has not been tested when multi-threading.
  // The stats are now kept in an EntropyDecoder (dec->edecs[0]),
  // not in the global 'param'.
  //
  // The DISABLE_ZPATT_OUT define has always been in effect,
  // so this code has never been used.  The initial comment
  // indicates there were crashes with the code.
  //SUBSTREAM_MT}}
#ifndef DISABLE_ZPATT_OUT
//  Bellow is comment out temporally.
//  Because error happens at "fprintf(fcomp, "%d\n", param.ztree[j]->stats[i]);"

  if (file_mode == TRUE) 
  {
    fcomp = fopen("zpatt.txt","w");
    for (j=0; j<NZTREES; j++) 
    {
      fprintf(fcomp, "%d\n", param.ztree[j]->nleaves);
      for (i=0; i<power2(param.ztree[j]->nleaves); i++)
      fprintf(fcomp, "%d\n", param.ztree[j]->stats[i]);
      fprintf(fcomp, "\n");
    }
    fclose(fcomp);
    
    if (param.trans8x8_enable) 
    {
      fcomp = fopen("zpatt_64.txt","w");
      
      for (j=0; j<NZTREES64; j++) 
      {
        fprintf(fcomp, "%d\n", param.dec->getzt64(j).s_nleaves);
        for (i=0; i<power2(param.dec->getzt64(j).s_nleaves); i++)
        fprintf(fcomp, "%d\n", param.dec->getzt64(j).stats[i]);
        fprintf(fcomp, "\n");
      }
      fclose(fcomp); 
    }
  }
#endif

  if (silent_mode == FALSE) 
  {
    printf("Number of I-frames: %d (%6.2f%%)\n",
    dec->get_frame_type_stat(ft_I_frame), 
    100.0*dec->get_frame_type_stat(ft_I_frame)/dec->get_total_frames());
    printf("Number of P-frames: %d (%6.2f%%)\n",
    dec->get_frame_type_stat(ft_P_frame), 
    100.0*dec->get_frame_type_stat(ft_P_frame)/dec->get_total_frames());
    printf("Number of B-frames: %d (%6.2f%%)\n",
    dec->get_frame_type_stat(ft_B_frame), 
    100.0*dec->get_frame_type_stat(ft_B_frame)/dec->get_total_frames());

#ifdef FRAME_SUBSTREAMS
    dec->collect_mlc_stats();
#define mlc mlcs[0] // The first decoder has the complete stats.
#endif
    
    for (j=0; j<LISTS; j++)
      for (i=0; i<MAX_REF_FRAMES; i++)
        printf("Ref index %d (LIST_%d): %d (%6.2f%%)\n", i, j,
        dec->mlc.ref_frame_stat[j][i], 
        100.0*dec->mlc.ref_frame_stat[j][i]/dec->mlc.ref_frame_stat_norm(j));
    
    printf("List_0: %d (%6.2f%%)\n",
    dec->mlc.list_stat[LIST_0], 
    100.0*dec->mlc.list_stat[LIST_0]/dec->mlc.list_stat_norm());
    printf("List_1: %d (%6.2f%%)\n",
    dec->mlc.list_stat[LIST_1], 
    100.0*dec->mlc.list_stat[LIST_1]/dec->mlc.list_stat_norm());
    printf("BiPred: %d (%6.2f%%)\n",
    dec->mlc.list_stat[BI_PRED], 
    100.0*dec->mlc.list_stat[BI_PRED]/dec->mlc.list_stat_norm());
#ifdef FRAME_SUBSTREAMS
#undef mlc
#endif
  }
  
  delete curr;
  delete curr_filtered;
  delete curr_pred;
  delete curr_qpe_m;
  delete curr_qpe_s;
  
  delete dec;
  delete buf;
#ifdef DEC_TRANSCODER
  delete buf_out;
#endif
  
  if (f_stat != NULL) fclose(f_stat);
  
  return EXIT_SUCCESS;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
