/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Param.cpp
*
* \brief Implementation of Param class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Param.H"
#include "EntropyModel.H"
#include "Global.H"
#include "LoopFilterOvercomplete3.H"

#include "Block.H" // cq_start, cq_end
//-----------------------------------------------------------------------------
// Local classes
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Integer parameter class
*
* Definition of an integer parameter in encoder configuration file.
*/
//-----------------------------------------------------------------------------
struct IntegerParameter 
{
public:
  char name[64];
  int *ptr;
  int default_val;
  int min_val;
  int max_val;
};

//-----------------------------------------------------------------------------
/*!
* \brief Floating point parameter class
*
* Definition of a floating point parameter in encoder configuration file.
*/
//-----------------------------------------------------------------------------
struct FloatParameter 
{
public:
  char name[64];
  double *ptr;
  double default_val;
  double min_val;
  double max_val;
};

//-----------------------------------------------------------------------------
/*!
* \brief String parameter class
*
* Definition of a string parameter in encoder configuration file.
*/
//-----------------------------------------------------------------------------
struct StringParameter 
{
public:
  char name[64];
  char *ptr;
  const char *default_val;
};


//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

Param param; // contains global parameters (nasty global variable)

Block qm_block; // cq_start, cq_end
//-----------------------------------------------------------------------------
/*!
* \brief Set of integer parameters recognized by encoder
*
*/
//-----------------------------------------------------------------------------
IntegerParameter encoder_int_param_list_ext[] = 
{
  // !! DO NOT CHANGE DEFAULT SETTINGS IN THIS LIST
  // !! THIS LIST SHOULD CONTAIN ONLY PARAMETERS THAT NEED TO BE TRANSMITTED
  { "bugfix87",         &param.bugfix87,         FALSE, FALSE, TRUE,
  },
  { "bugfix88",         &param.bugfix88,         FALSE, FALSE, TRUE,
  },
  { "bugfix89",         &param.bugfix89,         FALSE, FALSE, TRUE,
  },
  { "bugfix90",         &param.bugfix90,         FALSE, FALSE, TRUE,
  },
  { "bugfix93",         &param.bugfix93,         FALSE, FALSE, TRUE,
  },
  { "alt8x8_trans_idx", &param.alt8x8_trans_idx,  0,     0,     2
  },
  { "ndf_lc_luma",       &param.ndf_lc_luma,        0,     0,     4
  },
  { "ndf_lc_chroma",     &param.ndf_lc_chroma,      0,     0,     4
  },
#ifdef I16_IN_P_FRAME
  { "intra16x16p",      &param.intra16x16p,       FALSE, FALSE, TRUE,
  },
#endif
  { "",                 NULL,                     0,     0,     0              
  }
};

IntegerParameter encoder_int_param_list[] = 
{
  { "width",            &param.frame_width,      176,   8,     8191           
  },
  { "height",           &param.frame_height,     144,   8,     8191           
  },
  { "nframes",          &param.nframes,          1,     1,     INT_INF        
  },
  { "f_header",         &param.f_header,         0,     0,     INT_INF        
  },
  { "frame_offset",     &param.frame_offset,     0,     0,     INT_INF        
  },  
  { "pyramid_coding",   &param.pyramid_coding,   2,     0,     7              
  }, 
  { "no_reordering",    &param.no_reordering,    FALSE, FALSE, TRUE           
  },
  { "qp",               &param.qp,               64,    0,     127            
  },
  { "lambdaq_base",     &param.lambdaq_base,     -49,   -128,  127            
  },
  { "melq_base",        &param.melq_base,        -8,    -32,   31             
  }, 
  { "lambdaq_offset",   &param.lambdaq_offset,   16,    0,     64             
  },
  { "lambda_mode_chroma", &param.lambda_mode_chroma, 1, 0, 1
  },
  { "mel_offset",       &param.mel_offset,       -8,     -16,   16             
  }, 
  { "adapt_list_size",  &param.adapt_list_size,  TRUE, FALSE, TRUE           
  },
  { "chroma_me",        &param.use_chroma_me,    2,  0, 2           
  },
  { "chroma_on_fly",    &param.chroma_on_fly,    0, 0, 1
  },
  { "me_range",         &param.me_range,         256,    8,     1024           
  },
  { "lime_size",     &param.lime_size,     8,     1,     16             
  },
  { "me_refine_s0",     &param.me_refine_s0,     2,     0,     32             
  },
  { "me_refine_s1",     &param.me_refine_s1,     2,     0,     32             
  },
  { "me_refine_s2",     &param.me_refine_s2,     2,     0,     32             
  },
  { "me_trellis_s0",    &param.me_trellis_s0,    2,     0,     32             
  },
  { "me_trellis_s1",    &param.me_trellis_s1,    2,     0,     32             
  },
  { "me_trellis_s2",    &param.me_trellis_s2,    2,     0,     32             
  },
  { "me_trellis_o0",    &param.me_trellis_o0,    2,     0,     32             
  },
  { "me_trellis_o1",    &param.me_trellis_o1,    2,     0,     32             
  },
  { "bugfix80",         &param.bugfix80,         FALSE, FALSE, TRUE,
  },
  { "bugfix94",         &param.bugfix94,         FALSE, FALSE, TRUE,
  },
  { "me_region_s0",     &param.me_region_s0,     2,     0,     32             
  },
  { "me_region_s1",     &param.me_region_s1,     2,     0,     32             
  },
  { "me_region_s2",     &param.me_region_s2,     2,     0,     32             
  },
  { "me_region_o0",     &param.me_region_o0,     2,     0,     32             
  },
  { "me_region_o1",     &param.me_region_o1,     2,     0,     32             
  },
  { "open_loop",        &param.open_loop,        FALSE, FALSE, TRUE           
  },
#ifdef BETTER_CHROMA_DC
  { "better_color",     &param.better_color,     7,     0,     7              
  },
  { "chroma_dc_th",     &param.chroma_dc_th,     60,     0,     127              
  },
  { "chroma_dc_scale",  &param.chroma_dc_scale,     32,     0,     255              
  },
#else
  { "better_color",     &param.better_color,     6,     0,     6              
  },
#endif
  { "bc_offset",        &param.bc_offset,        4,     -16,   15             
  },  
  { "q_rdoffset",       &param.q_rdoffset,       1,     0,     2           
  },
  { "q_rdoffset8x8",    &param.q_rdoffset8x8,    0,     0,     2           
  },
  { "q_adaptoffset",    &param.q_adaptoffset,    4,     0,     4              
  },
  { "qp_frame_mod",     &param.qp_frame_mod,     TRUE,  FALSE, TRUE           
  },  
  { "q_offsetLI",       &param.q_offset[0],      30720, 0,     65536          
  },
  { "q_offsetCI",       &param.q_offset[1],      30720, 0,     65536          
  },
  { "q_offsetLIinP",    &param.q_offset[4],      21845, 0,     65536          
  },
  { "q_offsetCIinP",    &param.q_offset[5],      21845, 0,     65536          
  },
  { "q_offsetLP",       &param.q_offset[2],      10923, 0,     65536          
  },
  { "q_offsetCP",       &param.q_offset[3],      10923, 0,     65536          
  },
  { "qmatrix_enable_flag",           &(param.qmatrix_enable_flag[0]),         FALSE,     FALSE,     TRUE   
  }, // cq_start
  { "qoffsetmatrix_enable_flag",     &(param.qmatrix_enable_flag[1]),         FALSE,     FALSE,     TRUE   
  }, // cq_end
  { "loss_rate",        &param.loss_rate,        5,     2,     16             
  },
  { "ac_exp_limit",     &param.ac_exp_limit,  128,     0,     1023             
  },
  { "chroma_mc",        &param.chroma_mc,        1,     0,     1              
  },
  { "luma_mc",          &param.luma_mc,          4,     0,     4              
  },
  { "me_subpel",        &param.me_subpel,         7,     0,     7              
  },
  { "gen8pel_filter",   &param.gen8pel_filter,   0,     0,     2              
  },
  { "mc_round",         &param.mc_round,         TRUE,  FALSE, TRUE           
  },
  { "onlyI",            &param.onlyI,            0,     0,     INT_INF        
  }, 
  { "idr_interval",     &param.idr_interval,     0,     0,     INT_INF
  },
  { "deferred_decoder_refresh",     &param.deferred_decoder_refresh,     0,     0,     1
  },
  { "enableB",          &param.enableB,          TRUE, FALSE, TRUE            
  },
  { "adapt_frame_type", &param.adapt_frame_type, TRUE, FALSE, TRUE            
  },
  { "intra16x16",      &param.intra16x16,       1,    0,   1
  },
  { "transform_selection_16x16", &param.transform_selection_16x16,  1,     0,     1
  },
  { "chroma8x8",       &param.chroma8x8,       1,    0,   1
  },
  { "sb_height",&param.sb_height,  8,     2,     64             
  }, 
  { "sb_width",&param.sb_width,  8,     2,     64             
  }, 
  { "me_repeat_edge",   &param.me_repeat_edge,    0,     0,     4             
  },
  { "me_repeat_intref_p",   &param.me_repeat_intref_p,    0,     0,     4             
  },
  { "me_repeat_intref_b",   &param.me_repeat_intref_b,    1,     0,     4             
  },
  { "me_repeat_disposable_p",   &param.me_repeat_disposable_p,    1,     0,     4             
  },
  { "me_repeat_disposable_b",   &param.me_repeat_disposable_b,    1,     0,     4             
  },
  { "me_loops2_edge",        &param.me_loops2_edge,        2,     0,     32             
  },
  { "me_loops3_edge",        &param.me_loops3_edge,        2,     0,     32             
  },
  { "me_loops2_intref",        &param.me_loops2_intref,        2,     0,     32             
  },
  { "me_loops3_intref",        &param.me_loops3_intref,        2,     0,     32             
  },
  { "me_loops2_disposable",        &param.me_loops2_disposable,        2,     0,     32             
  },
  { "me_loops3_disposable",        &param.me_loops3_disposable,        2,     0,     32             
  },
  { "me_loops1",        &param.me_loops1,        0,     0,     32             
  },
  { "me_loops2",        &param.me_loops2,        2,     0,     32             
  },
  { "me_loops3",        &param.me_loops3,        2,     0,     32             
  },
  { "P_subpel",         &param.P_subpel,         3,     1,     4              
  },
  { "B_subpel",         &param.B_subpel,         1,     1,     4              
  },
  { "treillis_ncand",   &param.treillis_ncand,   128,   64,    MAXNCAND       
  },
  { "region_ncand",     &param.region_ncand,     256,   128,   MAXNCAND       
  },
  { "lumaDC",           &param.lumaDC,           6,     0,     6              
  },
  { "candidate_search_region_intra16x16",  &param.candidate_search_region_intra16x16,  12,  5, 255    
  },
  { "candidate_search_region",  &param.candidate_search_region,  8,  5, 255   
  },
  { "num_candidate_averaged",   &param.num_candidate_averaged,   4,  1,  16
  },
  { "candidate_sad_threshold",  &param.candidate_sad_threshold, 32,  0, 255
  }, 
  { "mvs_realcost",     &param.mvs_realcost,     TRUE,  FALSE, TRUE
  },
  { "ndf_mode",       &param.ndf_mode,       3,     0,     3              
  },
  { "deblockingfilter"  , &param.deblockingfilter , 1+4+0,     0,     15
  },
  { "dbf_offset_a"      , &param.dbf_offset_a     , 16,     1,     31
  },
  { "dbf_offset_b"      , &param.dbf_offset_b     , 16,     1,     31
  },
#ifdef WIENER_FILTER_INLOOP_SWITCH
  { "wienerfilter_inloop" , &param.wienerfilter_inloop  , 1,     0,     1
  },
#endif
  { "wienerfilter"        , &param.wienerfilter         , 1,     0,     1
  },
  { "wlf_enable_Y"        , &param.wlf_enable[0]         , 1,     0,     1
  },
  { "wlf_enable_U"        , &param.wlf_enable[1]         , 0,     0,     1
  },
  { "wlf_enable_V"        , &param.wlf_enable[2]         , 0,     0,     1
  },
  { "fs_max_rec"          , &param.fs_max_rec  , 7,     0,     MAX_WIENER_FILTER_LENGTH
  },
  { "fs_max_qpe"          , &param.fs_max_qpe  , 3,     0,     MAX_WIENER_FILTER_LENGTH
  },
  { "fs_max_pred"         , &param.fs_max_pred , 3,     0,     MAX_WIENER_FILTER_LENGTH
  },
  { "qp_off_intra",     &param.qp_off_intra,      0,  -64,     64             
  }, 
  { "qp_off_onlyI",     &param.qp_off_onlyI,     -8,  -64,     64             
  },  
  { "qp_off_intra_scene_cut",     &param.qp_off_intra_scene_cut,     -8,  -64,     64             
  },  
  { "qp_off_intra_disposable_pic",     &param.qp_off_intra_disposable_pic,      0,  -64,     64             
  },
  { "qp_off_intref",    &param.qp_off_intref,    10,   -64,     64             
  },
  { "qp_off_intdisp",   &param.qp_off_intdisp,   18,  -64,     64             
  },
  { "trellis_mode",     &param.trellis_mode,     2,     0,     2              
  },
  { "frame_skip",       &param.frame_skip,       0,     0,     INT_INF        
  },
  { "intermed_frame",   &param.intermed_frame,   5,     0,     INT_INF        
  },
  { "intermed_disp",    &param.intermed_disp,    TRUE,  FALSE, TRUE           
  },
  { "fastME",           &param.fastME,           5,     0,     5              
  },
  { "fastMEDual" ,      &param.fastMEDual,       5,     0,     5              
  },
  { "fastMEFixed",      &param.fastMEFixed,      TRUE,  FALSE, TRUE           
  },
  { "nref_frames",      &param.nref_frames,      4,     0,     MAX_LIST_FRAMES 
  },
  { "enable_intcomp",   &param.enable_intcomp,   TRUE,  FALSE, TRUE           
  },
  { "DC_mult",          &param.DC_mult,          1,     1,     8              
  },
  { "weight_diff_scale",&param.weight_diff_scale,0,     0,     3              
  },
  { "weight_diff_max",  &param.weight_diff_max,  16,    0,     24
  },
  { "ndf_two_rounds",    &param.ndf_two_rounds,    FALSE,  FALSE, TRUE           
  },
  { "ndf_weighting",     &param.ndf_weighting,     TRUE,  FALSE, TRUE           
  },
#ifdef AUTO_SELECT_HADAMARD
  { "hadamard",         &param.hadamard,         3,     0,     3
  },
#else
  { "hadamard",         &param.hadamard,         2,     0,     2
  },
#endif
  { "h0_mult",          &param.h0_mult,         48,     0,     64   // sad is first multiplied by param.h0_mult and then divided by 16 by shifting right by HADAMARD_SHIFT_0.
  },
#ifdef USE_SSD
  { "ssd_mult",         &param.ssd_mult,         8,     0,      64   // sad is first multiplied by param.ssd_mult and then divided by 256 by shifting right by SSD_SHIFT.
  },
  { "ssd_qp_P",         &param.ssd_qp_P,         127,   0,     127  // QP where the SAD for P pictures switches to using SSD..
  },
  { "ssd_qp_B",         &param.ssd_qp_B,         100,   0,     127  // QP where the SAD for B pictures switches to using SSD..
  },
  { "ssd_disp_only",    &param.ssd_disp_only,    FALSE, FALSE, TRUE           
  },
#endif
  { "copy_offset",      &param.copy_offset,      64,    0,     64
  },
  { "trans8x8_enable",  &param.trans8x8_enable,  TRUE,  FALSE, TRUE            
  },
  { "ndf_signif_cnt",    &param.ndf_signif_cnt,    8,     0,     32             
  },
  { "ndf_4x4_mult",      &param.ndf_4x4_mult,      3,     0,     32             
  },
  { "ndf_aggressive",    &param.ndf_aggressive,    FALSE, FALSE, TRUE           
  },
  { "ndf_ic_refine",     &param.ndf_ic_refine,     TRUE,  FALSE, TRUE           
  },
  { "ndf_mvd_th",        &param.ndf_mvd_th,        1,     0,     255            
  },
  { "ndf_s_pred_quant",  &param.ndf_s_pred_quant,  INVALID_THRES,     0, DENOMINATOR 
  },
  { "ndf_s_mot_disc",    &param.ndf_s_mot_disc,    INVALID_THRES,     0, DENOMINATOR 
  },
  { "ndf_s_intra_quant", &param.ndf_s_intra_quant, INVALID_THRES,     0, DENOMINATOR 
  },
  { "ndf_s_pred_signif", &param.ndf_s_pred_signif, INVALID_THRES,     0, DENOMINATOR 
  },
  { "ndf_s_pred_mot",    &param.ndf_s_pred_mot,    INVALID_THRES,     0, DENOMINATOR 
  },
  { "ndf_s_ic_strong",   &param.ndf_s_ic_strong,   INVALID_THRES,     0, DENOMINATOR 
  },
  { "ndf_s_ic_intermed", &param.ndf_s_ic_intermed, INVALID_THRES,     0, DENOMINATOR 
  },
  { "ndf_s_ic_weak",     &param.ndf_s_ic_weak,     INVALID_THRES,     0, DENOMINATOR 
  },
  { "ndf_s_chroma",      &param.ndf_s_chroma,      8,     0,     15             
  },
  { "ndf_x_pred_quant",  &param.ndf_x_pred_quant,  INVALID_EXTENT,     0, 4 
  },
  { "ndf_x_mot_disc",    &param.ndf_x_mot_disc,    INVALID_EXTENT,     0, 4 
  },
  { "ndf_x_intra_quant", &param.ndf_x_intra_quant, INVALID_EXTENT,     0, 4 
  },
  { "ndf_x_pred_signif", &param.ndf_x_pred_signif, INVALID_EXTENT,     0, 4 
  },
  { "ndf_x_pred_mot",    &param.ndf_x_pred_mot,    INVALID_EXTENT,     0, 4 
  },
  { "ndf_x_ic_strong",   &param.ndf_x_ic_strong,   INVALID_EXTENT,     0, 4 
  },
  { "ndf_x_ic_intermed", &param.ndf_x_ic_intermed, INVALID_EXTENT,     0, 4 
  },
  { "ndf_x_ic_weak",     &param.ndf_x_ic_weak,     INVALID_EXTENT,     0, 4 
  },
  { "motion_sharing_log",  &param.motion_sharing_log,  TRUE,  FALSE, TRUE
  },
  { "enable_motion_sharing",  &param.motion_sharing_enable,  TRUE,  FALSE, TRUE
  },
  { "enable_motion_sharing_me_init_offset",  &param.motion_sharing_me_init_offset,  FALSE,  FALSE, TRUE
  },
  { "enable_motion_sharing_me_refine",  &param.motion_sharing_me_refine,  TRUE,  FALSE, TRUE
  },
  { "enable_motion_sharing_me_refine_full",  &param.motion_sharing_me_refine_full,  TRUE,  FALSE, TRUE
  },
  { "enable_motion_sharing_me_trellis",  &param.motion_sharing_me_trellis,  TRUE,  FALSE, TRUE
  },
  { "enable_motion_sharing_me_trellis_full",  &param.motion_sharing_me_trellis_full,  TRUE,  FALSE, TRUE
  },
  { "enable_motion_sharing_me_region",  &param.motion_sharing_me_region,  TRUE,  FALSE, TRUE
  },
  { "mvshare_two_enable",  &param.mvshare_two_enable,  TRUE,  FALSE, TRUE
  },
  { "mvshare_two_enable_me",  &param.mvshare_two_enable_me,  TRUE,  FALSE, TRUE
  },
  { "AIF",              &param.aif,              TRUE,  FALSE, TRUE        
  }, 
  { "mpm_default",         &param.mpm_default,         2, 0, 15
  },
  { "edge_prediction_ipd", &param.edge_prediction_ipd, TRUE,  FALSE, TRUE        
  },
  { "bugfix92",         &param.bugfix92,         FALSE, FALSE, TRUE,
  },
  { "threshold_edge_vector", &param.threshold_edge_vector, 250,  0, 65500
  },  
#ifdef Q_OFFSET
  { "theta_intra",      &param.iTheta[0],         0,   -127,   128        
  }, 
  { "theta_inter",      &param.iTheta[1],         0,   -127,   128        
  }, 
  { "theta8x8_intra",   &param.iTheta8x8[0],      0,   -127,   128        
  }, 
  { "theta8x8_inter",   &param.iTheta8x8[1],      0,   -127,   128        
  }, 
#endif
  { "intra8x8",         &param.intra8x8,          1,     0,     2
  },
  { "transform_selection", &param.transform_selection,  1,     0,     1
  },
  { "ndf_trsf_8x8",        &param.ndf_trsf_8x8,            TRUE,    FALSE,     TRUE
  },
#ifdef MVCOMP_ZERORES
  { "mvcomp_default_config", &param.mv_default_comp_config_idx, 0, 0, 1
  },
  { "mvcomp0_zero_residue", &param.mv_comp_config[0].mv_zero_residue, 0, 0, 1
  },
  { "mvcomp0_temporal_proba", &param.mv_comp_config[0].mv_temporal_proba, MV_DEFAULT_TEMPORAL_PROBA, 1, 0xfffe // Default 1%
  },
  { "mvcomp0_max_temporal_diff", &param.mv_comp_config[0].mv_max_temporal_diff, 0, -32000, +32000
  },
#ifdef MVCOMP_HAVE_SYMBOL_COUNTS
  { "mvcomp0_auto_max_temporal_diff", &param.mv_comp_config[0].mv_auto_max_temporal_diff, 0, 0, 1
  },
  { "mvcomp0_keep_learned_movement", &param.mv_comp_config[0].mv_keep_learned_movement, 0, 0, 1
  },
#endif
  { "mvcomp1_zero_residue", &param.mv_comp_config[1].mv_zero_residue, 0, 0, 1
  },
  { "mvcomp1_temporal_proba", &param.mv_comp_config[1].mv_temporal_proba, MV_DEFAULT_TEMPORAL_PROBA, 1, 0xfffe // Default 1%
  },
  { "mvcomp1_max_temporal_diff", &param.mv_comp_config[1].mv_max_temporal_diff, 0, -32000, +32000
  },
  { "mvcomp1_auto_max_temporal_diff", &param.mv_comp_config[1].mv_max_temporal_diff, 0, 0, 1
  },
  { "mvcomp1_keep_learned_movement", &param.mv_comp_config[1].mv_keep_learned_movement, 0, 0, 1
  },
#endif
#ifdef CABAC_SYNC_POINTS
  { "superblock_synchro", &param.superblock_synchro, 1, 0, 1
  },
  { "superblock_nsubstreams", &param.superblock_nsubstreams, 0, 0, 0 // no encoder choice for the moment, always zero.
  },
#endif
  { "enable_simple_lambdas",      &param.enable_simple_lambdas,       FALSE, FALSE, TRUE
  },
  { "sl_trans_sel",      &param.sl_trans_sel, FALSE, FALSE, TRUE
  },
  { "",                 NULL,                     0,     0,     0              
  }
};

//-----------------------------------------------------------------------------
/*!
* \brief Set of floating point parameters recognized by encoder
*
*/
//-----------------------------------------------------------------------------
FloatParameter encoder_float_param_list[] = 
{
  { "learn_offset",     &param.learn_offset,     0.0,   0.0,   1.0            
  },
  { "sl_weight16x16", &param.sl_weight16x16, 1.05, 0.80, 1.25
  },
  { "sl_rate_offset", &param.sl_rate_offset, 3.0, 0.0, 4.0
  },
  { "sl_lambda_I_luma", &param.sl_lambda_I_luma, 0.4, 0.1, 1.0
  },
  { "sl_lambda_I_luma_rdoq", &param.sl_lambda_I_luma_rdoq, 0.4, 0.1, 1.0
  },
  { "sl_lambda_I_chroma", &param.sl_lambda_I_chroma, 0.3, 0.1, 1.0
  },
  { "sl_lambda_I_chroma_rdoq", &param.sl_lambda_I_chroma_rdoq, 0.3, 0.1, 1.0
  },
  { "sl_lambda_P_luma", &param.sl_lambda_P_luma, 0.8, 0.1, 1.0
  },
  { "sl_lambda_P_luma_rdoq", &param.sl_lambda_P_luma_rdoq, 0.6, 0.1, 1.0
  },
  { "sl_lambda_P_chroma", &param.sl_lambda_P_chroma, 0.8, 0.1, 1.0
  },
  { "sl_lambda_P_chroma_rdoq", &param.sl_lambda_P_chroma_rdoq, 0.5, 0.1, 1.0
  },
  { "",                 NULL,                    0.0,   0.0,   0.0            
  }
};

static char def_logfile[] = "log.txt";

//-----------------------------------------------------------------------------
/*!
* \brief Set of string parameters recognized by encoder
*
*/
//-----------------------------------------------------------------------------
StringParameter encoder_string_param_list[] = 
{
  { "srcfile",       param.srcfile,       NULL        
  },
  { "dstfile",       param.dstfile,       NULL        
  },
  { "recfile",       param.recfile,       NULL        
  },
  { "sadfile",       param.sadfile,       NULL        
  },
  { "reffile",       param.reffile,       NULL        
  },
  { "prob_file",     param.prob_file,     NULL        
  },
  { "ztreefile",     param.ztreefile,     NULL        
  },
  { "ztreefile_8x8", param.ztreefile_8x8,     NULL        
  },
  { "qmatrixfile",       param.qmatrixfile[0],  NULL        
  }, // cq_start
  { "qoffsetmatrixfile", param.qmatrixfile[1],  NULL        
  }, // cq_end
  { "qtabfile_4x4",  param.qtabfile_4x4,  "tab/quant_4x4.tab"        
  },   
  { "qtabfile_8x8",  param.qtabfile_8x8,  "tab/quant_8x8.tab"        
  }, 
  { "qtabfile_2x8",  param.qtabfile_2x8,  "tab/quant_2x8.tab"        
  },  
  { "qtabfile_8x2",  param.qtabfile_8x2,  "tab/quant_8x2.tab"        
  }, 
  { "parfile",       param.parfile,       NULL        
  },
  { "logfile",       param.logfile,       def_logfile 
  },
  { "sadlogfile",    param.sadlogfile,    NULL        
  },
  { "expPrmdFrmt",   param.expPrmdFrmt,   NULL        
  },
  { "expPrmdFrmtIdr",   param.expPrmdFrmtIdr,   NULL        
  },
#ifdef MVCOMP_ZERORES
  { "mvcomp0_predictor_0", &param.mv_comp_config[0].mv_predictor_strings[0][0], "A"
  },
  { "mvcomp0_predictor_1", &param.mv_comp_config[0].mv_predictor_strings[1][0], "B"
  },
  { "mvcomp0_predictor_2", &param.mv_comp_config[0].mv_predictor_strings[2][0], "temporal"
  },
  { "mvcomp1_predictor_0", &param.mv_comp_config[1].mv_predictor_strings[0][0], "A"
  },
  { "mvcomp1_predictor_1", &param.mv_comp_config[1].mv_predictor_strings[1][0], "B"
  },
  { "mvcomp1_predictor_2", &param.mv_comp_config[1].mv_predictor_strings[2][0], "none"
  },
#endif
  { "",              NULL,                NULL        
  }
};

static const char *CQ_TYPE[2] = { "INTRA", "INTER" }; // cq_start
static const char *CQ_SIZE[5] = { "4X4", "2X8", "8X2", "8X8", "16X16" };
static const char *CQ_YUV[3] = { "_LUMA", "_CHROMAU", "_CHROMAV" };
static const char *CQ_SETID[5] = { "_COMMON", "_1", "_2", "_3", "_4" };
static const char *CQ_VALUE[5] = { "_MOD_TYPE", "_SCALE", "_SLOPE", "_N_VALUES", "_VALUES" }; // cq_end

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
*/
//-----------------------------------------------------------------------------
Param::Param() 
{
  int i;

#ifdef USE_SSD
  for (i=0; i<256; i++) 
  {
    ssd_lookup[i] = (i * i) << SAMPLE_ADD_PREC;
  }
#endif

  for (i = ZERO; encoder_int_param_list[i].ptr != NULL; i++)
    *(encoder_int_param_list[i].ptr) = encoder_int_param_list[i].default_val;

  for (i = ZERO; encoder_float_param_list[i].ptr != NULL; i++)
    *(encoder_float_param_list[i].ptr) = encoder_float_param_list[i].default_val;

  for (i = ZERO; encoder_string_param_list[i].ptr != NULL; i++)
    if (encoder_string_param_list[i].default_val != NULL)
      strcpy(encoder_string_param_list[i].ptr, encoder_string_param_list[i].default_val);
    else
      encoder_string_param_list[i].ptr[0] = 0;

  for (i = ZERO; i < (param.intra16x16 ? NZTREES : NZTREES - NSUBZTREES); i++)
    ztree[i] = NULL;

  ztree_64 = NULL;
  if (param.chroma8x8)
    ztree_64_CT8 = NULL;

  num_ctx_read = 0;

#ifdef MVCOMP_ZERORES
  mv_pred_name[MV_PRED_LEFT] = "A";
  mv_pred_name[MV_PRED_TOP] = "B";
  mv_pred_name[MV_PRED_TEMPORAL] = "temporal";
  mv_pred_name[MV_PRED_ZERO] = "zero";
  mv_pred_name[MV_PRED_NONE] = "none";
  mv_pred_name[MV_MAX_PRED_INDEX+1] = NULL;
#endif
#ifdef CABAC_SYNC_POINTS
  decode_inputref = 0;
#endif
#ifdef DEC_TRANSCODER
  decode_transcode = 0;
  decode_transcode_nthreads = 0;
#endif
#ifdef DENOISE_MT
  decode_ndenoisethreads = 0;
#endif
#ifdef WIENER_MT
  decode_nwfthreads = 0;
#endif
#ifdef DEBLOCK_MT
  decode_ndeblockthreads = 0;
#endif

  initialProbEntropyModel = NULL;
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
Param::~Param() 
{
  int i;

  for (i = 0; i < (param.intra16x16 ? NZTREES : NZTREES - NSUBZTREES); i++)
  {
    if (ztree[i] != NULL)
      delete ztree[i];
  }

  if (ztree_64 != NULL)
    delete ztree_64;
  if (param.chroma8x8 && (ztree_64_CT8 != NULL))
    delete ztree_64_CT8;

  //FRAME_SUBSTREAMS: zt64 stats in EntropyDecoder now.

  delete initialProbEntropyModel;
}

//-----------------------------------------------------------------------------
/*!
* \brief Read probabilities for a context from file
*
* Reads pair of values from probability file and computes initial value of
* probability based on qp value.
*
* Now can take a name indicating the type of probability being read.  If the next
* line in the file does not correspond to this name, lines are skipped until the
* name is found.  This occurs when defines for some tools are deactivated
* (eg INTRA16x16).
*
* \param[in,out] f File pointer
* \param[in] name Tag indicating the type of proba being read.
* \param[in] qp
*
* \return Initial probability
*/
//-----------------------------------------------------------------------------
int Param::read_prob(FILE *f, char const *name, int qp) 
{
  int a,b;
  int w;
  double p;
  int r;

  int already_seen_eof = 0;
  // Scan the file completely if necessary, looking for the named proba.
  // This eliminates any ordering issues.
  // We could also check the ids, too, if we're sure that they're not going
  // to change.  For the moment we don't, allowing a more flexible prob.dat
  // file.
  // printf("reading %s\n", name);
  while (1)
  {
    char name_in_file[200];
    int res = fscanf(f, "%100s %d %d", &name_in_file[0], &a, &b);
    if (res <= 0)
    {
      if (already_seen_eof)
      {
        printf("Cannot find line in '.dat' for proba \"%s\"\n", name);
        exit(EXIT_FAILURE);
      }
      already_seen_eof = 1;
      // Reset to start of file.
      fseek(f, 0, SEEK_SET);
      continue;
    }

    if (res != 3)
    {
      printf("Did not read a complete line from the '.dat' probability file, read %d tokens.\n", res);
      exit(EXIT_FAILURE);
    }

    if (strcmp(name, name_in_file) == 0)
      break;
  }

  w = 4 * a + qp * b;
  p = (w - 4*4096) / 4 / 256.0;
  if (p < 0)
    p = 0.5 * pow(2.0, p);
  else
    p = 1.0 - 0.5 * pow(0.5, p);

  r = (int)(power2(PROB_BITS) * p + 0.5);
  r = clip(r, 32, 65504);

  num_ctx_read++;

  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Initialize internal fields
*
* Extends ME range to a multiple of 8 and initializes bitsF and bitsM arrays.
*/
//-----------------------------------------------------------------------------
void Param::update()
{
  int i;

  param.me_range = (param.me_range + 7) & 0xfffffff8;

  for (i = ONE; i < power2(PROB_BITS); i++) 
  {
    bitsF[i] = (float)(PROB_BITS - log((double)i)/log(2.0) + learn_offset);
    bitsM[0][i] = (int)(power2(RATE_BITS) * bitsF[i] + 0.5);
    bitsM[1][power2(PROB_BITS)-i] = bitsM[0][i];
  }
#ifdef CABAC_SYNC_POINTS
  if (!param.superblock_synchro)
      param.superblock_nsubstreams = 0;
  if (param.superblock_nsubstreams < 1)
      param.superblock_nsubstreams = 1;
#ifdef DEC_TRANSCODER
  if (param.superblock_nsubstreams > 1 && param.decode_transcode)
  {
      printf("superblock_nsubstreams=%d, param.decode_transcode=%d: can only transcode from single-substream input for the moment.\n",
             param.superblock_nsubstreams, param.decode_transcode);
      exit(1);
  }
#endif
  if (decode_nthreads <= 0)
      decode_nthreads = superblock_nsubstreams;
  else if (decode_nthreads > superblock_nsubstreams)
      decode_nthreads = superblock_nsubstreams;
#ifndef FRAME_SUBSTREAMS
  if (param.superblock_nsubstreams > 1)
  {
    printf("superblock_nsubstreams=%d > 1: but we're not compiled with FRAME_SUBSTREAMS.\n", param.superblock_nsubstreams);
    exit(-1);
  }
#endif
#endif
#ifdef DENOISE_MT
  if (decode_ndenoisethreads <= 0)
    decode_ndenoisethreads = 1;
#endif
#ifdef WIENER_MT
  if (decode_nwfthreads <= 0)
    decode_nwfthreads = 1;
#endif
#ifdef DEBLOCK_MT
  if (decode_ndeblockthreads <= 0)
    decode_ndeblockthreads = 1;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Read probability file
*
* Reads the probability file and assigns initial probability values to
* respective contexts.
*/
//-----------------------------------------------------------------------------
void Param::read_prob_file()
{  
  int i,j;
  FILE *f, *f2;
  FILE *fztree_8x8;
  int zt_size;
  int i_dummy;


  f = fopen(param.prob_file,"rt");
  if (f == NULL) 
  {
    printf("Could not open prob_file: <%s>\n", param.prob_file);
    exit(1);
  }

  // read ztree structure first, followed by probas using an EntropyModel.
  if (param.ztreefile[ZERO] == ZERO) 
  {
    fprintf(stderr, "A zerotree file must be specified.\n");
    exit(EXIT_FAILURE);
  }

  f2 = fopen(param.ztreefile,"rt");
  if (f2 == NULL) 
  {
    printf("Could not open ztreefile: <%s>\n", param.ztreefile);
    exit(EXIT_FAILURE);
  }
  for (i=0; i<NZTREES-NSUBZTREES; i++) 
  {
    fscanf(f2, "%d", &zt_size);
    param.ztree_pattern[i][0] = zt_size;

    for (j=1; j<3*zt_size; j++)
      fscanf(f2, "%d", &param.ztree_pattern[i][j]);
  }

  if (param.intra16x16)
  {
    for (i=NZTREES-NSUBZTREES; i<NZTREES; i++) 
    {
      fscanf(f2, "%d", &zt_size);
      param.ztree_pattern[i][0] = zt_size;

      for (j=1; j<3*zt_size; j++)
        fscanf(f2, "%d", &param.ztree_pattern[i][j]);
    }
  }
  else
  {
    for (i=NZTREES-NSUBZTREES; i<NZTREES; i++) 
    {
      fscanf(f2, "%d", &zt_size);

      for (j=1; j<3*zt_size; j++)
        fscanf(f2, "%d", &i_dummy);
    }
  }

  if (param.trans8x8_enable) 
  {
    fztree_8x8 = fopen(param.ztreefile_8x8,"rt");
    if (fztree_8x8 == NULL) 
    {
      printf("Could not open ztreefile: <%s>\n", param.ztreefile_8x8);
      exit(EXIT_FAILURE);
    }    

    fscanf(fztree_8x8, "%d", &zt_size);
    param.ztree_pattern_64[0] = zt_size;

    for (j=1; j<3*zt_size; j++)
    {
      fscanf(fztree_8x8, "%d", &param.ztree_pattern_64[j]);
    }
    fclose(fztree_8x8);
  }

  // For ztrees, we have to create the skeleton here.
#ifdef FRAME_SUBSTREAMS
  init_ztree(1);
#else
  init_ztree();
#endif

  initialProbEntropyModel = new EntropyModel();
  initialProbEntropyModel->read_prob_file(f);

  fclose(f);
  fclose(f2);

  // These values are forced to 32768 as before, regardless of qp.
#ifndef SIP_CTX_READ
  if (param.intra16x16)
  {
    for (i=0; i<NCTX_SIP_TYPE; i++)
    {
      initialProbEntropyModel->p_sip_type_b4[i].set_default_prob(32768);
      initialProbEntropyModel->p_sip_type_b4[i].restoreDefaultState();
    }
  }
#endif //SIP_CTX_READ

  // For ztrees, we have to copy the proba i_p values back into the pattern and into
  // i_prob[].  There should be better separation here between structure and probas,
  // but this minimizes code change for the moment.
  for (i=0; i<(param.intra16x16 ? NZTREES : NZTREES-NSUBZTREES); i++) 
  {
    ztree[i]->copy_initial_probs_back(initialProbEntropyModel, &ztree_pattern[i][0]);
  }
  if (param.trans8x8_enable) 
  {
    ztree_64->copy_initial_probs_back(initialProbEntropyModel, &ztree_pattern_64[0]);
#ifdef ZTREE_64_CT8_CTX_READ
    ztree_64_CT8->copy_initial_probs_back(initialProbEntropyModel, &ztree_pattern_64[0]);
#else
    ztree_64_CT8->copy_initial_probs_from(ztree_64, initialProbEntropyModel);
#endif //ZTREE_64_CT8_CTX_READ
  }

  // initialProbEntropyModel->printProbas("first");
}

//-----------------------------------------------------------------------------
/*!
* \brief Construct zerotrees
*
* Constructs zerotrees based on pattern.
*/
//-----------------------------------------------------------------------------
#ifdef FRAME_SUBSTREAMS
void Param::init_ztree(int encoder) 
#else
void Param::init_ztree() 
#endif
{
  int i;

  for (i=0; i<(param.intra16x16 ? NZTREES : NZTREES-NSUBZTREES); i++) 
#ifdef FRAME_SUBSTREAMS
    ztree[i] = new ZeroTree(ztree_pattern[i], 0, encoder ? -1 : (i == 0 ? 0 : (ztree[i-1]->first_ctx_id+ztree[i-1]->n_ctxs)));
#else
    ztree[i] = new ZeroTree(ztree_pattern[i], 0);
#endif

  if (param.trans8x8_enable)
  {
    // int add_node; // FRAME_SUBSTREAMS: moved next to stats.
    // FRAME_SUBSTREAMS: ztree_size copied to EntropyDecoder, too.
    int ztree_size[NZTREES64] = 
    {16, 16, 16, 16
    };
    int j;

    // Don't reset the new tree just yet.
#ifdef FRAME_SUBSTREAMS
    ztree_64 = new ZeroTree(ztree_pattern_64, 0, encoder ? -1 : ztree[i-1]->first_ctx_id+ztree[i-1]->n_ctxs);
#else
    ztree_64 = new ZeroTree(ztree_pattern_64, 0);
#endif
    if (param.chroma8x8)
      // Don't reset the new tree just yet.
#ifdef FRAME_SUBSTREAMS
      ztree_64_CT8 = new ZeroTree(ztree_pattern_64, 0, encoder ? -1 : ztree_64->first_ctx_id+ztree_64->n_ctxs);
#else
      ztree_64_CT8 = new ZeroTree(ztree_pattern_64, 0);
#endif

    j=0;
    for (i=0; i<NZTREES64; i++)
      j += ztree_size[i];

    if (j>ZTsize64)
    {
      printf("ztree split error, size exceeds!");
      exit(0);
    }

    //FRAME_SUBSTREAMS The zt64 stats are kept in the EntropyDecoder now.
  }  
}

//-----------------------------------------------------------------------------
/*!
* \brief Scan a matrix
*/
//-----------------------------------------------------------------------------
void Param::matrix_scan(int *qmatrixList, int quantization_list_id, int matrix_size) // cq_start
{
  int i;

  for(i = 0; i < matrix_size; i++)
    qm_block.val[i] = qmatrixList[i];
  if( (matrix_size == 16 && quantization_list_id < 6) || (matrix_size != 16) ) //!< 4x4(0~5), 8x8, 16x16
  {
    if(matrix_size == 16)
      qm_block.set_size(4, 4);
    else if(matrix_size == 64)
      qm_block.set_size(8, 8);
    else
      qm_block.set_size(16, 16);

    qm_block.zigzagscan();
  }
  else if(matrix_size == 16)
    qm_block.rescan(quantization_list_id == 7 ? 1 : 2, 0); //!< 2x8, 8x2
  for(i = 0; i < matrix_size; i++)
    qmatrixList[i] = qm_block.val[i];
}

//-----------------------------------------------------------------------------
/*!
* \brief Scan a matrix
*/
//-----------------------------------------------------------------------------
void Param::matrix_inv_scan(int *qmatrixList, int quantization_list_id, int matrix_size)
{
  int i;

  for(i = 0; i < matrix_size; i++)
    qm_block.val[i] = qmatrixList[i];
  if( (matrix_size == 16 && quantization_list_id < 6) || (matrix_size != 16) ) //!< 4x4(0~5), 8x8, 16x16
  {
    if(matrix_size == 16)
      qm_block.set_size(4, 4);
    else if(matrix_size == 64)
      qm_block.set_size(8, 8);
    else
      qm_block.set_size(16, 16);

    qm_block.inverse_zigzagscan();
  }
  else if(matrix_size == 16)
    qm_block.rescan(0, quantization_list_id == 7 ? 1 : 2); //!< 2x8, 8x2
  for(i = 0; i < matrix_size; i++)
    qmatrixList[i] = qm_block.val[i];
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate q matrix based on matrix mode type
*/
//-----------------------------------------------------------------------------
void Param::qmatrix_generate(int *qmatrixList_tmp, int *qmatrixList, int sizeOfQMatrixList, int matrix_type)
{
  int j;
  int qmatrix_mode;
  int scale_ratio;
  int slope_ratio;
  int delta_value_num;

  qmatrix_mode = qmatrixList_tmp[0];
  if (qmatrix_mode == 0)
  {
    //! update mode
    scale_ratio = qmatrixList_tmp[1];
    slope_ratio = qmatrixList_tmp[2];
    delta_value_num = qmatrixList_tmp[3];

    //! generate matrix using scale and slope ratio
    for (j = 0; j < sizeOfQMatrixList; j++)
      qmatrixList[j+4] = (qmatrixList[j+4] * (16+scale_ratio) + (qmatrixList[j+4]-16) * slope_ratio + 8) >>4;
 
    //! add delta
    if(matrix_type == 0)
    {
      for (j = 0; j < delta_value_num; j++)
        qmatrixList[j+4] = max(1, min(255, qmatrixList[j+4] + qmatrixList_tmp[j+4]));
    }
    else
    {
      for (j = 0; j < delta_value_num; j++)
        qmatrixList[j+4] += qmatrixList_tmp[j+4];

      for (j = 0; j < sizeOfQMatrixList; j++)
      {
        qmatrixList[j+4] = max(1, min(255, qmatrixList[j+4]));
        qmatrixList[j+4] = 128 - (qmatrixList[j+4]<<3);
      }
    }
  }
  else if(qmatrix_mode == 2 && matrix_type == 1)
  {
    //! flat mode
    for (j = 0; j < sizeOfQMatrixList; j++)
      qmatrixList[j+4] = qmatrixList_tmp[4];
  }
  else 
  {
    //! intra mode
    if(matrix_type == 0)
    {
      for (j = 0; j < sizeOfQMatrixList; j++)
        qmatrixList[j+4] = qmatrixList_tmp[j+4];
    }
    else
    {
      for (j = 0; j < sizeOfQMatrixList; j++)
        qmatrixList[j+4] = 128 - (qmatrixList_tmp[j+4]<<3);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Update q matrix
*/
//-----------------------------------------------------------------------------
void Param::update_qmatrix(int quantization_set_id, int matrix_type)
{
  int j;
  int k;
  int qmatrix_list_present_flag;
  int matrix_size;

  for (j = 0; j < 16; j++)
  {
    if (j<8)  //!< 4x4
      matrix_size = 16;
    else if (j<14) //!< 8x8
      matrix_size = 64;
    else //!< 16x16
      matrix_size = 256;

    qmatrix_list_present_flag = qmatrix_update_check[matrix_type][quantization_set_id][j];
    if (qmatrix_list_present_flag)
      qmatrix_generate(qmatrix_tmp[matrix_type][quantization_set_id][j], qmatrix[matrix_type][quantization_set_id][j], matrix_size, matrix_type);
    else
    {
      if(matrix_type == 0)
      {
        for (k = 0; k < matrix_size; k++)
          qmatrix[matrix_type][quantization_set_id][j][k+4] = qmatrix_tmp[matrix_type][quantization_set_id][j][k+4];
      }
      else
      {
        for (k = 0; k < matrix_size; k++)
          qmatrix[matrix_type][quantization_set_id][j][k+4] = 128 - (qmatrix_tmp[matrix_type][quantization_set_id][j][k+4]<<3);
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode values of a q matrix
*/
//-----------------------------------------------------------------------------
int Param::write_qmatrix_list(Buffer *buf, int *qmatrixList, int sizeOfQMatrixList, int *UseDefaultQMatrixFlag)
{
  int j;
  int k;
  int r;
  int lastScale;
  int nextScale;
  int delta_scale;

  r = 0;
  *UseDefaultQMatrixFlag = 0;
  if (qmatrixList[0] == 0)
  {
    r += buf->write_signed(-8); 
    *UseDefaultQMatrixFlag = 1;
    return r;
  }
  else
  {
    lastScale = 8;
    for (j = 0; j < sizeOfQMatrixList; j++)
    {
      nextScale = qmatrixList[j];
      delta_scale = nextScale - lastScale;
      r += buf->write_signed(delta_scale); 
      lastScale = nextScale;
      if (nextScale == 0)
      {
        for (k = j; k < sizeOfQMatrixList; k++)
          qmatrixList[k] = qmatrixList[j-1];
        return r;
      }
    }
  }
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode update values of a q matrix
*/
//-----------------------------------------------------------------------------
int Param::write_qmatrix_update_value(Buffer *buf, int *qmatrixList)
{
  int i;
  int r = 0;

  //! scale ratio, slope ratio, number of delta values, matrix values
  for(i = 1; i < qmatrixList[3] + 4; i++)
  {
    if (i != 3)
      r += buf->write_signed(qmatrixList[i]);
    else
      r += buf->write_unsigned(qmatrixList[i]);
  }
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode values of a q matrix set
*/
//-----------------------------------------------------------------------------
int Param::write_qmatrix_set(Buffer *buf, int quantization_set_id, int matrix_type)
{
  int j;
  int k;
  int r;
  int qmatrix_list_present_flag;
  int *qmatrixList;
  int qmatrix_modetype;
  int matrix_size;

  r = 0;
  for (j = 0; j < 16; j++)
  {
    if (j<8)  //!< 4x4
      matrix_size = 16;
    else if (j<14) //!< 8x8
      matrix_size = 64;
    else //!< 16x16
      matrix_size = 256;

    qmatrix_list_present_flag = qmatrix_update_check[matrix_type][quantization_set_id][j];
    r += buf->write_bits(qmatrix_list_present_flag, 1);
    qmatrixList = qmatrix_tmp[matrix_type][quantization_set_id][j];
    qmatrix_modetype = 1;
    if (qmatrix_list_present_flag)
    {
      qmatrix_modetype = qmatrixList[0];

      //! mode type
      r += buf->write_unsigned(qmatrix_modetype);
      if (qmatrix_modetype == 0) //!< update mode
      {
        matrix_scan(qmatrixList+4, j, matrix_size);
        r += write_qmatrix_update_value(buf, qmatrixList);
        matrix_inv_scan(qmatrixList+4, j, matrix_size);
      }
      else if (qmatrix_modetype == 2 && matrix_type == 1) //!< flat mode
        r += buf->write_signed(qmatrixList[4]);
    }
    if (qmatrix_modetype == 1) //!< matrix mode
    {
      qmatrix_list_present_flag = qmatrix_check[matrix_type][quantization_set_id][j];
      r += buf->write_bits(qmatrix_list_present_flag, 1);
      if (qmatrix_list_present_flag)
      {
        matrix_scan(qmatrixList+4, j, matrix_size);
        r += write_qmatrix_list(buf, qmatrixList+4, matrix_size, &(UseDefaultQMatrixFlag[matrix_type][j]));
        matrix_inv_scan(qmatrixList+4, j, matrix_size);
        if (UseDefaultQMatrixFlag[matrix_type][j])
        {
          for (k = 0; k < matrix_size; k++)
            qmatrixList[k+4] = 16;
        }
      }
    }
  }
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode values of common q matrix set and set values of all q matrix sets
*/
//-----------------------------------------------------------------------------
int Param::write_qmatrix_set_all(Buffer *buf, int matrix_type) // cq_end
{
  int r;
  
  r = write_qmatrix_set(buf, 0, matrix_type);
  update_qmatrix(0, matrix_type);
  dupli_qmatrix0(matrix_type);
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Write header to byte buffer
*
* Writes sequence header parameters to the byte buffer buf.
*
* \param[in,out] buf Byte buffer
*
* \return Number of bits written to buffer
*/
//-----------------------------------------------------------------------------
int Param::write(Buffer *buf) 
{
  int i,j;
  int zt_size;

  int r = 0;
  buf->reset();

  r += buf->write_bits(qp, 7);
  r += buf->write_bits(frame_height, 13);
  r += buf->write_bits(frame_width, 13);
  r += buf->write_bits(better_color, 3);
#ifdef BETTER_CHROMA_DC
  if (better_color==7)
  {
    r += buf->write_bits(chroma_dc_th, 8);
    r += buf->write_bits(chroma_dc_scale, 8);
  }
#endif
  r += buf->write_bits(bc_offset+16, 5);
  r += buf->write_bits(lumaDC, 3);
  r += buf->write_bits(loss_rate-2, 4);
  r += buf->write_bits(ac_exp_limit, 10);
  r += buf->write_bits(luma_mc, 3);
  r += buf->write_bits(chroma_mc, 2);
  r += buf->write_bits(gen8pel_filter, 2); 
  r += buf->write_bits(mc_round, 2);
  r += buf->write_bits(ndf_mode, 2);
  r += buf->write_bits(deblockingfilter, 4);
  if(deblockingfilter != 0)
  {
    r += buf->write_bits(dbf_offset_a, 5);
    r += buf->write_bits(dbf_offset_b, 5);
  }
  r += buf->write_bits(wienerfilter, 1);
#ifdef WIENER_FILTER_INLOOP_SWITCH
  if (wienerfilter)
  {
    r += buf->write_bits(wienerfilter_inloop, 1);
  }  
#endif
  r += buf->write_bits(use_extension, 1); // was lf_spthres

  r += buf->write_bits(0, 2); // reserved
  r += buf->write_bits(10, 4);
  r += buf->write_bits(10, 4);

  r += buf->write_bits(1, 1);
  r += buf->write_bits(candidate_search_region, 8);
  r += buf->write_bits(num_candidate_averaged, 4);
  r += buf->write_bits(candidate_sad_threshold, 8);

  r += buf->write_bits(0, 1); // was tma_enable
  r += buf->write_bits(0, 1); // was mva_enable
  r += buf->write_bits(nref_frames, 3);
  r += buf->write_bits(adapt_list_size, 1);
  r += buf->write_bits(enable_intcomp, 1);
  r += buf->write_bits(DC_mult - ONE, 3);
  r += buf->write_bits(7, 3);
  r += buf->write_bits(7, 3);
  r += buf->write_bits(param.weight_diff_scale-1, 3);

  r += buf->write_bits(trans8x8_enable, 1);
  r += buf->write_bits(motion_sharing_enable, 1);
  if(param.motion_sharing_enable)  
    r += buf->write_bits(mvshare_two_enable, 1);

  r += buf->write_bits(ndf_two_rounds, 1);
  r += buf->write_bits(ndf_weighting, 1);
  r += buf->write_bits(ndf_signif_cnt, 4);
  if (param.trans8x8_enable) r += buf->write_bits(ndf_4x4_mult, 4);
  r += buf->write_bits(ndf_aggressive, 1);
  r += buf->write_bits(ndf_ic_refine, 1);
  r += buf->write_bits(ndf_mvd_th, 8);
  r += buf->write_bits(ndf_s_pred_quant, 8);
  r += buf->write_bits(ndf_s_pred_signif, 8);
  r += buf->write_bits(ndf_s_pred_mot, 8);
  r += buf->write_bits(ndf_s_intra_quant, 8);
  r += buf->write_bits(ndf_s_mot_disc, 8);
  r += buf->write_bits(ndf_s_ic_strong, 8);
  r += buf->write_bits(ndf_s_ic_intermed, 8);
  r += buf->write_bits(ndf_s_ic_weak, 8);
  r += buf->write_bits(ndf_x_pred_quant, 8);
  r += buf->write_bits(ndf_x_pred_signif, 8);
  r += buf->write_bits(ndf_x_pred_mot, 8);
  r += buf->write_bits(ndf_x_intra_quant, 8);
  r += buf->write_bits(ndf_x_mot_disc, 8);
  r += buf->write_bits(ndf_x_ic_strong, 8);
  r += buf->write_bits(ndf_x_ic_intermed, 8);
  r += buf->write_bits(ndf_x_ic_weak, 8);

  r += buf->write_bits(ndf_s_chroma, 4);
  r += buf->write_bits(aif, 1);

  r += buf->write_bits(0, 1); // reserved
  r += buf->write_bits(mpm_default, 4);

  r += buf->write_bits(intra8x8, 2);
  r += buf->write_bits(transform_selection, 1);
  r += buf->write_bits(ndf_trsf_8x8, 1);
  r += buf->write_bits(1, 3); // reserved

  r += buf->write_bits(1, 1); // reserved
  r += buf->write_bits(sb_height, 8);
  r += buf->write_bits(sb_width, 8);
  r += buf->write_bits(intra16x16, 1);
  if (param.intra16x16)
  {
    r += buf->write_bits(candidate_search_region_intra16x16, 8);
    r += buf->write_bits(transform_selection_16x16, 1);
    r += buf->write_bits(chroma8x8, 1);
  }

  for (i=0; i<NZTREES-NSUBZTREES; i++) 
  {
    zt_size = ztree_pattern[i][0];
    r += buf->write_bits(zt_size, 8);

    for (j=1; j<2*zt_size; j++)
      r += buf->write_bits(ztree_pattern[i][j], 1);

    for (j=2*zt_size; j<3*zt_size; j++)
      r += buf->write_bits(ztree_pattern[i][j], 4);

    // Don't write the probas here -- they're written later.
  }

  if (param.trans8x8_enable) 
  {
    zt_size = ztree_pattern_64[0];
    r += buf->write_bits(zt_size, 8);

    for (j=1; j<2*zt_size; j++)
    {
      r += buf->write_bits(ztree_pattern_64[j], 1);    
    }

    for (j=2*zt_size; j<3*zt_size; j++)
    {
      r += buf->write_bits(ztree_pattern_64[j], 8); 
    }

    // Don't write the probas here -- they're written later.
  }

  // Write all further non-proba info...
  if (param.trans8x8_enable) 
  {
    r += buf->write_bits(4,3);  
  }  

  if (param.intra16x16)
  {
    for (i=NZTREES-NSUBZTREES; i<NZTREES; i++) 
    {
      zt_size = ztree_pattern[i][0];
      r += buf->write_bits(zt_size, 8);

      for (j=1; j<2*zt_size; j++)
        r += buf->write_bits(ztree_pattern[i][j], 1);

      for (j=2*zt_size; j<3*zt_size; j++)
        r += buf->write_bits(ztree_pattern[i][j], 5); // Changed number of bits to describe leaves index

      //for (j=0; j<2*(zt_size-1); j++)
      //  r += buf->write_bits(ztree_pattern[i][j+3*zt_size], PROB_BITS);
    }
  }
#ifdef MVCOMP_ZERORES
  for (i = 0; i < 2; i++)
  {
      MVCompConfig *pconfig = &mv_comp_config[i];
      buf->write_bits(pconfig->mv_npredictors, 4);
      for (int p = 0; p < pconfig->mv_npredictors; p++)
          buf->write_bits(pconfig->mv_predictors[p], 4);
      if (pconfig->mv_predictor_used[MV_PRED_TEMPORAL])
      {
          buf->write_bits(pconfig->mv_temporal_proba, PROB_BITS);
          buf->write_bits(pconfig->mv_auto_max_temporal_diff, 1);
          if (!pconfig->mv_auto_max_temporal_diff)
              buf->write_bits(pconfig->mv_max_temporal_diff&0xffff, 16);
      }
  }
#endif
#ifdef CABAC_SYNC_POINTS
  buf->write_bits(superblock_synchro, 1);
  // Write the number of threads indicator.  The basic encoder is
  // just single threaded, and so writes zero.  To minimise output,
  // we write a bit indicator first -- zero implies zero, 1 implies
  // the number of threads follows in the next 8 bits.
  if (superblock_synchro)
  {
      if (superblock_nsubstreams <= 1)
          buf->write_bits(0, 1);
      else
      {
          buf->write_bits(1, 1);
          buf->write_bits(superblock_nsubstreams, 8);
      }
  }
#endif

  max_quantization_set_num_minus_one = MAX_QUANT_SET_NUM - 1; // cq_start
  r += buf->write_bits(max_quantization_set_num_minus_one, 8);
  r += buf->write_bits(qmatrix_enable_flag[0], 1);
  if (qmatrix_enable_flag[0])
    r += write_qmatrix_set_all(buf, 0);
  r += buf->write_bits(qmatrix_enable_flag[1], 1);
  if (qmatrix_enable_flag[1])
    r += write_qmatrix_set_all(buf, 1); // cq_end

  // Now, write the probabilities using the initial entropy model.
  r += initialProbEntropyModel->write_stream_probs(buf);

  if (use_extension)
  {
    r += buf->write_bits(tag_bugfix87, 8);
    r += buf->write_bits(bugfix87, 1);

    r += buf->write_bits(tag_bugfix88, 8);
    r += buf->write_bits(bugfix88, 1);

    r += buf->write_bits(tag_bugfix89, 8);
    r += buf->write_bits(bugfix89, 1);

    r += buf->write_bits(tag_alt8x8_trans_idx, 8);
    r += buf->write_bits(alt8x8_trans_idx, 4);

    r += buf->write_bits(tag_ndf_lc, 8);
    r += buf->write_bits(ndf_lc_luma, 4);
    r += buf->write_bits(ndf_lc_chroma, 4);

    r += buf->write_bits(tag_edge_pred_bug_fix, 8);
    r += buf->write_bits(bugfix90, 1);
    
#ifdef I16_IN_P_FRAME
    r += buf->write_bits(tag_intra16x16p, 8);
    r += buf->write_bits(intra16x16p, 1);
#endif

    r += buf->write_bits(tag_bugfix93, 8);
    r += buf->write_bits(bugfix93, 1);

    r += buf->write_bits(tag_end, 8);
  }
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode values of a q matrix
*/
//-----------------------------------------------------------------------------
void Param::read_qmatrix_list(Buffer *buf, int *qmatrixList, int sizeOfQMatrixList, int *UseDefaultQMatrixFlag) // cq_start
{
  int j;
  int delta_scale;
  int lastScale = 8;
  int nextScale = 8;

  for (j = 0; j < sizeOfQMatrixList; j++)
  {
    if (nextScale != 0)
    {
      delta_scale = buf->read_signed();
      nextScale = ( lastScale + delta_scale + 256 ) % 256;
      *UseDefaultQMatrixFlag = ((j==0) && (nextScale==0));
    }
    qmatrixList[j] = (nextScale == 0) ? lastScale : nextScale;
    lastScale = qmatrixList[j];
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode udapte values of a q matrix
*/
//-----------------------------------------------------------------------------
void Param::read_qmatrix_update_value(Buffer *buf, int *qmatrixList)
{
  int i;

  //! scale ratio, slope ratio, number of delta values, matrix values
  for(i = 1; i < 260; i++)
  {
    if( (i >= 4) && ((i - 4) >= qmatrixList[3]) ) 
      break;
    if (i != 3)
      qmatrixList[i] = buf->read_signed();
    else
      qmatrixList[i] = buf->read_unsigned();
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode values of a q matrix set
*/
//-----------------------------------------------------------------------------
void Param::read_qmatrix_set(Buffer *buf, int quantization_set_id, int matrix_type)
{
  int j;
  int k;
  int qmatrix_list_present_flag;
  int *qmatrixList;
  int qmatrix_modetype;
  int matrix_size;

  for (j = 0; j < 16; j++)
  {
    if (j<8) //!< 4x4
      matrix_size = 16;
    else if (j<14) //!< 8x8
      matrix_size = 64;
    else //!< 16x16
      matrix_size = 256;
        
    qmatrix_list_present_flag = buf->read_bits(1);
    qmatrixList = qmatrix_tmp[matrix_type][quantization_set_id][j];
    qmatrix_modetype = 1;
    if (qmatrix_list_present_flag)
    {
      //! mode type
      qmatrixList[0] =  buf->read_unsigned();
      qmatrix_modetype = qmatrixList[0];

      if (qmatrix_modetype == 0) //!< update mode
      {
        read_qmatrix_update_value(buf, qmatrixList);
        matrix_inv_scan(qmatrixList+4, j, matrix_size);
      }
      else if (qmatrix_modetype == 2 && matrix_type == 1) //!< flat mode
        qmatrixList[4] = buf->read_signed();
    }
    if (qmatrix_modetype == 1) //!< matrix mode
    {
      qmatrix_list_present_flag = buf->read_bits(1);
      if (qmatrix_list_present_flag)
      {
        read_qmatrix_list(buf, qmatrixList+4, matrix_size, &(UseDefaultQMatrixFlag[matrix_type][j]));
        matrix_inv_scan(qmatrixList+4, j, matrix_size);
        if (UseDefaultQMatrixFlag[matrix_type][j])
        {
          for (k = 0; k < matrix_size; k++)
            qmatrixList[k+4] = 16;
        }
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode values of common q matrix and set values of all q matrix sets
*/
//-----------------------------------------------------------------------------
void Param::read_qmatrix_set_all(Buffer *buf, int matrix_type) // cq_end
{
  read_qmatrix_set(buf, 0, matrix_type);
  update_qmatrix(0, matrix_type);
  dupli_qmatrix0(matrix_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Read header from byte buffer
*
* Reads sequence header parameters from the byte buffer buf.
*
* \param[in,out] buf Byte buffer
*/
//-----------------------------------------------------------------------------
void Param::read(Buffer *buf) 
{
  int i,j;
  int zt_size;
  int reserved;

  qp = buf->read_bits(7);
  frame_height = buf->read_bits(13);
  frame_width = buf->read_bits(13);
  better_color = buf->read_bits(3);
#ifdef BETTER_CHROMA_DC
  if (better_color==7)
  {
    chroma_dc_th = buf->read_bits(8);
    chroma_dc_scale = buf->read_bits(8);
  }
#endif
  bc_offset = buf->read_bits(5) - 16;    
  lumaDC = buf->read_bits(3);
  loss_rate = 2+buf->read_bits(4);
  ac_exp_limit = buf->read_bits(10);
  luma_mc = buf->read_bits(3);  
  chroma_mc = buf->read_bits(2);
  gen8pel_filter = buf->read_bits(2); 
  mc_round = buf->read_bits(2);
  ndf_mode = buf->read_bits(2);  
  deblockingfilter = buf->read_bits(4);
  if (deblockingfilter != 0)
  {
    dbf_offset_a = buf->read_bits(5);
    dbf_offset_b = buf->read_bits(5);
  }
  wienerfilter = buf->read_bits(1);
#ifdef WIENER_FILTER_INLOOP_SWITCH
  if (wienerfilter)
  {
    wienerfilter_inloop = buf->read_bits(1);
  }  
#endif

  use_extension = buf->read_bits(1); // was lf_spthres

  reserved = buf->read_bits(2); assert(reserved == 0);
  reserved = buf->read_bits(4); assert(reserved == 10);
  reserved = buf->read_bits(4); assert(reserved == 10);

  reserved = buf->read_bits(1); assert(reserved == 1);
  candidate_search_region = buf->read_bits(8);
  num_candidate_averaged = buf->read_bits(4);
  candidate_sad_threshold = buf->read_bits(8);

  reserved = buf->read_bits(1); assert(reserved == 0);
  reserved = buf->read_bits(1); assert(reserved == 0);
  
  nref_frames = buf->read_bits(3);
  adapt_list_size = buf->read_bits(1);
  enable_intcomp = buf->read_bits(1);
  DC_mult = buf->read_bits(3) + ONE;
  reserved = (buf->read_bits(3) + 1) & 7; assert(reserved == 0);
  reserved = (buf->read_bits(3) + 1) & 7; assert(reserved == 0);
  param.weight_diff_scale = (buf->read_bits(3) + 1) & 7;

  trans8x8_enable = buf->read_bits(1);
  motion_sharing_enable = buf->read_bits(1);
  if (param.motion_sharing_enable)
    mvshare_two_enable = buf->read_bits(1);

  ndf_two_rounds = buf->read_bits(1);
  ndf_weighting = buf->read_bits(1);
  ndf_signif_cnt = buf->read_bits(4);
  if (param.trans8x8_enable)
    ndf_4x4_mult = buf->read_bits(4);
  ndf_aggressive = buf->read_bits(1);
  ndf_ic_refine = buf->read_bits(1);
  ndf_mvd_th = buf->read_bits(8);
  ndf_s_pred_quant = buf->read_bits(8);
  ndf_s_pred_signif = buf->read_bits(8);
  ndf_s_pred_mot = buf->read_bits(8);
  ndf_s_intra_quant = buf->read_bits(8);
  ndf_s_mot_disc = buf->read_bits(8);
  ndf_s_ic_strong = buf->read_bits(8);
  ndf_s_ic_intermed = buf->read_bits(8);
  ndf_s_ic_weak = buf->read_bits(8);
  ndf_x_pred_quant = buf->read_bits(8);
  ndf_x_pred_signif = buf->read_bits(8);
  ndf_x_pred_mot = buf->read_bits(8);
  ndf_x_intra_quant = buf->read_bits(8);
  ndf_x_mot_disc = buf->read_bits(8);
  ndf_x_ic_strong = buf->read_bits(8);
  ndf_x_ic_intermed = buf->read_bits(8);
  ndf_x_ic_weak = buf->read_bits(8);

  ndf_s_chroma = buf->read_bits(4);
  aif = buf->read_bits(1);

  reserved = buf->read_bits(1); assert(reserved == 0);
  mpm_default = buf->read_bits(4);

  intra8x8 = buf->read_bits(2);
  transform_selection = buf->read_bits(1);
  ndf_trsf_8x8 = buf->read_bits(1);

  reserved = buf->read_bits(3); assert(reserved == 1);
  reserved = buf->read_bits(1); assert(reserved == 1);

  sb_height= buf->read_bits(8);
  sb_width= buf->read_bits(8);
  
  intra16x16 = buf->read_bits(1);
  if (param.intra16x16)
  {
    candidate_search_region_intra16x16 = buf->read_bits(8);
    transform_selection_16x16 = buf->read_bits(1);
    chroma8x8 = buf->read_bits(1);
  }
  else
  {
    transform_selection_16x16 = 0;
    chroma8x8 = 0;
  }
  param.intra16x16_org = param.intra16x16;

  for (i=0; i<NZTREES-NSUBZTREES; i++) 
  {
    zt_size = buf->read_bits(8);
    ztree_pattern[i][0] = zt_size;

    for (j=1; j<2*zt_size; j++)
      ztree_pattern[i][j] = buf->read_bits(1);

    for (j=2*zt_size; j<3*zt_size; j++)
      ztree_pattern[i][j] = buf->read_bits(4);

    // Don't read the probas here, they're read later.
  }

  if (param.trans8x8_enable) 
  {
    zt_size = buf->read_bits(8);
    ztree_pattern_64[0] = zt_size;

    for (j=1; j<2*zt_size; j++)
      ztree_pattern_64[j] = buf->read_bits(1);

    for (j=2*zt_size; j<3*zt_size; j++)
      ztree_pattern_64[j] = buf->read_bits(8);

    // Don't read the probas here, they're read later.
  }

  // We read all remaining non-proba info here.

  if (param.trans8x8_enable) 
  {
    reserved = buf->read_bits(3);
    assert(reserved == 4);
  } 

 if (param.intra16x16)
 {
   for (i=NZTREES-NSUBZTREES; i<NZTREES; i++) 
   {
     zt_size = buf->read_bits(8);
     ztree_pattern[i][0] = zt_size;

     for (j=1; j<2*zt_size; j++)
       ztree_pattern[i][j] = buf->read_bits(1);

     for (j=2*zt_size; j<3*zt_size; j++)
       ztree_pattern[i][j] = buf->read_bits(5); // Changed number of bits to describe leaves index
   }
 }
#ifdef MVCOMP_ZERORES
  mv_default_comp_config_idx = 0;
  for (i = 0; i < 2; i++)
  {
      MVCompConfig *pconfig = &mv_comp_config[i];
      pconfig->mv_npredictors = buf->read_bits(4);
      if (pconfig->mv_npredictors > MV_MAX_NPREDICTORS)
      {
          fprintf(stderr, "too many predictors, config %d: %d\n", MV_MAX_NPREDICTORS, pconfig->mv_npredictors);
          exit(1);
      }
      memset(pconfig->mv_predictor_used, 0, sizeof(pconfig->mv_predictor_used));
      for (int p = 0; p < pconfig->mv_npredictors; p++)
      {
          pconfig->mv_predictors[p] = buf->read_bits(4);
          if (pconfig->mv_predictors[p] > MV_MAX_PRED_INDEX)
          {
              fprintf(stderr, "invalid mv comp predictor, config %d, predictor %d: %d\n", i, p, pconfig->mv_predictors[p]);
              exit(1);
          }
          pconfig->mv_predictor_used[pconfig->mv_predictors[p]] = 1;
      }
      if (pconfig->mv_predictor_used[MV_PRED_TEMPORAL])
      {
          pconfig->mv_temporal_proba = buf->read_bits(PROB_BITS);
          pconfig->mv_auto_max_temporal_diff = buf->read_bits(1);
          if (!pconfig->mv_auto_max_temporal_diff)
              pconfig->mv_max_temporal_diff = (int)((short)buf->read_bits(16)); // signed.
      }
#ifdef VERBOSE_LOG
      pconfig->print(i);
#endif
  }
#endif
#ifdef CABAC_SYNC_POINTS
  // We expect a synchro bit in the stream unless the decoder's been
  // fired up with -inputref.  '-inputref' means we're decoding from a reference
  // stream, so the stream contains no synchro indication and no synchro.
  if (param.decode_inputref)
  {
      // Ensure superblock_synchro is 0 by default.
      superblock_synchro = 0;
      superblock_nsubstreams = 0;
  }
  else
  {
    superblock_synchro = buf->read_bits(1);
    if (superblock_synchro && buf->read_bits(1) != 0)
        superblock_nsubstreams = buf->read_bits(8);
    else
        superblock_nsubstreams = 0;
  }
#ifdef VERBOSE_LOG
  printf("superblock_synchronisation=%d superblock_nsubstreams=%d\n", superblock_synchro, superblock_nsubstreams);
#endif
#endif

  max_quantization_set_num_minus_one = buf->read_bits(8); // cq_start
  qmatrix_enable_flag[0] = buf->read_bits(1);
  if (qmatrix_enable_flag[0])
    read_qmatrix_set_all(buf, 0);
  qmatrix_enable_flag[1] = buf->read_bits(1);
  if (qmatrix_enable_flag[1])
    read_qmatrix_set_all(buf, 1); // cq_end

  // Create an entropy model which we'll use to read the initial probas.  We
  // have read all the necessary param info to push the correct set of probas
  // to read into.
  // For ztrees, we have to create the skeleton here.
#ifdef FRAME_SUBSTREAMS
  init_ztree(0);
#else
  init_ztree();
#endif

  initialProbEntropyModel = new EntropyModel();
  initialProbEntropyModel->read_stream_probs(buf);

  // Special (non-written) initialisations.
#ifndef SIP_CTX_READ
  if (param.intra16x16)
  {
    for (i=0; i<NCTX_SIP_TYPE; i++)
    {
      initialProbEntropyModel->p_sip_type_b4[i].set_default_prob(32768);
    }
  }
#endif //SIP_CTX_READ

  // For ztrees, we have to copy the proba i_p values back into the pattern and into
  // i_prob[].  There should be better separation here between structure and probas,
  // but this minimizes code change for the moment.
  for (i=0; i<(param.intra16x16 ? NZTREES : NZTREES-NSUBZTREES); i++) 
  {
    ztree[i]->copy_initial_probs_back(initialProbEntropyModel, &ztree_pattern[i][0]);
  }
  if (param.trans8x8_enable) 
  {
    ztree_64->copy_initial_probs_back(initialProbEntropyModel, &ztree_pattern_64[0]);
#ifdef ZTREE_64_CT8_CTX_READ
    ztree_64_CT8->copy_initial_probs_back(initialProbEntropyModel, &ztree_pattern_64[0]);
#else
    // ztree_64_CT8 is a copy of ztree_64.
    ztree_64_CT8->copy_initial_probs_from(ztree_64, initialProbEntropyModel);
#endif //ZTREE_64_CT8_CTX_READ
  }

  //initialProbEntropyModel->printProbas("first");
  
  if (use_extension)
  {
    bool done = false;
    while (!done)
    {
      int code = buf->read_bits(8);
      
      switch (code)
      {
        case tag_end:
          done = true;
          break;
          
        case tag_bugfix87:
          bugfix87 = buf->read_bits(1);
          break;
          
        case tag_bugfix88:
          bugfix88 = buf->read_bits(1);
          break;

        case tag_bugfix89:
          bugfix89 = buf->read_bits(1);
          break;

        case tag_bugfix93:
          bugfix93 = buf->read_bits(1);
          break;

        case tag_alt8x8_trans_idx:
          alt8x8_trans_idx = buf->read_bits(4);
          break;

        case tag_ndf_lc:
          ndf_lc_luma = buf->read_bits(4);
          ndf_lc_chroma = buf->read_bits(4);
          break;

        case tag_edge_pred_bug_fix:
          bugfix90 = buf->read_bits(1);
          break;
          
#ifdef I16_IN_P_FRAME
        case tag_intra16x16p:
          intra16x16p = buf->read_bits(1);
          break;
#endif
          
        default:
          printf("Unknown extension tag:0x%02x. Quitting...\n", code);
          exit(EXIT_FAILURE);
          break;
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Parse encoder configuration file
*
* Parses content of encoder configuration file and assigns values to respective
* parameters.
*
* \param[in] filename File name of configuration file.
*
* \return ONE if success, ZERO if error
*/
//-----------------------------------------------------------------------------
int Param::read_config_file(char *filename) 
{
  FILE *f;
  char line[MAX_LINE_LEN];
  int equal_pos;
  int i;
  char *p;
  int atoi_p;
  int matched;
#ifdef MVCOMP_ZERORES
  int conf;
#endif

  f = fopen(filename, "rt");
  if (!f)
    return ZERO;

  use_extension = FALSE;
  
  while (TRUE) 
  {
    fgets(line, MAX_LINE_LEN, f);
    if (feof(f))
      break;

    equal_pos = UNDEFINED;

    for (i = ZERO; i < MAX_LINE_LEN; i++) 
    {
      if (line[i] == 0)
        break;
      else if (line[i] == '=') 
      {
        line[i] = 0;
        equal_pos = i;
      }
      else if (line[i] == '#') 
      {
        line[i] = 0;
        break;
      }
      else if (line[i] == 0x0d || line[i] == 0x0a)
        line[i] = 0;
      else if (line[i] == '"')
        line[i] = 0;
    }

    if (equal_pos != UNDEFINED) 
    {
      matched = FALSE;

      for (i = 0; encoder_int_param_list_ext[i].ptr != NULL; i++) 
      {
        if (strcmp(line, encoder_int_param_list_ext[i].name) == 0) 
        {
          p = line+equal_pos+1;
          atoi_p = atoi(p);
          atoi_p = imax(encoder_int_param_list_ext[i].min_val, imin(encoder_int_param_list_ext[i].max_val,atoi_p));
          *(encoder_int_param_list_ext[i].ptr) = atoi_p;
          matched = TRUE;
          use_extension = TRUE;
        }
      }
      for (i = 0; encoder_int_param_list[i].ptr != NULL; i++) 
      {
        if (strcmp(line, encoder_int_param_list[i].name) == 0) 
        {
          p = line+equal_pos+1;
          atoi_p = atoi(p);
          atoi_p = imax(encoder_int_param_list[i].min_val, imin(encoder_int_param_list[i].max_val,atoi_p));
          *(encoder_int_param_list[i].ptr) = atoi_p;
          matched = TRUE;
        }
      }
      for (i = 0; encoder_float_param_list[i].ptr != NULL; i++) 
      {
        if (strcmp(line, encoder_float_param_list[i].name) == 0) 
        {
          p = line+equal_pos+1;

          *(encoder_float_param_list[i].ptr) = atof(p);
          matched = TRUE;
        }
      }
      for (i = 0; encoder_string_param_list[i].ptr != NULL; i++) 
      {
        if (strcmp(line, encoder_string_param_list[i].name) == 0) 
        {
          p = line+equal_pos+1;
          strcpy(encoder_string_param_list[i].ptr, p+1);
          matched = TRUE;
        }
      }
      if (matched == FALSE) 
      {
        printf("Unknown parameter <%s>\n", line);
      }
    }
  }

  fclose(f);

// The following is where paremeters are initialized or overriden, be careful.

#ifdef AUTO_SELECT_HADAMARD
  if (hadamard == 3)
  {
    if (frame_width>ASH_WIDTH && frame_height>ASH_HEIGHT)
      hadamard=2;
    else
      hadamard=1;
  }
#endif

#ifdef MVCOMP_ZERORES
  // Map predictor strings to internal values.
  for (conf = 0; conf < 2; conf++)
  {
      MVCompConfig *pconfig = &param.mv_comp_config[conf];
      int pred;
      for (pred = 0; pred < MV_MAX_NPREDICTORS; pred++)
      {
        pconfig->mv_predictors[pred] = get_mv_predictor(pconfig->mv_predictor_strings[pred]);
        if (pconfig->mv_predictors[pred] == MV_PRED_NONE)
            break;
      }
      // Add "none" predictors if necessary, to have at least two predictors.
      while (pred < 2)
          pconfig->mv_predictors[pred++] = MV_PRED_NONE;
      pconfig->mv_npredictors = pred;
      // Genearate set of preds used.
      memset(pconfig->mv_predictor_used, 0, sizeof(pconfig->mv_predictor_used));
      for (int i = 0; i < pconfig->mv_npredictors; i++)
          pconfig->mv_predictor_used[pconfig->mv_predictors[i]] = 1;
  }
#ifdef VERBOSE_LOG
  for (conf = 0; conf < 2; conf++)
      param.mv_comp_config[conf].print(conf);
#endif
#endif
  if(motion_sharing_enable==FALSE)
  {
    mvshare_two_enable=FALSE;
    mvshare_two_enable_me=FALSE;
    motion_sharing_me_init_offset=FALSE;
    motion_sharing_me_refine=FALSE;
    motion_sharing_me_refine_full=FALSE;
    motion_sharing_me_trellis=FALSE;
    motion_sharing_me_trellis_full=FALSE;
    motion_sharing_me_region=FALSE;
  }

  if(param.pyramid_coding==7)
    sprintf(param.expPrmdFrmt,"b3r%db1r%db0e%db2e%db5r%db4e%db6e%d",
    param.qp+param.qp_off_intref,param.qp+2*param.qp_off_intref,
    param.qp+param.qp_off_intref+param.qp_off_intdisp,
    param.qp+param.qp_off_intref+param.qp_off_intdisp,
    param.qp+2*param.qp_off_intref,
    param.qp+param.qp_off_intref+param.qp_off_intdisp,
    param.qp+param.qp_off_intref+param.qp_off_intdisp);

   if(param.pyramid_coding==7)
    sprintf(param.expPrmdFrmtIdr,"b3r%db1r%db0e%db2e%db5r%db4e%db6e%d",
    param.qp+param.qp_off_intref,param.qp+2*param.qp_off_intref,
    param.qp+param.qp_off_intref+param.qp_off_intdisp,
    param.qp+param.qp_off_intref+param.qp_off_intdisp,
    param.qp+2*param.qp_off_intref,
    param.qp+param.qp_off_intref+param.qp_off_intdisp,
    param.qp+param.qp_off_intref+param.qp_off_intdisp);

  return ONE;
}

#ifdef MVCOMP_ZERORES

char const *Param::mv_pred_name[MV_MAX_PRED_INDEX+2];
//-----------------------------------------------------------------------------
/*!
* \brief Map a mv predictor name to its internal value.
*
* \return value of predictor
*/
//-----------------------------------------------------------------------------
int Param::get_mv_predictor(char const *str)
{
    for (int i = 0; mv_pred_name[i] != NULL; i++)
    {
        if (strcmp(str, mv_pred_name[i]) == 0)
            return i;
    }
    fprintf(stderr, "unsupported mv predictor: '%s'\n", str);
    exit(-1);
    return -1;
}

MVCompConfig::MVCompConfig()
{
  mv_zero_residue = 0;
  mv_npredictors = 0;
  mv_temporal_proba = 0;
  mv_keep_learned_movement = 0;
  mv_max_temporal_diff = 0;
  mv_auto_max_temporal_diff = 0;
  memset(mv_predictor_used, 0, sizeof(mv_predictor_used));
}

MVCompConfig::~MVCompConfig()
{
}

void MVCompConfig::print(int i)
{
  printf("MVCOMP: config %d: #preds=%d (", i, mv_npredictors);
  for (int p = 0; p < mv_npredictors; p++)
      printf(" %d", mv_predictors[p]);
  printf(" ): zero-residue=%d temp_proba=%d max-diff=%d auto-diff=%d keep-learned-movement=%d\n",
         mv_zero_residue,
         mv_temporal_proba,
         mv_max_temporal_diff,
         mv_auto_max_temporal_diff,
         mv_keep_learned_movement);
}

#endif

//-----------------------------------------------------------------------------
/*!
* \brief Initialize values of all q matrix sets
*/
//-----------------------------------------------------------------------------
void Param::ini_qmatrix(int matrix_type) // cq_start
{
  int i, j, k;
  int max_quantization_set_num = MAX_QUANT_SET_NUM;

  for (j = 0; j < max_quantization_set_num; j++)
  {
    qmatrix_overwrite_flag[matrix_type][j] = 0;
    for (i = 0; i < 16; i++)
    {
      qmatrix_check[matrix_type][j][i] = 0;
      qmatrix_update_check[matrix_type][j][i] = 0;
      qmatrix_tmp[matrix_type][j][i][0] = 1;
      qmatrix[matrix_type][j][i][0] = 1;
      for (k = 1; k < 4; k++)
      {
        qmatrix_tmp[matrix_type][j][i][k] = 0;
        qmatrix[matrix_type][j][i][k] = 0;
      }
      for (k = 4; k < 260; k++)
      {
        qmatrix_tmp[matrix_type][j][i][k] = 16;
        qmatrix[matrix_type][j][i][k] = 16;
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Duplicate common q matrix set to other q matrix sets
*/
//-----------------------------------------------------------------------------
void Param::dupli_qmatrix0(int matrix_type)
{
  int i, j, k;
  int max_quantization_set_num = MAX_QUANT_SET_NUM;

  for (j = 1; j < max_quantization_set_num; j++)
  {
    for (i = 0; i < 16; i++)
    {
      if (!qmatrix_check[matrix_type][j][i])
      {
        for (k = 4; k < 260; k++)
        {
          qmatrix_tmp[matrix_type][j][i][k] = qmatrix_tmp[matrix_type][0][i][k];
          qmatrix[matrix_type][j][i][k] = qmatrix_tmp[matrix_type][0][i][k];
        }
      }
      else
      {
        if (qmatrix_tmp[matrix_type][j][i][0] == 0)
        {
          for (k = 4; k < 260; k++)
            qmatrix[matrix_type][j][i][k] = qmatrix_tmp[matrix_type][0][i][k];
        }
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Allocates memory buf, opens file Filename in f, reads contents into
*/
//-----------------------------------------------------------------------------
char* Param::GetConfigFileContent (char *Filename)
{
  long FileSize;
  FILE *f;
  char *buf;

  if (NULL == (f = fopen (Filename, "r")))
  {
    printf( "Cannot open configuration file %s.", Filename);
    return NULL;
  }
  if (0 != fseek (f, 0, SEEK_END))
  {
    printf( "Cannot fseek in configuration file %s.", Filename);
    return NULL;
  }
  FileSize = ftell (f);
  if (FileSize < 0 || FileSize > 100000)
  {
    printf( "Unreasonable Filesize %ld reported by ftell for configuration file %s.", FileSize, Filename);
    return NULL;
  }
  if (0 != fseek (f, 0, SEEK_SET))
  {
    printf(  "Cannot fseek in configuration file %s.", Filename);
    return NULL;
  }
  if ((buf = (char*)malloc (FileSize + 1))==NULL) 
  {
    printf(  "Cannot allocate memory at GetConfigFileContent: buf");
    return NULL;
  }
  FileSize = fread (buf, 1, FileSize, f);
  buf[FileSize] = '\0';
  fclose (f);
  return buf;
}

//-----------------------------------------------------------------------------
/*!
* \brief Parse q matrix configuration file
*/
//-----------------------------------------------------------------------------
int   Param::read_qmatrix_file(char* filename, int matrix_type)
{
  char *content;
  int j;
  int max_quantization_set_num = MAX_QUANT_SET_NUM;

  if (qmatrix_enable_flag[0] || qmatrix_enable_flag[1])
  {
    content = GetConfigFileContent(filename);
    if (content!='\0')
      parse_qmatrix(content, strlen (content), matrix_type);
    else
      printf("\nError: \nProceeding with default values for all matrices.");

    printf("\n");

    free(content);
  }
  for (j=0; j<max_quantization_set_num; j++)
    patch_qmatrix(j, matrix_type);
  return ONE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get array id for storing matrix
*/
//-----------------------------------------------------------------------------
int Param::get_array_id(int type, int size, int yuv)
{
  int aid = -1;

  if(type == 0 && size == 0 && yuv == 0) //!< 4x4, 2x8, 8x2
    aid = 0;
  else if(type == 0 && size == 0 && yuv == 1)
    aid = 1;
  else if(type == 0 && size == 0 && yuv == 2)
    aid = 2;
  else if(type == 1 && size == 0 && yuv == 0)
    aid = 3;
  else if(type == 1 && size == 0 && yuv == 1)
    aid = 4;
  else if(type == 1 && size == 0 && yuv == 2)
    aid = 5;
  else if(type == 0 && size == 1 && yuv == 0)
    aid = 6;
  else if(type == 0 && size == 2 && yuv == 0)
    aid = 7;

  else if(type == 0 && size == 3 && yuv == 0) //!< 8x8
    aid = 8;
  else if(type == 1 && size == 3 && yuv == 0)
    aid = 9;
  else if(type == 0 && size == 3 && yuv == 1)
    aid = 10;
  else if(type == 0 && size == 3 && yuv == 2)
    aid = 11;
  else if(type == 1 && size == 3 && yuv == 1)
    aid = 12;
  else if(type == 1 && size == 3 && yuv == 2)
    aid = 13;

  else if(type == 0 && size == 4 && yuv == 0) //!< 16x16
    aid = 14;
  else if(type == 1 && size == 4 && yuv == 0)
    aid = 15;

  return aid;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get value range for storing matrix
*/
//-----------------------------------------------------------------------------
int Param::get_value_arrage(int value, int size)
{
  int range;

  if(value < 4)
    range = 1;
  else
  {
    if(size == 4) //!< 16x16
      range = 256;
    else if(size == 3) //!< 8x8
      range = 64;
    else //!< 4x4, 2x8, 8x2 
      range = 16;
  }
  return range;
}

//-----------------------------------------------------------------------------
/*!
* \brief Find substring in string
*/
//-----------------------------------------------------------------------------
int Param::strindex(char *str,const char *substr)
{
  int end,i,j;

  end = strlen(str) - strlen(substr);
  if ( end > 0 )
  {
    for ( i = 0; i <= end; i++ )
      for ( j = i; str[j] == substr[j-i]; j++ )
        if ( substr[j-i+1] == '\0' )
          return i+1;
  }
  return -1;
}

//-----------------------------------------------------------------------------
/*!
* \brief Check the parameter name of q matrix
*/
//-----------------------------------------------------------------------------
void Param::check_qmatrix_para (char *s, int *set_id, int *array_id, int *array_off, int *range)
{
  int i;
  int type = -1;
  int size = -1;
  int yuv = -1;
  int sid = -1;
  int value = -1;

  for(i = 0; i < 2 && type < 0; i++)
    if(strindex(s, CQ_TYPE[i]) > 0)
      type = i;
  for(i = 0; i < 5 && size < 0; i++)
    if(strindex(s, CQ_SIZE[i]) > 0)
      size = i;
  for(i = 0; i < 3 && yuv < 0; i++)
    if(strindex(s, CQ_YUV[i]) > 0)
      yuv = i;
  for(i = 0; i < 5 && sid < 0; i++)
    if(strindex(s, CQ_SETID[i]) > 0)
      sid = i;
  for(i = 0; i < 5 && value < 0; i++)
    if(strindex(s, CQ_VALUE[i]) > 0)
      value = i;
  *set_id = sid;
  *array_id = get_array_id(type, size, yuv);
  *range = get_value_arrage(value, size);
  *array_off = value;
}


//-----------------------------------------------------------------------------
/*!
* \brief Parse the q matrix values read from cfg file.
*/
//-----------------------------------------------------------------------------
void Param::parse_qmatrix (char *buf, int bufsize, int matrix_type)
{
  char *items[10000] = {NULL};
  int item = 0;
  int InString = 0, InItem = 0;
  char *p = buf;
  char *bufend = &buf[bufsize];
  int IntContent;
  int i, j, range = 0, cnt;
  int set_id, array_id, array_off;
  int *ScalingList = NULL;

  while (p < bufend)
  {
    switch (*p)
    {
      case 13:
        p++;
        break;
      case '#':               //!< Found comment
        *p = '\0';            //!< Replace '#' with '\0' in case of comment immediately following integer or string
        while (*p != '\n' && p < bufend)  //!< Skip till EOL or EOF, whichever comes first
          p++;
        InString = 0;
        InItem = 0;
        break;
      case '\n':
        InItem = 0;
        InString = 0;
        *p++='\0';
        break;
      case ' ':
      case '\t':              //!< Skip whitespace, leave state unchanged
        if (InString)
          p++;
        else
        {                     //!< Terminate non-strings once whitespace is found
          *p++ = '\0';
          InItem = 0;
        }
        break;
      case '"':               //!< Begin/End of String
        *p++ = '\0';
        if (!InString)
        {
          items[item++] = p;
          InItem = ~InItem;
        }
        else
          InItem = 0;
        InString = ~InString; //!< Toggle
        break;
      case ',':
        p++;
        InItem = 0;
        break;
      default:
        if (!InItem)
        {
          items[item++] = p;
          InItem = ~InItem;
        }
        p++;
    }
  }

  item--;

  for (i=0; i<item; i+=cnt)
  {
    cnt=0;
    check_qmatrix_para (items[i+cnt], &set_id, &array_id, &array_off, &range);
    if (set_id < 0 || array_id < 0)
    {
      printf (" Parsing error in quantization matrix config file: Parameter Name '%s' not recognized.\n", items[i+cnt]);
      exit(-1);
    }
    cnt++;
    if (strcmp ("=", items[i+cnt]))
    {
      printf ( " Parsing error in quantization matrix config file: '=' expected as the second token in each item.\n");
      exit(-1);
    }
    cnt++;
    ScalingList = qmatrix_tmp[matrix_type][set_id][array_id]+array_off;
    qmatrix_overwrite_flag[matrix_type][set_id] = set_id > 0 ? 1 : 0;
    if(range == 1)
      qmatrix_update_check[matrix_type][set_id][array_id] = 1;
    else
      qmatrix_check[matrix_type][set_id][array_id] = 1;
    for (j=0; j<range; j++)
    {
      if (1 != sscanf (items[i+cnt+j], "%d", &IntContent))
      {
        printf (" Parsing error in quantization matrix file: Expected numerical value for Parameter of %s, found '%s'.\n", items[i], items[i+cnt+j]);
        exit(-1);
      }
      ScalingList[j] = (short)IntContent; //!< save value in matrix
    }
    cnt+=j;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Patch q matrix values
*/
//-----------------------------------------------------------------------------
void Param::patch_qmatrix(int quantization_set_id, int matrix_type) // cq_end
{
  int *qmatrix;
  int i, cnt;
  int matrix_size;

  for (i=0; i<16; i++)
  {
    if (i<8)  //!< 4x4
      matrix_size = 16;
    else if (i<14) //!< 8x8
      matrix_size = 64;
    else //!< 16x16
      matrix_size = 256;

    //! patch q matrix update values
    qmatrix = qmatrix_tmp[matrix_type][quantization_set_id][i];
    if (qmatrix_update_check[matrix_type][quantization_set_id][i])
    {
      if(matrix_type == 0)
        qmatrix[0] = max(0, min(1, qmatrix[0]));
      else
        qmatrix[0] = max(0, min(2, qmatrix[0]));
      qmatrix[3] = max(0, min(matrix_size, qmatrix[3]));
      if (qmatrix[0] == 2 && matrix_type == 1) //!< flat mode
      {
        qmatrix[3] = 1;
        if (abs(qmatrix[4]) > 127) //!< [-127, 127]
          qmatrix[4] = 0;
      }
    }
    else
    {
      qmatrix[0] = 1;
      qmatrix[1] = qmatrix[2] = qmatrix[3] = 0;
    }
    //! patch q matrix
    qmatrix = qmatrix_tmp[matrix_type][quantization_set_id][i]+4;
    if (!qmatrix_check[matrix_type][quantization_set_id][i]) //!< matrix not found, pad with default value
    {
      for (cnt=0; cnt<matrix_size; cnt++)
       qmatrix[cnt] = 16;
    }
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
